/*
 *      Copyright (c) 2003 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: bex.c,v 1.5 2007/03/26 23:19:35 nis Exp $";
#endif

#include <config.h>

#include <sys/types.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>

#include "bex.h"

/*

expr	: a_expr
	;

a_expr	: m_expr
	| m_expr '|' a_expr
	;

m_expr	: n_expr
	| n_expr '&' n_expr

n_expr	: f_expr
	| '!' f_expr
	;

f_expr	: elem
	| '(' expr ')'
	;

elem	: id
	| T
	| F
	;
*/

struct a_expr {
	struct m_expr *e;
	struct a_expr *next;
};

struct m_expr {
	struct n_expr *e;
	struct m_expr *next;
};

struct n_expr {
	struct f_expr *e;
	int not;
};

struct f_expr {
	struct a_expr *e;
	size_t i;
};

static struct a_expr bex_true_0, *bex_true = &bex_true_0;
static struct a_expr bex_false_0, *bex_false = &bex_false_0;

static int eval_a(struct a_expr *, struct bex *);
/*static int eval_m(struct m_expr *, struct bex *);*/

static struct a_expr *conv_a(int *, size_t, size_t *);
static struct m_expr *conv_m(int *, size_t, size_t *);
static struct n_expr *conv_n(int *, size_t, size_t *);
static struct f_expr *conv_f(int *, size_t, size_t *);

static void free_a_expr(struct a_expr *);
static void free_m_expr(struct m_expr *);
static void free_n_expr(struct n_expr *);
static void free_f_expr(struct f_expr *);

struct bex {
	size_t len;
	int *in, *attrs;
	struct a_expr *expr;
};

struct bex *
bex_compile(attrs, len)
int *attrs;
size_t len;
{
	struct bex *b;
	size_t j;
	if (!(b = calloc(1, sizeof *b))) {
		return NULL;
	}
	b->in = NULL;
	b->expr = NULL;
	b->len = len;
	if (!(b->in = calloc(len, sizeof *b->in))) {
		goto error;
	}
	b->attrs = attrs;

	if (attrs[0]!=BEX_START || attrs[len-1]!=BEX_END) {
		goto error;
	}
	if (!(b->expr = conv_a(attrs, 1, &j))) {
		goto error;
	}
	if (j!=len-1) {
		goto error;
	}
	return b;

error:
	bex_free(b);
	return NULL;
}

void
bex_free(b)
struct bex *b;
{
	if (!b) {
		return;
	}
	free(b->in);
	free_a_expr(b->expr);
	free(b);
}

void
bex_clear(b)
struct bex *b;
{
/*
	size_t i;
	for (i=0; i<b->len; i++) {
		b->in[i] = 0;
	}
*/
	bzero(b->in, b->len * sizeof *b->in);
}

void
bex_setnth(b, n)
struct bex *b;
size_t n;
{
	assert(n>0);
	assert(n<b->len);
	assert(b->in[n]==0);
	assert(b->attrs[n]==BEX_ELEM);
	b->in[n]=1;
}

int
bex_check(b)
struct bex *b;
{
	return eval_a(b->expr, b);
}

static int
eval_a(a, b)
struct a_expr *a;
struct bex *b;
{
	for (; a; a=a->next) {
		struct m_expr *m = a->e;
		for (; m; m=m->next) {
			int v;
			struct n_expr *n = m->e;
			struct f_expr *f = n->e;
			if (f->e) {
				assert(f->i==0);
				if (f->e == bex_true) {
					v = 1;
				}
				else if (f->e == bex_false) {
					v = 0;
				}
				else {
					v = eval_a(f->e, b);
				}
			}
			else {
				v = b->in[f->i];
			}
			if (v == n->not) {
				goto cont;
			}
		}
		return 1;
	cont: ;
	}
	return 0;
}

#if 0
static
eval_a(a, b)
struct a_expr *a;
struct bex *b;
{
	for (; a; a=a->next) {
		if (eval_m(a->e, b)) {
			return 1;
		}
	}
	return 0;
}

static
eval_m(m, b)
struct m_expr *m;
struct bex *b;
{
	for (; m; m=m->next) {
		int v;
		struct n_expr *n = m->e;
		struct f_expr *f = n->e;
		if (f->e) {
			assert(f->i==0);
			v = eval_a(f->e, b);
		}
		else {
			v = b->in[f->i];
		}
		if (v == n->not) {
			return 0;
		}
	}
	return 1;
}
#endif

static struct a_expr *
conv_a(attr, i, j)
int *attr;
size_t i, *j;
{
	struct a_expr *a, *n;
	if (!(n = a = calloc(1, sizeof *a))) {
		return NULL;
	}
	n->next = NULL;
	if (!(n->e = conv_m(attr, i, &i))) {
		goto error;
	}
	while (attr[i]==BEX_OR) {
		if (!(n = n->next = calloc(1, sizeof *n->next))) {
			goto error;
		}
		n->next = NULL;
		if (!(n->e = conv_m(attr, i+1, &i))) {
			goto error;
		}
	}
	*j = i;
	return a;
error:
	for (; a; ) {
		n = a;
		a = a->next;
		free_m_expr(n->e);
		free(n);
	}
	return NULL;
}

static struct m_expr *
conv_m(attr, i, j)
int *attr;
size_t i, *j;
{
	struct m_expr *m, *n;
	if (!(n = m = calloc(1, sizeof *m))) {
		return NULL;
	}
	n->next = NULL;
	if (!(n->e = conv_n(attr, i, &i))) {
		goto error;
	}
	while (attr[i]==BEX_AND) {
		if (!(n = n->next = calloc(1, sizeof *n->next))) {
			goto error;
		}
		n->next = NULL;
		if (!(n->e = conv_n(attr, i+1, &i))) {
			goto error;
		}
	}
	*j = i;
	return m;
error:
	for (; m; ) {
		n = m;
		m = m->next;
		free_n_expr(n->e);
		free(n);
	}
	return NULL;
}

struct n_expr *
conv_n(attr, i, j)
int *attr;
size_t i, *j;
{
	struct n_expr *n;
	if (!(n = calloc(1, sizeof *n))) {
		return NULL;
	}
	if (attr[i]==BEX_NOT) {
		i++;
		n->not = 1;
	}
	else {
		n->not = 0;
	}
	if (!(n->e = conv_f(attr, i, j))) {
		free(n);
		return NULL;
	}
	return n;
}

static struct f_expr *
conv_f(attr, i, j)
int *attr;
size_t i, *j;
{
	struct f_expr *f;
	if (!(f = calloc(1, sizeof *f))) {
		return NULL;
	}
	f->i = 0;
	f->e = NULL;
	if (attr[i]==BEX_ELEM) {
		f->i = i;
		*j = i+1;
		return f;
	}
	else if (attr[i]==BEX_T) {
		f->e = bex_true;
		*j = i+1;
		return f;
	}
	else if (attr[i]==BEX_F) {
		f->e = bex_false;
		*j = i+1;
		return f;
	}
	else if (attr[i]==BEX_LPAR) {
		if (!(f->e = conv_a(attr, i+1, &i))) {
			goto error;
		}
		if (attr[i]!=BEX_RPAR) {
			goto error;
		}
		*j = i+1;
		return f;
	}
error:
	assert(f->e != bex_false);
	assert(f->e != bex_true);
	free_a_expr(f->e);
	free(f);
	return NULL;
}

static void
free_a_expr(e)
struct a_expr *e;
{
	struct a_expr *n;
	for (; e; ) {
		n = e;
		e = e->next;
		free_m_expr(n->e);
		free(n);
	}
}

static void
free_m_expr(e)
struct m_expr *e;
{
	struct m_expr *n;
	for (; e; ) {
		n = e;
		e = e->next;
		free_n_expr(n->e);
		free(n);
	}
}

static void
free_n_expr(e)
struct n_expr *e;
{
	if (e) {
		free_f_expr(e->e);
	}
	free(e);
}

static void
free_f_expr(e)
struct f_expr *e;
{
	if (e && e->e != bex_true && e->e != bex_false) {
		free_a_expr(e->e);
	}
	free(e);
}
