/*
 *	Copyright (c) 2004 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: bitlist.h,v 1.6 2007/03/26 07:32:14 nis Exp $";
#endif

struct _bitlist_t {
	bitlist_reg_t q;
	unsigned char *p, *l;
	int m;
};

/*
 * p から 4 byte 取ってきて network order -> host order する.
 * ILP32 : OK
 * LP64 : OK
 * ILP64 : NG --- 未対応
 * LLP64 : -
 */
#if 0
#define NPTOHL(p)	ntohl(*(u_int *)(p))
#else
#define NPTOHL(p)	((u_int)be32ptoh(p))
#if 1
#define be32ptoh(p)	\
	((*((p) + 0) & 0xff) << 24) |	\
	((*((p) + 1) & 0xff) << 16) |	\
	((*((p) + 2) & 0xff) << 8) |	\
	(*((p) + 3) & 0xff)
#else
#define be32ptoh(p)	\
	((*((p) + 3) & 0xff) << 24) |	\
	((*((p) + 2) & 0xff) << 16) |	\
	((*((p) + 1) & 0xff) << 8) |	\
	(*((p) + 0) & 0xff)
#endif
#endif

/*
 * q に p から BITLIST_BPW byte 取ってきてセットする.
 * p は BITLIST_BPW byte 進む.
 * p は align されていなくても良い.
 */
#if BITLIST_BPW == 8
#define	BITLIST_GETWORD(b)	\
	((b)->q = (bitlist_reg_t)NPTOHL((b)->p) \
	, (b)->q <<= 32 \
	, (b)->q |= (bitlist_reg_t)NPTOHL((b)->p + 4) \
	, (b)->p += BITLIST_BPW)
#else
#define	BITLIST_GETWORD(b)	\
	((b)->q = (bitlist_reg_t)NPTOHL((b)->p) \
	, (b)->p += BITLIST_BPW)
#endif

/*
 * int n
 *
 * q + n が l 未満なら, q の上位 n byte を p に書き出す.
 * そうでなければ, なにもしない.
 * いずれの場合でも p は n byte 進む.
 */
#if BITLIST_BPW == 8
#define	BITLIST_PUTWORDN(n, b) do { \
	if ((b)->p + (n) < (b)->l) { 	/* kokoha, <= de yoi noka? */ \
		if (1 <= n) *((b)->p + 0) = (b)->q >> 56; \
		if (2 <= n) *((b)->p + 1) = (b)->q >> 48; \
		if (3 <= n) *((b)->p + 2) = (b)->q >> 40; \
		if (4 <= n) *((b)->p + 3) = (b)->q >> 32; \
		if (5 <= n) *((b)->p + 4) = (b)->q >> 24; \
		if (6 <= n) *((b)->p + 5) = (b)->q >> 16; \
		if (7 <= n) *((b)->p + 6) = (b)->q >>  8; \
		if (8 <= n) *((b)->p + 7) = (b)->q; \
	} \
	(b)->p += (n); \
} while (0)
#else
#define	BITLIST_PUTWORDN(n, b) do { \
	if ((b)->p + (n) < (b)->l) { 	/* kokoha, <= de yoi noka? */ \
		if (1 <= n) *((b)->p + 0) = (b)->q >> 24; \
		if (2 <= n) *((b)->p + 1) = (b)->q >> 16; \
		if (3 <= n) *((b)->p + 2) = (b)->q >>  8; \
		if (4 <= n) *((b)->p + 3) = (b)->q; \
	} \
	(b)->p += (n); \
} while (0)
#endif

/*
 * void *p0
 *
 * 初期状態にする.
 */
#define BITLIST_INIT_GETBITS(b, p0) do { \
	(b)->p = (u_char *)(p0); \
	(b)->q = 0; \
	(b)->m = 0; \
} while (0)

/*
 * lvalue bitlist_reg_t r
 * int n
 *
 * 初期状態: q の右詰めの m bits が有効である.
 * r に q から n bits 取って来て, 右詰めでセットする. (0 < n <= BITLIST_BPW * 8)
 * r のそれ以外の bit は変化しない.
 * 必要なら, *p から q に BITLIST_BPW byte 取ってくる.
 * q, m は変化し, 新たな状態になる.
 * この操作の結果, 0 <= m < BITLIST_BPW * 8 になる.
 *
 * なお, p からの fetch は BITLIST_BPW byte 単位で行われるので,
 * p の有効な byte 数が BITLIST_BPW で割り切れない場合に,
 * 最後の 1..7 byte への不正なアクセスが起こり得る
 * ことに注意しなければならない.
 */
#define	BITLIST_GETBITS(r, n, b) do { \
	register bitlist_reg_t BLTMP_MASK = ~(~(bitlist_reg_t)1 << ((n) - 1)); \
	register bitlist_reg_t BLTMP_R = 0; \
	if ((n) <= (b)->m) { \
		(b)->m = (b)->m - (n); \
	} \
	else { \
		(b)->m = (n) - (b)->m; \
		if ((b)->m != BITLIST_BPW * 8) { \
			BLTMP_R |= (b)->q << (b)->m; \
		} \
		BITLIST_GETWORD((b)); \
		(b)->m = BITLIST_BPW * 8 - (b)->m; \
	} \
	BLTMP_R |= (b)->q >> (b)->m; \
	(r) = ((r) & ~BLTMP_MASK) | (BLTMP_R & BLTMP_MASK); \
} while (0)

/*
 * void *p0
 * void *l0
 *
 * 初期状態にする.
 *
 * buffer overrun について:
 * p が l0 を越える場合, p の内容は保証されないが,
 * l0 以降 (l0 を含む) に書き込みが起こることはない.
 * その場合でも, p は進むことに注意せよ.
 * 配列外を指すポインタを作ることは C 言語の規格では違反だが,
 * たいていの処理系では問題を起こさない. (XXX いずれ修正します...)
 */
#define	BITLIST_INIT_PUTBITS(b, p0, l0) do { \
	(b)->p = (u_char *)(p0); \
	(b)->q = 0; \
	(b)->m = 0; \
	(b)->l = (u_char *)(l0); \
} while (0)

/*
 * bitlist_reg_t r
 * int n
 *
 * r の右詰め n bits を q に左詰めでセットする. (0 < n <= BITLIST_BPW * 8)
 * r は変化しない.
 *
 * q には既に m bits が左詰めでセットされている.
 * 必要に応じて q は *p に書き出される.
 * この操作の結果, 0 <= m < BITLIST_BPW * 8 になる.
 */
#define	BITLIST_PUTBITS(r, n, b) do { \
	register bitlist_reg_t BLTMP_R = (r) << (BITLIST_BPW * 8 - (n)); \
	(b)->q |= BLTMP_R >> (b)->m; \
	(b)->m += (n); \
	if ((b)->m >= BITLIST_BPW * 8) { \
		BITLIST_PUTWORDN(BITLIST_BPW, (b)); \
		(b)->m -= BITLIST_BPW * 8; \
		if ((b)->m) { \
			(b)->q = BLTMP_R << ((n) - (b)->m); \
		} \
		else { \
			(b)->q = 0; \
		} \
	} \
} while (0)

#define	BITLIST_PUTBIT(r, b)	BITLIST_PUTBITS(r, 1, b)

/*
 * q に残っている左詰め m bits を p に書き出す (byte に切り上げ).
 * p は書き出した byte 数だけ進む.
 * p - p0 が書き込んだ byte 数である.
 */
#define	BITLIST_FINISH_PUTBITS(b) do { \
	int BLTMP_N = ((b)->m + 7) / 8; \
	BITLIST_PUTWORDN(BLTMP_N, (b)); \
} while (0)
