/*
 *	Copyright (C) 1996 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: bset.c,v 1.2 2007/03/21 03:17:41 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/bset.h>

typedef unsigned int bset_elem_t;
typedef unsigned int bset_summary_t;

struct _bset_t {
	bset_summary_t *summary;
	size_t summary_wsize;
	size_t nbits;
	size_t wsize;
	bset_elem_t bits[1];
};

#define	BITS_PER_BYTE		8
#define	BITS_PER_ELEM		(sizeof(bset_elem_t)*BITS_PER_BYTE)
#define	BITS_PER_SUMMARY	(sizeof(bset_summary_t)*BITS_PER_BYTE)

static size_t numbit[256] = {
	0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
	1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
};

/* allocate bitset and initialize to empty set */
struct _bset_t *
alloc_set(size)
size_t size;
{
	size_t wsize;
	struct _bset_t *s;

	wsize = (size+BITS_PER_ELEM-1)/BITS_PER_ELEM;
	if (!(s = malloc(offsetof(struct _bset_t, bits[wsize])))) {
		return NULL;
	}
	s->nbits = 0;
	s->wsize = wsize;
	bzero(s->bits, wsize * sizeof *s->bits);
	s->summary_wsize = (wsize+BITS_PER_SUMMARY-1)/BITS_PER_SUMMARY;
	if (!(s->summary = malloc(s->summary_wsize * sizeof *s->summary))) {
		free(s);
		return NULL;
	}
	bzero(s->summary, s->summary_wsize * sizeof *s->summary);
	return s;
}

void
free_set(s)
struct _bset_t *s;
{
	free(s->summary);
	free(s);
}

void
set_bit(s, n)
struct _bset_t *s;
size_t n;
{
	register bset_elem_t old, new;
	size_t i, j;

	i = n/BITS_PER_ELEM;
	j = n%BITS_PER_ELEM;
	old = s->bits[i];
	new = old | (bset_elem_t)1 << j;
	if (old!=new) {
		s->nbits++;
		s->bits[i] = new;
		if (!old) {
			size_t yd, yo;
			yd = i/BITS_PER_SUMMARY;
			yo = i%BITS_PER_SUMMARY;
			s->summary[yd] |= (bset_summary_t)1<<yo;
		}
	}
}

size_t
countbit(s)
struct _bset_t *s;
{
	return s->nbits;
}

size_t
set_and_count_bit(s1, s2)
struct _bset_t *s1, *s2;
{
	size_t i;
	size_t r;
	size_t summary_wsize;
	size_t yd;

	if (s1->wsize != s2->wsize) {
		return 0;
	}
	summary_wsize = s1->summary_wsize;
	for (r=0, yd=0, i=0; yd<summary_wsize; yd++, i+=BITS_PER_SUMMARY) {
		bset_summary_t sumbits;
		size_t j;
		sumbits = s1->summary[yd] & s2->summary[yd];
		for (j=i; sumbits; j++, sumbits>>=1) {
			/* j >= wsize to narukotoha arienai.
			   nazenaraba, souiu tokoro deha,
			   sumbits == 0 dakara. */
			if (sumbits & 1) {
				register bset_elem_t c;

				c = s1->bits[j] & s2->bits[j];
				for (; c; c>>=8) {
					r += numbit[c&0xff];
				}
			}
		}
	}
	return r;
}
