/*
 *	Copyright (C) 1996 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: cm.c,v 1.5 2007/03/24 02:52:58 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>
#include <geta/bset.h>
#include <geta/nlog.h>
#include <geta/dci.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/daui.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwsched.h>
#include <geta/dwam.h>

static int ecompar_static(void const *, void const *);

void *
dw_matrix(nrow, ncol, size)
ssize_t nrow, ncol;
size_t size;
{
	void **m, *base;
	ssize_t index_size, i;
	size_t pad = 0;

	index_size = nrow;

	if (sizeof *m < size) {
		size_t a, n;
		a = index_size*sizeof *m;
		n = (a + size - 1)/size;
		pad = n*size - a;
	}

	if (!(m = malloc(index_size*sizeof *m + pad + nrow*ncol*size))) {
		return NULL;
	}
	base = (char *)(m+index_size)+pad;
	for (i = 0; i<nrow; i++) {
		m[i] = (char *)base + size*i*ncol;
	}
	return m;
}

/*
 * nwords * nwords no matrix wo kaesu.
 * youso (i,j) ni ha,
 * w[i] ga arawareru buysyo syuugou wo A(w[i]),
 * w[j] ga arawareru buysyo syuugou wo A(w[j]) to sita tokini, 
 * A(w[i]) \intersection A(w[j]) \intersection d
 * no youso suu dearu.
 * toku ni taikaku youso (i,i) niha, 
 * A(w[i]) \intersection d
 * ga hairu.
 * 
 * kaette kita hairetu ha, free(3) de zentai wo free suru kotoga dekiru.
 */
ssize_t **
dw_cm(w, nwords, d, ndocs, dxr)
ssize_t nwords, ndocs;
struct syminfo const *w;
struct syminfo const *d;
DWFILE *dxr;
{
	struct _bset_t **words;
	ssize_t i, j;

	ssize_t **c_matrix;

	if (!(words = malloc(nwords * sizeof *words))) {
		return NULL;
	}

	if (!(c_matrix = dw_matrix(nwords, nwords, sizeof **c_matrix))) {
		return NULL;
	}

	for (j = 0; j<nwords; j++) {
		ssize_t e;
		idx_t word = w[j].id;
		struct xr_vec const *vec;
#if defined INCR_CM
		size_t k;
#endif
		struct syminfo const *dp;
		struct syminfo const *de;

		if ((e=dxr_get_vec(dxr, word, &vec))<0) {
			continue;
		}
		if (!(words[j] = alloc_set(ndocs))) {
			return NULL;
		}
#if defined INCR_CM
		for (k=0, dp=d, de=d+ndocs; k<vec->elem_num; k++) {
			idx_t id;
			id = vec->elems[k].id;

			while (dp<de && dp->id<id) dp++;

			if (dp->id==id) {
				set_bit(words[j], dp - d);
			}
		}
#else
		for (dp=d, de=d+ndocs; dp<de; dp++) {
			struct xr_elem key;
			key.id = dp->id;
			if (bsearch(&key, vec->elems, vec->elem_num, sizeof key, ecompar_static)) {
				set_bit(words[j], dp - d);
			}
		}
#endif
	}

	for (i = 0; i<nwords; i++) {
		c_matrix[i][i] = countbit(words[i]);
		for (j=i+1; j<nwords; j++) {
			c_matrix[i][j] = c_matrix[j][i] = set_and_count_bit(words[i], words[j]);
		}
	}

	for (j = 0; j<nwords; j++) {
		free_set(words[j]);
	}

	free(words);

	return c_matrix;
}

static int
ecompar_static(va, vb)
void const *va, *vb;
{
	struct xr_elem const *a, *b;
	a = va;
	b = vb;
	if (a->id < b->id) {
		return -1;
	}
	else if (a->id > b->id) {
		return 1;
	}
	return 0;
}
