#include <config.h>

#include <sys/types.h>
#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dci.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/daui.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwam.h>

#include "dcsP.h"
#include "priq.h"

static freq_t W;
static ssize_t dno;
static freq_t *wfq;
static ssize_t wfq_max;
static freq_t *c_wfq;
static freq_t *c_dno;
static float *c_scr;
static struct dcell **dmem;
static struct dcell *dcells;

static struct syminfo_list *symlp;

static float *pc;
static float **sim_matrix;
static float **triangle_matrix;
static float *unit_matrix;

static ssize_t *centroid;

static int bias;

#ifdef CS_FAST
static ssize_t *matrix_max_id;
static float *matrix_max_score;
#endif

static int
set_global_info(cs_type)
{
	ssize_t i, j;
  struct syminfo_cell *scp;
  struct syminfo *sp;

  W = 0;
  for (i = 0, scp = symlp->scp; i < symlp->elem_num; i++, scp++) {
    c_wfq[i] = scp->freq_sum;
    c_dno[i] = 1;
    c_scr[i] = (float)scp->weight;
    if (cs_type == CS_HBC) {
#ifdef CS_HBC_EVEN_PC
      pc[i] = 1.0;
#else
      pc[i] = (float)scp->weight;
#endif
    }
    dmem[i] = &(dcells[i]);
    dmem[i]->id = dmem[i]->cid = i;
    dmem[i]->height = 0;
    dmem[i]->num = 1;
    dmem[i]->next = dmem[i]->up = dmem[i]->rd = dmem[i]->ld = NULL;

    for (j = 0, sp = scp->sp; j < scp->elem_num; j++, sp++) {
      idx_t wid;
      freq_t wfreq;
      wid = sp->id;
      wfreq = sp->TF_d;
      if (!(wfq = check_n_realloc_frq_array(wfq, sp->id, &wfq_max, 0))) {
	perror("check_n_realloc_int_array");
	return -1;
      }
      if (cs_type == CS_HBC) {
	wfq[wid] += wfreq;
      } else {
	wfq[wid]++;  /* document frequency */
      }
      W += wfreq;
    }
  }
#ifdef CS_GLOBAL_WINFO
  W = wam_total_freq_sum(wam, dir);
#endif
  return 0;
}

static float
get_triangle_matrix(i, j)
ssize_t i, j;
{
  if (i < j) return triangle_matrix[i][j-i-1];
  else return triangle_matrix[j][i-j-1];
}

static void
put_triangle_matrix(i, j, v)
     float v;
{
  if (i < j) triangle_matrix[i][j-i-1] = v;
  else triangle_matrix[j][i-j-1] = v;
}

static float
flabs(v)
     float v;
{
  if (v > 0) return v; else return -v;
}

static int
cs_alloc()
{
  wfq = NULL;
  wfq_max = 0;
  
  if (!(c_wfq = calloc(dno, sizeof *c_wfq))) {
    perror("calloc");
    return -1;
  }
  if (!(c_dno = calloc(dno, sizeof *c_dno))) {
    perror("calloc");
    return -1;
  }
  if (!(c_scr = calloc(dno, sizeof *c_scr))) {
    perror("calloc");
    return -1;
  }
  if (!(pc = calloc(dno, sizeof *pc))) {
    perror("calloc");
    return -1;
  }
  if (!(dmem = calloc(dno, sizeof *dmem))) {
    perror("calloc");
    return -1;
  }
  if (!(dcells = calloc(dno*2, sizeof *dcells))) {
    perror("calloc");
    return -1;
  }
  return 0;
}

static void
clear1()
{
  zfree(wfq);
}

static void
clear2()
{
  zfree(c_wfq);
  zfree(c_dno);
  zfree(c_scr);
  zfree(pc);
}

static void
clear3()
{
  zfree(dcells);
  zfree(dmem);
}

static float
calc_sim_pair_hbc(x, y)
ssize_t x, y;
{
  ssize_t x_i, y_i, x_wno, y_wno;
  freq_t x_wFQ, y_wFQ;
  idx_t x_wid = 0, y_wid = 0;
  float x_wfq, y_wfq;
  freq_t wFQ;
  float out;

  x_wno = symlp->scp[x].elem_num;
  y_wno = symlp->scp[y].elem_num;
  x_wFQ = symlp->scp[x].freq_sum;
  y_wFQ = symlp->scp[y].freq_sum;

  for (x_i = 0, y_i = 0, out = 0.0; x_i < x_wno && y_i < y_wno; x_i++) {
    x_wid = symlp->scp[x].sp[x_i].id;
    for (; (y_wid = symlp->scp[y].sp[y_i].id) < x_wid && y_i < y_wno; y_i++) ;
    if (x_wid == y_wid) {
      x_wfq = symlp->scp[x].sp[x_i].TF_d;
      y_wfq = symlp->scp[y].sp[y_i].TF_d;
#ifdef CS_GLOBAL_WINFO
      wFQ = wam_freq_sum(wam, WAM_REVERT_DIRECTION(dir), x_wid);
#else
      wFQ = wfq[x_wid];
#endif
      out += (x_wfq / x_wFQ) * (y_wfq / y_wFQ) / ((float)wFQ / W);
    }
  }
  return out;
}

static float
calc_sim_pair_euclid(x, y)
ssize_t x, y;
{
  ssize_t x_i, y_i, x_wno, y_wno;
  idx_t x_wid = 0, y_wid = 0;
  float out;

  x_wno = symlp->scp[x].elem_num;
  y_wno = symlp->scp[y].elem_num;

  for (x_i = 0, y_i = 0, out = 0.0; x_i < x_wno || y_i < y_wno;) {
    if (x_i < x_wno) x_wid = symlp->scp[x].sp[x_i].id;
    if (y_i < y_wno) y_wid = symlp->scp[y].sp[y_i].id;
    if (y_i == y_wno || (x_i != x_wno && x_wid < y_wid)) {
      float lsc;
      lsc = symlp->scp[x].sp[x_i].weight;
      out += (lsc * lsc);
      x_i++;
    } else if (x_i == x_wno || (y_i != y_wno && y_wid < x_wid)) {
      float lsc;
      lsc = symlp->scp[y].sp[y_i].weight;
      out += (lsc * lsc);
      y_i++;
    } else if (x_wid == y_wid) {
      float lsc;
      lsc = symlp->scp[x].sp[x_i].weight - symlp->scp[y].sp[y_i].weight;
      out += (lsc * lsc);
      x_i++;
      y_i++;
    }
  }
  return out;
}

static float
calc_sim_pair_cosin(x, y)
ssize_t x, y;
{
  ssize_t x_i, y_i, x_wno, y_wno;
  idx_t x_wid, y_wid;
  float out;

  x_wno = symlp->scp[x].elem_num;
  y_wno = symlp->scp[y].elem_num;
  
  for (x_i = 0, y_i = 0, out = 0.0; x_i < x_wno && y_i < y_wno; x_i++) {
    x_wid = symlp->scp[x].sp[x_i].id;
    for (; (y_wid = symlp->scp[y].sp[y_i].id) < x_wid && y_i < y_wno; y_i++) ;
    if (x_wid == y_wid) {
      out += (symlp->scp[x].sp[x_i].weight * symlp->scp[y].sp[y_i].weight);
    }
  }
  return out;
}

static int init_sim_matrix(void);

static int
init_sim_matrix_old()
{
  ssize_t i, j;

  if (!(sim_matrix = calloc(dno, sizeof *sim_matrix))) {
    perror("calloc");
    return -1;
  }
  for (i = 0; i < dno; i++) {
    sim_matrix[i] = (float *)NULL;
    if (!(sim_matrix[i] = calloc(dno, sizeof *(sim_matrix[i])))) {
      perror("calloc");
      return -1;
    }
  }

  if (!(unit_matrix = calloc(dno, sizeof *unit_matrix))) {
    perror("calloc");
    return -1;
  }

  for (i = 0; i < dno; i++) {
    float wi, wq, w;
    for (j = i; j < dno; j++) {
#ifdef CS_VERBOSE
      fprintf(stderr, "%d %d   \r", i, j);
#endif
      sim_matrix[i][j] = sim_matrix[j][i] = calc_sim_pair_hbc(i, j);
    }
    wi = (float)symlp->scp[i].freq_sum;
    wq = (float)bias;
    w = wi + wq;
    wi /= w;
    wq /= w;
    unit_matrix[i] = log(wi * sim_matrix[i][i] + wq * pc[i]);
  }
  return 0;
}

static void
clear_sim_matrix()
{
  zfree(unit_matrix);
  if (sim_matrix) {
    ssize_t i;
    for (i = 0; i < dno; i++) zfree(sim_matrix[i]);
    zfree(sim_matrix);
  }
}

static int
init_triangle_matrix(cs_type)
{
  ssize_t i, j;

  if (!(centroid = calloc(dno, sizeof *centroid))) {
    perror("calloc");
    return -1;
  }
  for (i = 0; i < dno; i++) centroid[i] = i;

  if (!(triangle_matrix = calloc(dno-1, sizeof *triangle_matrix))) {
    perror("calloc");
    return -1;
  }
  for (i = 0; i < dno-1; i++) {
    triangle_matrix[i] = (float *)NULL;
    if (!(triangle_matrix[i] = calloc(dno-i-1, sizeof *(triangle_matrix[i])))) {
      perror("calloc");
      return -1;
    }
  }

#ifdef CS_FAST
  if (!(matrix_max_id = calloc(dno, sizeof *matrix_max_id))) {
    perror("calloc");
    return -1;
  }
  if (!(matrix_max_score = calloc(dno, sizeof *matrix_max_score))) {
    perror("calloc");
    return -1;
  }
#endif

  for (i = 0; i < dno - 1; i++) {
#ifdef CS_FAST
    ssize_t mid = -1;
    float mscore;
    mscore = - HUGE;
#endif
    for (j = i + 1; j < dno; j++) {
      float sc = 0.0;
      switch (cs_type) {
      case CS_HBC:
	{
	  float wi, wj, wq, w, pcij, sci, scj;
	  wi = (float)c_wfq[i];
	  wj = (float)c_wfq[j];
	  w = wi + wj;
	  wi /= w;
	  wj /= w;
	  pcij = wi * pc[i] + wj * pc[j];
	  sci  = wi * sim_matrix[i][i] + wj * sim_matrix[i][j];
	  scj  = wi * sim_matrix[j][i] + wj * sim_matrix[j][j];
	  wi = (float)symlp->scp[i].freq_sum;
	  wq = (float)bias;
	  w = wi + wq;
	  wi /= w;
	  wq /= w;
	  sc = log(wi * sci + wq * pcij);
	  wj = (float)symlp->scp[j].freq_sum;
	  wq = (float)bias;
	  w = wj + wq;
	  wj /= w;
	  wq /= w;
	  sc += log(wj * scj + wq * pcij);
	  sc -= (unit_matrix[i] + unit_matrix[j]);
	  break;
	}
      case CS_WARD:
	sc = - calc_sim_pair_euclid(i, j);
	if (sc == 0.0) sc = - HUGE;
	break;
      case CS_GAVE:
      case CS_SLINK:
      case CS_CLINK:
	sc = calc_sim_pair_cosin(i, j);
	break;
      }
      triangle_matrix[i][j-i-1] = sc;
#ifdef CS_FAST
      if (sc > mscore) {
	mid = j;
	mscore = sc;
      }
#endif
    }
#ifdef CS_FAST
    matrix_max_id[i] = mid;
    matrix_max_score[i] = mscore;
#endif
  }
  return 0;
}

static void
clear_triangle_matrix()
{
  zfree(centroid);
  if (triangle_matrix) {
    ssize_t i;
    for (i = 0; i < dno-1; i++) zfree(triangle_matrix[i]);
    zfree(triangle_matrix);
  }
#ifdef CS_FAST
  zfree(matrix_max_id);
  zfree(matrix_max_score);
#endif
}

static ssize_t
merge_pair_hbc(step, c1, c2)
size_t step;
ssize_t c1, c2;
{
  ssize_t cc1, cc2, ccr, ccl, ccu;
  struct dcell *dp;
  float w1, w2, wq, wi, w, pc12, u, sc, cent_sc;
  ssize_t c;
	ssize_t cent_id = -1;

  cc1 = c1;
  cc2 = c2;
  if (c_scr[c1]/c_dno[c1] < c_scr[c2]/c_dno[c2]) {
    cc1 = c2;
    cc2 = c1;
  }
  dmem[cc2]->height = step;

  w1 = (float)c_wfq[c1];
  w2 = (float)c_wfq[c2];
  w = w1 + w2;
  w1 /= w;
  w2 /= w;
  pc12 = w1 * pc[c1] + w2 * pc[c2];
  u = 0.0;
  cent_sc = - HUGE;
  for (dp = dmem[cc1]; dp->next != NULL; dp = dp->next) {
    sc = w1 * sim_matrix[dp->id][c1] + w2 * sim_matrix[dp->id][c2];
    sim_matrix[dp->id][c1] = sc;
    wi = (float)symlp->scp[dp->id].freq_sum;
    wq = (float)bias;
    w = wi + wq;
    wi /= w;
    wq /= w;
    sc = log(wi * sc + wq * pc12);
    if (sc > cent_sc) {
      cent_sc = sc;
      cent_id = dp->id;	/* XXX dp->id => ssize_t */
    }
    u += sc;
  }
  dp->next = dmem[cc2];
  for (; dp != NULL; dp = dp->next) {
    sc = w1 * sim_matrix[dp->id][c1] + w2 * sim_matrix[dp->id][c2];
    sim_matrix[dp->id][c1] = sc;
    wi = (float)symlp->scp[dp->id].freq_sum;
    wq = (float)bias;
    w = wi + wq;
    wi /= w;
    wq /= w;
    sc = log(wi * sc + wq * pc12);
    if (sc > cent_sc) {
      cent_sc = sc;
      cent_id = dp->id;
    }
    u += sc;
  }

  ccu = dno + step - 1;
  ccr = dmem[cc1]->cid;
  ccl = dmem[cc2]->cid;
  dcells[ccu].num = dcells[ccr].num + dcells[ccl].num;
  dcells[ccu].rd = &(dcells[ccr]);
  dcells[ccu].ld = &(dcells[ccl]);
  dcells[ccu].up = NULL;
  dcells[ccr].up = dcells[ccl].up = &(dcells[ccu]);
  dmem[cc1]->cid = ccu;
  dcells[ccu].cid = -1;
  
  dmem[c1] = dmem[cc1];
  dmem[c2] = NULL;
  unit_matrix[c1] = u;
  unit_matrix[c2] = 0.0;
  c_wfq[c1] += c_wfq[c2];
  c_wfq[c2] = 0;
  c_dno[c1] += c_dno[c2];
  c_dno[c2] = 0;
  c_scr[c1] += c_scr[c2];
  c_scr[c2] = 0.0;
  pc[c1] = w1 * pc[c1] + w2 * pc[c2];
  
  for (c = 0; c < dno; c++) {
    if (c != c1 && c != c2 && c_dno[c]) {
      for (dp = dmem[c]; dp != NULL; dp = dp->next) {
	sim_matrix[dp->id][c1] = w1 * sim_matrix[dp->id][c1] + w2 * sim_matrix[dp->id][c2];
      }
    }
  }
  
  return cent_id;
}

static ssize_t	/* XXX idx_t ni sitai tokoro... -1 wo dou atukau bekika. */
merge_pair_dmem(step, c1, c2)
size_t step;
ssize_t c1, c2;
{
  ssize_t cc1, cc2, ccr, ccl, ccu;
  struct dcell *dp;

  cc1 = c1;
  cc2 = c2;
  if (c_scr[c1]/c_dno[c1] < c_scr[c2]/c_dno[c2]) {
    cc1 = c2;
    cc2 = c1;
  }
  dmem[cc2]->height = step;
  for (dp = dmem[cc1]; dp->next != NULL; dp = dp->next) ;
  dp->next = dmem[cc2];

  ccu = dno + step - 1;
  ccr = dmem[cc1]->cid;
  ccl = dmem[cc2]->cid;
  dcells[ccu].num = dcells[ccr].num + dcells[ccl].num;
  dcells[ccu].rd = &(dcells[ccr]);
  dcells[ccu].ld = &(dcells[ccl]);
  dcells[ccu].up = NULL;
  dcells[ccr].up = dcells[ccl].up = &(dcells[ccu]);
  dmem[cc1]->cid = ccu;
  dcells[ccu].cid = -1;	/* XXX koko ga geyin de cid ha ssize_t */
  
  dmem[c1] = dmem[cc1];
  dmem[c2] = NULL;
  return -1;
}

static ssize_t
merge_pair(step, c1, c2, cs_type)
size_t step;
ssize_t c1, c2;
{
  if (cs_type == CS_HBC) return merge_pair_hbc(step, c1, c2);
  else return merge_pair_dmem(step, c1, c2);
}

static void
update_triangle_matrix_hbc(c)
ssize_t c;
{
  ssize_t i;

  for (i = 0; i < dno; i++) {
    if ((i != c) && c_dno[i]) {
      ssize_t c1, c2;
      float w1, w2, wq, wi, w, pc12, sc;
      struct dcell *dp;
      if (i < c) {
	c1 = i; c2 = c;
      } else {
	c1 = c; c2 = i;
      }
      w1 = (float)c_wfq[c1];
      w2 = (float)c_wfq[c2];
      w = w1 + w2;
      w1 /= w;
      w2 /= w;
      pc12 = w1 * pc[c1] + w2 * pc[c2];
      sc = 0.0;
      for (dp = dmem[c1]; dp != NULL; dp = dp->next) {
	float scid;
	wi = (float)symlp->scp[dp->id].freq_sum;
	wq = (float)bias;
	w = wi + wq;
	wi /= w;
	wq /= w;
	scid = w1 * sim_matrix[dp->id][c1] + w2 * sim_matrix[dp->id][c2];
	sc += log(wi * scid + wq * pc12);
      }
      for (dp = dmem[c2]; dp != NULL; dp = dp->next) {
	float scid;
	wi = (float)symlp->scp[dp->id].freq_sum;
	wq = (float)bias;
	w = wi + wq;
	wi /= w;
	wq /= w;
	scid = w1 * sim_matrix[dp->id][c1] + w2 * sim_matrix[dp->id][c2];
	sc += log(wi * scid + wq * pc12);
      }
      sc -= (unit_matrix[c1] + unit_matrix[c2]);
      triangle_matrix[c1][c2-c1-1] = sc;
    }
  }
}

static void
update_triangle_matrix_general(ci, cj, cs_type)
ssize_t ci, cj;
{
  ssize_t ck;

  for (ck = 0; ck < dno; ck++) {
    if ((ck != ci) && (ck != cj) && c_dno[ck]) {
      freq_t mi, mj, mk;
      float ai = 0.0, aj = 0.0, b = 0.0, c = 0.0, d_ik, d_jk, d_ij, d_ikjk;
      mi = c_dno[ci];
      mj = c_dno[cj];
      mk = c_dno[ck];
      switch (cs_type) {
      case CS_WARD:
	ai = (float)(mi + mk) / (mi + mj + mk);
	aj = (float)(mj + mk) / (mi + mj + mk);
	b = (float)(- mk) / (mi + mj + mk);
	c = 0.0;
	break;
      case CS_GAVE:
	ai = (float)(mi) / (mi + mj);
	aj = (float)(mj) / (mi + mj);
	b = 0.0;
	c = 0.0;
	break;
      case CS_SLINK:
	ai = aj = 0.5;
	b = 0.0;
	c = - 0.5;
	break;
      case CS_CLINK:
	ai = aj = 0.5;
	b = 0.0;
	c = 0.5;
	break;
      }
      d_ij = get_triangle_matrix(ci, cj);
      d_ik = get_triangle_matrix(ci, ck);
      d_jk = get_triangle_matrix(cj, ck);
      d_ikjk = flabs(d_ik - d_jk);
      put_triangle_matrix(ci, ck, ai * d_ik + aj * d_jk + b * d_ij + c * d_ikjk);
    }
  }
  c_wfq[ci] += c_wfq[cj];
  c_wfq[cj] = 0;
  c_dno[ci] += c_dno[cj];
  c_dno[cj] = 0;
  c_scr[ci] += c_scr[cj];
  c_scr[cj] = 0.0;
}

static void
update_triangle_matrix(c1, c2, cs_type)
ssize_t c1, c2;
{
  if (cs_type == CS_HBC) update_triangle_matrix_hbc(c1);
  else update_triangle_matrix_general(c1, c2, cs_type);
}

#ifdef CS_FAST
static void
update_matrix_max_info(c1, c2)
ssize_t c1, c2;
{
  ssize_t i, j;
  ssize_t max_id = -1;
  float score, max_score;

  for (i = 0; i < dno - 1; i++) {
    if (c_dno[i]) {
      if (i < c1) {
	score = triangle_matrix[i][c1-i-1];
      }
      if (i < c1 && score > matrix_max_score[i]) {
	matrix_max_id[i] = c1;
	matrix_max_score[i] = score;
      } else if (i == c1 || matrix_max_id[i] == c1 || matrix_max_id[i] == c2) {
	max_score = - HUGE;
	for (j = i + 1; j < dno; j++) {
	  if (c_dno[j]) {
	    if ((score = triangle_matrix[i][j-i-1]) > max_score) {
	      max_id = j;
	      max_score = score;
	    }
	  }
	}
	matrix_max_id[i] = max_id;
	matrix_max_score[i] = max_score;
      }
    }
  }
}
#endif

static struct dcell *
merge(cs_type)
{
  size_t step;
  float max_score, score;
  ssize_t i;
	ssize_t cent;
	ssize_t c1 = -1, c2 = -1;
  struct dcell *rdp = NULL;

  for (step = 1; step < dno ; step++) {
#ifdef CS_FAST
    max_score = - HUGE;
    for (i = 0; i < dno - 1; i++) {
      if (c_dno[i]) {
	if ((score = matrix_max_score[i]) > max_score) {
	  c1 = i;
	  c2 = matrix_max_id[i];
	  max_score = score;
	}
      }
    }
#else
    max_score = -HUGE;
    for (i = 0; i < dno - 1; i++) {
      if (c_dno[i]) {
        ssize_t j;
	for (j = i + 1; j < dno; j++) {
	  if (c_dno[j]) {
	    if ((score = triangle_matrix[i][j-i-1]) > max_score) {
	      c1 = i;
	      c2 = j;
	      max_score = score;
	    }
	  }
	}
      }
    }
#endif
    cent = merge_pair(step, c1, c2, cs_type);
    if (cs_type == CS_HBC) {
      centroid[c1] = cent;
      centroid[c2] = -1;
    }
#ifdef CS_VERBOSE
    fprintf(stdout, "%d %d %d %d %e\n", step, c1, c2, cent, max_score);
#endif
    update_triangle_matrix(c1, c2, cs_type);
#ifdef CS_FAST
    update_matrix_max_info(c1, c2);
#endif
    rdp = dmem[c1];
  }
  return rdp;
}

static int
dendro_sort(s, root)
     struct syminfo_list *s;
     struct dcell *root;
{
  struct syminfo_cell *qtmp, *q;
  struct dcell *dp;
  ssize_t i;

  if (!(qtmp = (struct syminfo_cell *)bdup(s->scp, s->elem_num * sizeof(*qtmp)))) {
    perror("bdup");
    return -1;
  }

  for (dp = root, q = s->scp, i = 0; dp != NULL; dp = dp->next, i++) {
    ssize_t dpid;
    dpid = dp->id;
    bcopy(&(qtmp[dpid]), &(q[i]), sizeof qtmp[dpid]);
    q[i].height = dp->height;
    dcells[dpid].cid = i;
  }

  zfree(qtmp);
  return 0;
}

void
sort_cw_syminfo_list(q, qno, cno, type)
     struct syminfo *q;
size_t qno, cno;
{
	size_t i, len, thresh, height;
  struct syminfo *begin;

  begin = q;
  len = 0;
  height = 0;
  if ((thresh = qno - cno) < 0) return;
  for (i = 0; i < qno; i++) {
    if (thresh < q[i].u.s.i) {
      if (type == CS_SORT_WEIGHT) {
	qsort(begin, len, sizeof *begin, sym_compar);
      }
      begin = &(q[i]);
      len = 0;
      height++;
    }
    q[i].u.s.i = height;
    len++;
  }
  if (type == CS_SORT_WEIGHT) {
    qsort(begin, len, sizeof *begin, sym_compar);
  }
}

static ssize_t
get_leaf_id(dp)
     struct dcell *dp;
{
  struct dcell *p;
  for (p = dp; p->rd != NULL; p = p->rd) ;
  return p->cid;
}

static struct cs_elem *
mk_cslist(q, cno, thresh, cswmax, wam, dir, w_type)
     struct syminfo *q;
     size_t *cno;
     WAM *wam;
size_t thresh, cswmax;
{
  struct cs_elem *cep;
  struct dcell **dcpp;
  ssize_t dcppn;
  size_t step, i;

  cep = NULL;

  step = dno * 2 - 2;
  if (dcells[step].num < thresh) {
    return NULL;
  }
  if (!(dcpp = calloc((*cno)+1, sizeof *dcpp))) {
    perror("calloc");
    goto mk_cslist_error;
  }
  dcppn = 0;
  dcpp[dcppn] = &(dcells[step]);
  dcells[step].cid = dcppn;
  dcppn++;
  for (; dcppn < *cno && step >= dno; step--) {
    struct dcell *e = &dcells[step];
    if (e->num >= thresh) {
      int replace;
      replace = 1;
      if (e->rd->num >= thresh) {  /* right down */
	dcpp[e->cid] = e->rd;
	if (e->rd->cid == -1) {
	  e->rd->cid = e->cid;
        }
	replace = 0;
      }
      if (e->ld->num >= thresh) {  /* left down */
	if (replace) {
	  dcpp[e->cid] = e->ld;
	  if (e->ld->cid == -1) {
	    e->ld->cid = e->cid;
          }
	} else {
	  dcpp[dcppn] = e->ld;
	  if (e->ld->cid == -1) {
	    e->ld->cid = dcppn;
          }
	  dcppn++;
	}
      }
    }
#if 0
    if (e->num >= thresh && e->cid >= 0) {
      if (dcell[step].rd->num >= thresh && dcell[step].ld->num >= thresh) {
	dcpp[e->cid] = e->rd;
	e->rd->cid = e->cid;
	dcpp[dcppn] = e->ld;
	e->ld->cid = dcppn;
	dcppn++;
      }
    }
#endif
  }
  *cno = dcppn;
  
  if (!(cep = calloc((*cno)+1, sizeof *cep))) {
    perror("calloc");
    goto mk_cslist_error;
  }
  for (i = 0; i < (*cno)+1; i++) {
    cep[i].csw = cep[i].csa = NULL;
  }

  for (i = 0; i < *cno; i++) {
    struct syminfo *a, *r;
    ssize_t alen, rlen;
    
    alen = dcpp[i]->num;
    rlen = cswmax;
    
    if (!(a = (struct syminfo *)bdup(&(q[get_leaf_id(dcpp[i])]), alen * sizeof(*a)))) {
      perror("bdup");
      goto mk_cslist_error;
    }
    
    if (!(r = wsh(a, alen, wam, dir, w_type, &rlen, NULL))) {
      perror("wsh");
      zfree(a);
      goto mk_cslist_error;
    }
    
    qsort(r, rlen, sizeof *r, sym_compar);
    
    cep[i].csa = a;
    cep[i].csalen = alen;
    cep[i].csw = r;
    cep[i].cswlen = rlen;
  }

  zfree(dcpp);

  return cep;
mk_cslist_error:
  if (cep) {
    for (i = 0; cep[i].csa != NULL; i++) zfree(cep[i].csa);
    for (i = 0; cep[i].csw != NULL; i++) zfree(cep[i].csw);
    zfree(cep);
  }
  zfree(dcpp);
  
  return NULL;
}

static void
cs_init()
{
  W = 0;
  dno = 0;
  bias = 0;
  wfq = NULL;
  wfq_max = 0;
  c_wfq = NULL;
  c_dno = NULL;
  c_scr = NULL;
  pc = NULL;
  dmem = NULL;
  dcells = NULL;
  centroid = NULL;
  unit_matrix = NULL;
  sim_matrix = NULL;
  triangle_matrix = NULL;
  symlp = NULL;
#ifdef CS_FAST
  matrix_max_id = NULL;
  matrix_max_score = NULL;
#endif
}

static void
cs_error()
{
  W = 0;
  dno = 0;
  bias = 0;

  zfree(wfq);
  wfq_max = 0;

  zfree(c_wfq);
  zfree(c_dno);
  zfree(c_scr);
  zfree(pc);
  zfree(dmem);
  zfree(dcells);
  
  free_syminfo_list(symlp);  
  clear_sim_matrix();
  clear_triangle_matrix();
}

void
free_cs_elems(cep)
     struct cs_elem *cep;
{
  struct cs_elem *p;

  if (cep) {
    for (p = cep; p->csw != NULL; p++) {
      zfree(p->csw);
      zfree(p->csa);
    }
    zfree(cep);
  }
}

int cs_use_new_ism = 1;

static int
cs_core(slp, qbias, cs_type)
     struct syminfo_list *slp;
{
  struct dcell *root;

  cs_init();

  symlp = slp;
  dno = slp->elem_num;
  bias = qbias;

  if (cs_alloc() != 0) {
    cs_error();
    return -1;
  }
  
  set_global_info(cs_type);
  
  if (cs_type == CS_HBC) {
    if ((*(cs_use_new_ism?init_sim_matrix:init_sim_matrix_old))() != 0) {
      cs_error();
      return -1;
    }
  }
  
  if (init_triangle_matrix(cs_type) != 0) {
    cs_error();
    return -1;
  }

  root = merge(cs_type);

  clear_triangle_matrix();
  if (cs_type == CS_HBC) {
    clear_sim_matrix();
  }

  if (dendro_sort(symlp, root) != 0) {
    cs_error();
    return -1;
  }

  clear1();
  clear2();
  
  return 0;
}

int
cs(slp, qbias, cs_type)
     struct syminfo_list *slp;
{
  int r;

  r = cs_core(slp, qbias, cs_type);
  clear3();

  return r;
}

struct cs_elem *
csb(q, qno, qbias, wam, dir, cs_type, weight_type, elemn, cno, thresh, cswmax, r)
     struct syminfo *q, **r;
     WAM *wam;
     size_t *cno;
size_t thresh, cswmax;
ssize_t qno, elemn;
{
  size_t i;
  struct syminfo *qtmp;
  struct syminfo_list sl;
  struct cs_elem *rc;

  if (alloc_syminfo_list(&sl, qno) != 0) {
    cs_error();
    return NULL;
  }
  if (init_syminfo_list(q, qno, wam, dir, &sl, cs_type, weight_type, elemn) != 0) {
    cs_error();
    return NULL;
  };

  if (cs_core(&sl, qbias, cs_type) != 0) {
    cs_error();
    return NULL;
  }

  if (!(qtmp = calloc(qno, sizeof *qtmp))) {
    perror("calloc");
    cs_error();
    return NULL;
  }
  for (i = 0; i < sl.elem_num; i++) {
    bcopy(sl.scp[i].orig, &(qtmp[i]), sizeof *(sl.scp[i].orig));
    qtmp[i].u.s.i = sl.scp[i].height;
  }

  free_syminfo_list(&sl);

  rc = mk_cslist(qtmp, cno, thresh, cswmax, wam, dir, weight_type);

  if (r) {
    *r = qtmp;
  } else {
    free(qtmp);
  }
  
  clear3();

  return rc;
}

struct container {
	ssize_t index, wno, ptr;
	freq_t wFQ;
	struct syminfo *sp;
	idx_t wid;
};

static void eject(struct container *, size_t);
static int compar(const void *, const void *);

static int
init_sim_matrix()
{
#define	G	* 1024 * 1024 * 1024
	ssize_t i, j;
	size_t nh;
	struct container *c, *h;
	struct priq *q;
	size_t total, tally = 0;

	if (!(sim_matrix = malloc(dno * sizeof *sim_matrix))) {
		perror("calloc");
		return -1;
	}
	total = dno * (dno * sizeof *sim_matrix[0]);
	if (total > 1 G) {
		fprintf(stderr, "init_sim_matrix: allocating %dG bytes", (int)(total / (1 G)));
	}
	for (i = 0; i < dno; i++) {
		if (total > 1 G) {
			for (tally += dno * sizeof *sim_matrix[i]; tally > 1 G; tally -= 1 G) {
				fprintf(stderr, ".");
			}
		}
		if (!(sim_matrix[i] = malloc(dno * sizeof *sim_matrix[i]))) {
			perror("calloc");
			return -1;
		}
		for (j = 0; j < dno; j++) {
			sim_matrix[i][j] = 0;
		}
	}
	if (total > 1 G) {
		fprintf(stderr, "\n");
	}

	if (!(unit_matrix = calloc(dno, sizeof *unit_matrix))) {
		perror("calloc");
		return -1;
	}

	if (!(c = calloc(dno, sizeof *c))) {
		perror("calloc");
		return -1;
	}

	if (!(h = calloc(dno, sizeof *h))) {
		perror("calloc");
		return -1;
	}

	if (!(q = priq_creat(sizeof (struct container *), compar))) {
		perror("priq_creat");
		return -1;
	}

	for (i = 0; i < dno; i++) {
		struct container *t = &c[i];
		t->index = i;
		t->wno = symlp->scp[i].elem_num;
		t->wFQ = symlp->scp[i].freq_sum;
		t->ptr = 0;
		t->sp = &symlp->scp[i].sp[0];
		t->wid = t->sp->id;
		if (priq_enq(q, &t) == -1) {
			perror("priq_enq");
			return -1;
		}
	}

	nh = 0;
	while (!priq_emptyp(q)) {
		struct container *t = *(struct container **)priq_top(q);

		if (nh != 0 && t->wid != h[0].wid) {
			eject(h, nh);
			nh = 0;
		}

		h[nh++] = *t;

		if (++t->ptr < t->wno) {
			t->sp++;
			t->wid = t->sp->id;
			priq_rplatop(q, &t);
		}
		else {
			priq_deq(q);
		}
	}

	if (nh != 0) {
		eject(h, nh);
	}

	priq_free(q);
	free(c);
	free(h);

	for (i = 0; i < dno; i++) {
		float wi, wq, w;
		for (j = i + 1; j < dno; j++) {
#ifdef CS_VERBOSE
			fprintf(stderr, "%d %d   \r", i, j);
#endif
			sim_matrix[i][j] = sim_matrix[j][i];
		}
		wi = symlp->scp[i].freq_sum;
		wq = bias;
		w = wi + wq;
		wi /= w;
		wq /= w;
		unit_matrix[i] = log(wi * sim_matrix[i][i] + wq * pc[i]);
	}
	return 0;
}

static void
eject(h, nh)
struct container *h;
size_t nh;
{
	ssize_t i, j;
	idx_t wid = h[0].wid;
	for (i = 0; i < nh; i++) {
		for (j = 0; j <= i; j++) {
			struct container *hx, *hy;
			ssize_t x, y;
			float wFQ;
			float x_wfq, y_wfq;
			float x_wFQ, y_wFQ;
			hx = &h[i];
			hy = &h[j];
			x = hx->index;
			y = hy->index;
			x_wfq = hx->sp->TF_d;
			y_wfq = hy->sp->TF_d;
			x_wFQ = hx->wFQ;
			y_wFQ = hy->wFQ;
#ifdef CS_GLOBAL_WINFO
			wFQ = wam_freq_sum(wam, WAM_REVERT_DIRECTION(dir), wid);
#else
			wFQ = wfq[wid];
#endif
			sim_matrix[x][y] += (x_wfq / x_wFQ) * (y_wfq / y_wFQ) / (wFQ / W);
		}
	}
}

static int
compar(va, vb)
const void *va;
const void *vb;
{
	const struct container **a;
	const struct container **b;

	a = (const struct container **)va;
	b = (const struct container **)vb;

	if ((*a)->wid == (*b)->wid) {
		if ((*a)->index == (*b)->index) {
			return 0;
		}
		else if ((*a)->index < (*b)->index) {
			return -1;
		}
		else {
			return 1;
		}
	}
	else if ((*a)->wid < (*b)->wid) {
		return -1;
	}
	else {
		return 1;
	}
}
