#include <config.h>

#include <sys/types.h>
#include <sys/time.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dci.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/daui.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwam.h>

#include "dcsP.h"

static int
init_syminfo_list_hbc(q, qno, wam, dir, sl, cs_type, w_type)
     struct syminfo *q;
ssize_t qno;
     WAM *wam;
     struct syminfo_list *sl;
{
  size_t i, j;
  ssize_t en;
  struct xr_vec const *wbp;
  struct syminfo_cell *scellp;
  struct syminfo *symp;

  sl->elem_num = qno;
  for (i = 0, scellp = sl->scp; i < qno; i++, scellp++) {
    if (wam_get_vec(wam, dir, q[i].id, &wbp)==-1) {
      perror("wam_get_vec");
      return -1;
    }
    scellp->elem_num = en = wbp->elem_num;
    scellp->freq_sum = wbp->freq_sum;
    scellp->weight = q[i].weight;
    scellp->orig = &(q[i]);
    scellp->sp = NULL;

    if (!(scellp->sp = calloc(en, sizeof *(scellp->sp)))) {
      perror("calloc");
      return -1;
    }

    for (j = 0, symp = scellp->sp; j < en; j++, symp++) {
      symp->id = wbp->elems[j].id;
      symp->TF_d = wbp->elems[j].freq;
      symp->weight = symp->TF_d;
    }
  }
  return 0;
}

extern int cs_use_new_ism;

static int
init_syminfo_list_wsh(q, qno, wam, dir, sl, cs_type, w_type, elemn)
     struct syminfo *q;
ssize_t qno;
     WAM *wam;
     struct syminfo_list *sl;
ssize_t elemn;
{
  size_t i, j;
  ssize_t en;
  freq_t fsum;
  double wsum;
  struct syminfo_cell *scellp;
  struct syminfo *symp;
  
  sl->elem_num = qno;
  if (elemn <= 0) {
    elemn = wam_size(wam, WAM_REVERT_DIRECTION(dir));
  }
  for (i = 0, scellp = sl->scp; i < qno; i++, scellp++) {
    en = elemn;
    if (!(scellp->sp = wsh(&(q[i]), 1, wam, dir, w_type, &en, NULL))) {
      perror("wsh");
      return -1;
    }
    if (cs_use_new_ism) qsort(scellp->sp, en, sizeof *scellp->sp, sym_icompar);
    scellp->elem_num = en;
    scellp->weight = q[i].weight;
    scellp->orig = &(q[i]);
    
    fsum = 0;
    wsum = 0.0;
    for (j = 0, symp = scellp->sp; j < en; j++, symp++) {
      fsum += symp->TF_d;
      wsum += symp->weight;
    }
    scellp->freq_sum = fsum;
    for (j = 0, symp = scellp->sp; j < en; j++, symp++) {
      symp->weight /= wsum;
    }
  }
  return 0;
}
    
int
init_syminfo_list(q, qno, wam, dir, sl, cs_type, w_type, wno)
     struct syminfo *q;
ssize_t qno;
     WAM *wam;
     struct syminfo_list *sl;
ssize_t wno;
{
  return init_syminfo_list_wsh(q, qno, wam, dir, sl, cs_type, w_type, wno);
}

