/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: csche.c,v 1.10 2007/03/29 02:10:01 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <assert.h>
#include <netdb.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwc.h>

#include "dxrP.h"
#include "iosched.h"
#include "csche.h"
#include "sdup.h"
#include "nht.h"

static ssize_t selp_eqv_area(struct sched *, idx_t);
static ssize_t selp_ex_7(struct sched *, idx_t);
static ssize_t selp_ex_14(struct sched *, idx_t);

/*
 * iosched->custom ha, iosched->{scc,tg,mkrhost,local} wo henkou site yoi.
 * sorerani taisuru henkou no kekka ha subete yuukou to naru.
 */
static int get_hps(char const **, char const **, int, ssize_t, char const *);

/* get host-prot spec from `hosts' file. */
/* kono kansuu ha, yobareru kaisuu ga sukunainode, 
  genzai ha, maikai file wo scan suru youni zissou sarete iru.
  sonouti, tyanto tukuri naosu beki de arou. 02/14/02.

  `articles' server to, 'words' server wo betu no haiti ni sitai baai,
   3rd hikisuu no type de kotonaru file wo yomu youuni sureba yoi.
 */
static int
get_hps(hostname, protspec, type, fragment, param)
char const **hostname, **protspec;
char const *param;
ssize_t fragment;
{
	ssize_t i;
	char const *path;
	char buf[512], buf2[sizeof buf];
	FILE *f;
	char *p, *q;
	ssize_t fidx = fragment+1;

	path = param;

	if (!(f = fopen(path, "r"))) {
		perror(path);
		return -1;
	}
	for (i=0; i<=fidx; i++) {
		if (!fgets(buf, sizeof buf, f)) {
			break;
		}
		if (i>0 && *buf=='+' && (*(buf+1)=='\n' || *(buf+1)=='\r')) {
			bcopy(buf2, buf, sizeof buf);
			i = fidx+1;
			break;
		}
		bcopy(buf, buf2, sizeof buf);
	}
	fclose(f);
	if (i!=fidx+1) {
		fprintf(stderr, "too few servers defined in %s\n", path);
		return -1;
	}
	if (p = index(buf, '\n')) {
		*p = '\0';
	}
	if (!(p = index(buf, ':'))) {
		fprintf(stderr, "parse error: buf=[%s]\n", buf);
		return -1;
	}
	*p++ = '\0';
	if (q = index(p, ':')) {
		*q = '\0';
	}
	*hostname = strdup(buf);
	*protspec = strdup(p);
	return 0;
}

int
mkrhost_hosts(s, iocb, type, fragment)
struct iosched *s;
struct sched_iocb *iocb;
ssize_t fragment;
{
	char const *hostname, *protspec;

	if (get_hps(&hostname, &protspec, type, fragment, s->param)==-1) {
		return -1;
	}		/* hostname and protspec are immutable. */

	iocb->R.host = hostname;
	iocb->R.serv = protspec;
	return 0;
}

/* mkrhost_localhost ha, sched.c ni onazi mono ga aru. riyuu ha, 
 * sotirano comment wo sansyou saretai */
int
mkrhost_localhost(s, iocb, type, fragment)
struct iosched *s;
struct sched_iocb *iocb;
ssize_t fragment;
{
	iocb->R.host = "localhost";
#if defined DWD_PORT
	iocb->R.serv = DWD_PORT;
#else
	iocb->R.serv = DWD_SERV;
#endif
	return 0;
}

/* localhost:6376- */
int
mkrhost_smp_local(s, iocb, type, fragment)
struct iosched *s;
struct sched_iocb *iocb;
ssize_t fragment;
{
	char buf[32];
	iocb->R.host = "localhost";
	snprintf(buf, sizeof buf, "%s", sizetostr(6376+fragment+1));
	iocb->R.serv = strdup(buf);
	return 0;
}

int
custom_local(s, arg)
struct iosched *s;
char *arg;
{
	printf("***** scheduling: custom_local\n");
	s->local = 1;
	return 0;
}

int
custom_word_eqv_area(s, arg)
struct iosched *s;
char *arg;
{
	size_t p;
	struct sched *sc;
	ssize_t df, pdf, DF;
	idx_t i;
	idx_t *pp;
	size_t nc;
	struct ht *c;

	printf("***** scheduling: custom_word_eqv_area\n");
	sc = &s->scc[1];
	c = s->tg[1].h;
	nc = s->tg[1].n;
	sc->selpp = selp_eqv_area;
	if (!(pp = calloc(sc->onp, sizeof *pp))) {
		return -1;
	}
	sc->pp = pp;
	DF = 0;
	for (i = 0; i<nc; i++) {
		struct hte *e = hta(c, i);
		DF += e->i.df;
	}
	printf("df=%s, nc=%s, onp=%s\n", ssizetostr(DF), sizetostr(nc), sizetostr(sc->onp));
	df = pdf = 0;
	for (p=0, i = 0; p<sc->onp; p++) {
		ssize_t th;
		if (p<sc->onp-1) {
			th = (DF * (p+1)) / sc->onp;
		}
		else {
			th = DF;
		}
		for (; i<nc && df<=th; i++) {
			struct hte *e = hta(c, i);
			df += e->i.df;
		}
		pp[p] = i;
		printf("%s/%s ", idxtostr(i), sizetostr(df - pdf));
		pdf = df;
	}
	printf("\n");
	assert(i==nc);
	assert(DF==df);
	return 0;
}

static ssize_t
selp_eqv_area(sc, id)
struct sched *sc;
idx_t id;
{
	if (sc->pp) {
		size_t p;
		for (p=0; p<sc->onp; p++) {
			if (id <= sc->pp[p]) {
				break;
			}
		}
		if (p==sc->onp) {
			fprintf(stderr, "selp failed\n");
			return -1;
		}
		return sc->onp;
	}
	return -1;
}

int
custom_ex_7(s, arg)
struct iosched *s;
char *arg;
{
	printf("***** scheduling: custom_ex_7\n");
	s->scc[0].selpp = selp_ex_7;
	s->scc[1].selpp = selp_ex_7;
	if (s->scc[0].np>7) {
		s->scc[0].np = 7;
	}
	if (s->scc[1].np>7) {
		s->scc[1].np = 7;
	}
	return 0;
}

static ssize_t
selp_ex_7(sc, id)
struct sched *sc;
idx_t id;
{
	size_t i;
	if ((i = id % sc->onp)>=7) {
		return 0;
	}
	return i;
}

int
custom_ex_14(s, arg)
struct iosched *s;
char *arg;
{
	printf("***** scheduling: custom_ex_14\n");
	s->scc[0].selpp = selp_ex_14;
	s->scc[1].selpp = selp_ex_14;
	if (s->scc[0].np>14) {
		s->scc[0].np = 14;
	}
	if (s->scc[1].np>14) {
		s->scc[1].np = 14;
	}
	return 0;
}

static ssize_t
selp_ex_14(sc, id)
struct sched *sc;
idx_t id;
{
	size_t i;
	if ((i = id % sc->onp)>=14) {
		return 0;
	}
	return i;
}
