/*	$Id: dlmgr.c,v 1.4 2007/03/29 08:46:33 nis Exp $	*/

/*-
 * Copyright (c) 2007 Shingo Nishioka.
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Id: dlmgr.c,v 1.4 2007/03/29 08:46:33 nis Exp $";
#endif

#include <config.h>
#include <sys/param.h>

#include <assert.h>
#include <fcntl.h>
#include <libgen.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>

#if ! defined O_EXLOCK
#define O_EXLOCK 0
#endif

static int read_dl(void);
static int rdd(FILE *);
static int wrd(FILE *);

struct dle {
	char *host, *serv;
};

static int has_downlist = 0;
static struct dle *downlist = NULL;
static size_t downlist_size = 0;

extern char dl_path[];

int
dlm_match(host, serv)
char const *host, *serv;
{
	size_t i;
	if (!has_downlist) {
		(void)read_dl();
	}
	for (i = 0; i < downlist_size; i++) {
		struct dle *e = &downlist[i];
		if (!e->host || !e->serv) {
			continue;
		}
		if (!strcmp(e->host, host) && !strcmp(e->serv, serv)) {
			return 1;
		}
	}
	return 0;
}

int
dlm_add(host, serv)
char const *host, *serv;
{
	int d;
	FILE *f;
	has_downlist = 1;
	if ((d = open(dl_path, O_RDWR | O_CREAT | O_EXLOCK, 0666)) == -1) {
		perror(dl_path);
		return -1;
	}
	if (!(f = fdopen(d, "r+"))) {
		perror("fdopen");
		close(d);
		return -1;
	}
	if (rdd(f) == -1) {
		fclose(f);
		return -1;
	}
	if (dlm_match(host, serv)) {
		fprintf(stderr, "%s:%s already is in %s\n", host, serv, dl_path);
	}
	else {
		rewind(f);
		fprintf(f, "%s %s\n", host, serv);
		if (wrd(f) == -1) {
			fclose(f);
			return -1;
		}
		rewind(f);
		if (rdd(f) == -1) {
			fclose(f);
			return -1;
		}
	}
	fclose(f);
	return 0;
}

static int
read_dl()
{
	int d;
	FILE *f;
	has_downlist = 1;
	if ((d = open(dl_path, O_RDONLY | O_EXLOCK, 0666)) == -1) {
		return -1;
	}
	if (!(f = fdopen(d, "r"))) {
		perror("fdopen");
		close(d);
		return -1;
	}
	if (rdd(f) == -1) {
		fclose(f);
		return -1;
	}
	fclose(f);
	return 0;
}

static int
rdd(f)
FILE *f;
{
	size_t i, n;
	char buf[MAXPATHLEN * 2];
	for (i = 0; i < downlist_size; i++) {
		struct dle *e = &downlist[i];
		free(e->host);
		free(e->serv);
	}
	free(downlist);
	for (n = 0; fgets(buf, sizeof buf, f); n++) ;
	rewind(f);
	if (!(downlist = calloc(n, sizeof *downlist))) {
		return -1;
	}
	for (i = 0; fgets(buf, sizeof buf, f); ) {
		char *p;
		struct dle *e;
		if (p = index(buf, '\n')) {
			*p = '\0';
		}
		else {
			continue;
		}
		if (p = index(buf, ' ')) {
			*p++ = '\0';
		}
		else {
			continue;
		}
		assert(i < n);
		e =  &downlist[i++];
		e->host = strdup(buf);
		e->serv = strdup(p);
	}
	assert(i <= n);
	downlist_size = i;
	return 0;
}

static int
wrd(stream)
FILE *stream;
{
	size_t i;
	for (i = 0; i < downlist_size; i++) {
		struct dle *e = &downlist[i];
		fprintf(stream, "%s %s\n", e->host, e->serv);
	}
	return 0;
}

#if 0
int
main(argc, argv)
char *argv[];
{
	if (argc != 2) {
		fprintf(stderr, "usage: %s r|w\n", basename(argv[0]));
		return 1;
	}
	if (*argv[1] == 'w') {
		printf("dlm_add\n");
		dlm_add("host2", "serv1");
	}
	else {
		printf("read_dl\n");
		read_dl();
	}
	printf("close\n");
	return 0;
}
#endif
