/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dumpdcx.c,v 1.5 2007/03/24 02:52:58 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <assert.h>
#if defined HAVE_LIBGEN_H
#include <libgen.h>
#endif
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwc.h>
#include <geta/dwam.h>

#include "tools.h"

/* tango no namae hyou wo, TF, DF tuki de insastu suru */

static char *progname = NULL;

void dump_dcx(FILE *, DWFILE *, DWFILE *);
static void usage(void);

static void
usage()
{
	fprintf(stderr, "usage: %s handle\n", progname);
	exit(1);
}

int
main(argc, argv)
char *argv[];
{
	char *dirname, *handle;
	DWFILE *f, *g;
	char path[MAXPATHLEN];
	int type;
	ssize_t frag = 0;

	progname = basename(argv[0]);

	if (!(dirname = getenv(PROOTENV))) {
		fprintf(stderr, "environment variable " PROOTENV " is not set");
		return 1;
	}

	if (argc!=2) {
		usage();
	}

	handle = argv[1];

	type = DCW_COL;
	frag = 0;
	dw_mkpathname(path, sizeof path, dirname, handle, type, frag, 0);
	fprintf(stderr, "path=[%s]\n", path);

	if (!(g = dcw_open(path, handle))) {
		nperror(path);
		return 1;
	}

	type = DXR_COL;
	frag = 0;
	dw_mkpathname(path, sizeof path, dirname, handle, type, frag, 0);
	fprintf(stderr, "path=[%s]\n", path);
	if (!(f = dxr_open(path, handle))) {
		nperror(path);
		return 1;
	}
	dump_dcx(stdout, f, g);
	dcw_close(g);
	dxr_close(f);
	return 0;
}

void
dump_dcx(stream, f, g)
FILE *stream;
DWFILE *f, *g;
{
	idx_t id;
	ssize_t ne;

	ne = dxr_size(f);
	for (id = 1; id<=ne; id++) {
		char const *name;
		idx_t nid;
		ssize_t elem_num;
		freq_t freq_sum;

		name = dcw_id2name(g, id);
		assert(name);
		nid = dcw_name2id(g, name);
		assert(id==nid);
		elem_num = dxr_elem_num(f, id);
		freq_sum = dxr_freq_sum(f, id);
		fprintf(stream, "%s[%s,%s]%s\n", idxtostr(id), ssizetostr(elem_num), freqtostr(freq_sum), name);
	}
}
