/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dumpdw.c,v 1.2 2007/03/20 03:41:49 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <assert.h>
#if defined HAVE_LIBGEN_H
#include <libgen.h>
#endif
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwc.h>
#include <geta/dwam.h>

#include "tools.h"

static char *progname = NULL;

static void usage(void);
static int encode_type(char const *);

static void
usage()
{
	fprintf(stderr, "usage: %s [-q|-quick] [-s|-summary] handle type [fragment]\n", progname);
	fprintf(stderr, "\tquick: do not show vector (DXR_*)\n");
	fprintf(stderr, "\tsummary: only summary information is printed\n");
	exit(1);
}

int
main(argc, argv)
char *argv[];
{
	char *dirname, *handle;
	DWFILE *f;
	char path[MAXPATHLEN];
	int type;
	ssize_t frag = 0;
	int summary = 0;
	int quick = 0;

	progname = basename(argv[0]);

	if (!(dirname = getenv(PROOTENV))) {
		fprintf(stderr, "environment variable " PROOTENV " is not set");
		return 1;
	}

	for (argc--, argv++; argc>0; argc--, argv++) {
		if (!strcmp(argv[0], "-s") || !strcmp(argv[0], "-summary")) {
			summary=1;
		}
		else if (!strcmp(argv[0], "-q") || !strcmp(argv[0], "-quick")) {
			quick=1;
		}
		else {
			break;
		}
	}

	if (argc<2) {
		usage();
	}

	handle = argv[0];
	argc--, argv++;
	if (!(type = encode_type(argv[0]))) {
		usage();
	}
	argc--, argv++;
	if ((type==DXR_ROW || type==DXR_COL) && argc==1) {
		char *e;
		frag = strtosize(argv[0], &e, 10);
		if (*e) {
			usage();
		}
		argc--, argv++;
	}
	else if (argc!=0) {
		usage();
	}

	dw_mkpathname(path, sizeof path, dirname, handle, type, frag, 0);

	switch (type) {
	case DW_SCHE:
		fprintf(stderr, "sorry, not implemented\n");
		return 1;
	case DCW_ROW:
	case DCW_COL:
		if (!(f = dcw_open(path, handle))) {
			nperror(path);
			return 1;
		}
		if (summary) {
			dcw_info(stdout, f);
		}
		else {
			dump_dcw(stdout, f);
		}
		dcw_close(f);
		break;
	case DXR_ROW:
	case DXR_COL:
		if (!(f = dxr_open(path, handle))) {
			nperror(path);
			return 1;
		}
		if (summary) {
			dxr_info(stdout, f);
		}
		else {
			dump_dxr(stdout, f, quick);
		}
		dxr_close(f);
		break;
	default:
		fprintf(stderr, "should not happen\n");
		return 1;
	}
	return 0;
}

static int
encode_type(s)
char const *s;
{
	size_t i;
	static struct {
		char *name;
		int type;
	} types[] = {
#define	FOO(x)	{#x, x}
		FOO(DW_SCHE),
		FOO(DCW_ROW),
		FOO(DCW_COL),
		FOO(DXR_ROW),
		FOO(DXR_COL),
#undef FOO
	};
	for (i=0; i<nelems(types); i++) {
		if (!strcasecmp(s, types[i].name)) {
			return types[i].type;
		}
	}
	return 0;
}
