/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dumptools.c,v 1.7 2007/03/26 00:17:57 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <assert.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwsh.h>
#include <geta/dwwt.h>
#include <geta/dwc.h>

#include "tools.h"

static size_t kstrlen(char const *);

void
dump_dcw(stream, f)
FILE *stream;
DWFILE *f;
{
	idx_t id, nid;
	ssize_t ne;

	ne = dcw_size(f);
	for (id = 1; id<=ne; id++) {
		char const *name;
		name = dcw_id2name(f, id);
		assert(name);
		nid = dcw_name2id(f, name);
		assert(id==nid);
		fprintf(stream, "%s[%s]\n", idxtostr(id), name);
	}
}

void
dump_dxr(stream, f, quick)
FILE *stream;
DWFILE *f;
{
	idx_t id;
	ssize_t ne;

	ne = dxr_size(f);
	for (id = 1; id<=ne; id++) {
		ssize_t elem_num;
		freq_t freq_sum;
		size_t i;
		struct xr_vec const *v;
		if ((elem_num = dxr_get_vec(f, id, &v))<0) {
			continue;
		}
		assert(elem_num == dxr_elem_num(f, id));
		freq_sum = dxr_freq_sum(f, id);
		fprintf(stream, "%s[%s,%s]", idxtostr(id), ssizetostr(elem_num), freqtostr(freq_sum));
		if (!quick && v->elem_num) {
			assert(v->elem_num == elem_num);
			assert(v->freq_sum == freq_sum);
			for (i=0; i<v->elem_num; i++) {
				fprintf(stream, "(%s,%s)", idxtostr(v->elems[i].id), freqtostr(v->elems[i].freq));
			}
		}
		fprintf(stream, "\n");
	}
}

void
dcw_info(stream, f)
FILE *stream;
DWFILE *f;
{
	fprintf(stream, "size = %s\n", ssizetostr(dcw_size(f)));
	fprintf(stream, "mtime = %x\n", dcw_mtime(f));
}

void
dxr_info(stream, f)
FILE *stream;
DWFILE *f;
{
	fprintf(stream, "size = %s\n", ssizetostr(dxr_size(f)));
	fprintf(stream, "mtime = %x\n", dxr_mtime(f));
	fprintf(stream, "max_elem_num = %s\n", ssizetostr(dxr_max_elem_num(f)));
	fprintf(stream, "max_freq_sum = %s\n", freqtostr(dxr_max_freq_sum(f)));
	fprintf(stream, "total_elem_num = %s\n", freqtostr(dxr_total_elem_num(f)));
	fprintf(stream, "total_freq_sum = %s\n", freqtostr(dxr_total_freq_sum(f)));
	fprintf(stream, "compress = %s\n", dxr_compress_type(f));
}

void
dump_syminfo(stream, r, nr, dcw)
FILE *stream;
struct syminfo *r;
size_t nr;
DWFILE *dcw;
{
	size_t i;
	size_t maxlen = 0;

	if (dcw) {
		for (i=0; i<nr; i++) {
			char const *p;
			struct syminfo *t = &r[i];
			size_t len;
			if (!(p = dcw_id2name(dcw, t->id))) {
				p = "*";
			}
			len = kstrlen(p);
			if (maxlen < len) {
				maxlen = len;
			}
		}
	}
	for (i=0; i<nr; i++) {
		struct syminfo *t = &r[i];
#define FREECOLUMN
#if defined FREECOLUMN
		fprintf(stream, "%s: ", sizetostr(i));
		fprintf(stream, "ID=%s ", idxtostr(t->id));
		fprintf(stream, "TF=%s ", freqtostr(t->TF));
		fprintf(stream, "TF_d=%s ", freqtostr(t->TF_d));
		fprintf(stream, "DF=%s ", freqtostr(t->DF));
		fprintf(stream, "DF_d=%s ", freqtostr(t->DF_d));
		fprintf(stream, "weight=%g ", t->weight);
		fprintf(stream, "\n");
#else
		fprintf(stream, "%4d: ", i);
		fprintf(stream, "ID=%5x ", t->id);

		if (dcw) {
			char const *p;
			size_t len, j;
			if (!(p = dcw_id2name(dcw, t->id))) {
				p = "*";
			}
			len = kstrlen(p);
			fprintf(stream, "%s ", p);
			for (j=len; j<maxlen; j++) {
				putc(' ', stream);
			}
		}
		fprintf(stream, "TF=%-3s ", freqtostr(t->TF));
		fprintf(stream, "TF_d=%-3s ", freqtostr(t->TF_d));
		fprintf(stream, "DF=%-3s ", freqtostr(t->DF));
		fprintf(stream, "DF_d=%-3s ", freqtostr(t->DF_d));
		fprintf(stream, "weight=%-12.9f ", t->weight);
		fprintf(stream, "\n");
#endif
	}
}

static size_t
kstrlen(p)
char const *p;
{
	size_t len;
	len = strlen(p);
	if (*p == 033 && len>6) {
		len -= 6;
	}
	return len;
}
