/*
 *	Copyright (C) 2001 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dw.c,v 1.7 2007/03/26 23:07:29 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>

static int mkdirp(char const *, mode_t);

#define WDIR	"dwam/"


int
dw_mkpathname(path, size, dirname, handle, type, fragment, create_dir)
char *path;
char const *dirname, *handle;
size_t size;
ssize_t fragment;
{
	char dir[MAXPATHLEN];
	snprintf(dir, sizeof dir, "%s/" WDIR "%s", dirname, handle);
	switch (type) {
	case DW_SCHE:
		sprintf(path, "%s/sche", dir);
		break;
	case DCW_ROW:
		sprintf(path, "%s/dcwr", dir);
		break;
	case DCW_COL:
		sprintf(path, "%s/dcwc", dir);
		break;
	case DXR_ROW:
		if (fragment==-1) {
			sprintf(path, "%s/dxrr.s", dir);
		}
		else {
			sprintf(path, "%s/dxrr.%s", dir, sizetostr(fragment));
		}
		break;
	case DXR_COL:
		if (fragment==-1) {
			sprintf(path, "%s/dxrc.s", dir);
		}
		else {
			sprintf(path, "%s/dxrc.%s", dir, sizetostr(fragment));
		}
		break;
	default:
		dwerrno = DWE_INVAL;
		fprintf(stderr, "DWE_INVAL\n");
		return -1;
	}
	if (create_dir) {
		if (mkdirp(path, 0777)==-1) {
			perror(path);
			return -1;
		}
	}
	return 0;
}

void
dw_mkauipath(path, size, dirname, handle, abase, frag, create_dir)
char *path;
size_t size;
ssize_t frag;
char const *dirname, *handle, *abase;
{
	char p[32];
	char dir[MAXPATHLEN];
	snprintf(dir, sizeof dir, "%s/" WDIR "%s", dirname, handle);
	if (frag==-1) {
		strlcpy(p, "i", sizeof p);
	}
	else {
		snprintf(p, sizeof p, "%s", sizetostr(frag));
	}
	snprintf(path, size, "%s/%s.%s", dir, abase, p);
	if (create_dir) {
		if (mkdirp(path, 0777)==-1) {
			return;
		}
	}
}

static int
mkdirp(path, mode)
char const *path;
mode_t mode;
{
	size_t l;
	char *p, *q;

	if (*path=='\0' || !strcmp(path, "/")) {
		dwerrno = DWE_NOTDIR;
		return -1;
	}
	l = strlen(path);
	if (!(q = strdup(path))) {
		dwerrno = errno;
		return -1;
	}
	if (q[l-1] != '/') {
		if (!(p = rindex(q, '/'))) {
			free(q);
			dwerrno = DWE_IERROR;
			return -1;
		}
		*(p+1) = '\0';
	}

	for (p=index(q+1, '/'); p; p=index(p+1, '/')) {
		*p = '\0';
		mkdir(q, mode);
		*p = '/';
	}
	free(q);
	return 0;
}

char *
udate()
{
	static char buf[32];
	struct timeval t;
	gettimeofday(&t, NULL);
	sprintf(buf, "#%ld%05ld", t.tv_sec-1119570000, t.tv_usec / 10);
	return buf;
}
