/*
 *	Copyright (C) 1996 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dwam.h,v 1.7 2007/03/29 02:10:01 nis Exp $"; */

#define	PROOTENV	"GETAROOT"

struct wam_d;
typedef struct wam_d WAM;

int wam_init(char const *);
WAM *wam_open(char const *);
void wam_close(WAM *);

WAM *wam_prop_open(char const *, int, char const *);
WAM *wam_prop_cmd(char const *, int, char const *);
WAM *wam_prop_dl(char const *, int, char const *, char const *, int);
ssize_t wam_prop_gets(WAM *, idx_t, char const **);
ssize_t wam_prop_nentries(WAM *);

#define	wam_aux_open(handle, abase) wam_prop_open((handle), WAM_ROW, (abase))
#define	wam_aux_cmd(handle, cmd) wam_prop_cmd((handle), WAM_ROW, (cmd))
#define	wam_aux_dl(handle, path, name, mode) wam_prop_dl((handle), WAM_ROW, (path), (name), (mode))
#define wam_aux_gets(f, id, p)	wam_prop_gets((f), (id), (p))
#define wam_aux_nentries(f)	wam_prop_nentries((f))

#define	WAM_ROW	0	/* row */
#define	WAM_COL	1	/* column */

#define	WAM_REVERT_DIRECTION(dir)	(1-(dir))

struct syminfo;

ssize_t wam_size(WAM *, int);
char const *wam_id2name(WAM *, int, idx_t);
idx_t wam_name2id(WAM *, int, char const *);

ssize_t wam_elem_num(WAM *, int, idx_t);
freq_t wam_freq_sum(WAM *, int, idx_t);
ssize_t wam_max_elem_num(WAM *, int);
freq_t wam_max_freq_sum(WAM *, int);
freq_t wam_total_elem_num(WAM *, int);
freq_t wam_total_freq_sum(WAM *, int);

struct syminfo *wsh(struct syminfo const *, ssize_t, WAM *, int, int, ssize_t *, ssize_t *);
struct syminfo *wwsh(struct syminfo const *, ssize_t, WAM *, int, int, ssize_t *, ssize_t *, void *);
struct syminfo *wstem(char const *, WAM *, int, char const *, ssize_t *);
struct syminfo *wstemv(char const *, WAM *, int, char const *, char * const [], ssize_t *);
struct syminfo *bex_wsh(struct syminfo const *, ssize_t, WAM *, int, int, ssize_t *, ssize_t *, struct syminfo const *, ssize_t);

int wam_setopt(WAM *, int, char *);

ssize_t wam_get_vec(WAM *, int, idx_t, struct xr_vec const **);
char const *wam_handle(WAM *);

#define	WAM_OCASCADE		1
#define	WAM_OSUMI		2
#define	WAM_OPROBM		3
#define	WAM_OCACHE		4
#define	WAM_OFAULTTOLERANT	5

/* cm.c */
#if 0
freq_t **wcm0(struct syminfo *, int, struct syminfo *, int, WAM *, int);
#endif
ssize_t **wcm(struct syminfo const *, ssize_t, struct syminfo const *, ssize_t, WAM *, int);
void *dw_matrix(ssize_t, ssize_t, size_t);
ssize_t **dw_cm(struct syminfo const *, ssize_t, struct syminfo const *, ssize_t, DWFILE *);

/* setmask */
int wam_setmask(WAM *, int, idx_t *, size_t, idx_t *, size_t);


/***** CS *****/

/* fast but using more memory */
#define CS_FAST

/* verbose mode */
/*#define CS_VERBOSE*/

/* even apriori probabilitiy (for only HBC) */
/*#define CS_HBC_EVEN_PC*/

/* use global word statistics */
/*#define CS_GLOBAL_WINFO*/

/***** DO NOT MODIFY THE FOLLOWING LINES *****/

#define CS_HBC 1    /* hbc */
#define CS_WARD 2   /* Ward's method */
#define CS_GAVE 3   /* group average link */
#define CS_SLINK 4  /* single link */
#define CS_CLINK 5  /* complete link */

#define CS_SORT_DENDRO 0
#define CS_SORT_WEIGHT 1

struct syminfo_cell {
  ssize_t elem_num;
  freq_t freq_sum;
  double weight;
  int height;
  struct syminfo *orig;
  struct syminfo *sp;
};

struct syminfo_list {
  ssize_t elem_num;
  struct syminfo_cell *scp;
};

struct cs_elem {
  struct syminfo *csw;  /* pointer to word list */
  struct syminfo *csa;  /* pointer to article list */
  ssize_t cswlen;    /* length of csw */
  ssize_t csalen;    /* length of csa */
};

/* alloc.c */
int *check_n_realloc_int_array(int *, size_t, size_t *, int);
freq_t *check_n_realloc_frq_array(freq_t *, size_t, ssize_t *, int);

/* syminfo_list.c */
int alloc_syminfo_list(struct syminfo_list *, ssize_t);
void free_syminfo_list(struct syminfo_list *);

/* init_syminfo_list.c */
int init_syminfo_list(struct syminfo *, ssize_t, WAM *, int, struct syminfo_list *, int, int, ssize_t);

/* cs.c */
int cs(struct syminfo_list *, int, int);
struct cs_elem *csb(struct syminfo *, ssize_t, int, WAM *, int, int, int, ssize_t, size_t *, size_t, size_t, struct syminfo **);
void free_cs_elems(struct cs_elem *);
void sort_cw_syminfo_list(struct syminfo *, size_t, size_t, int);
