/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dwc.c,v 1.17 2007/03/26 00:17:57 nis Exp $"; */

#include <config.h>

#define DMKXRC

#if 1
#define SSI	/* small syminfo */
#else
ERROR Error error : SSI must be enabled to support heterogeneious cluster
#endif

#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>
#include <geta/nlog.h>
#include <geta/dxr.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwam.h>

#include "dxrP.h"
#include "dwcP.h"
#include "pkt.h"

#define	BLKSIZE	8192

#define	BFSC	2048

struct dw_FILE {
	int magic;
	int s, sid;
	struct com_dxr_hdr h;

	char *msgbuf;
	size_t msgbuf_len;

	idx_t seq;
	ssize_t t_size, r_size;

	size_t remote_msgbuf_len;	/* used by dwc_wshS. */
	ssize_t req_n;		/* used by dwc_wshS/R, dwc_mcS/R. */

	asig_t mysg, peersg;
};

static int check_rem_dxr_hdr(struct com_dxr_hdr *);
static int rc_handshake(int, DWFILE *);
static int rc_sync(int);
static int rc_tcp_nodelay(int);
static int rc_bfs(int, size_t);
static int rc_open(int, int, char const *, int, ssize_t, int, int, void *, size_t);
static int rc_close(int, int);
static void rc_disconnect(int);
static ssize_t rc_get_vec(int, int, idx_t, char *, size_t, int);
static int rc_set_mask(int, int, idx_t *, size_t, idx_t *, size_t, DWFILE *);
static int rc_create(int, int, char const *, size_t, int, int, ssize_t, size_t, int, int, int);
static int rc_write_vec(int, int, idx_t, struct xr_vec const *, ssize_t, int);
static int rc_flush(int, int);
static int rc_flush_nowait(int, int);
static int rc_flush_wait(int);
static ssize_t rc_seq_vec(int, char *, size_t);
static ssize_t rc_seq_blk(int, char *, size_t);
static int rc_write_blk(int, int, char const *, size_t);

#define	ptrdiff(a, b)	((char *)(a) - (char *)(b))

DWFILE *
dwc_dxr_open(s, handle, type, fragment, cascade)
char const *handle;
ssize_t fragment;
{
	DWFILE *f = NULL;

	if (!(f = calloc(1, sizeof *f))) {
		dwerrno = errno;
		goto fail;
	}

	f->s = s;
	if ((f->sid = rc_handshake(f->s, f))==-1) {
		goto fail;
	}

	if (rc_open(f->s, f->sid, handle, type, fragment, 0, cascade, &f->h, sizeof f->h)==-1) {
		goto fail;
	}

	if (check_rem_dxr_hdr(&f->h)==-1) {
		goto fail;
	}

	if ((f->msgbuf_len = sizeof_dxr_vec(f->h.max_elem_num))<MINBUFLEN) {
		f->msgbuf_len = MINBUFLEN;
	}

	if (!(f->msgbuf = malloc(f->msgbuf_len))) {
		dwerrno = errno;
		goto fail;
	}

	f->seq = 0;
	f->remote_msgbuf_len = 0;

	f->magic = DWC_DXR_MAGIC;
	return f;
fail:
	free(f);
	return NULL;
}

static int
check_rem_dxr_hdr(h)
struct com_dxr_hdr *h;
{
	if (h->type != DWD_REMOTE) {
		dwerrno = DWE_BADF;
		return -1;
	}
	return 0;
}

DWFILE *
dwc_dxr_create(s, handle, n, type, mtime, fragment, svec_size, trc_vec, compress)
char const *handle;
size_t n;
ssize_t fragment;
size_t svec_size;
{
	DWFILE *f = NULL;

	if (!(f = calloc(1, sizeof *f))) {
		dwerrno = errno;
fprintf(stderr, "!calloc 1 %s\n", sizetostr(sizeof *f));
		goto fail;
	}

	f->s = s;
	if ((f->sid = rc_handshake(f->s, f))==-1) {
fprintf(stderr, "!rc_handshake\n");
		goto fail;
	}

	if (rc_create(f->s, f->sid, handle, n, type, mtime, fragment, svec_size, 0, trc_vec, compress)==-1) {
fprintf(stderr, "!rc_create\n");
		goto fail;
	}

	f->msgbuf_len = MINBUFLEN;

	if (!(f->msgbuf = malloc(f->msgbuf_len))) {
		dwerrno = errno;
fprintf(stderr, "!malloc %s\n", sizetostr(f->msgbuf_len));
		goto fail;
	}

	f->magic = DWC_DXRC_MAGIC;
	return f;
fail:
	free(f);
	return NULL;
}

int
dwc_flush(f)
DWFILE *f;
{
	switch (f->magic) {
	case DWC_DXR_MAGIC:
	case DWC_RFILE_MAGIC:
		return 0;
	case DWC_DXRC_MAGIC:
	case DWC_WFILE_MAGIC:
		return rc_flush(f->s, f->sid);
	default:
		dwerrno = DWE_INVAL;
		return -1;
	}
}

int
dwc_flush_nowait(f)
DWFILE *f;
{
	switch (f->magic) {
	case DWC_DXR_MAGIC:
	case DWC_RFILE_MAGIC:
		return 0;
	case DWC_DXRC_MAGIC:
	case DWC_WFILE_MAGIC:
		return rc_flush_nowait(f->s, f->sid);
	default:
		dwerrno = DWE_INVAL;
		return -1;
	}
}

int
dwc_flush_wait(f)
DWFILE *f;
{
	switch (f->magic) {
	case DWC_DXR_MAGIC:
	case DWC_RFILE_MAGIC:
		return 0;
	case DWC_DXRC_MAGIC:
	case DWC_WFILE_MAGIC:
		return rc_flush_wait(f->s);
	default:
		dwerrno = DWE_INVAL;
		return -1;
	}
}

int
dwc_close(f)
DWFILE *f;
{
	if (rc_close(f->s, f->sid) == -1) {
		fprintf(stderr, "dwc_close: warning: communication failed\n");
		/* warning only */
	}
rc_disconnect(f->s);
	close(f->s);	/* XXX: ok? */
	f->magic = 0;
	free(f);
	return 0;
}

void
dwc_disconnect(s)
{
	rc_disconnect(s);
}

int
dwc_tcp_nodelay(f)
DWFILE *f;
{
	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}

	if (rc_tcp_nodelay(f->s)==-1) {
		return -1;
	}
	return 0;
}

ssize_t
dwc_dxr_size(f)
DWFILE *f;
{
	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}
	return f->h.nentries;
}

ssize_t
dwc_dxr_get_vec(f, id, v)
DWFILE *f;
idx_t id;
struct xr_vec const **v;
{
	size_t len;
	ssize_t e;

	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}
	if (id==0 || id>f->h.nentries) {
		dwerrno = DWE_BADR;
		return -1;
	}
	if ((len = rc_get_vec(f->s, f->sid, id, f->msgbuf, f->msgbuf_len, 0))==-1) {
		return -1;
	}
	if (len==0) {
		dwerrno = DWE_IERROR;
		return -1;
	}
	*v = (struct xr_vec *)f->msgbuf;
	e = (*v)->elem_num;

#if defined XDEBUG
fprintf(stderr, "%s: dwc: len = %s, sizeof_dxr_vec(e) = %s\n", udate(), freqtostr(len), sizetostr(sizeof_dxr_vec(e)));
#endif

	assert(len == sizeof_dxr_vec(e));

	return e;
}

int
dwc_dxr_write_vec(f, id, v, fragment)
DWFILE *f;
idx_t id;
struct xr_vec const *v;
ssize_t fragment;
{
	if (f->magic != DWC_DXRC_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}
	if (v->elem_num) {
		if (rc_write_vec(f->s, f->sid, id, v, fragment, 0)==-1) {
			return -1;
		}
	}
	return 0;
}

#if defined DMKXRC
int
dwc_add_trc_vec(f, id, v, fragment)
DWFILE *f;
idx_t id;
struct xr_vec const *v;
{
	if (f->magic != DWC_DXRC_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}
	if (v->elem_num) {
		if (rc_write_vec(f->s, f->sid, id, v, fragment, 1)==-1) {
			return -1;
		}
	}
	return 0;
}
#endif

static int
rc_handshake(s, f)
DWFILE *f;
{
	struct msg m;
	asig_t tmpsg;

#if defined XDEBUG
fprintf(stderr, "%s: dwc: handshake [%d]\n", udate(), s);
#endif
	f->mysg = asig();
	tmpsg = htonl(f->mysg);
	if (send_pkt_c(s, -1, DWCC_HND, &tmpsg, sizeof tmpsg)==-1) {
fprintf(stderr, "dwc: rc_handshake: send_pkt_c failed\n");
		return -1;
	}

	m.u.data = &tmpsg;
	m.u.size = sizeof tmpsg;
	if (recv_pkt_c(s, DWCC_HND | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_handshake: recv_pkt_c failed\n", udate());
#endif
		return -1;
	}
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_handshake succeeded\n", udate());
#endif
	if (m.h.len != sizeof tmpsg) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_handshake: short reply\n", udate());
#endif
		dwerrno = DWE_SHORTR;
		return -1;
	}
	f->peersg = ntohl(tmpsg);
	if (f->peersg != f->mysg) {
		fprintf(stderr, "sig mismatch: %08x / %08x\n", f->peersg, f->mysg);
		return -1;
	}
	return m.h.session_id;
}

static int
rc_sync(s)
{
	struct msg m;

#if defined XDEBUG
fprintf(stderr, "%s: dwc: sync [%d]\n", udate(), s);
#endif
	if (send_pkt_c(s, -1, DWCC_SYN, NULL, 0)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_sync: send_pkt_c failed\n", udate());
#endif
		return -1;
	}

	m.u.data = NULL;
	m.u.size = 0;
	if (recv_pkt_c(s, DWCC_SYN | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_sync: recv_pkt_c failed\n", udate());
#endif
		return -1;
	}
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_sync succeeded\n", udate());
#endif
	return 0;
}

static int
rc_tcp_nodelay(s)
{
	struct msg m;
	int tmp=1;

#if defined XDEBUG
fprintf(stderr, "%s: dwc: tcp_nodelay [%d]\n", udate(), s);
#endif
	if (send_pkt_c(s, -1, DWCC_NDL, NULL, 0)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_tcp_nodelay: send_pkt_c failed\n", udate());
#endif
		return -1;
	}

	m.u.data = NULL;
	m.u.size = 0;
	if (recv_pkt_c(s, DWCC_NDL | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_tcp_nodelay: recv_pkt_c failed\n", udate());
#endif
		return -1;
	}
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_tcp_nodelay succeeded\n", udate());
#endif
#if 0
fprintf(stderr, "dwc: TCP_NODELAY\n");
#endif
#if defined TCP_NODELAY
	(void)setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *)&tmp, sizeof tmp);
#endif
	return 0;
}

static int
rc_bfs(s, bfs)
size_t bfs;
{
	struct msg m;
	pintcom_t cbfs = bfs;

#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_bfs[%d]\n", udate(), s);
#endif
	if (send_pkt_c(s, -1, DWCC_BFS, &cbfs, sizeof cbfs)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_bfs failed\n", udate());
#endif
		return -1;
	}

	m.u.data = NULL;
	m.u.size = 0;
	if (recv_pkt_c(s, DWCC_BFS | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_bfs failed\n", udate());
#endif
		return -1;
	}
	return 0;
}

static int
rc_open(s, sid, handle, type, fragment, plain, cascade, v, size)
char const *handle;
ssize_t fragment;
void *v;
size_t size;
{
	struct msg m;
	size_t ac;
	pintcom_t p[4];
	void *e;
	size_t psize;
	struct iovec iov[2];

	if ((ac = pack_ints_dtdd(p, sizeof p, &e, "dtdd", &type, &fragment, &plain, &cascade)) != 4) {
		return -1;
	}
	psize = ptrdiff(e, p);
	iov[0].iov_base = (char *)p;
	iov[0].iov_len = psize;
	iov[1].iov_base = (char *)handle;
	iov[1].iov_len = strlen(handle)+1;
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_open[%d]\n", udate(), s);
#endif
	if (send_pkt_v(s, sid, DWCC_OPN, iov, 2)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_open: send_pkt_v failed\n", udate());
#endif
		return -1;
	}

	if (plain) {
		pintcom_t *ans_buf;

		if (2 * sizeof *ans_buf != size) {
			return -1;
		}

		ans_buf = v;
		m.u.data = ans_buf;
		m.u.size = 2 * sizeof *ans_buf;
		if (recv_pkt_c(s, DWCC_OPN | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_open: recv_pkt_c failed\n", udate());
#endif
			return -1;
		}
		if (m.h.len != 2 * sizeof *ans_buf) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_open: short reply\n", udate());
#endif
			dwerrno = DWE_SHORTR;
			return -1;
		}
	}
	else {
		struct com_dxr_hdr *h = v;

		if (sizeof *h != size) {
			return -1;
		}

		m.u.data = h;
		m.u.size = sizeof *h;
		if (recv_pkt_c(s, DWCC_OPN | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_open: recv_pkt_c failed\n", udate());
#endif
			return -1;
		}
		if (m.h.len != sizeof *h) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_open: short reply\n", udate());
#endif
			dwerrno = DWE_SHORTR;
			return -1;
		}
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_open: succeeded\n", udate());
#endif
	}

	return 0;
}

static int
rc_close(s, sid)
{
	struct msg m;

#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_close[%d]\n", udate(), s);
#endif
	if (send_pkt_c(s, sid, DWCC_CLO, NULL, 0)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_close: send_pkt failed\n", udate());
#endif
		return -1;
	}

	m.u.data = NULL;
	m.u.size = 0;
	if (recv_pkt_c(s, DWCC_CLO | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_close: recv_pkt_c failed\n", udate());
#endif
		return -1;
	}
	return 0;
}

static void
rc_disconnect(s)
{
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_disconnect: %d\n", udate(), s);
#endif
	if (send_pkt_c(s, -1, DWCC_DIS, NULL, 0)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_disconnect: send_pkt failed\n", udate());
#endif
		return;
	}
}

static ssize_t
rc_get_vec(s, sid, id, msgbuf, msgbuf_len, get_summary)
idx_t id;
char *msgbuf;
size_t msgbuf_len;
{
	struct msg m;
	size_t ac;
	pintcom_t p[2];
	void *e;
	size_t psize;

	if ((ac = pack_ints_id(p, sizeof p, &e, "id", &id, &get_summary)) != 2) {
		return -1;
	}
	psize = ptrdiff(e, p);

#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_get_vec[%d]\n", udate(), s);
#endif
	if (send_pkt_c(s, sid, DWCC_VEC, p, psize)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_get_vec: send_pkt_c failed\n", udate());
#endif
		return -1;
	}

	m.u.data = msgbuf;
	m.u.size = msgbuf_len;
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_get_vec: msgbuf_len = %s\n", udate(), sizetostr(msgbuf_len));
#endif
	if (recv_pkt_c(s, DWCC_VEC | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_get_vec: recv_pkt_c failed\n", udate());
#endif
		return -1;
	}
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_get_vec: succeeded\n", udate());
#endif

	return m.h.len;
}

/*
 * XXX pmask, nmask ga dummy pointer no toki okasiku naranaika 
 * kakunin suru koto. (coding ha sita ga, kensyou ha siteinai)
 */
static int
rc_set_mask(s, sid, pmask, plen, nmask, nlen, f)
idx_t *pmask, *nmask;
size_t plen, nlen;
DWFILE *f;	/* ad hock? */
{
	struct msg m;
	size_t pmask_size, nmask_size;
	size_t request_bfs;
	size_t ac;
	pintcom_t p[4];
	void *e;
	size_t psize;
	struct iovec iov[3];
	int ispmask = 1, isnmask = 1;

	if (!pmask) {
		ispmask = 0;
		plen = 0;
	}
	if (!nmask) {
		isnmask = 0;
		nlen = 0;
	}

	pmask_size = plen * sizeof *pmask;
	nmask_size = nlen * sizeof *nmask;

	request_bfs = sizeof p + pmask_size + nmask_size;
	if (request_bfs%BFSC) {
		request_bfs += BFSC - request_bfs%BFSC;
	}
	if (f->remote_msgbuf_len < request_bfs) {
		if (rc_bfs(f->s, request_bfs)==-1) {
			return -1;
		}
		f->remote_msgbuf_len = request_bfs;
	}

	if ((ac = pack_ints_ssdd(p, sizeof p, &e, "ssdd", &plen, &nlen, &ispmask, &isnmask)) != 4) {
		return -1;
	}
	psize = ptrdiff(e, p);

	iov[0].iov_base = (char *)p;
	iov[0].iov_len = psize;

	iov[1].iov_base = (char *)pmask;
	iov[1].iov_len = pmask_size;

	iov[2].iov_base = (char *)nmask;
	iov[2].iov_len = nmask_size;

	if (send_pkt_v(s, sid, DWCC_MSK, iov, 3)==-1) {
		fprintf(stderr, "rc_set_mask: send packet failed\n");
		return -1;
	}

	m.u.data = NULL;
	m.u.size = 0;
	if (recv_pkt_c(s, DWCC_MSK | DWCC_ACK, &m)==-1) {
		fprintf(stderr, "rc_set_mask: recv packet failed\n");
		return -1;
	}
	return 0;
}

#define	WVE_ARGNUM	2

static int
rc_create(s, sid, handle, n, type, mtime, fragment, bfs, plain, trc_vec, compress)
size_t n, bfs;
ssize_t fragment;
char const *handle;
{
	struct msg m;
	size_t ac;
	pintcom_t p[7];
	void *e;
	size_t psize;
	struct iovec iov[2];

	if ((ac = pack_ints_sddtddd(p, sizeof p, &e, "sddtddd", &n, &type, &mtime, &fragment, &plain, &trc_vec, &compress)) != 7) {
fprintf(stderr, "!pack_ints (1)\n");
		return -1;
	}
	psize = ptrdiff(e, p);
	iov[0].iov_base = (char *)p;
	iov[0].iov_len = psize;
#if defined XDEBUG
fprintf(stderr, "%s: rc_create: handle = [%s]\n", udate(), handle);
#endif
	iov[1].iov_base = (char *)handle;
	iov[1].iov_len = strlen(handle)+1;
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_create[%d]\n", udate(), s);
#endif
	if (send_pkt_v(s, sid, DWCC_CRE, iov, 2)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_create: send_pkt_v(cre) failed\n", udate());
#endif
fprintf(stderr, "!send_pkt_v (1)\n");
		return -1;
	}

	m.u.data = NULL;
	m.u.size = 0;
	if (recv_pkt_c(s, DWCC_CRE | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_create: recv_pkt_c (cre)failed\n", udate());
#endif
fprintf(stderr, "!recv_pkt_c (1)\n");
		return -1;
	}
	if (rc_bfs(s, bfs + WVE_ARGNUM * sizeof (pintcom_t))==-1) {
fprintf(stderr, "!rc_bfs (1)\n");
		return -1;
	}

	return 0;
}

static int
rc_write_vec(s, sid, id, v, fragment, atv)
idx_t id;
struct xr_vec const *v;
ssize_t fragment;
{
	size_t ac;
	pintcom_t p[WVE_ARGNUM];
	void *e;
	size_t psize;
	struct iovec iov[2];

#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_write_vec: fragment=%s, atv=%d\n", udate(), sizetostr(fragment), atv);
#endif
	if ((ac = pack_ints_it(p, sizeof p, &e, "it", &id, &fragment)) != 2) {
		return -1;
	}
	psize = ptrdiff(e, p);
	iov[0].iov_base = (char *)p;
	iov[0].iov_len = psize;
	iov[1].iov_base = (char *)v;
	if (fragment == WVE_WRITE || atv) {
		iov[1].iov_len = sizeof_dxr_vec(v->elem_num);
	}
	else {
		iov[1].iov_len = sizeof_dxr_vec(0);
	}

#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_write_vec(bfs)\n", udate());
#endif
#if defined DMKXRC
	if (atv) {
		if (send_pkt_v(s, sid, DWCC_ATV, iov, 2)==-1) {
			return -1;
		}
	} else
#endif
	if (send_pkt_v(s, sid, DWCC_WVE, iov, 2)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_write_vec: send_pkt_v failed\n", udate());
#endif
		return -1;
	}
	return 0;
}

static int
rc_flush(s, sid)
{
	if (rc_flush_nowait(s, sid)==-1) {
		return -1;
	}
	return rc_flush_wait(s);
#if 0

	struct msg m;

#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_flush[%d]\n", udate(), s);
#endif
	if (send_pkt_c(s, sid, DWCC_FLU, NULL, 0)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_flush: send_pkt failed\n", udate());
#endif
		return -1;
	}

	m.u.data = NULL;
	m.u.size = 0;
	if (recv_pkt_c(s, DWCC_FLU | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_flush: recv_pkt_c failed\n", udate());
#endif
		return -1;
	}
	return 0;
#endif
}

static int
rc_flush_nowait(s, sid)
{
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_flush_nowait[%d]\n", udate(), s);
#endif
	if (send_pkt_c(s, sid, DWCC_FLU, NULL, 0)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_flush_nowait: send_pkt failed\n", udate());
#endif
		return -1;
	}
	return 0;
}

static int
rc_flush_wait(s)
{
	struct msg m;

	m.u.data = NULL;
	m.u.size = 0;
	if (recv_pkt_c(s, DWCC_FLU | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_flush_wait: recv_pkt_c failed\n", udate());
#endif
		return -1;
	}
	return 0;
}

int
dwc_dxr_seq_init(f)
DWFILE *f;
{
	struct msg m;
	int s = f->s;

	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}

#if defined XDEBUG
fprintf(stderr, "%s: dwc: sec_vec init [%d]\n", udate(), s);
fprintf(stderr, "%s: dwc: sec_vec init [s=%d, sid=%d]\n", udate(), s, f->sid);
fprintf(stderr, "%s: dwc: sec_vec init [s=%d, sid=%d]\n", udate(), s, f->sid);
#endif
	if (send_pkt_c(s, f->sid, DWCC_SVI, NULL, 0)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: sec_vec init: send_pkt failed\n", udate());
#endif
		return -1;
	}

	m.u.data = NULL;
	m.u.size = 0;
	if (recv_pkt_c(s, DWCC_SVI | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: sec_vec init: recv_pkt_c failed\n", udate());
#endif
		return -1;
	}
#if defined XDEBUG
fprintf(stderr, "%s: dwc: sec_vec init succeeded\n", udate());
#endif
	f->seq = 1;
	return 0;
}

ssize_t
dwc_dxr_seq_get_vec(f, id, v)
DWFILE *f;
idx_t id;
struct xr_vec const **v;
{
	ssize_t e;
	ssize_t len;

	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}

	if (f->seq == 0 || f->seq != id) {
		f->seq = 0;
		dwerrno = DWE_IERROR;
		return -1;
	}

	if ((len = rc_seq_vec(f->s, f->msgbuf, f->msgbuf_len))==-1) {
		f->seq = 0;
		return -1;
	}
	if (len==0) {
		f->seq = 0;
		dwerrno = DWE_IERROR;
		return -1;
	}
	*v = (struct xr_vec *)f->msgbuf;
	e = (*v)->elem_num;

	f->seq++;
	return e;
}

int
dwc_dxr_seq_finish(f)
DWFILE *f;
{
	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}

	while (f->seq != 0 && f->seq <= f->h.nentries) {
		ssize_t len;
		if ((len = rc_seq_vec(f->s, f->msgbuf, f->msgbuf_len))==-1) {
			f->seq = 0;
			return -1;
		}
		if (len==0) {
			f->seq = 0;
			dwerrno = DWE_IERROR;
			return -1;
		}
		f->seq++;
	}

	f->seq = 0;

	if (rc_sync(f->s)==-1) {
		return -1;
	}
	return 0;
}

static ssize_t
rc_seq_vec(s, msgbuf, msgbuf_len)
char *msgbuf;
size_t msgbuf_len;
{
	struct msg m;

	m.u.data = msgbuf;
	m.u.size = msgbuf_len;
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_seq_vec: msgbuf_len = %s\n", udate(), sizetostr(msgbuf_len));
#endif
	if (recv_pkt_c(s, DWCC_SVR | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_seq_vec: recv_pkt_c failed\n", udate());
#endif
		return -1;
	}
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_seq_vec: succeeded\n", udate());
#endif

	return m.h.len;
}

DWFILE *
dwc_file_open(s, handle, type, fragment)
char const *handle;
ssize_t fragment;
{
	DWFILE *f = NULL;
	pintcom_t p[2];
	size_t blksize;

	if (!(f = calloc(1, sizeof *f))) {
		dwerrno = errno;
		goto fail;
	}

	f->s = s;
	if ((f->sid = rc_handshake(f->s, f))==-1) {
		goto fail;
	}

	if (rc_open(f->s, f->sid, handle, type, fragment, 1, 0, p, sizeof p)==-1) {
		goto fail;
	}

	CKOVFL(p[0]);
	CKOVFL(p[1]);
	f->t_size = p[0];
	blksize = p[1];
#if defined XDEBUG
fprintf(stderr, "%s: rc_open: size=%s, blksize=%s\n", udate(), sizetostr(f->t_size), sizetostr(blksize));
#endif

	f->msgbuf_len = MAX(MINBUFLEN, blksize);

	if (!(f->msgbuf = malloc(f->msgbuf_len))) {
		dwerrno = errno;
		goto fail;
	}

	f->r_size = -1;

	f->magic = DWC_RFILE_MAGIC;
	return f;
fail:
	free(f);
	return NULL;
}

DWFILE *
dwc_file_create(s, handle, type, fragment)
char const *handle;
ssize_t fragment;
{
	DWFILE *f = NULL;

	if (!(f = calloc(1, sizeof *f))) {
		dwerrno = errno;
		goto fail;
	}

	f->s = s;
	if ((f->sid = rc_handshake(f->s, f))==-1) {
		goto fail;
	}

	if (rc_create(f->s, f->sid, handle, 0, type, 0, fragment, BLKSIZE + sizeof (pintcom_t), 1, 0, 0)==-1) {
		goto fail;
	}

	f->msgbuf_len = MINBUFLEN;

	if (!(f->msgbuf = malloc(f->msgbuf_len))) {
		dwerrno = errno;
		goto fail;
	}

	f->magic = DWC_WFILE_MAGIC;
	return f;
fail:
	free(f);
	return NULL;
}

int
dwc_file_write_blk(f, b, size)
DWFILE *f;
char const *b;
size_t size;
{
	if (f->magic != DWC_WFILE_MAGIC) {
		fprintf(stderr, "dwc_file_write_blk: invalid magic\n");
		dwerrno = DWE_INVAL;
		return -1;
	}

	while (size > 0) {
		size_t siz;
		siz = MIN(size, BLKSIZE);

		if (rc_write_blk(f->s, f->sid, b, siz)==-1) {
			return -1;
		}

		b += siz;
		size -= siz;
	}

	return 0;
}

static int
rc_write_blk(s, sid, b, size)
char const *b;
size_t size;
{
	struct iovec iov[2];
	pintcom_t csize = size;

	if (size > BLKSIZE) {
		fprintf(stderr, "dwc: rc_write_blk: too large size(%s)\n", sizetostr(size));
		dwerrno = DWE_INVAL;
		return -1;
	}

	iov[0].iov_base = (char *)&csize;
	iov[0].iov_len = sizeof csize;
	iov[1].iov_base = (char *)b;
	iov[1].iov_len = size;

#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_write_blk(bfs)\n", udate());
#endif
	if (send_pkt_v(s, sid, DWCC_BLK, iov, 2)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_write_blk: send_pkt_v failed\n", udate());
#endif
		return -1;
	}
	return 0;
}

int
dwc_file_seq_init(f)
DWFILE *f;
{
	struct msg m;
	int s = f->s;

	if (f->magic != DWC_RFILE_MAGIC) {
		fprintf(stderr, "dwc_file_write_blk: invalid magic\n");
		dwerrno = DWE_INVAL;
		return -1;
	}

#if defined XDEBUG
fprintf(stderr, "%s: dwc: sec_file init [%d]\n", udate(), s);
#endif
	if (send_pkt_c(s, f->sid, DWCC_SBI, NULL, 0)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: sec_file init: send_pkt failed\n", udate());
#endif
		return -1;
	}

	m.u.data = NULL;
	m.u.size = 0;
	if (recv_pkt_c(s, DWCC_SBI | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: sec_file init: recv_pkt_c failed\n", udate());
#endif
		return -1;
	}
#if defined XDEBUG
fprintf(stderr, "%s: dwc: sec_file init succeeded\n", udate());
#endif
	f->r_size = 0;
#if defined XDEBUG
fprintf(stderr, "%s: f->t_size = %s\n", udate(), sizetostr(f->t_size));
#endif
	return 0;
}

int
dwc_file_seq_recv_blk(f, bp, sizep)
DWFILE *f;
char const **bp;
size_t *sizep;
{
	ssize_t len;

	if (f->magic != DWC_RFILE_MAGIC) {
		fprintf(stderr, "dwc_file_write_blk: invalid magic\n");
		dwerrno = DWE_INVAL;
		return -1;
	}

	if (f->t_size == f->r_size) {
		*bp = NULL;
		*sizep = 0;
		return 0;
	}

	if (f->r_size==-1 || f->t_size <= f->r_size) {
		dwerrno = DWE_IERROR;
		return -1;
	}

	if ((len = rc_seq_blk(f->s, f->msgbuf, f->msgbuf_len))==-1) {
		f->r_size = f->t_size = -1;
		return -1;
	}
	if (len==0 || f->t_size < f->r_size+len) {
		f->r_size = f->t_size = -1;
		dwerrno = DWE_IERROR;
		return -1;
	}

	f->r_size += len;
	*bp = f->msgbuf;
	*sizep = len;
	return 0;
}

int
dwc_file_seq_finish(f)
DWFILE *f;
{
	if (f->magic != DWC_RFILE_MAGIC) {
		fprintf(stderr, "dwc_file_write_blk: invalid magic\n");
		dwerrno = DWE_INVAL;
		return -1;
	}

	while (f->r_size != -1 && f->r_size < f->t_size) {
		ssize_t len;
		if ((len = rc_seq_blk(f->s, f->msgbuf, f->msgbuf_len))==-1) {
			f->r_size = f->t_size = -1;
			return -1;
		}
		if (len==0 || f->t_size < f->r_size+len) {
			f->r_size = f->t_size = -1;
			dwerrno = DWE_IERROR;
			return -1;
		}
		f->r_size += len;
	}

	f->r_size = f->t_size = -1;

	if (rc_sync(f->s)==-1) {
		return -1;
	}

	return 0;
}

static ssize_t
rc_seq_blk(s, msgbuf, msgbuf_len)
char *msgbuf;
size_t msgbuf_len;
{
	struct msg m;

	m.u.data = msgbuf;
	m.u.size = msgbuf_len;
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_seq_blk: msgbuf_len = %s\n", udate(), sizetostr(msgbuf_len));
#endif
	if (recv_pkt_c(s, DWCC_SBR | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_seq_blk: recv_pkt_c failed\n", udate());
#endif
		return -1;
	}
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_seq_blk: succeeded\n", udate());
#endif

	return m.h.len;
}

int
dwc_wshS(f, q, nq, type, n, tn, sum_i, prob_mode)
DWFILE *f;
struct syminfo const *q;
ssize_t nq, n, tn, sum_i;
{
#if defined SSI
	static struct s_syminfo *r = NULL;
	static size_t maxnrq = 0;
	size_t rsize;
#else
	size_t qsize;
#endif
	size_t oe_qsize;
	size_t request_bfs;
	size_t ac;
	pintcom_t p[6];
	void *e;
	size_t psize;
	struct iovec iov[2];

assert(sizeof (struct s_syminfo) <= sizeof (struct syminfo));
assert(sizeof (struct syminfo) <= sizeof (struct oe_syminfo));

	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}

	if (sum_i) {
		f->magic = DWC_DXRS_MAGIC;
	}
	else {
		f->magic = DWC_DXRW_MAGIC;
	}

	if ((ac = pack_ints_jdjjjd(p, sizeof p, &e, "jdjjjd", &nq, &type, &n, &tn, &sum_i, &prob_mode)) != 6) {
		return -1;
	}
	psize = ptrdiff(e, p);

#if defined SSI
	rsize = nq * sizeof *r;
#else
	qsize = nq * sizeof *q;
#endif

/*
 * request_bfs ha, server ga uketotta syminfo (or s_syminfo) wo,
 * server gawa no memory kouzou no henkan surutoki ni
 * realloc sezuni sumu you, over estimate siteoku
 */
	oe_qsize = nq * sizeof (struct oe_syminfo);
	request_bfs = psize + oe_qsize;

	f->req_n = n;

	if (request_bfs%BFSC) {
#if defined XDEBUG
fprintf(stderr, "%s: dwcc_wsh: request_bfs = %s => ", udate(), sizetostr(request_bfs));
#endif
		request_bfs += BFSC - request_bfs%BFSC;
#if defined XDEBUG
fprintf(stderr, "%s\n", sizetostr(request_bfs));
#endif
	}
	if (f->remote_msgbuf_len < request_bfs) {
#if defined XDEBUG
fprintf(stderr, "%s: dwcc_wsh: request_bfs = %s\n", udate(), sizetostr(request_bfs));
#endif
		if (rc_bfs(f->s, request_bfs)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwcc_wsh: rc_bfs failed\n", udate());
#endif
			return -1;
		}
		f->remote_msgbuf_len = request_bfs;
	}

	iov[0].iov_base = (char *)p;
	iov[0].iov_len = psize;
#if defined SSI
	if (nq>maxnrq) {
		void *new;
		size_t newsize = NALIGN(nq, 256);
		if (!(new = realloc(r, newsize*sizeof *r))) {
			return -1;
		}
		maxnrq = newsize;
		r = new;
	}
	pack_ssi(r, q, nq);
	iov[1].iov_base = (char *)r;
	iov[1].iov_len = rsize;
#else
	iov[1].iov_base = (char *)q;
	iov[1].iov_len = qsize;
#endif
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_wsh: send_pkt_v : %p\n", udate(), iov);
#endif

	if (send_pkt_v(f->s, f->sid, DWCC_WSH, iov, 2)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_wsh: send_pkt_v failed\n", udate());
#endif
		return -1;
	}

	return 0;
}

struct syminfo *
dwc_wshR(f, n, total, elapsed)
DWFILE *f;
ssize_t *n, *total;
struct timeval *elapsed;
{
	struct msg m;
	size_t rsize;
	size_t reply_bfs;
	ssize_t dlen;
	struct syminfo *d;
	size_t ac;
	void *o;
	size_t olen;
	pintcom_t p[4];
	void *e;
	int tv_sec, tv_usec;

	if (f->magic != DWC_DXRW_MAGIC) {
		dwerrno = DWE_INVAL;
		return NULL;
	}

	f->magic = DWC_DXR_MAGIC;

	rsize = f->req_n * sizeof *d;
#if defined SSI
/*
 * rsize ha, syminfo no size de keisan site oku.
 * korede, unpack_ssi no toki ni realloc sinakute sumu.
 */
#endif
	if ((ac = pack_ints_jjdd(p, sizeof p, &e, "jjdd", n, total, &tv_sec, &tv_usec)) != 4) {
		return NULL;
	}
	olen = ptrdiff(e, p);

	reply_bfs = rsize + olen;

	if (!(d = malloc(reply_bfs))) {
		dwerrno = errno;
		return NULL;
	}

	m.u.data = d;
	m.u.size = reply_bfs;
	if (recv_pkt_c(f->s, DWCC_WSH | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_wsh: recv_pkt_c failed\n", udate());
#endif
		return NULL;
	}

	if (m.h.len < olen) {
		dwerrno = DWE_IERROR;
		return NULL;
	}
	dlen = m.h.len - olen;
	o = (char *)d + dlen;

	if ((ac = unpack_ints_jjdd(o, olen, NULL, "jjdd", n, total, &tv_sec, &tv_usec)) != 4) {
		return NULL;
	}

	assert(ac == 4);
#if defined SSI
	assert((char *)(d+(*n)) <= (char *)d + reply_bfs);
	unpack_ssi(d, (struct s_syminfo *)d, *n);
	assert((*n) == dlen / sizeof (struct s_syminfo));
#else
	assert((*n) == dlen / sizeof *d);
#endif
	assert((*total) >= (*n));
	elapsed->tv_sec = tv_sec;
	elapsed->tv_usec = tv_usec;

#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_wsh succeeded\n", udate());
#endif
	return d;
}

ssize_t
dwc_wshRs(f, dp, size)
DWFILE *f;
float **dp;
ssize_t size;
{
	struct msg m;
	float *d;
	ssize_t r;

	if (f->magic != DWC_DXRS_MAGIC) {
		fprintf(stderr, "dwc_wshRs: invalid magic");
		dwerrno = DWE_INVAL;
		return -1;
	}

	f->magic = DWC_DXRN_MAGIC;

	if (!(d = malloc(size * sizeof *d))) {
		dwerrno = errno;
		return -1;
	}

	m.u.data = d;
	m.u.size = size * sizeof *d;
	if (recv_pkt_c(f->s, DWCC_WSHS | DWCC_ACK, &m)==-1) {
		fprintf(stderr, "dwc: rc_wsh: recv_pkt_c failed");
		return -1;
	}

	if (m.h.len % sizeof *d) {
		fprintf(stderr, "dwc: rc_wsh: short reply(%s)", sizetostr(m.h.len));
		dwerrno = DWE_IERROR;
		return -1;
	}

#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_wsh succeeded\n", udate());
#endif
	*dp = d;
	r = m.h.len / sizeof *d;
	return r;
}

int
dwc_wshRn(f, n)
DWFILE *f;
ssize_t n;
{
	pintcom_t crqn = n;
	if (f->magic != DWC_DXRN_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}

	f->magic = DWC_DXRW_MAGIC;

#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_wshRn[%d]\n", udate(), /*s*/0);
#endif
	if (send_pkt_c(f->s, f->sid, DWCC_RQN, &crqn, sizeof crqn)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_wshRn failed\n", udate());
#endif
		return -1;
	}

	/* no ack */
	return 0;
}

ssize_t
dwc_dxr_elem_num(f, id)
DWFILE *f;
idx_t id;
{
	size_t len;
	struct xr_vec const *v;

	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}
	if (id==0 || id>f->h.nentries) {
		dwerrno = DWE_IERROR;
		return -1;
	}
	if ((len = rc_get_vec(f->s, f->sid, id, f->msgbuf, f->msgbuf_len, 1))==-1) {
		return -1;
	}
	if (len==0) {
		dwerrno = DWE_IERROR;
		return -1;
	}
	v = (struct xr_vec *)f->msgbuf;
	assert(len == sizeof_dxr_vec(0));
	return v->elem_num;
}

freq_t
dwc_dxr_freq_sum(f, id)
DWFILE *f;
idx_t id;
{
	size_t len;
	struct xr_vec const *v;

	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}
	if (id==0 || id>f->h.nentries) {
		dwerrno = DWE_IERROR;
		return -1;
	}
	if ((len = rc_get_vec(f->s, f->sid, id, f->msgbuf, f->msgbuf_len, 1))==-1) {
		return -1;
	}
	if (len==0) {
		dwerrno = DWE_IERROR;
		return -1;
	}
	v = (struct xr_vec *)f->msgbuf;
	assert(len == sizeof_dxr_vec(0));
	return v->freq_sum;
}

ssize_t
dwc_dxr_max_elem_num(f)
DWFILE *f;
{
	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}

	return f->h.max_elem_num;
}

freq_t
dwc_dxr_max_freq_sum(f)
DWFILE *f;
{
	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}

	return f->h.max_freq_sum;
}

freq_t
dwc_dxr_total_elem_num(f)
DWFILE *f;
{
	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}

	return f->h.total_elem_num;
}

freq_t
dwc_dxr_total_freq_sum(f)
DWFILE *f;
{
	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}

	return f->h.total_freq_sum;
}

int
dwc_wcmS(f, words, nwords, docs, ndocs)
DWFILE *f;
struct syminfo const *words, *docs;
ssize_t nwords, ndocs;
{
#if defined SSI
	static struct s_syminfo *r1 = NULL;
	static struct s_syminfo *r2 = NULL;
	static ssize_t maxnr1 = 0;
	static ssize_t maxnr2 = 0;
	size_t r1size, r2size;
#else
	size_t words_size, docs_size;
#endif
	size_t oe_r1size, oe_r2size;
	size_t request_bfs;
	size_t ac;
	pintcom_t p[2];
	void *e;
	size_t psize;
	struct iovec iov[3];
#if defined XDEBUG
fprintf(stderr, "%s: dwcc_wcm: start.\n", udate());
#endif

	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}

	f->magic = DWC_DXRW_MAGIC;


#if defined SSI
	r1size = nwords * sizeof *r1;
	r2size = ndocs * sizeof *r2;
#else
	words_size = nwords * sizeof *words;
	docs_size = ndocs * sizeof *docs;
#endif

	if ((ac = pack_ints_jj(p, sizeof p, &e, "jj", &nwords, &ndocs)) != 2) {
		return -1;
	}
	psize = ptrdiff(e, p);

	f->req_n = nwords;

/*
 * request_bfs ha, server ga uketotta syminfo (or s_syminfo) wo,
 * server gawa no memory kouzou no henkan surutoki ni
 * realloc sezuni sumu you, over estimate siteoku
 */
	oe_r1size = nwords * sizeof (struct oe_syminfo);
	oe_r2size = ndocs * sizeof (struct oe_syminfo);
	request_bfs = psize + oe_r1size + oe_r2size;

	if (request_bfs%BFSC) {
#if defined XDEBUG
fprintf(stderr, "%s: dwcc_wcm: request_bfs = %s => ", udate(), sizetostr(request_bfs));
#endif
		request_bfs += BFSC - request_bfs%BFSC;
#if defined XDEBUG
fprintf(stderr, "%s\n", sizetostr(request_bfs));
#endif
	}
	if (f->remote_msgbuf_len < request_bfs) {
#if defined XDEBUG
fprintf(stderr, "%s: dwcc_wcm: request_bfs = %s\n", udate(), sizetostr(request_bfs));
#endif
		if (rc_bfs(f->s, request_bfs)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwcc_wcm: rc_bfs failed\n", udate());
#endif
			return -1;
		}
		f->remote_msgbuf_len = request_bfs;
	}

	iov[0].iov_base = (char *)p;
	iov[0].iov_len = psize;
#if defined SSI
	if (nwords>maxnr1) {
		void *new;
		size_t newsize = NALIGN(nwords, 256);
		if (!(new = realloc(r1, newsize*sizeof *r1))) {
			return -1;
		}
		maxnr1 = newsize;
		r1 = new;
	}
	if (ndocs>maxnr2) {
		void *new;
		size_t newsize = NALIGN(ndocs, 256);
		if (!(new = realloc(r2, newsize*sizeof *r2))) {
			return -1;
		}
		maxnr2 = newsize;
		r2 = new;
	}
	pack_ssi(r1, words, nwords);
	pack_ssi(r2, docs, ndocs);
	iov[1].iov_base = (char *)r1;
	iov[1].iov_len = r1size;
	iov[2].iov_base = (char *)r2;
	iov[2].iov_len = r2size;
#else
	iov[1].iov_base = (char *)words;
	iov[1].iov_len = words_size;
	iov[2].iov_base = (char *)docs;
	iov[2].iov_len = docs_size;
#endif
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_wcm: send_pkt_v : %p\n", udate(), iov);
#endif

	if (send_pkt_v(f->s, f->sid, DWCC_CM, iov, 3)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_wcm: send_pkt_v failed\n", udate());
#endif
		return -1;
	}

	return 0;
}

ssize_t **
dwc_wcmR(f)
DWFILE *f;
{
	struct msg m;
	size_t rsize;
	size_t reply_bfs;
	ssize_t **d;
	ssize_t nwords;

	if (f->magic != DWC_DXRW_MAGIC) {
		dwerrno = DWE_INVAL;
		return NULL;
	}

	f->magic = DWC_DXR_MAGIC;
	nwords = f->req_n;

	rsize = nwords * nwords * sizeof **d;
	reply_bfs = rsize;

	if (!(d = dw_matrix(nwords, nwords, sizeof **d))) {
		dwerrno = errno;
		return NULL;
	}

	m.u.data = &d[0][0];
	m.u.size = reply_bfs;
	if (recv_pkt_c(f->s, DWCC_CM | DWCC_ACK, &m)==-1) {
#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_wcm: recv_pkt_c failed\n", udate());
#endif
		return NULL;
	}

	if (m.h.len != reply_bfs) {
		return NULL;
	}

#if defined XDEBUG
fprintf(stderr, "%s: dwc: rc_wcm succeeded\n", udate());
#endif
	return d;
}

int
dwc_dxr_set_mask(f, pmask, plen, nmask, nlen)
DWFILE *f;
idx_t *pmask, *nmask;
size_t plen, nlen;
{
	if (f->magic != DWC_DXR_MAGIC) {
		dwerrno = DWE_INVAL;
		return -1;
	}
	if (rc_set_mask(f->s, f->sid, pmask, plen, nmask, nlen, f)==-1) {
		return -1;
	}

	return 0;
}
