/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dwdserv.c,v 1.19 2007/03/29 07:37:38 nis Exp $"; */

#include <config.h>

#define DMKXRC

#if 1
#define SSI	/* small syminfo */
#else
ERROR Error error : SSI must be enabled to support heterogeneious cluster
#endif

#include <sys/param.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/file.h>

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/timer.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwam.h>
#include <geta/dwsched.h>

#include "mmalloc.h"
#include "dwsys.h"
#include "pkt.h"
#include "dxrP.h"
#include "dwcP.h"
#include "tutil.h"
#if defined DMKXRC
#include "cel.h"
#endif

#if HAVE_IEEEFP_H
#include <ieeefp.h>
#endif

#include "zvec.h"

#define	BLKSIZE	8192

#define	PAGESIZE	8192

struct remote {
	int mode;

	/* dxr, plain file */
	FILE *stream;

	/* dxr */
	size_t addr;
	struct dxr_entry *bindex;
	struct dxr_hdr h;
	int compress;

#if defined DMKXRC
	int trc_vec;
#endif
};

struct rdxr {
	int mode;

	char handle[MAXPATHLEN];
	char dirname[MAXPATHLEN];

	/* dxr */
	DWFILE *xr;
	struct com_dxr_hdr h;

	/* plain file */
	FILE *f;
	size_t size;

	/* for wsh */
	int wsh_initialized;

	DWFILE *qs, *ds;
	char qs_path[MAXPATHLEN], ds_path[MAXPATHLEN];

	/* for cascaded wsh */
	int cascade;
	ssize_t nodeid;
	struct sym_vec *sv;
	ssize_t np;
	int type;
	struct dwsched *df;

	/* for dwcc_vec */
	struct xr_vec v0;

	/* for heterogenious environment */
	asig_t mysg, peersg, tmpsg;
};

struct dwdd {
	char const *dirname;
	struct remote r;
	struct rdxr q;
	int in_use;
	struct dwdd *next;
};

static int dwcc_hnd(struct msg *, struct msg *, struct rdxr *);
static int dwcc_bfs(struct msg *, char **, size_t *);
static int dwcc_opn(struct msg *, struct msg *, struct dwdd *, pintcom_t *);
static int dwcc_cre(struct msg *, char const *, struct remote *);
static int dwcc_vec(struct msg *, struct msg *, struct rdxr *);
static int dwcc_wve(struct msg *, struct remote *, int);
static int dwcc_blk(struct msg *, struct remote *);
static ssize_t dwd_open_plain(int, ssize_t, struct rdxr *);
static int dwd_open_and_fill(int, ssize_t, struct rdxr *);
static int dwd_create(char *, size_t, int, int, ssize_t, int, struct remote *, int, int);
static int dwd_write_vec(idx_t, struct xr_vec *, ssize_t, struct remote *);
#if defined DMKXRC
static int dwd_add_trc_vec(idx_t, struct xr_vec *, ssize_t, struct remote *);
#endif
static int dwcc_make_summary_information(struct remote *);
#if defined DMKXRC
static int dwcc_flush_trc_vec(struct remote *);
#endif
static int dwcc_svi(int, struct rdxr *);
static int dwcc_flush(struct remote *);
static int dwd_write_blk(char const *, size_t, struct remote *);
static int dwcc_sbi(int, struct rdxr *);
static int dwcc_wsh(int, struct rdxr *, void *, size_t, size_t);
static int dwcc_wsh_init(struct rdxr *);
static int cascade_init(struct rdxr *);

static int dwcc_cm(int, struct rdxr *, void *, size_t, size_t);
static int dwcc_msk(int, struct rdxr *, void *, size_t, size_t);

#define	M_DXRW		'W'
#define	M_DXRR		'R'
#define	M_PLAINW	'w'
#define	M_PLAINR	'r'

#define	MAXSESSION	6

#define	ptrdiff(a, b)	((char *)(a) - (char *)(b))

#define CHKSESID(s) do { \
	if (m.h.session_id<0 || nelems(dwdd) <= m.h.session_id) { \
		syslog(LOG_DEBUG, "session id out of range: %d, cmd=%d", m.h.session_id, m.h.cmd); \
		goto error; \
	} \
	if (!dwdd[m.h.session_id].in_use) { \
		syslog(LOG_DEBUG, "session %d is not open, cmd=%d", m.h.session_id, m.h.cmd); \
	} \
} while(0)

void
dwd_serve(s, dirname)
char const *dirname;
{
	struct msg m, r;
	char *msgbuf;
	size_t msgbuf_len = MINBUFLEN;
	size_t i;
	static struct dwdd dwdd[MAXSESSION];
	struct dwdd *p, *freelist;

	freelist = NULL;
	for (i=0; i<nelems(dwdd); i++) {
		p = &dwdd[i];
		p->r.mode = 0;
		p->q.mode = 0;
		p->q.df = NULL;
		p->dirname = dirname;
		p->next = freelist;
		p->in_use = 0;
		freelist = p;
	}

	if (!(msgbuf = malloc(msgbuf_len))) {
		syslog(LOG_DEBUG, "malloc: %s", strerror(errno));
		return;
	}

	for (;;) {
		pintcom_t ans_buf[2];
		int sid = -1;
		m.h.cmd = 0;
		m.h.session_id = -2;
		m.u.data = msgbuf;
		m.u.size = msgbuf_len;
nlogflush();
		if (recv_pkt(s, &m)==-1) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: recv_pkt failed: cmd=%d\n", udate(), m.h.cmd); 
#endif
			break;
		}
		r.h.len = 0;
		switch (m.h.cmd) {
		case DWCC_HND:
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: handshake\n", udate());
#endif
			if (!(p = freelist)) {
				goto error;
				sid = -1;
			}
			else {
				assert(!p->in_use);
				freelist = freelist->next;
				sid = p - dwdd;
				p->in_use = 1;
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: new session: %d\n", udate(), sid);
#endif
			}
			if (dwcc_hnd(&m, &r, &p->q)==-1) {
				goto error;
			}
			break;
		case DWCC_SYN:
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: sync\n", udate());
#endif
			break;
		case DWCC_NDL: {
			int tmp=1;
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: TCP_NODELAY\n", udate());
#endif
#if defined TCP_NODELAY
			(void)setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *)&tmp, sizeof tmp);
#endif
			}
			break;
		case DWCC_BFS:
			if (dwcc_bfs(&m, &msgbuf, &msgbuf_len)==-1) {
				goto error;
			}
			break;
		case DWCC_OPN:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_opn(&m, &r, p, ans_buf)==-1) {
				goto error;
			}
			break;
		case DWCC_CRE:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_cre(&m, p->dirname, &p->r)==-1) {
				goto error;
			}
			break;
		case DWCC_CLO:
			CHKSESID(m.h.session_id);
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: close session: %d\n", udate(), m.h.session_id);
#endif
			p = &dwdd[m.h.session_id];
			p->r.mode = 0;
			p->q.mode = 0;
			if (p->q.df) {
				clear_dwsched(p->q.df);
				free(p->q.df); /* all right? */
				p->q.df = NULL;
			}
/*
			(void)send_pkt_c(s, -1, DWCC_CLO | DWCC_ACK, NULL, 0);
*/
			p->in_use = 0;
			p->next = freelist;
			freelist = p;
			break;
		case DWCC_DIS:
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: disconnect\n", udate());
#endif
			goto end;
		case DWCC_VEC:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_vec(&m, &r, &p->q)==-1) {
				goto error;
			}
			break;
		case DWCC_FLU:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_flush(&p->r)==-1) {
				goto error;
			}
			break;
		case DWCC_WVE:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_wve(&m, &p->r, 0)==-1) {
				goto error;
			}
			continue;	/* no ack */
#if defined DMKXRC
		case DWCC_ATV:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_wve(&m, &p->r, 1)==-1) {
				goto error;
			}
			continue;	/* no ack */
#endif
		case DWCC_SVI:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_svi(s, &p->q)==-1) {
				goto error;
			}
			continue;	/* no ack */
		case DWCC_BLK:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_blk(&m, &p->r)==-1) {
				goto error;
			}
			continue;	/* no ack */
		case DWCC_SBI:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_sbi(s, &p->q)==-1) {
				goto error;
			}
			continue;	/* no ack */
		case DWCC_WSH:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_wsh(s, &p->q, m.u.data, m.h.len, msgbuf_len)==-1) {
				goto error;
			}
			continue;	/* ack is sent by dwcc_wsh */
		case DWCC_CM:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_cm(s, &p->q, m.u.data, m.h.len, msgbuf_len)==-1) {
				goto error;
			}
			continue;	/* ack is sent by dwcc_cm */
		case DWCC_MSK:
			CHKSESID(m.h.session_id);
			p = &dwdd[m.h.session_id];
			if (dwcc_msk(s, &p->q, m.u.data, m.h.len, msgbuf_len)==-1) {
				goto error;
			}
			break;
		default:
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: unknown command: cmd=%d\n", udate(), m.h.cmd);
#endif
			goto error;
		}
#if 0
		syslog(LOG_DEBUG, "reply: %x %p %d\n", m.h.cmd | DWCC_ACK, r.u.data, r.h.len);
#endif
		if (send_pkt_c(s, sid, m.h.cmd | DWCC_ACK, r.u.data, r.h.len)==-1) {
			syslog(LOG_DEBUG, "send_pkt_c failed\n");
			break;
		}
	}
error:
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: reply: %d\n", udate(), r.h.cmd);
#endif
	(void)send_pkt_c(s, -1, DWCC_NAK, NULL, 0);
end:
	for (i=0; i<nelems(dwdd); i++) {
		p = &dwdd[i];
		if (p->in_use && p->q.df) {
			clear_dwsched(p->q.df);
			free(p->q.df); /* all right? */
			p->q.df = NULL;
		}
	}
	free(msgbuf);
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: serve finish\n", udate());
#endif
}

static int
dwcc_hnd(m, r, q)
struct msg *m, *r;
struct rdxr *q;
{
	if (m->h.len != sizeof q->tmpsg) {
		return -1;
	}
	q->tmpsg = *(asig_t *)m->u.data;
	q->peersg = ntohl(q->tmpsg);
	q->mysg = asig();
	q->tmpsg = htonl(q->mysg);
	r->u.data = &q->tmpsg;
	r->h.len = sizeof q->tmpsg;
	return 0;
}

static int
dwcc_bfs(m, msgbuf, msgbuf_len)
struct msg *m;
char **msgbuf;
size_t *msgbuf_len;
{
	size_t len;
	pintcom_t cbfs;
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: set buffer size\n", udate());
#endif
	if (m->h.len != sizeof cbfs) {
		return -1;
	}
	if ((cbfs = *(pintcom_t *)m->u.data)<=0) {
		return -1;
	}
	CKOVFL(cbfs);
	len = cbfs;

	if (*msgbuf_len < len) {
		void *new;
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: bfs: msgbuf_len=%s, newlen=%s\n", udate(), sizetostr(*msgbuf_len), sizetostr(len));
#endif
		if (!(new = realloc(*msgbuf, len))) {
			syslog(LOG_DEBUG, "malloc: %s", strerror(errno));
			return -1;
		}
		*msgbuf = new;
		*msgbuf_len = len;
	}
	return 0;
}

static int
dwcc_opn(m, r, dwdd, ans_buf)
struct msg *m, *r;
struct dwdd *dwdd;
pintcom_t *ans_buf;
{
	int type, plain, cascade;
	size_t handle_len;
	ssize_t fragment;
	size_t ac;
	void *p;
	size_t psize;
	char const *handle;

#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: open\n", udate());
#endif
	p = m->u.data;

	if ((ac = unpack_ints_dtdd(p, m->h.len, &handle, "dtdd", &type, &fragment, &plain, &cascade)) != 4) {
		return -1;
	}
	psize = ptrdiff(handle, p);

	dwdd->q.type = type;
	dwdd->q.nodeid = fragment;
	dwdd->q.cascade = cascade;
/*
	handle = (char *)(p+ac);
*/
	if (m->h.len <= psize + 1) {
		return -1;
	}
	handle_len = m->h.len - psize - 1;

	if (handle_len <= 0 || handle[handle_len] != '\0') {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: open: handle termination check failed\n", udate());
#endif
		return -1;
	}
	if (strstr(handle, "..")) {
		return -1;
	}
	dwdd->q.sv = NULL;
	dwdd->q.df = NULL;

	/* dwdd->q.{handle,dirname} ha dwd_open_plain, dwd_open_{plain,and_fill} no naka de sansyou sarerunode, maemotte copy site oku */
	strlcpy(dwdd->q.handle, handle, sizeof dwdd->q.handle);
	strlcpy(dwdd->q.dirname, dwdd->dirname, sizeof dwdd->q.dirname);
	if (plain) {
		size_t size;

#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: open(plain): handle=[%s]\n", udate(), handle);
#endif
		if ((size = dwd_open_plain(type, fragment, &dwdd->q))==-1) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: open_plain failed\n", udate());
#endif
			return -1;
		}
		dwdd->q.cascade = 0;

		ans_buf[0] = size;
		ans_buf[1] = BLKSIZE;
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: open(plain): size=[%s]\n", udate(), sizetostr(size));
#endif

		r->h.len = 2 * sizeof ans_buf[0];
		r->u.data = ans_buf;
	}
	else {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: open: handle=[%s]\n", udate(), handle);
#endif
		/* dwd_open_and_fill ha, dxr sika hirakenai koto ni tyuui. (dcw ha open dekinai) */
		if (dwd_open_and_fill(type, fragment, &dwdd->q)==-1) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: fill_info failed\n", udate());
#endif
			return -1;
		}

		r->h.len = sizeof dwdd->q.h;
		r->u.data = &dwdd->q.h;

	}
	return 0;
}

static int
dwcc_cre(m, dirname, r)
struct msg *m;
char const *dirname;
struct remote *r;
{
	int type, mtime, plain, trc_vec, compress;
	size_t n, handle_len;
	ssize_t fragment;
	size_t ac;
	void *p;
	size_t psize;
	char const *handle;
	char path[MAXPATHLEN];
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: create\n", udate());
#endif

	p = m->u.data;

	if ((ac = unpack_ints_sddtddd(p, m->h.len, &handle, "sddtddd", &n, &type, &mtime, &fragment, &plain, &trc_vec, &compress)) != 7) {
syslog(LOG_DEBUG, "unpack ints failed\n");
		return -1;
	}
	psize = ptrdiff(handle, p);
/*
	if (n < 0) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: n is out of range: %s\n", udate(), sizetostr(n));
#endif
		return -1;
	}
*/
/*
	handle = (char *)(p+ac);
*/
	if (m->h.len <= psize + 1) {
syslog(LOG_DEBUG, "invalid psize\n");
		return -1;
	}
	handle_len = m->h.len - psize - 1;

	if (handle_len <= 0 || handle[handle_len] != '\0') {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: handle termination check error: %s\n", udate(), sizetostr(handle_len));
#endif
		return -1;
	}
	if (strstr(handle, "..")) {
syslog(LOG_DEBUG, "invalid path (..)\n");
		return -1;
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: create: handle=[%s]\n", udate(), handle);
syslog(LOG_DEBUG, "%s: dwd: create: n=%s, type=%d, fragment=%s, plain=%d\n", udate(), sizetostr(n), type, sizetostr(fragment), plain);
#endif
	if (dw_mkpathname(path, sizeof path, dirname, handle, type, fragment, 1)==-1) {
		syslog(LOG_DEBUG, "mkpathname failed");
		return -1;
	}
	if (dwd_create(path, n, type, mtime, fragment, plain, r, trc_vec, compress)==-1) {
syslog(LOG_DEBUG, "dwd_create failed\n");
		return -1;
	}
	return 0;
}

static int
dwcc_vec(m, r, q)
struct msg *m, *r;
struct rdxr *q;
{
	struct xr_vec const *v;
	idx_t id;
	size_t ac;
	void *p;
	int get_summary;
	ssize_t e;
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: vec\n", udate());
#endif
	p = m->u.data;
	if ((ac = unpack_ints_id(p, m->h.len, NULL, "id", &id, &get_summary)) != 2) {
		return -1;
	}
/*
	if (m->h.len != sizeof id) {
		return -1;
	}
*/
	if (q->mode != M_DXRR) {
		return -1;
	}
/*
	id = *(u_int *)m->u.data;
*/
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: vec: id=%s\n", udate(), idxtostr(id));
#endif
	if (get_summary) {
		q->v0.elem_num = dxr_elem_num(q->xr, id);
		q->v0.freq_sum = dxr_freq_sum(q->xr, id);
		v = &q->v0;
		r->h.len = sizeof_dxr_vec(0);
	}
	else {
		if ((e=dxr_get_vec(q->xr, id, &v))==-1) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: dxr_get_vec failed\n", udate());
#endif
			return -1;
		}
		r->h.len = sizeof_dxr_vec(e);
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: vec: elem_num=%s, len=%s\n", udate(), freqtostr(v->elem_num), sizetostr(r->h.len));
#endif
	r->u.data = (void *)v;
	return 0;
}

static int
dwcc_wve(m, r, atv)
struct msg *m;
struct remote *r;
{
	idx_t id;
	ssize_t fragment;
	size_t ac;
	void *p;
	size_t psize;
	struct xr_vec *v;

#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_vec: start.\n", udate());
#endif

	p = m->u.data;
	if ((ac = unpack_ints_it(p, m->h.len, &v, "it", &id, &fragment)) != 2) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_vec: !unpack_ints\n", udate());
#endif
		return -1;
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_vec: fragment=%s, atv=%d\n", udate(), sizetostr(fragment), atv);
#endif
	psize = ptrdiff(v, p);
/*
	v = (struct xr_vec *)(p+ac);
*/

	if (m->h.len < psize + sizeof_dxr_vec(0)) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_vec: short packet\n", udate());
#endif
		return -1;
	}

	if (fragment == WVE_WRITE || atv) {
		if (m->h.len != psize + sizeof_dxr_vec(v->elem_num)) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_vec: error! e\n", udate());
#endif
			return -1;
		}
	}
	else {
		if (m->h.len != psize + sizeof_dxr_vec(0)) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_vec: error! 0\n", udate());
#endif
			return -1;
		}
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_vec: id=%s, elem_num=%s\n", udate(), idxtostr(id), dftostr(v->elem_num));
#endif
#if defined DMKXRC
	if (atv) {
		if (dwd_add_trc_vec(id, v, fragment, r)==-1) {
			return -1;
		}
	} else
#endif
	if (dwd_write_vec(id, v, fragment, r)==-1) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_vec: !write_vec\n", udate());
#endif
		return -1;
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_vec: ok.\n", udate());
#endif
	return 0;
}

static int
dwcc_blk(m, r)
struct msg *m;
struct remote *r;
{
	size_t size;
	pintcom_t csize;
	char const *b;

	if (m->h.len < sizeof csize) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_blk: size error!: m->h.len = %s\n", udate(), sizetostr(m->h.len));
#endif
		return -1;
	}
	csize = *(pintcom_t *)m->u.data;
	CKOVFL(csize);
	size = csize;
	b = (char *)((pintcom_t *)m->u.data+1);
	if (m->h.len != sizeof csize + size) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_blk: size error!: m->h.len = %s, size=%s\n", udate(), sizetostr(m->h.len), sizetostr(size));
#endif
		return -1;
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: write_blk: size=%s\n", udate(), sizetostr(size));
#endif
	if (dwd_write_blk(b, size, r)==-1) {
		return -1;
	}
	return 0;
}

static ssize_t
dwd_open_plain(type, fragment, q)
struct rdxr *q;
ssize_t fragment;
{
	char path[MAXPATHLEN];
	FILE *f;
	ssize_t size;

	if (dw_mkpathname(path, sizeof path, q->dirname, q->handle, type, fragment, 0)==-1) {
		syslog(LOG_DEBUG, "mkpathname[dxr p] failed");
		return -1;
	}

	if (!(f = fopen(path, "r"))) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: fill_info: dwd_open_plain failed: %s\n", udate(), path);
#endif
		return -1;
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: fill_info: dwd_open_plain: %s => %p\n", udate(), path, f);
#endif
	if (fseeko(f, 0, SEEK_END)==-1) {
		return -1;
	}
	if ((size = ftello(f))==-1) {
		return -1;
	}
	rewind(f);
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: fill_info: dwd_open_plain: size = %s\n", udate(), sizetostr(size));
#endif

	q->f = f;
	q->size = size;
	q->mode = M_PLAINR;
	return size;
}

static int
dwd_open_and_fill(type, fragment, q)
struct rdxr *q;
ssize_t fragment;
{
	DWFILE *d;
	struct com_dxr_hdr *h;
	char path[MAXPATHLEN];
	int rtype;

	switch (type) {
	case DXR_ROW:
		rtype = DXR_COL;
		break;
	case DXR_COL:
		rtype = DXR_ROW;
		break;
	default:
		syslog(LOG_DEBUG, "dwd_open_and_fill: invalid type: %d", type);
		return -1;
	}

	if (dw_mkpathname(path, sizeof path, q->dirname, q->handle, type, fragment, 0)==-1) {
		syslog(LOG_DEBUG, "mkpathname[dxr] failed");
		return -1;
	}

	if (dw_mkpathname(q->qs_path, sizeof q->qs_path, q->dirname, q->handle, type, -1, 0)==-1) {
		syslog(LOG_DEBUG, "mkpathname[qs_path] failed");
		return -1;
	}

	if (dw_mkpathname(q->ds_path, sizeof q->ds_path, q->dirname, q->handle, rtype, -1, 0)==-1) {
		syslog(LOG_DEBUG, "mkpathname[ds_path] failed");
		return -1;
	}

	if (!(q->xr = dxr_open(path, q->handle))) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: fill_info: dxr_open failed: %s\n", udate(), path);
#endif
		return -1;
	}

	q->qs = q->ds = NULL;
	q->wsh_initialized = 0;
	d = q->xr;
	h = &q->h;
	memset(h, 0, sizeof *h);
	h->type = DWD_REMOTE;
	/*h->mtime = dxr_mtime(d);*/
	h->nentries = dxr_size(d);
	h->max_elem_num = dxr_max_elem_num(d);
	h->max_freq_sum = dxr_max_freq_sum(d);
	h->total_elem_num = dxr_total_elem_num(d);
	h->total_freq_sum = dxr_total_freq_sum(d);
	q->mode = M_DXRR;
	return 0;
}

static int
dwd_create(path, n, type, mtime, fragment, plain, r, trc_vec, compress)
size_t n;
char *path;
struct remote *r;
ssize_t fragment;
{
	void *addr;
	memset(r, 0, sizeof *r);

	if (!(r->stream = fopen(path, "w"))) {
		syslog(LOG_DEBUG, "%s: %s", path, strerror(errno));
		return -1;
	}

	if (plain) {
		r->mode = M_PLAINW;
	}
	else {
		struct xr_vec zero = ZEROVEC;
		size_t vec_offset;
		size_t entry_table_size;
		struct dxr_entry bi;
		struct dxr_hdr *h;
		r->compress = compress;
		h = &r->h;

		if (fill_dxr_hdr(h, type, mtime, fragment)==-1) {
			syslog(LOG_DEBUG, "fill_dxr_hdr failed");
			return -1;
		}
		h->nentries = n;
		h->max_elem_num = 0;
		h->max_freq_sum = 0;
		h->total_elem_num = 0;
		h->total_freq_sum = 0;
		h->compress_mode = r->compress;

		entry_table_size = (n+1)*sizeof bi;
		vec_offset = sizeof *h + entry_table_size;

		h->entry_table_offset = sizeof *h;
		h->vector_table_offset = vec_offset;

		if (fseeko(r->stream, vec_offset, SEEK_SET)==-1) {
			return -1;
		}

		if (fwrite(&zero, sizeof zero, 1, r->stream)!=1) {
			return -1;
		}
		r->addr = sizeof zero;

		addr = (void *)BI_BASE;
		if (addr) {
			r->bindex = mmalc((void*)addr, (n+1) * sizeof *r->bindex);
		}
		else {
			r->bindex = malloc((n+1) * sizeof *r->bindex);
		}

		if (!r->bindex) {
			syslog(LOG_DEBUG, "calloc: %s", strerror(errno));
			return -1;
		}

		bzero(r->bindex, (n+1) * sizeof *r->bindex);
#if defined DMKXRC
		r->trc_vec = trc_vec;
		if (trc_vec) {
			size_t i;
			for (i=1; i<=n; i++) {
				struct cel_t **p;

				p = (struct cel_t **)&r->bindex[i].addr;
				*p = NULL;
			}
		}
#endif

		r->mode = M_DXRW;
	}

	return 0;
}

static int
dwcc_flush(r)
struct remote *r;
{
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: dwcc_flush: start\n", udate());
#endif

	switch (r->mode) {
	case M_DXRW:
#if defined DMKXRC
		if (r->trc_vec) {
			if (dwcc_flush_trc_vec(r)==-1) {
syslog(LOG_DEBUG, "%s: dwd: dwcc_flush: dwcc_flush_trc_vec failed\n", udate());
				return -1;
			}
		}
#endif
		fwrite("\n#\n#\n#\n", 7, 1, r->stream);
		if (dwcc_make_summary_information(r) == -1) {
			return -1;
		}
		rewind(r->stream);
		if (fwrite(&r->h, sizeof r->h, 1, r->stream)!=1) {
syslog(LOG_DEBUG, "%s: dwd: dwcc_flush: fwrite(h) failed\n", udate());
			return -1;
		}
		if (fwrite(r->bindex, sizeof *r->bindex, r->h.nentries+1, r->stream)!=r->h.nentries+1) {
syslog(LOG_DEBUG, "%s: dwd: dwcc_flush: fwrite(r) failed\n", udate());
			return -1;
		}
		break;
	case M_PLAINW:
		/* do nothing */
		break;
	default:
		return -1;
	}

	if (fclose(r->stream)==EOF) {
syslog(LOG_DEBUG, "%s: dwd: dwcc_flush: fclose failed\n", udate());
		return -1;
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: dwcc_flush: finish\n", udate());
#endif
	return 0;
}

static int
dwd_write_vec(id, v, fragment, r)
idx_t id;
struct xr_vec *v;
struct remote *r;
ssize_t fragment;
{
/*
	struct dxr_hdr *h;
*/

	if (r->mode != M_DXRW) {
		return -1;
	}

	if (fragment == WVE_WRITE) {
		size_t size;
		if ((size = wvefc(v, r->stream, r->compress))==-1) {
			return -1;
		}
		r->bindex[id].addr = r->addr;
		r->addr += size;
	}
	else {
		r->bindex[id].addr = fragment;
	}
	r->bindex[id].elem_num = v->elem_num;
	r->bindex[id].freq_sum = v->freq_sum;

#if 0
	h = &r->h;
	if (h->max_elem_num < v->elem_num) {
		h->max_elem_num = v->elem_num;
	}
	if (h->max_freq_sum < v->freq_sum) {
		h->max_freq_sum = v->freq_sum;
	}
	h->total_elem_num += v->elem_num;
	h->total_freq_sum += v->freq_sum;
#endif
	return 0;
}

#if defined DMKXRC
static int
dwd_add_trc_vec(id, v, fragment, r)
idx_t id;
struct xr_vec *v;
struct remote *r;
ssize_t fragment;
{
	size_t i;
	if (v->elem_num) {
		if (fragment == WVE_WRITE) {
			assert(r->trc_vec);
			for (i=0; i<v->elem_num; i++) {
				idx_t j = v->elems[i].id;
				struct cel_t *cel, **p;

				p = (struct cel_t **)&r->bindex[j].addr;

				if (!(cel = new_cel())) {
					return -1;
				}
				cel->id = id;
				cel->freq = v->elems[i].freq;
				cel->n = *p;
				*p = cel;
			}
		}
		else {
			assert(!r->trc_vec);
			for (i=0; i<v->elem_num; i++) {
				idx_t j = v->elems[i].id;

				r->bindex[j].elem_num++;
				r->bindex[j].freq_sum += v->elems[i].freq;
if (r->bindex[j].elem_num < 0) {
	return -1;
}
if (r->bindex[j].freq_sum < 0) {
	return -1;
}
			}
		}
	}
	return 0;
}
#endif

static int
dwcc_make_summary_information(r)
struct remote *r;
{
	struct dxr_hdr *h;
	struct dxr_entry *b;
	idx_t id;
	h = &r->h;
	b = r->bindex;
	for (id = 1; id<=h->nentries; id++) {
		ssize_t elem_num = b[id].elem_num;
		freq_t freq_sum = b[id].freq_sum;
		if (h->max_elem_num < elem_num) {
			h->max_elem_num = elem_num;
		}
		if (h->max_freq_sum < freq_sum) {
			h->max_freq_sum = freq_sum;
		}
		h->total_elem_num += elem_num;
		h->total_freq_sum += freq_sum;
if (h->total_elem_num < 0) {
	return -1;
}
if (h->total_freq_sum < 0) {
	return -1;
}
	}
	return 0;
}

#if defined DMKXRC
static int
dwcc_flush_trc_vec(r)
struct remote *r;
{
	idx_t id;
	struct xr_vec *v = NULL;
	size_t vsize = 0;
	for (id = 1; id<=r->h.nentries; id++) {
		ssize_t i;
		size_t size;
		struct cel_t *q, **p;
		p = (struct cel_t **)&r->bindex[id].addr;
		for (q=*p, i=0; q; q = q->n, i++) ;
		if ((size = sizeof_dxr_vec(i))>vsize) {
			struct xr_vec *new;
			size = NALIGN(size, PAGESIZE);
			if (!(new = realloc(v, size))) {
				free(v);
/* syslog(LOG_DEBUG, "dwcc_flush_trc_vec: realloc failed\n"); */
				return -1;
			}
			v = new;
			vsize = size;
		}
		v->freq_sum = 0;
		for (q=*p, i=0; q; q = q->n, i++) {
if (i < 0) {
	return -1;
}
			v->elems[i].id = q->id;
			v->elems[i].freq = q->freq;
			v->freq_sum += q->freq;
if (v->freq_sum < 0) {
	return -1;
}
		}
if (i < 0) {
	return -1;
}
		v->elem_num = i;
		if (v->elem_num) {
			qsort(v->elems, v->elem_num, sizeof *v->elems, ecompar);
			if ((size = wvefc(v, r->stream, r->compress))==-1) {
/* syslog(LOG_DEBUG, "dwcc_flush_trc_vec: wvefc: failed\n"); */
				return -1;
			}
			r->bindex[id].addr = r->addr;
			r->addr += size;
		}
		else {
			r->bindex[id].addr = 0;
		}
		r->bindex[id].elem_num = v->elem_num;
		r->bindex[id].freq_sum = v->freq_sum;
	}
	free(v);
	return 0;
}
#endif

static int
dwcc_svi(s, q)
struct rdxr *q;
{
	idx_t id;
/*
	struct msg r;
*/

	if (q->mode != M_DXRR) {
		return -1;
	}

#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: dwcc_svi: %d\n", udate(), /*r.h.cmd*/0);
#endif
	if (send_pkt_c(s, -1, DWCC_SVI | DWCC_ACK, NULL, 0)==-1) {
		return -1;
	}

/*
	r.h.cmd = DWCC_SVR | DWCC_ACK;
*/
	for (id = 1; id<=q->h.nentries; id++) {
		struct xr_vec const *v;
		ssize_t e;
		if ((e = dxr_get_vec(q->xr, id, &v))==-1) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: dwcc_svi: dxr_get_vec failed\n", udate());
#endif
			return -1;
		}
#if 0
		r.h.len = sizeof_dxr_vec(e);
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: dwcc_svi: elem_num=%s, len=%s\n", udate(), dftostr(v->elem_num), sizetostr(r.h.len));
#endif
		r.u.data = v;
#endif
		if (send_pkt_c(s, -1, DWCC_SVR | DWCC_ACK, v, sizeof_dxr_vec(e))==-1) {
			return -1;
		}
	}
	return 0;
}

static int
dwd_write_blk(b, size, r)
char const *b;
size_t size;
struct remote *r;
{
	if (r->mode != M_PLAINW) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd_write_blk: failed: invalid magic\n", udate());
#endif
		return -1;
	}
	if (fwrite(b, 1, size, r->stream)!=size) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd_write_blk: fwrite failed\n", udate());
#endif
		return -1;
	}
	return 0;
}

static int
dwcc_sbi(s, q)
struct rdxr *q;
{
/*
	struct msg r;
*/
	size_t size, sz;
	char *buf;

#if defined XDEBUG
syslog(LOG_DEBUG, "%s: DWCC_SBI: start\n", udate());
#endif
	if (q->mode != M_PLAINR) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: DWCC_SBI: error: !M_PLAINR\n", udate());
#endif
		return -1;
	}

	if (!(buf = malloc(BLKSIZE))) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: DWCC_SBI: error: !malloc\n", udate());
#endif
		return -1;
	}

/*
	r.h.len = 0;
	r.h.cmd = DWCC_SBI | DWCC_ACK;
*/
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: dwcc_sbi: %d\n", udate(), /*r.h.cmd*/0);
#endif
	if (send_pkt_c(s, -1, DWCC_SBI | DWCC_ACK, NULL, 0)==-1) {
		return -1;
	}

/*
	r.h.cmd = DWCC_SBR | DWCC_ACK;
*/
	rewind(q->f);
	for (size = 0; size<q->size; size += sz) {
		if ((sz = fread(buf, 1, BLKSIZE, q->f))<=0) {
			syslog(LOG_DEBUG, "DWCC_SBI: fread: %p, %s", q->f, strerror(errno));
			break;
		}
/*
		r.h.len = sz;
		r.u.data = buf;
*/
		if (send_pkt_c(s, -1, DWCC_SBR | DWCC_ACK, buf, sz)==-1) {
			return -1;
		}
	}
	free(buf);
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: DWCC_SBI: finish\n", udate());
#endif
	return 0;
}

static int
dwcc_wsh(s, q, v, size, bufsize)
struct rdxr *q;
void *v;
size_t size, bufsize;
{
	struct timeval elapsed;
	size_t ac;
	void *p;
	size_t psize, osize;
	int type, prob_mode;
	ssize_t n, nq, total, tn, sum_i;
	struct syminfo *query;

	struct syminfo *d = NULL;

	pintcom_t o[4];
	void *e;
	struct iovec iov[2];

	struct timeval start;
	int tv_sec, tv_usec;
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh\n", udate());
#endif

	if (q->mode != M_DXRR) {
		goto fail;
	}

	gettimeofday(&start, NULL);

	if (!q->wsh_initialized) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: wsh_init\n", udate());
#endif
		if (dwcc_wsh_init(q)==-1) {
			goto fail;
		}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: cascade=%d\n", udate(), q->cascade);
#endif
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh: 2\n", udate());
#endif

	p = v;

	if ((ac = unpack_ints_jdjjjd(p, size, &query, "jdjjjd", &nq, &type, &n, &tn, &sum_i, &prob_mode)) != 6) {
		goto fail;
	}
/*
	query = (struct syminfo *)(p+ac);
*/
	psize = ptrdiff(query, p);
#if defined XDEBUG
syslog(LOG_DEBUG, "dwcc_wsh: %s %d %s %s %s %d\n",
	ssizetostr(nq),
	type,
	ssizetostr(n),
	ssizetostr(tn),
	ssizetostr(sum_i),
	prob_mode);
#endif

#if defined SSI
	if (!((char *)(query+nq) <= (char *)v + bufsize)) {
		syslog(LOG_DEBUG, "dwcc_wsh assertion failed: buffer overrun");
		goto fail;
	}
	unpack_ssi(query, (struct s_syminfo *)query, nq);
#endif
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh: 3\n", udate());
#endif

#if defined SSI
	if (nq != (size - psize) / sizeof (struct s_syminfo)) {
#else
	if (nq != (size - psize) / sizeof *query) {
#endif
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh: invalid nq (nq=%s, size=%s)\n", udate(), dftostr(nq), sizetostr(size));
#endif
		goto fail;
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh: nq=%s, type=%d, n=%s\n", udate(), dftostr(nq), type, sizetostr(n));
syslog(LOG_DEBUG, "%s: dwcc_wsh: 4\n", udate());
#endif

	if (q->cascade) {
		d = rdwsh(query, nq, q->sv, q->np, type, &n, tn, &total, sum_i, prob_mode, 0);
	}
	else {
		d = dw_wsh(query, nq, q->xr, q->qs, q->ds, type, &n, &total);
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh: 5\n", udate());
#endif

	if (!d) {
		goto fail;
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh: 6\n", udate());
#endif

	if (sum_i) {
/*
		struct msg m, r;
*/
		struct msg m;
		ssize_t nsum;
		float *sum;
		ssize_t rqn;
		pintcom_t crqn;

		if ((nsum = mksum(d, n, &sum, sum_i))==-1) {
			goto fail;
		}

/*
		r.h.cmd = DWCC_WSHS | DWCC_ACK;
		r.h.len = nsum * sizeof *sum;
		r.u.data = sum;
*/
		if (send_pkt_c(s, -1, DWCC_WSHS | DWCC_ACK, sum, nsum * sizeof *sum)==-1) {
			goto fail;
		}

		m.h.cmd = 0;
		m.u.data = &crqn;
		m.u.size = sizeof crqn;
		if (recv_pkt_c(s, DWCC_RQN, &m)==-1) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: recv_pkt_c failed: cmd=%d\n", udate(), m.h.cmd);
#endif
			goto fail;
		}
		if (m.h.len != sizeof crqn) {
			goto fail;
		}
		CKOVFL(crqn);
		rqn = crqn;
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: rqn=%s\n", udate(), dftostr(rqn));
#endif

		if (n > rqn) {
			n = rqn;
		}
		free(sum);
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh: 7\n", udate());
#endif

	gettimeofday(&elapsed, NULL);
	timersub(&elapsed, &start, &elapsed);

	tv_sec = elapsed.tv_sec;
	tv_usec = elapsed.tv_usec;

	if ((ac = pack_ints_jjdd(o, sizeof o, &e, "jjdd", &n, &total, &tv_sec, &tv_usec)) != 4) {
		goto fail;
	}
	osize = ptrdiff(e, o);
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh: 8\n", udate());
#endif
#if defined SSI
	pack_ssi((struct s_syminfo *)d, d, n);
#endif
	iov[0].iov_base = (char *)d;
#if defined SSI
	iov[0].iov_len = n * sizeof (struct s_syminfo);
#else
	iov[0].iov_len = n * sizeof *d;
#endif
	iov[1].iov_base = (char *)o;
	iov[1].iov_len = osize;
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh: 9\n", udate());
#endif

	if (send_pkt_v(s, -1, DWCC_WSH | DWCC_ACK, iov, 2)==-1) {
		goto fail;
	}

	free(d);
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh: finish\n", udate());
#endif
	return 0;

fail:
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_wsh failed\n", udate());
#endif
	free(d);
	return -1;
}

static int
dwcc_wsh_init(q)
struct rdxr *q;
{
	if (q->wsh_initialized) {
		return -1;
	}

	if (wsh_init()==-1) {
		return -1;
	}

	q->wsh_initialized = 1;

#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: fill_info: dxr_open: %s\n", udate(), q->qs_path);
#endif
	if (!(q->qs = dxr_open(q->qs_path, q->handle))) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: fill_info: dxr_open failed: %s\n", udate(), q->qs_path);
#endif
		q->mode = 0;
		return -1;
	}

#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: fill_info: dxr_open: %s\n", udate(), q->ds_path);
#endif
	if (!(q->ds = dxr_open(q->ds_path, q->handle))) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwd: fill_info: dxr_open failed: %s\n", udate(), q->ds_path);
#endif
		q->mode = 0;
		return -1;
	}

	if (q->cascade) {
		if (cascade_init(q)==-1) {
			q->mode = 0;
			return -1;
		}
	}

	return 0;
}

static int
cascade_init(q)
struct rdxr *q;
{
	struct qqs *qq;
	struct rcxs *rc;

#if defined XDEBUG
syslog(LOG_DEBUG, "%s: reading schedule\n", udate());
#endif
	if (!(q->df = calloc(1, sizeof *q->df))) {
		syslog(LOG_DEBUG, "calloc: %s", strerror(errno));
		return -1;
	}
	if (read_dwsched(q->df, q->dirname, q->handle)) {
		char cwd[MAXPATHLEN];
		syslog(LOG_DEBUG, "read_dwsched failed: %s, %s", q->dirname, q->handle);
		getcwd(cwd, sizeof cwd);
		syslog(LOG_DEBUG, "read_dwsched failed: cwd=%s", cwd);
		return -1;
	}

#if defined XDEBUG
syslog(LOG_DEBUG, "%s: nodeid=%s\n", udate(), sizetostr(q->nodeid));
syslog(LOG_DEBUG, "%s: initializing sym_vec\n", udate());
#endif
	if (!(qq = calloc(1, sizeof *qq))) {
		syslog(LOG_DEBUG, "calloc: %s", strerror(errno));
		return -1;
	}
	qq->xr = q->xr;
	qq->qs = q->qs;
	qq->ds = q->ds;
	if (q->type == DXR_ROW) {
		rc = &q->df->rcx[0];
	}
	else if (q->type == DXR_COL) {
		rc = &q->df->rcx[1];
	}
	else {
		syslog(LOG_DEBUG, "unknown type: %d", q->type);
		return -1;
	}
	if ((q->np = makesv(q->nodeid, rc->np, rc->rems, &q->sv, qq, 0))==-1) {
		syslog(LOG_DEBUG, "makesv: %s", strerror(errno));
		return -1;
	}

	if (q->np==1) {	/* no child. back to single mode */
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: cascade off\n", udate());
#endif
		zfree(q->sv);
		q->cascade = 0;
	}

	return 0;
}

static int
dwcc_cm(s, q, v, size, bufsize)
struct rdxr *q;
void *v;
size_t size, bufsize;
{
	size_t ac;
	void *p;
/*	size_t psize;*/
	ssize_t nwords, ndocs;
	struct syminfo *words, *docs;
	ssize_t nwd;

	ssize_t **d = NULL;

	struct iovec iov[1];

#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_cm\n", udate());
#endif

	if (q->mode != M_DXRR) {
		goto fail;
	}

	if (!q->wsh_initialized) {
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: wsh_init\n", udate());
#endif
		if (dwcc_wsh_init(q)==-1) {
			goto fail;
		}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: cascade=%d\n", udate(), q->cascade);
#endif
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_cm: 2\n", udate());
#endif

	p = v;

	if ((ac = unpack_ints_jj(p, size, &words, "jj", &nwords, &ndocs)) != 2) {
		goto fail;
	}
/*
	words = (struct syminfo *)(p+ac);
*/
/*	psize = ptrdiff(words, p);*/
	docs = words + nwords;
	nwd = nwords + ndocs;
#if defined SSI
	if (!((char *)(words+nwd) <= (char *)v + bufsize)) {
		syslog(LOG_DEBUG, "dwcc_cm assertion failed: buffer overrun");
		goto fail;
	}
	unpack_ssi(words, (struct s_syminfo *)words, nwd);
#endif
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_cm: 3\n", udate());
#endif

#if defined SSI
	if (nwd != (size - ac * sizeof *p) / sizeof (struct s_syminfo)) {
#else
	if (nwd != (size - ac * sizeof *p) / sizeof *words) {
#endif
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_cm: invalid nq (nq=%s, size=%s)\n", udate(), sizetostr(nwords+ndocs), sizetostr(size));
#endif
		goto fail;
	}
#if defined XDEBUG
/*syslog(LOG_DEBUG, "%s: dwcc_cm: nq=%d, type=%d, n=%d\n", udate(), nq, type, n);*/
syslog(LOG_DEBUG, "%s: dwcc_cm: 4\n", udate());
#endif

	if (q->cascade) {
		d = rdwcm(words, nwords, docs, ndocs, q->sv, q->np, 0);
	}
	else {
		d = dw_cm(words, nwords, docs, ndocs, q->xr);
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_cm: 5\n", udate());
#endif

	if (!d) {
		goto fail;
	}
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_cm: 6\n", udate());
#endif

	iov[0].iov_base = (char *)&d[0][0];
	iov[0].iov_len = nwords * nwords * sizeof **d;
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_cm: 7\n", udate());
#endif

	if (send_pkt_v(s, -1, DWCC_CM | DWCC_ACK, iov, 1)==-1) {
		goto fail;
	}

	free(d);
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_cm: finish\n", udate());
#endif
	return 0;

fail:
#if defined XDEBUG
syslog(LOG_DEBUG, "%s: dwcc_cm failed\n", udate());
#endif
	free(d);
	return -1;
}

static int
dwcc_msk(s, q, v, size, bufsize)
struct rdxr *q;
void *v;
size_t size, bufsize;
{
	size_t ac;
	void *p;
	size_t psize;
	size_t plen, nlen;
	idx_t *pmask, *nmask;
	int ispmask, isnmask;
	idx_t dummy;

	if (q->mode != M_DXRR) {
		return -1;
	}

	if (!q->wsh_initialized) {
		if (dwcc_wsh_init(q)==-1) {
			return -1;
		}
	}

	if (q->cascade) {
		return -1;
	}

	p = v;

	if ((ac = unpack_ints_ssdd(p, size, &pmask, "ssdd", &plen, &nlen, &ispmask, &isnmask)) != 4) {
		return -1;
	}
	psize = ptrdiff(pmask, p);
	nmask = pmask + plen;
	if (plen == 0) {
		pmask = ispmask ? &dummy : NULL;
	}
	if (nlen == 0) {
		nmask = isnmask ? &dummy : NULL;
	}

	if (plen + nlen != (size - psize) / sizeof *pmask) {
		return -1;
	}

	if (dxr_set_mask(q->xr, pmask, plen, nmask, nlen)==-1) {
		return -1;
	}

	return 0;
}
