/*
 *	Copyright (C) 2001 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dwsched.c,v 1.4 2007/03/29 02:10:01 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <errno.h>
#if defined HAVE_LIBGEN_H
#include <libgen.h>
#endif
#include <netdb.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dwerrno.h>
#include <geta/dxr.h>
#include <geta/dwc.h>
#include <geta/dwsched.h>

static void clear_rcxs(struct rcxs *);
static int parse_rcxs_entry(struct rcxsp *, char *, ssize_t, int, int, char const *, char const *, int);
static int read_rcxs(FILE *, struct rcxs *, int, char const *, char const *, int, int);
static int make_prob_list_0(struct rcxs *, int);
static int scan_line(char *, char *, size_t, char *, size_t, freq_t *, freq_t *);
static char *fgets_chop(char *, size_t, FILE *);
#if ! defined HAVE_BASENAME
static char *basename(const char *);
#endif

int
read_dwsched(d, dirname, handle)
struct dwsched *d;
char const *dirname, *handle;
{
	FILE *f;
	char buf[8];
	int use_total = 0;
	char base[MAXPATHLEN];

	strlcpy(d->handle, handle, sizeof d->handle);
	dw_mkpathname(d->path, sizeof d->path, dirname, handle, DW_SCHE, 0, 0);

	if (!(f = fopen(d->path, "r"))) {
		fprintf(stderr, "%s: %s", d->path, strerror(errno));
		dwerrno = errno;
		return -1;
	}

	if (!fgets_chop(buf, sizeof buf, f)) {
		fprintf(stderr, "!local/remote");
		return -1;
	}
	if (!strcmp(buf, "L")) {
		d->local = 1;
	}
	else if (!strcmp(buf, "N")) {
		d->local = 0;
	}
	else {
		fprintf(stderr, "!local/remote");
		dwerrno = DWE_BADF;
		return -1;
	}

	if (!fgets_chop(base, sizeof base, f)) {
		fprintf(stderr, "!r.cw_path");
		return -1;
	}
	dw_mkpathname(d->rcx[0].cw_path, sizeof d->rcx[0].cw_path, dirname, handle, DCW_ROW, 0, 0);
	if (strcmp(base, basename(d->rcx[0].cw_path))) {
		fprintf(stderr, "!%s.r.cw_path (basename mismatch)", handle);
		return -1;
	}

	if (!fgets_chop(base, sizeof base, f)) {
		fprintf(stderr, "!c.cw_path");
		return -1;
	}
	dw_mkpathname(d->rcx[1].cw_path, sizeof d->rcx[1].cw_path, dirname, handle, DCW_COL, 0, 0);
	if (strcmp(base, basename(d->rcx[1].cw_path))) {
		fprintf(stderr, "!%s.c.cw_path (basename mismatch)", handle);
		return -1;
	}

	if (read_rcxs(f, &d->rcx[0], d->local, dirname, d->handle, DXR_ROW, use_total)==-1) {
		return -1;
	}

	if (read_rcxs(f, &d->rcx[1], d->local, dirname, d->handle, DXR_COL, use_total)==-1) {
		return -1;
	}

	fclose(f);
	return 0;
}

void
clear_dwsched(d)
struct dwsched *d;
{
	clear_rcxs(&d->rcx[0]);
	clear_rcxs(&d->rcx[1]);
}

static void
clear_rcxs(r)
struct rcxs *r;
{
	size_t i, c;
	struct rcxsp *rp;
	rp = &r->s;
	if (rp->f) {
		dwc_close(rp->f);
		rp->f = NULL;
		rp->s = -1;
	}
	for (c=0; c<2; c++) {
		if (!r->rems[c]) {
			continue;
		}
		for (i=0; i<r->np; i++) {
			rp = &r->rems[c][i];
			if (rp->f) {
				dwc_close(rp->f);
				rp->f = NULL;
				rp->s = -1;
			}
		}
		free(r->rems[c]);
	}
}

static int
parse_rcxs_entry(rp, buf, p, c, local, dirname, handle, type)
struct rcxsp *rp;
char *buf;
ssize_t p;
char const *dirname, *handle;
{
	if (scan_line(buf, rp->host, sizeof rp->host, rp->serv, sizeof rp->serv, &rp->total_elem_num, &rp->elem_num)==-1) {
		return -1;
	}
	if (local) {
		char buf2[MAXPATHLEN];
		strlcpy(buf2, rp->host, sizeof buf2); /* zitu ha hituyou nai. */
		dw_mkpathname(rp->host, sizeof rp->host, dirname, handle, type, p, 0);
		if (strcmp(buf2, basename(rp->host))) {
			fprintf(stderr, "!%s.%d.%s.path (basename mismatch)", handle, type, p == -1 ? "s" : sizetostr(p));
			return -1;
		}
	}
	rp->i.fragment = p;
	rp->i.handle = handle;
	rp->i.type = type;
	rp->i.cascade = c;
	rp->p = 0;
	rp->f = NULL;
	rp->s = -1;
	return 0;
}

static int
read_rcxs(f, r, local, dirname, handle, type, use_total)
FILE *f;
struct rcxs *r;
char const *dirname, *handle;
{
	char buf[MAXPATHLEN];
	char *endp;
	size_t p;
	struct rcxsp *rp;

	r->rems[0] = r->rems[1] = NULL;

	if (!fgets_chop(buf, sizeof buf, f)) {
		fprintf(stderr, "!np");
		goto error;
	}
	r->np = strtosize(buf, &endp, 10);
	if (!(*buf && *endp=='\0')) {
		fprintf(stderr, "!np(strtosize)");
		dwerrno = DWE_BADF;
		goto error;
	}
	if (!fgets_chop(buf, sizeof buf, f)) {
		fprintf(stderr, "!s.host");
		goto error;
	}
	rp = &r->s;
	if (parse_rcxs_entry(rp, buf, -1, -1, local, dirname, handle, type) == -1) {
		goto error;
	}
	if (!(r->rems[0] = calloc(r->np, sizeof *r->rems[0]))) {
		fprintf(stderr, "calloc rems[0]: %s", strerror(errno));
		dwerrno = errno;
		goto error;
	}
	for (p=0; p<r->np; p++) {
		if (!fgets_chop(buf, sizeof buf, f)) {
			fprintf(stderr, "!rems[0][%s].host", sizetostr(p));
			goto error;
		}
		rp = &r->rems[0][p];
		if (parse_rcxs_entry(rp, buf, p, 0, local, dirname, handle, type) == -1) {
			goto error;
		}
	}
	if (make_prob_list_0(r, use_total)==-1) {
		goto error;
	}
	if (!(r->rems[1] = calloc(r->np, sizeof *r->rems[1]))) {
		fprintf(stderr, "calloc rems[1]: %s", strerror(errno));
		dwerrno = errno;
		goto error;
	}
	memmove(r->rems[1], r->rems[0], r->np * sizeof *r->rems[1]);
	for (p=0; p<r->np; p++) {
		rp = &r->rems[1][p];
		rp->i.cascade = 1;
	}
	return 0;

error:

	free(r->rems[0]);
	free(r->rems[1]);
	r->rems[0] = r->rems[1] = NULL;
	return -1;
}

static int
make_prob_list_0(r, use_total)
struct rcxs *r;
{
	freq_t itotal;
	double total;
	size_t p;
	freq_t tally;

	if (r->np==0) {
		return 0;
	}

	r->s.p = 1.0;

	itotal = use_total?r->s.total_elem_num:r->s.elem_num;

	if (itotal==0) {
		double prob = 1.0 / r->np;
		for (p=0; p<r->np; p++) {
			r->rems[0][p].p = prob;
		}
		return 0;
	}
	total = itotal;
	tally = 0;
	for (p=0; p<r->np; p++) {
		struct rcxsp *rp = &r->rems[0][p];
		freq_t par;
		par = use_total?rp->total_elem_num:rp->elem_num;
		tally += par;
		rp->p = par/total;
	}
	if (tally != itotal) {
		fprintf(stderr, "inconsistency found in schedule file: tally error\n");
		dwerrno = DWE_BADF;
		return -1;
	}
	return 0;
}

static int
scan_line(buf, host, host_size, serv, serv_size, total_elem_num, elem_num)
char *buf;
char *host;
size_t host_size, serv_size;
char *serv;
freq_t *total_elem_num, *elem_num;
{
	char *p, *q, *e;

	q = buf;
	if (!buf || !*buf) {
		fprintf(stderr, __FILN__ ": !buf");
		dwerrno = DWE_BADF;
		return -1;
	}
	if (q = index(q, ' ')) {
		*q++ = '\0';
	}
	strlcpy(host, buf, host_size);

	p = q;
	if (!p || !*p) {
		fprintf(stderr, __FILN__ ": !serv");
		dwerrno = DWE_BADF;
		return -1;
	}
	if (q = index(q, ' ')) {
		*q++ = '\0';
	}
	strlcpy(serv, p, serv_size);

	p = q;
	if (!p || !*p) {
		fprintf(stderr, __FILN__ ": !total_elem_num");
		dwerrno = DWE_BADF;
		return -1;
	}
	if (q = index(q, ' ')) {
		*q++ = '\0';
	}
	*total_elem_num = strtofreq(p, &e, 10);
	if (*e) {
		fprintf(stderr, __FILN__ ": parse_error: total_elem_num is not integer");
		return -1;
	}

	p = q;
	if (!p || !*p) {
		fprintf(stderr, __FILN__ ": !elem_num");
		dwerrno = DWE_BADF;
		return -1;
	}
	if (q = index(q, ' ')) {
		fprintf(stderr, "trailing garbage ignored\n");
		*q++ = '\0';
	}
	*elem_num = strtofreq(p, &e, 10);
	if (*e) {
		fprintf(stderr, __FILN__ ": parse_error: total_elem_num is not integer");
		return -1;
	}
	return 0;
}

static char *
fgets_chop(str, size, stream)
char *str;
size_t size;
FILE *stream;
{
	size_t l;
	if (!fgets(str, size, stream)) {
		dwerrno = DWE_BADF;
		return NULL;
	}
	l = strlen(str);
	if (l==0 || str[l-1]!='\n') {
		dwerrno = DWE_BADF;
		return NULL;
	}
	str[l-1] = '\0';
	return str;
}

#if ! defined HAVE_BASENAME
static char *
basename(p)
const char *p;
{
	char *r;
	if (r = rindex(p, '/')) {
		return r+1;
	}
	return p;
}
#endif
