/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: dwsys.c,v 1.2 2007/03/21 02:19:52 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>
#include <geta/dxr.h>

#include "dwsys.h"
#include "dcwP.h"
#include "dxrP.h"
#include "dauiP.h"

int
fill_dcw_hdr(h, type, mtime)
struct dcw_hdr *h;
{
	memset(h, 0, sizeof *h);
	switch (type) {
	case DCW_ROW:
		strcpy(h->ident, "!dcw.row");
		break;
	case DCW_COL:
		strcpy(h->ident, "!dcw.col");
		break;
	default:
		errno = EINVAL;
		return -1;
	}
	h->type = type;
	h->version = DCW_VERSION;
	h->mtime = mtime;
	return 0;
}

int
fill_dxr_hdr(h, type, mtime, fragment)
struct dxr_hdr *h;
ssize_t fragment;
{
	memset(h, 0, sizeof *h);
	switch (type) {
	case DXR_ROW:
		if (fragment==-1) {
			sprintf(h->ident, "!dxr.row.s");
		}
		else {
			sprintf(h->ident, "!dxr.row.%s", sizetostr(fragment));
		}
		break;
	case DXR_COL:
		if (fragment==-1) {
			sprintf(h->ident, "!dxr.col.s");
		}
		else {
			sprintf(h->ident, "!dxr.col.%s", sizetostr(fragment));
		}
		break;
	default:
		errno = EINVAL;
		return -1;
	}
	h->type = type;
	h->version = DXR_VERSION;
	h->mtime = mtime;
	h->fragment = fragment;
	return 0;
}

int
fill_daui_hdr(h, mtime, fragment)
struct daui_hdr *h;
ssize_t fragment;
{
	memset(h, 0, sizeof *h);
	strcpy(h->ident, "!daui");
	h->type = 0;
	h->version = DAUI_VERSION;
	h->mtime = mtime;
	h->fragment = fragment;
	return 0;
}

int
ecompar(va, vb)
void const *va, *vb;
{
	struct xr_elem const *a, *b;
	a = va;
	b = vb;
	if (a->id < b->id) {
		return -1;
	}
	else if (a->id > b->id) {
		return 1;
	}
	return 0;
}

void
refresh_mem(p, size)
void *p;
size_t size;
{
	size_t i;
	char *q;
	int v = 0;
	for (i = 0; i < size; i += 32) {
		q = (char *)p + i;
		v += *q;
	}
}
