/*
 *	Copyright (C) 2001 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: estimz.c,v 1.7 2007/03/24 02:52:58 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <assert.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwsh.h>
#include <geta/dwwt.h>
#include <geta/dwc.h>
#include <geta/dwam.h>

static int foo(char const *, char const *, int, ssize_t, int, char const *);
static int mkDF(char const *, char const *, int);
static void try_all_partitioning(ssize_t, size_t *, ssize_t);
static void check_memory_usage(size_t, size_t *);
static void printp(size_t, size_t *);

/*
 kono program ha, 
  dxrr.0, dxrc.0 wo kara, sorewo saidai gen ni assyuku surutameno,
  assyuku no tameno, bit patter wo keisan suru.

 usage: dwam/`handle'/dxrc.0 no aru directory wo GETAROOT to site,
   $ estimz `handle'
   wo zikkou suru.
*/

freq_t nDF;	/* XXX: freq_t dato guai ga warui? */
size_t *DF;	/* zitu ha ruiseki DF ga haitte iru */

#define	MAXNPART	8

struct {
	size_t i;
	size_t men;
	size_t dbits[MAXNPART];
	size_t nbits[MAXNPART];
	size_t range[MAXNPART];
	size_t df[MAXNPART];
} min;

int
main(argc, argv)
char *argv[];
{
	int type;
	ssize_t frag;
	char const *dirname, *handle;

	if (argc!=2) {
		fprintf(stderr, "usage: %s handle\n", argv[0]);
		return 1;
	}

	if (!(dirname = getenv(PROOTENV))) {
		fprintf(stderr, "environment variable " PROOTENV " is not set");
		return 1;
	}
	frag = 0;
	handle = argv[1];
	printf("%s %s\n", dirname, handle);

	type = DXR_ROW;
	foo(dirname, handle, type, frag, 1, "XR FREQ");

	type = DXR_ROW;
	foo(dirname, handle, type, frag, 0, "XR ROW ID");

	type = DXR_COL;
	foo(dirname, handle, type, frag, 0, "XR COL ID");

	return 0;
}

static int
foo(dirname, handle, type, frag, mode, clas)
char const *dirname, *handle, *clas;
ssize_t frag;
{
	size_t j;
	char path[MAXPATHLEN];
	size_t dbits[MAXNPART];

	printf("%s\n", clas);

	dw_mkpathname(path, sizeof path, dirname, handle, type, frag, 0);

	if (mkDF(path, handle, mode)==-1) {
		return -1;
	}

	min.men = ~0;
	try_all_partitioning(0, dbits, -1);

	printp(min.i, min.dbits);
	for (j=0; j<min.i; j++) {
		printf("%s/%s/%s,", sizetostr(min.range[j]), sizetostr(min.nbits[j]), sizetostr(min.df[j]));
	}
	printf("\n");
	printf("men=%s (byte)\n", sizetostr(min.men/8));

	free(DF);
	DF=NULL;
	nDF = 0;
	return 0;
}

static void
try_all_partitioning(i, dbits, p)
ssize_t i, p;
size_t *dbits;
{
	ssize_t j, s;
	if (i<MAXNPART) {
		if (p<=0) {
			s = p+1;
		}
		else {
			s = p;
		}
		for (j=s; j<32; j++) {
			dbits[i] = j;
			try_all_partitioning(i+1, dbits, j);
		}
		dbits[i] = j;
		check_memory_usage(i+1, dbits);
	}
}

static void
check_memory_usage(i, dbits)
size_t i;
size_t *dbits;
{
	size_t men;
	size_t v = 1;
	size_t j;
	size_t range[MAXNPART];	/* range of each partition */
	size_t nbits[MAXNPART];	/* # of bits which belongs to the partition */
	size_t df[MAXNPART];	/* df for each partition */

	for (j=0; j<i; j++) {
		int lastp = j==i-1;
		size_t k;
		size_t l, n;
		size_t ib = j+1;
		size_t db = dbits[j];
		if (lastp) {
			nbits[j] = ib-1+db;
			range[j] = ~0;
			df[j] = DF[nDF];
		}
		else {
			size_t ll;
			for (n=1, k=0; k<db; k++, n*=2) ;
			l = v+n;
			nbits[j] = ib+db;
			range[j] = l;
			ll = l-1;
			if (ll<nDF) {
				df[j] = DF[ll];
			}
			else {
				df[j] = DF[nDF];
			}
			v = l;
		}
	}
	for (j=i-1; j>0; j--) {
		df[j] -= df[j-1];
	}
	for (men=0, j=0; j<i; j++) {
		men += df[j] * nbits[j];
	}
	if (men<min.men) {
		min.i = i;
		min.men = men;
		memcpy(min.dbits, dbits, sizeof min.dbits);
		memcpy(min.range, range, sizeof min.range);
		memcpy(min.nbits, nbits, sizeof min.nbits);
		memcpy(min.df, df, sizeof min.df);
	}
}

static void
printp(i, dbits)
size_t i;
size_t *dbits;
{
	size_t j;
	for (j=0; j<i; j++) {
		printf("%s, ", sizetostr(dbits[j]));
	}
	printf("-1\n");
}

static int
mkDF(path, handle, mode)
char const *path, *handle;
{
	idx_t id;
	ssize_t ne;
	DWFILE *f;
	size_t i;

	if (!(f = dxr_open(path, handle))) {
		nperror(path);
		return -1;
	}

	ne = dxr_size(f);
	nDF = 0;
	for (id = 1; id<=ne; id++) {
		ssize_t elem_num;
		size_t i;
		struct xr_vec const *v;
		if ((elem_num = dxr_get_vec(f, id, &v))>0) {
			for (i=0; i<v->elem_num; i++) {
				struct xr_elem const *p = &v->elems[i];
				freq_t f;
				f = mode?p->freq:p->id;
				if (nDF < f) {
					nDF = f;
				}
			}
		}
	}
	printf("nDF=%s\n", sizetostr(nDF));
	if (!(DF = calloc(nDF+1, sizeof *DF))) {
		nperror("calloc");
		return -1;
	}
	for (id = 1; id<=ne; id++) {
		ssize_t elem_num;
		size_t i;
		struct xr_vec const *v;
		if ((elem_num = dxr_get_vec(f, id, &v))>0) {
			for (i=0; i<v->elem_num; i++) {
				struct xr_elem const *p = &v->elems[i];
				size_t f = mode?p->freq:p->id;
				DF[f]++;
			}
		}
	}
	dxr_close(f);

	/* ruiseki ni henkan */
	for (i=1; i<=nDF; i++) {
		DF[i] = DF[i-1]+DF[i];
	}

	return 0;
}
