/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: wam.xs,v 1.10 2007/03/29 07:37:38 nis Exp $"; */

#include <config.h>

#include <sys/types.h>

#include <stdlib.h>

#undef JIS_WAM

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include <geta/dtypes.h>
#include <geta/dwut.h>
#include <geta/dw.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/daui.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwsched.h>
#include <geta/dci.h>
#include <geta/dwam.h>

typedef struct syminfo syminfo;

static SV *vec2rh(WAM *, struct xr_vec const *, size_t, int);
static SV *s2rah(syminfo *, size_t, WAM *, int, int);
static syminfo *rah2s(AV *, size_t *);
static char **ra2av(AV *);
static int xs_wam_open(char *);
static int xs_wam_close(int);
static WAM *xs_wam_desc(int);
static size_t iwl(syminfo *, size_t, size_t, WAM *, int);
static SV *wshb0(SV *, SV *, int, int, int, size_t, size_t, int);
static SV *idl2namel0(int, int, SV *);
static SV *namel2idl0(int, int, SV *);

static ssize_t lastnd = 0;

static char const *weight_types[] = WEIGHT_TYPE_NAMES;

#if defined JIS_WAM
#include "jis.c"
#endif

static int
not_here(char *s)
{
	croak("%s not implemented on this architecture", s);
	return -1;
}

static int
constant(char *name, int arg)
{
	size_t i;
	errno = 0;

	for (i=0; i < nelems(weight_types); i++) {
		if (strEQ(name, weight_types[i])) {
			return i;
		}
	}

	switch (*name) {
	case 'A':
		break;
	case 'B':
		break;
	case 'C':
		break;
	case 'D':
		break;
	case 'E':
		break;
	case 'F':
		break;
	case 'G':
		break;
	case 'H':
		break;
	case 'I':
		break;
	case 'J':
		break;
	case 'K':
		break;
	case 'L':
		break;
	case 'M':
		break;
	case 'N':
		break;
	case 'O':
		break;
	case 'P':
		break;
	case 'Q':
		break;
	case 'R':
		break;
	case 'S':
		break;
	case 'T':
		break;
	case 'U':
		break;
	case 'V':
		break;
	case 'W':
	if (strEQ(name, "WAM_ROW"))
#if defined WAM_ROW
		return WAM_ROW;
#else
		goto not_there;
#endif
	if (strEQ(name, "WAM_COL"))
#if defined WAM_COL
		return WAM_COL;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_OR"))
#if defined WSH_OR
		return WSH_OR;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_AND"))
#if defined WSH_AND
		return WSH_AND;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_NOT"))
#if defined WSH_NOT
		return WSH_NOT;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_ORG"))
#if defined WSH_ORG
		return WSH_ORG;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_NF_CE"))
#if defined WSH_NF_CE
		return WSH_NF_CE;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_NF_NOT"))
#if defined WSH_NF_NOT
		return WSH_NF_NOT;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_NF_OR"))
#if defined WSH_NF_OR
		return WSH_NF_OR;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_NF_DE"))
#if defined WSH_NF_DE
		return WSH_NF_DE;
#else
		goto not_there;
#endif
	if (strEQ(name, "WAM_OCASCADE"))
#if defined WAM_OCASCADE
		return WAM_OCASCADE;
#else
		goto not_there;
#endif
	if (strEQ(name, "WAM_OSUMI"))
#if defined WAM_OSUMI
		return WAM_OSUMI;
#else
		goto not_there;
#endif
	if (strEQ(name, "WAM_OPROBM"))
#if defined WAM_OPROBM
		return WAM_OPROBM;
#else
		goto not_there;
#endif
	if (strEQ(name, "WAM_OCACHE"))
#if defined WAM_OCACHE
		return WAM_OCACHE;
#else
		goto not_there;
#endif
	if (strEQ(name, "WAM_OFAULTTOLERANT"))
#if defined WAM_OFAULTTOLERANT
		return WAM_OFAULTTOLERANT;
#else
		goto not_there;
#endif
		break;
	case 'X':
		break;
	case 'Y':
		break;
	case 'Z':
		break;
	}
	errno = EINVAL;
	return 0;

	not_there:
	errno = ENOENT;
	return 0;
}

#if defined JIS_WAM

static SV *newSVpvJE(char const *);

static SV *
newSVpvJE(p)
char const *p;
{
	char *q;
	SV *s;
	size_t l;

	l = strlen(p);
	if (!(q = malloc(l+1))) {
		perror("malloc");
		return NULL;
	}
	jtoes(p, q, l+1);
	s = newSVpv(q, 0);
	free(q);
	return s;
}

#define	my_newSVpv(p) newSVpvJE(p)

#else

#define	my_newSVpv(p) newSVpv((p), 0)

#endif

static SV *
vec2rh(w, vec, e, r)
WAM *w;
struct xr_vec const *vec;
size_t e;
{
	AV *na;
	struct xr_elem const *ep, *ee;

	na = newAV();
	av_extend(na, e);
	for (ep = vec->elems, ee = ep+e; ep < ee; ep++) {
		HV *nh;
		SV *sv;
		char const *p;
		nh = newHV();

#define	my_hv_store(nh, key, sv) hv_store((nh), #key, sizeof #key-1, (sv), 0)
		sv = newSViv(ep->id);
		my_hv_store(nh, id, sv);

		p = wam_id2name(w, r, ep->id);
		sv = my_newSVpv(p);
		my_hv_store(nh, name, sv);

		sv = newSViv(ep->freq);
		my_hv_store(nh, freq, sv);

		sv = newRV_noinc((SV *)nh);
		av_push(na, sv);
	}
	return newRV_noinc((SV *)na);
}

static SV *
s2rah(dp, nd, w, dir, include_df)
syminfo *dp;
size_t nd;
WAM *w;
{
	AV *na;
	ssize_t i;

	na = newAV();
	av_extend(na, nd);
	for (i=0; i<nd; i++) {
		HV *nh;
		SV *sv;
		char const *p;
		nh = newHV();

		sv = newSViv(dp[i].id);
		my_hv_store(nh, id, sv);

		p = wam_id2name(w, dir, dp[i].id);
		sv = my_newSVpv(p);
		my_hv_store(nh, name, sv);

		sv = newSVnv(dp[i].weight);
		my_hv_store(nh, weight, sv);

		sv = newSViv(dp[i].TF_d);
		my_hv_store(nh, TF_d, sv);

		sv = newSViv(dp[i].TF);
		my_hv_store(nh, TF, sv);

		if (include_df) {
			sv = newSViv(dp[i].DF_d);
			my_hv_store(nh, DF_d, sv);

			sv = newSViv(dp[i].DF);
			my_hv_store(nh, DF, sv);
		}

#undef	my_hv_store

		sv = newRV_noinc((SV *)nh);
		av_push(na, sv);
	}
	return newRV_noinc((SV *)na);
}

static syminfo *
rah2s(a, lenp)
AV *a;
size_t *lenp;
{
	I32 len = 0;
	size_t i;
	syminfo *q, *qp;

	len = av_len(a) + 1;
	if (!(q = calloc(len, sizeof *q))) {
		perror("calloc");
		return NULL;
	}
	for (i = 0, qp = q; i<len; i++, qp++) {
		SV **sp;

		sp = av_fetch(a, i, 0);
		if (SvROK(*sp) && SvTYPE(SvRV(*sp))==SVt_PVHV) {
			HV *h;
			SV **s;
			h = (HV *)SvRV(*sp);
#define	my_hv_fetch(h, key) hv_fetch((h), #key, sizeof #key-1, 0)

			s = my_hv_fetch(h, id);
			qp->id = s ? SvIV(*s) : 0;

			s = my_hv_fetch(h, TF_d);
			qp->TF_d = s ? SvIV(*s) : 0;

			s = my_hv_fetch(h, TF);
			qp->TF = s ? SvIV(*s) : 0;

			s = my_hv_fetch(h, weight);
			qp->weight = s ? SvNV(*s) : 0.0;

			s = my_hv_fetch(h, attr);
			qp->attr = s ? SvIV(*s) : WSH_OR;
#undef	my_hv_fetch
		}
		else {
			qp->id = 0;
			qp->TF_d = 0;
			qp->TF = 0;
			qp->attr = WSH_OR;
		}
	}
	*lenp = len;
	return q;
}

static char **
ra2av(a)
AV *a;
{
	I32 len = 0;
	size_t i;
	char **argv;

	len = av_len(a) + 1;
	if (!(argv = calloc(len, sizeof *argv))) {
		perror("calloc");
		return NULL;
	}
	for (i = 0; i<len; i++) {
		SV **sp;

		sp = av_fetch(a, i, 0);
		if (SvTYPE(*sp)==SVt_PV) {
			unsigned int l;
			argv[i] = SvPV(*sp, l); /* l: ignored */
		}
	}
	argv[i] = NULL;
	return argv;
}

static WAM *dt[256];

static int
xs_wam_open(handle)
char *handle;
{
	int d;
	static int initialized = 0;
	if (!initialized) {
		for (d=0; d<nelems(dt); d++) {
			dt[d] = NULL;
		}
		initialized = 1;
	}
	for (d=1; d<nelems(dt); d++) {
		if (!dt[d]) {
			if ((dt[d] = wam_open(handle))!=NULL) {
				return d;
			}
			else {
				return 0;
			}
		}
	}
	return 0;
}

static int
xs_wam_close(d)
{
	if (d<=0 || nelems(dt)<=d || dt[d]==NULL) {
		return -1;
	}
	wam_close(dt[d]);
	dt[d] = NULL;
	return 0;
}

static WAM *
xs_wam_desc(d)
{
	if (d<=0 || nelems(dt)<=d) {
		return NULL;
	}
	return dt[d];
}

#if ! defined MAX
#define	MAX(a, b)	((a)>(b)?(a):(b))
#endif

static size_t
iwl(terms, nkwords, takano_iw_limit, w, dir)
syminfo *terms;
size_t nkwords, takano_iw_limit;
WAM *w;
{
	size_t i, j;

	if (nkwords<=0) {
		return nkwords;
	}
	/* note: terms[j..nkwords] ni gomi ga nokoru. */
	for (i=0, j=0; i<nkwords; i++) {
		if (wam_elem_num(w, dir, terms[i].id)<=takano_iw_limit) {
			terms[j++] = terms[i];
		}
	}
	return MAX(j, 1);
}

static SV *
wshb0(s, e, d, mode, weight_type, maxnd, takano_iw_limit, target_wam)
SV *s;
SV *e;
int d;
int mode;
int weight_type;
size_t maxnd;
size_t takano_iw_limit;
int target_wam;
{
	AV *a, *b;
	syminfo *q = NULL;
	syminfo *qe = NULL;
	ssize_t len, elen;
	syminfo *dp = NULL;
	ssize_t nd, total;
	WAM *w, *wt;
	SV *r;

	if (!(SvROK(s) && SvTYPE(SvRV(s))==SVt_PVAV)) {
		return NULL;
	}
	if (e && !(SvROK(e) && SvTYPE(SvRV(e))==SVt_PVAV)) {
		return NULL;
	}
	a = (AV *)SvRV(s);
	nd = maxnd;
	if (!(w = xs_wam_desc(d))) {
		return NULL;
	}
	if (!(q = rah2s(a, &len))) {
		return NULL;
	}

	if (e) {
		b = (AV *)SvRV(e);
		if (!(qe = rah2s(b, &elen))) {
			free(q);
			return NULL;
		}
		dp = bex_wsh(q, len, w, mode, weight_type, &nd, &total, qe, elen);
		free(qe);
	}
	else {
		dp = wsh(q, len, w, mode, weight_type, &nd, &total);
	}
	free(q);

	if (!dp) {
		return NULL;
	}

	qsort(dp, nd, sizeof *dp, sym_compar);
	if (takano_iw_limit) {
		if (wt = xs_wam_desc(target_wam)) {
			nd = iwl(dp, nd, takano_iw_limit, wt, WAM_REVERT_DIRECTION(mode));
		}
	}
/*
* lastnd no atai ha, perl karaha &wam::get_last_nd de toreru.
* wsa deha, ws to ha kotornari, lastnd ha nd to onazi ni naru.
*/
	lastnd = total;
	r = s2rah(dp, nd, w, WAM_REVERT_DIRECTION(mode), 1);

	free(dp);
	return r;
}

static SV *
idl2namel0(d, mode, ra)
int d;
int mode;
SV *ra;
{
	size_t len;
	size_t i;
	AV *na;
	AV *a;
	WAM *w;
	SV *r;

	r = NULL;
	if (SvROK(ra) && SvTYPE(SvRV(ra))==SVt_PVAV && (w = xs_wam_desc(d))) {
		a = (AV *)SvRV(ra);
		na = newAV();
		len = av_len(a) + 1;
		av_extend(na, len);
		for (i=0; i<len; i++) {
			SV **sp;
			sp = av_fetch(a, i, 0);
			if (SvIOK(*sp)) {
				char const *p;
				SV *sv = NULL;
				idx_t id;
				id = SvIV(*sp);
				if ((p = wam_id2name(w, mode, id))!=NULL) {
					sv = my_newSVpv(p);
				}
				av_push(na, sv);
			}
		}
		r = newRV_noinc((SV *)na);
	}
	return r;
}

static SV *
namel2idl0(d, mode, ra)
int d;
int mode;
SV *ra;
{
	size_t len;
	size_t i;
	AV *na;
	AV *a;
	WAM *w;
#if defined JIS_WAM
	char *q;
#endif
	SV *r;

	r = NULL;
	if (SvROK(ra) && SvTYPE(SvRV(ra))==SVt_PVAV && (w = xs_wam_desc(d))) {
		a = (AV *)SvRV(ra);
		na = newAV();
		len = av_len(a) + 1;
		av_extend(na, len);
		for (i=0; i<len; i++) {
			SV **sp;
			sp = av_fetch(a, i, 0);
			if (SvTYPE(*sp)==SVt_PV) {
				SV *sv;
				char *name;
				idx_t id;
				unsigned int l;
				name = SvPV(*sp, l); /* l: ignored */
#if defined JIS_WAM
				l = etojs(name, NULL, 0);
				if (!(q = malloc(l+1))) {
					perror("malloc");
					id = 0;
				}
				else {
					etojs(name, q, l+1);
					id = wam_name2id(w, mode, q);
					free(q);
				}
#else
				id = wam_name2id(w, mode, name);
#endif
				sv = newSViv(id);
				av_push(na, sv);
			}
		}
		r = newRV_noinc((SV *)na);
	}
	return r;
}


MODULE = wam		PACKAGE = wam

int
constant(name,arg)
	char *name
	int arg


#if USE_CICONF
int
set_ciconf(path)
	char *path

	CODE:
	if (!strEQ(path, "")) {
		ci_set_ciconf(path);
	}
	RETVAL = 0;

	OUTPUT:
	RETVAL

#endif


int
init(projectroot)
	char *projectroot

	CODE:
	if (strEQ(projectroot, "")) {
		projectroot = NULL;
	}
	RETVAL = wam_init(projectroot);

	OUTPUT:
	RETVAL


int
open(handle)
	char *handle

	CODE:
	RETVAL = xs_wam_open(handle);

	OUTPUT:
	RETVAL


int
close(d)
	int d;

	CODE:
	RETVAL = xs_wam_close(d);

	OUTPUT:
	RETVAL


ssize_t
size(d, mode)
	int d;
	int mode;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_size(w, mode);
	}

	OUTPUT:
	RETVAL

ssize_t
elem_num(d, mode, id)
	int d;
	int mode;
	size_t id;

	PREINIT:
	WAM *w;
	idx_t idx;

	CODE:
	idx = id;
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_elem_num(w, mode, idx);
	}

	OUTPUT:
	RETVAL


ssize_t
freq_sum(d, mode, id)
	int d;
	int mode;
	size_t id;

	PREINIT:
	WAM *w;
	idx_t idx;

	CODE:
	idx = id;
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_freq_sum(w, mode, idx);
	}

	OUTPUT:
	RETVAL


ssize_t
max_elem_num(d, mode)
	int d;
	int mode;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_max_elem_num(w, mode);
	}

	OUTPUT:
	RETVAL

ssize_t
max_freq_sum(d, mode)
	int d;
	int mode;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_max_freq_sum(w, mode);
	}

	OUTPUT:
	RETVAL


ssize_t
total_elem_num(d, mode)
	int d;
	int mode;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_total_elem_num(w, mode);
	}

	OUTPUT:
	RETVAL


ssize_t
total_freq_sum(d, mode)
	int d;
	int mode;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_total_freq_sum(w, mode);
	}

	OUTPUT:
	RETVAL


SV *
id2name(d, mode, id)
	int d;
	int mode;
	size_t id;

	PREINIT:
	char const *name;
	WAM *w;
	idx_t idx;

	CODE:
	idx = id;
	RETVAL = NULL;
	if ((w = xs_wam_desc(d)) && (name = wam_id2name(w, mode, idx))!=NULL) {
		RETVAL = my_newSVpv(name);
	}

	OUTPUT:
	RETVAL


size_t
name2id(d, mode, name)
	int d;
	int mode;
	char *name;

	PREINIT:
	WAM *w;
#if defined JIS_WAM
	char *q;
#endif

	CODE:
	RETVAL = 0;
	if (w = xs_wam_desc(d)) {
#if defined JIS_WAM
		size_t l;
		l = etojs(name, NULL, 0);
		if (!(q = malloc(l+1))) {
			perror("malloc");
		}
		else {
			if (etojs(name, q, l+1)!=-1) {
				RETVAL = wam_name2id(w, mode, q);
			}
			free(q);
		}
#else
		RETVAL = wam_name2id(w, mode, name);
#endif
	}

	OUTPUT:
	RETVAL


SV *
get_vec(d, mode, id)
	int d;
	int mode;
	size_t id;

	PREINIT:
	ssize_t e;
	struct xr_vec const *vec;
	WAM *w;
	idx_t idx;

        CODE:
	idx = id;
	RETVAL = NULL;
	if ((w = xs_wam_desc(d)) && (e = wam_get_vec(w, mode, idx, &vec))!=-1) {
		RETVAL = vec2rh(w, vec, e, WAM_REVERT_DIRECTION(mode));
	}

	OUTPUT:
	RETVAL


SV *
wsh(s, d, mode, weight_type, maxnd, takano_iw_limit, target_wam)
	SV *s;
	int d;
	int mode;
	int weight_type;
	size_t maxnd;
	size_t takano_iw_limit;
	int target_wam;

	CODE:
	RETVAL = wshb0(s, NULL, d, mode, weight_type, maxnd, takano_iw_limit, target_wam);

	OUTPUT:
	RETVAL


SV *
wshb(s, e, d, mode, weight_type, maxnd, takano_iw_limit, target_wam)
	SV *s;
	SV *e;
	int d;
	int mode;
	int weight_type;
	size_t maxnd;
	size_t takano_iw_limit;
	int target_wam;

	CODE:
	RETVAL = wshb0(s, e, d, mode, weight_type, maxnd, takano_iw_limit, target_wam);

	OUTPUT:
	RETVAL


size_t
get_last_nd()

	CODE:
	RETVAL = lastnd;

	OUTPUT:
	RETVAL


SV *
getaux(handle, path, id)
	char *handle;
	char *path;
	size_t id;

	PREINIT:
	char const *p;
	WAM *f;
	idx_t idx;

	CODE:
	idx = id;
	RETVAL = NULL;
	if ((f = wam_aux_open(handle, path))!=NULL) {
		if (wam_aux_gets(f, idx, &p)>0) {
			RETVAL = my_newSVpv(p);
		}
		wam_close(f);
	}

	OUTPUT:
	RETVAL


SV *
getprop(handle, dir, path, id)
	char *handle;
	int dir;
	char *path;
	size_t id;

	PREINIT:
	char const *p;
	WAM *f;
	idx_t idx;

	CODE:
	idx = id;
	RETVAL = NULL;
	if ((f = wam_prop_open(handle, dir, path))!=NULL) {
		if (wam_prop_gets(f, idx, &p)>0) {
			RETVAL = my_newSVpv(p);
		}
		wam_close(f);
	}

	OUTPUT:
	RETVAL


SV *
wstem(text, d, mode, path)
	char *text;
	int d;
	int mode;
	char *path;

	PREINIT:
	ssize_t nd;
	syminfo *dp;
	WAM *w;

	CODE:
	RETVAL = NULL;
	if ((w = xs_wam_desc(d)) && (dp = wstem(text, w, mode, path, &nd))) {
		RETVAL = s2rah(dp, nd, w, mode, 0);
		free(dp);
	}

	OUTPUT:
	RETVAL


SV *
wstemv(text, d, mode, path, argv)
	char *text;
	int d;
	int mode;
	char *path;
	SV *argv;

	PREINIT:
	ssize_t nd;
	syminfo *dp = NULL;
	char **av = NULL;

	CODE:
        if (SvROK(argv) && SvTYPE(SvRV(argv))==SVt_PVAV) {
		AV *a;
		if (a = (AV *)SvRV(argv)) {
			av = ra2av(a);
		}
	}
	if (av && (dp = wstemv(text, dt[d], mode, path, av, &nd))) {
		RETVAL = s2rah(dp, nd, dt[d], mode, 0);
	}
	else {
		RETVAL = NULL;
	}
	free(dp);
	free(av);

	OUTPUT:
	RETVAL


SV *
get_weight_types()
	PREINIT:
	size_t i;
	AV *na;

	CODE:
	na = newAV();
	for (i=0; i < nelems(weight_types); i++) {
		SV *sv;
		sv = newSVpv(weight_types[i], 0); /* my_newSVpv ha tukawanai */
		av_push(na, sv);
	}
	RETVAL = newRV_noinc((SV *)na);

	OUTPUT:
	RETVAL


#if USE_CICONF
SV *
ci_value(handle, name)
	char *handle;
	char *name;

	PREINIT:
	char const *p;

	CODE:
	RETVAL = NULL;
	if ((p = ci_value(handle, name)) != NULL) {
		RETVAL = my_newSVpv(p);
	}

	OUTPUT:
	RETVAL

#endif


SV *
idl2namel(d, mode, ra)
	int d;
	int mode;
	SV *ra;

	CODE:
	RETVAL = idl2namel0(d, mode, ra);

	OUTPUT:
	RETVAL


SV *
namel2idl(d, mode, ra)
	int d;
	int mode;
	SV *ra;

	CODE:
	RETVAL = namel2idl0(d, mode, ra);

	OUTPUT:
	RETVAL


int
set_prob_mode(d, e)
	int d;
	int e;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if ((w = xs_wam_desc(d))) {
		RETVAL = wam_setopt(w, WAM_OPROBM, (char *)&e);
	}

	OUTPUT:
	RETVAL


int
set_opt(d, o, v)
	int d;
	int o;
	int v;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if ((w = xs_wam_desc(d))) {
		RETVAL = wam_setopt(w, o, (char *)&v);
	}

	OUTPUT:
	RETVAL
