--
--	Copyright (C) 2002 Hitachi, Ltd.
--	All rights reserved.
--	Author: Shingo NISHIOKA
--		nis@harl.hitachi.co.jp
--

-- rcsid="$Id: Wam.hs,v 1.1.1.1 2005/06/20 07:13:48 nis Exp $"

module Wam(Wam,
	WamAux,
	WamProp,
	Direction(Row, Col),
	XrVec(elem_num, freq_sum, elems),

	initialize,
	open,
	auxOpen,
	auxClose,
	auxGets,
	propOpen,
	propClose,
	propGets,
	close,

	size,
	elemNum,
	freqSum,
	maxElemNum,
	maxFreqSum,
	totalElemNum,
	totalFreqSum,

	idToName,
	nameToId,
	handle,

	getVec,
	wsh,

	wStem,

	ciValue,
	ciFlags,
	) where

import CForeign
import CString
import MarshalAlloc
import MarshalArray
import MarshalUtils
import Ptr
import Storable

import WeightType
import XrElem
import SymInfo

type CWam = Ptr ()

foreign import ccall "wam_init" wam_init :: CString -> IO Int
foreign import ccall "wam_open" wam_open :: CString -> IO CWam
foreign import ccall "wam_aux_open" wam_aux_open :: CString -> CString -> IO CWam
foreign import ccall "wam_prop_open" wam_prop_open :: CString -> Int -> CString -> IO CWam
foreign import ccall "wam_close" wam_close :: CWam -> IO ()

foreign import ccall "wam_size" wam_size :: CWam -> Int -> IO Int
foreign import ccall "wam_elem_num" wam_elem_num :: CWam -> Int -> Int -> IO Int
foreign import ccall "wam_freq_sum" wam_freq_sum :: CWam -> Int -> Int -> IO Int
foreign import ccall "wam_max_elem_num" wam_max_elem_num :: CWam -> Int -> IO Int
foreign import ccall "wam_max_freq_sum" wam_max_freq_sum :: CWam -> Int -> IO Int
foreign import ccall "wam_total_elem_num" wam_total_elem_num :: CWam -> Int -> IO Int
foreign import ccall "wam_total_freq_sum" wam_total_freq_sum :: CWam -> Int -> IO Int

foreign import ccall "wam_id2name" wam_id2name :: CWam -> Int -> Int -> IO CString
foreign import ccall "wam_name2id" wam_name2id :: CWam -> Int -> CString -> IO Int
foreign import ccall "wam_handle" wam_handle :: CWam -> IO CString

foreign import ccall "wam_get_vec" wam_get_vec :: CWam -> Int -> Int -> Ptr (Ptr a) -> IO Int
foreign import ccall "wsh" _wsh :: Ptr a -> Int -> CWam -> Int -> Int -> Ptr Int -> Ptr Int -> IO (Ptr a)
foreign import ccall "wam_aux_gets" wam_aux_gets :: CWam -> Int -> Ptr (Ptr a) -> IO Int
foreign import ccall "wam_prop_gets" wam_prop_gets :: CWam -> Int -> Ptr (Ptr a) -> IO Int

foreign import ccall "wstem" wstem :: CString -> CWam -> Int -> CString -> Ptr Int -> IO CString

foreign import ccall "ci_value" ci_value :: CString -> CString -> IO CString
foreign import ccall "ci_flags" ci_flags :: CString -> IO Int

data Wam = Wam CWam deriving (Show)
data WamAux = WamAux CWam
data WamProp = WamProp CWam

data Direction = Row | Col deriving (Show, Read, Eq)

instance Enum Direction where
	fromEnum Col = 1
	fromEnum Row = 0
	toEnum 0 = Row
	toEnum 1 = Col

data XrVec = XrVec {elem_num, freq_sum :: Int, elems :: [XrElem]}

instance Show XrVec where
	showsPrec _ x = showsXrVec x

showsXrVec :: XrVec -> ShowS
showsXrVec v =
	(shows (length (elems v))) .
	(", {"++) .
	(shows (elem_num v)) .
	(", "++) .
	(shows (freq_sum v)) .
	("}"++) .
	(foldr1 (.) (map (\ x -> (", " ++) . (shows x)) (elems v)))

initialize :: Maybe String -> IO ()
initialize projectroot = do
	r <- maybe (wam_init (castPtr nullPtr))
		   (\ projectroot -> withCString projectroot wam_init)
		projectroot
	if r == -1
		then fail "WAM init failed"
		else return ()

open :: String -> IO Wam
open handle = withCString handle opn
	where
		opn handle = do
			r <- wam_open handle
			if r == nullPtr
				then fail "WAM open failed"
				else return (Wam r)

auxOpen :: String -> String -> IO WamAux
auxOpen handle path = withCString2 handle path aopn
	where
		aopn handle path = do
			r <- wam_aux_open handle path
			if r == nullPtr
				then fail "AUX open failed"
				else return (WamAux r)

propOpen :: String -> Direction -> String -> IO WamProp
propOpen handle d path = withCString2 handle path aopn
	where
		aopn handle path = do
			r <- wam_prop_open handle (fromEnum d) path
			if r == nullPtr
				then fail "PROP open failed"
				else return (WamProp r)

close :: Wam -> IO ()
close (Wam wp) = wam_close wp

auxClose :: WamAux -> IO ()
auxClose (WamAux wp) = wam_close wp

propClose :: WamProp -> IO ()
propClose (WamProp wp) = wam_close wp

size :: Wam -> Direction -> IO Int
size (Wam wp) d = wam_size wp (fromEnum d)

elemNum :: Wam -> Direction -> Int -> IO Int
elemNum (Wam wp) d i = wam_elem_num wp (fromEnum d) i

freqSum :: Wam -> Direction -> Int -> IO Int
freqSum (Wam wp) d i = wam_freq_sum wp (fromEnum d) i

maxElemNum :: Wam -> Direction -> IO Int
maxElemNum (Wam wp) d = wam_max_elem_num wp (fromEnum d)

maxFreqSum :: Wam -> Direction -> IO Int
maxFreqSum (Wam wp) d = wam_max_freq_sum wp (fromEnum d)

totalElemNum :: Wam -> Direction -> IO Int
totalElemNum (Wam wp) d = wam_total_elem_num wp (fromEnum d)

totalFreqSum :: Wam -> Direction -> IO Int
totalFreqSum (Wam wp) d = wam_total_freq_sum wp (fromEnum d)

idToName :: Wam -> Direction -> Int -> IO String
idToName (Wam wp) d id = do
	name <- wam_id2name wp (fromEnum d) id
	if name == nullPtr
		then fail ("no such ID: " ++ show d ++ " " ++ show id)
		else peekCString name

nameToId :: Wam -> Direction -> String -> IO Int
nameToId (Wam wp) d name = withCString name (wam_name2id wp (fromEnum d))

handle :: Wam -> IO String
handle (Wam wp) = do
	handle <- wam_handle wp
	if handle == nullPtr
		then fail "handle failed for unknown reason (should not happen)"
		else peekCString handle

getVec :: Wam -> Direction -> Int -> IO XrVec
getVec (Wam wp) d id =
	withArray [wp] (\ vp -> do
		elem_num <- wam_get_vec wp dir id vp
		if elem_num == -1
			then fail "getVec failed"
			else peekVec elem_num vp)
	where
		dir = fromEnum d
		peekVec elem_num vp = do
			v <- peek vp
			h:t <- peekArray (elem_num + 1) (castPtr v)
			return (XrVec (elemId h) (elemFreq h) t)

wsh :: Wam -> Direction -> [SymInfo] -> WeightType -> Int -> IO ((Int, Int), [SymInfo])
wsh (Wam wp) d q wt nd =
	withArray2 q [nd, 0] (\ query ndp -> do
		let totalp = plusPtr ndp (sizeOf nd)
		r <- _wsh query nq wp dir wet ndp totalp
		if r == nullPtr
			then fail "wsh failed"
			else peekSymInfo r ndp totalp)
	where
		nq = length q
		dir = fromEnum d
		wet = fromEnum wt
		peekSymInfo r ndp totalp = do
			nnd <- peek ndp
			ntotal <- peek totalp
			nr <- peekArray nnd (castPtr r)
			return ((nnd, ntotal), nr)

wStem :: Wam -> Direction -> String -> String -> IO [SymInfo]
wStem (Wam wp) d text path = withCString2 path text ws
	where
		ws path text =
			withArray [0] (\ ndp -> do
				r <- wstem text wp (fromEnum d) path ndp
				if r == nullPtr
					then fail "wStem failed"
					else peekSVec r ndp)
		peekSVec r ndp = do
			nnd <- peek ndp
			nr <- peekArray nnd (castPtr r)
			return nr

auxGets :: WamAux -> Int -> IO String
auxGets (WamAux wp) id = withArray [wp] aget
	where
		aget bufp = do
			length <- wam_aux_gets wp id bufp
			if length == -1
				then fail "auxGets failed"
				else peekDCString bufp
		peekDCString bufp = do
			buf <- peek bufp
			peekCString (castPtr buf)

propGets :: WamProp -> Int -> IO String
propGets (WamProp wp) id = withArray [wp] aget
	where
		aget bufp = do
			length <- wam_prop_gets wp id bufp
			if length == -1
				then fail "propGets failed"
				else peekDCString bufp
		peekDCString bufp = do
			buf <- peek bufp
			peekCString (castPtr buf)

ciValue :: String -> String -> IO String
ciValue handle attribute = withCString2 handle attribute cv
	where cv handle attribute = do
		r <- ci_value handle attribute
		if r == nullPtr
			then fail "ciValue failed"
			else peekCString (castPtr r)

ciFlags :: String -> IO Int
ciFlags handle = withCString handle ci_flags

withCString2 :: String -> String -> (CString -> CString -> IO a) -> IO a
withCString2 a b f = withCString a (\ a -> withCString b (\ b -> f a b))

withArray2 :: (Storable a, Storable b) => [a] -> [b] -> (Ptr a -> Ptr b -> IO c) -> IO c
withArray2 a b f = withArray a (\ a -> withArray b (\ b -> f a b))

-- flip
