/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: heapselect.c,v 1.2 2007/03/21 03:17:41 nis Exp $"; */

#include <config.h>

#include <sys/types.h>
#include <sys/time.h>

#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>

/*
	base[0..nmemb-1]	2*i+1 2*i+2 (nmemb+1)/2 .. nmemb-1
*/

#define	copy(s, t, size) { \
	a = s; \
	b = t; \
	count=size; \
	do { \
		*b++ = *a++; \
	} while (--count); \
}

#define	swap(s, t, size) { \
	a = s; \
	b = t; \
	count=size; \
	do { \
		c = *a; \
		*a++ = *b; \
		*b++ = c; \
	} while (--count); \
}

int
heapselect(vbase, nrepr, nmemb, size, compar)
void *vbase;
size_t nrepr, nmemb, size;
int (*compar)(void const *, void const *);
{
	char *base;
	char *a, *b, c;
	size_t count;
	void *x;
	size_t i, j, m;
	size_t l, r;
	size_t is, js, ms;

	base = vbase;
	if (size==0 || nrepr > nmemb) {
		dwerrno = EINVAL;
		return -1;
	}
	if (nmemb==0 || nrepr==0 || nrepr==nmemb) {
		return 0;
	}

	if (!(x = calloc(1, size))) {
		dwerrno = errno;
		return -1;
	}

	l = (nrepr+1)/2;

	r = nrepr-1;

	/* l .. r is heap. */

	assert(2*l+1 > r);

	while (0 < l) {
		l--;
		i = l;
		is = i*size;
		copy(base+is, x, size);
		while ((j = 2*i+1) <= r) {
			js = j*size;
			if (j < r && (*compar)(base+js, base+js+size)<0) {
				j++;
				js+=size;
			}
			if ((*compar)(base+js, x)<0) {
				break;
			}
			copy(base+js, base+is, size);
			i = j;
			is = js;
		}
		if (i!=l) {
			copy(x, base+is, size);
		}
	}

	/* now, 0 .. r is heap. */

	for (m=r+1; m<nmemb; m++) {
		ms = m*size;
		if ((*compar)(base+ms, base)<0) {
			swap(base+ms, base, size);

			/* now 1 .. r is heap. */

			i = 0;
			is = 0;
			copy(base+is, x, size);
			while ((j = 2*i+1) <= r) {
				js = j*size;
				if (j < r && (*compar)(base+js, base+js+size)<0) {
					j++;
					js+=size;
				}
				if ((*compar)(base+js, x)<0) {
					break;
				}
				copy(base+js, base+is, size);
				i = j;
				is = js;
			}
			if (i!=0) {
				copy(x, base+is, size);
			}
		}

		/* now 0 .. r is heap. */
	}

	free(x);
	return 0;
}
