/*
 *	Copyright (c) 2004 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: int.c,v 1.9 2007/03/26 07:32:14 nis Exp $";
#endif

#include <config.h>
#include <sys/types.h>

#include <stdarg.h>
#include <stddef.h>

#include <stdlib.h>
#include <limits.h>
#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>
#include <geta/dwsh.h>

#define nelems(e)	(sizeof (e) / sizeof *(e))

char const *
freqtostr(n)
freq_t n;
{
	static int i = 0;
	static char s[8][22];
	char *r, *p, *q;
	int neg = 0;

	if (n == 0) {
		return "0";
	}

	if (sizeof n == 4 && n == 0x80000000ul) {
		return "-2147483648";
	}
	else if (sizeof n == 8 && n == 0x8000000000000000ull) {
		return "-9223372036854775808";
	}

	r = s[i++];

	if (i >= nelems(s)) {
		i = 0;
	}

	if (n < 0) {
		n = -n;
		neg = 1;
	}
	for (p = r; n; p++) {
		int c = n % 10;
		*p = c + '0';
		n = n / 10;
	}
	if (neg) {
		*p = '-';
		*(p + 1) = '\0';
	}
	else {
		*p-- = '\0';
	}
	for (q = r; q < p; p--, q++) {
		int t = *p;
		*p = *q;
		*q = t;
	}

	return r;
}

char const *
ltostr(n)
long n;
{
	static int i = 0;
	static char s[8][22];
	char *r, *p, *q;
	int neg = 0;

	if (n == 0) {
		return "0";
	}

	if (sizeof n == 4 && n == 0x80000000ul) {
		return "-2147483648";
	}
	else if (sizeof n == 8 && n == 0x8000000000000000ull) {
		return "-9223372036854775808";
	}

	r = s[i++];

	if (i >= nelems(s)) {
		i = 0;
	}

	if (n < 0) {
		n = -n;
		neg = 1;
	}
	for (p = r; n; p++) {
		int c = n % 10;
		*p = c + '0';
		n = n / 10;
	}
	if (neg) {
		*p = '-';
		*(p + 1) = '\0';
	}
	else {
		*p-- = '\0';
	}
	for (q = r; q < p; p--, q++) {
		int t = *p;
		*p = *q;
		*q = t;
	}

	return r;
}

char const *
ultostr(n)
unsigned long n;
{
	static int i = 0;
	static char s[8][22];
	char *r, *p, *q;

	if (n == 0) {
		return "0";
	}

	r = s[i++];

	if (i >= nelems(s)) {
		i = 0;
	}

	for (p = r; n; p++) {
		int c = n % 10;
		*p = c + '0';
		n = n / 10;
	}
	*p-- = '\0';
	for (q = r; q < p; p--, q++) {
		int t = *p;
		*p = *q;
		*q = t;
	}

	return r;
}

char const *
ultoxstr(n)
unsigned long n;
{
	static int i = 0;
	static char s[8][22];
	char *r, *p, *q;

	if (n == 0) {
		return "0";
	}

	r = s[i++];

	if (i >= nelems(s)) {
		i = 0;
	}

	for (p = r; n; p++) {
		int c = n % 16;
		if (c < 10) {
			*p = c + '0';
		}
		else {
			*p = c + 'A' - 10;
		}
		n = n / 16;
	}
	*p-- = '\0';
	for (q = r; q < p; p--, q++) {
		int t = *p;
		*p = *q;
		*q = t;
	}

	return r;
}

#if 0
main()
{
	long n;
	n = 100000*100000L*100000; printf("%s\n", ltostr(n));
	n = 0xffffffffffffffff; printf("%s\n", ltostr(n));
	n = 0x8000000000000001; printf("%s\n", ltostr(n));
	n = 0x8000000000000000; printf("%s\n", ltostr(n));
	n = 0x7fffffffffffffff; printf("%s %s\n", ultostr(n), ultostr(100));
}
#endif

freq_t
strtofreq(nptr, endptr, base)
char const *nptr;
char **endptr;
{
	freq_t e;
	e = strtoul(nptr, endptr, base);
	return e;
}

ssize_t
strtodf(nptr, endptr, base)
char const *nptr;
char **endptr;
{
	ssize_t e;
	e = strtoul(nptr, endptr, base);
	return e;
}

idx_t
strtoidx(nptr, endptr, base)
char const *nptr;
char **endptr;
{
	idx_t e;
	e = strtoul(nptr, endptr, base);
	return e;
}

asig_t
asig()
{
#if defined WORDS_BIGENDIAN 
#define	ENDIAN	1
#else
#define	ENDIAN	0
#endif
#define	X(type, shift)	((type) << (shift) * 4)
	return X(sizeof(struct s_syminfo), 6) |
	       X(ENDIAN, 5) |
	       X(sizeof(double), 4) |
	       X(sizeof(float), 3) |
	       X(sizeof(freq_t), 2) |
	       X(sizeof(df_t), 1) |
	       X(sizeof(idx_t), 0);
#undef X
}
