/*
 *	Copyright (C) 2001 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: merge.c,v 1.4 2007/03/24 00:32:56 nis Exp $"; */

#include <config.h>

#include <sys/types.h>
#include <sys/time.h>

#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dwerrno.h>

#define	copy(s, t, size) { \
	a = s; \
	b = t; \
	count=size; \
	do { \
		*b++ = *a++; \
	} while (--count); \
}

struct he {
	void const *p;
	void const *e;
};

inline static void reheap(struct he *, size_t, size_t, int (*)(const void *, const void *));
static ssize_t hmerge(size_t, void const **, size_t *, void *, size_t, size_t, int (*)(const void *, const void *));
static ssize_t qmerge(size_t, void const **, size_t *, void *, size_t, size_t, int (*)(const void *, const void *));

/* n-way merge */
static ssize_t
hmerge(nbase, base, belems, result, nmemb, size, compar)
size_t nbase;
void const **base;
size_t *belems;
void *result;
size_t nmemb, size;
int (*compar)(const void *, const void *);
{
	struct he h[1024];
	char const *a;
	char *b;
	size_t count;
	size_t i, j;
	size_t l;
	size_t total;

	if (nbase>=nelems(h)) {
		errno = EINVAL;
		return -1;
	}

	for (i=1, j=0, total=0; j<nbase; j++) {
		if (belems[j]>0) {
			h[i].p = base[j];
			h[i].e = (char *)h[i].p+belems[j]*size;
			total += belems[j];
			i++;
		}
	}
	nbase = i-1;
	if (nbase==0) {
		/* nothing to do */
		return 0;
	}
	if (total>nmemb) {
		errno = EINVAL;
		return -1;
	}

	for (l=nbase/2; l>0; l--) {
		reheap(h, l, nbase, compar);
	}

	for (; nbase>0; ) {
		copy(h[1].p, (char *)result, size);
		result = (char *)result + size;

		if ((h[1].p = (char *)h[1].p+size) >= h[1].e) {	/* exhausted */
			h[1] = h[nbase--];
		}
		reheap(h, 1, nbase, compar);
	}
	return total;
}

inline static void
reheap(a, l, na, compar)
struct he *a;
size_t l, na;
int (*compar)(const void *, const void *);
{
	size_t i, j;
	struct he x;
	i = l;
	x = a[i];
	while ((j=2*i) <= na) {
		if (j<na && compar(a[j].p, a[j+1].p)>0) {
			j++;
		}
		if (compar(a[j].p, x.p)>=0) {
			break;
		}
		a[i] = a[j];
		i = j;
	}
	a[i] = x;
}

static ssize_t
qmerge(n, base, belems, result, nmemb, size, compar)
size_t n;
void const **base;
size_t *belems;
void *result;
size_t nmemb, size;
int (*compar)(const void *, const void *);
{
	size_t i;
	size_t total;

	for (i=0, total=0; i<n; i++) {
		if (total+belems[i]>nmemb) {
			return -1;
		}
		memmove((char *)result+total*size, base[i], belems[i]*size);
		total += belems[i];
	}
	qsort(result, total, size, compar);
	return total;
}

#if 0

typedef size_t typ;
int typ_compar(const void *, const void *);

main(argc, argv)
char *argv[];
{
	ssize_t total, total2;
	typ **t = NULL, *result = NULL, *result2 = NULL;
	typ *telems = NULL;
	size_t cnt = 0;
	size_t r_nmemb = 0;
	size_t t_n = 0;

	size_t nmemb = 0;

	size_t n = random()%1024;
	size_t i, j;
	struct timeval start, end;
	gettimeofday(&start, NULL);

again:

	if (n>t_n) {
		free(telems);
		free(t);
		if (!(telems = malloc(n * sizeof *telems))) {
			nperror("calloc");
			return 1;
		}

		if (!(t = malloc(n * sizeof *t))) {
			nperror("calloc");
			return 1;
		}
		t_n = n;
	}
	for (i=0; i<n; i++) {
		telems[i] = random()%128;
		if (!(t[i] = calloc(telems[i], sizeof *t[i]))) {
			nperror("calloc");
			return 1;
		}
		for (j=0; j<telems[i]; j++) {
			t[i][j] = random()%1024;
		}
		nmemb += telems[i];
		qsort(t[i], telems[i], sizeof *t[i], typ_compar);
	}

	if (nmemb > r_nmemb) {
		free(result);
		free(result2);
		if (!(result = malloc(nmemb * sizeof *result))) {
			nperror("calloc");
			return 1;
		}
		if (!(result2 = malloc(nmemb * sizeof *result2))) {
			nperror("calloc");
			return 1;
		}
		r_nmemb = nmemb;
	}

	if (argc==1) {
		if ((total = hmerge(n, t, telems, result, nmemb, sizeof *result, typ_compar))==-1) {
			nperror("hmerge");
			return 1;
		}
	}
	else {
		if ((total2 = qmerge(n, t, telems, result2, nmemb, sizeof *result2, typ_compar))==-1) {
			nperror("qmerge");
		}
	}

#if 0
	assert(total==total2);
	if (bcmp(result, result2, total * sizeof *result2)) {
		return 1;
	}
#endif

	if (cnt++<512) {
		goto again;
	}
	gettimeofday(&end, NULL);
	timersub(&end, &start, &end);
	printf("tim: %d.%03d\n", end.tv_sec, end.tv_usec/1000);

	return 0;
}

typ_compar(va, vb)
const void *va, *vb;
{
	typ const *a, *b;
	a = va;
	b = vb;

	if (*a<*b) {
		return -1;
	}
	else if (*a>*b) {
		return 1;
	}
	return 0;
}

#endif
