/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: mkdw.c,v 1.18 2007/03/29 07:37:38 nis Exp $"; */

#include <config.h>

#define DMKXRC
#define MKDXRCF

#include <sys/param.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/uio.h>

#include <assert.h>
#include <fcntl.h>
#if defined HAVE_LIBGEN_H
#include <libgen.h>
#endif
#include <netdb.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/timer.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwc.h>

#include "dwsys.h"
#include "sdup.h"
#include "nht.h"
#include "dcwP.h"
#include "dxrP.h"
#include "mmalloc.h"
#include "nsxhash.h"
#include "rht.h"
#include "nnet.h"

#include "iosched.h"
#if defined DMKXRC
#include "cel.h"
#endif

#include "zvec.h"

#define	PAGESIZE	8192

struct srv {
	char const *host;
	/*int port;*/
	char const *serv;
	DWFILE *f;
};

#if defined X_READ_RAW_VEC

struct xread_param {
	struct {
		int m;
		double p;
	} a, w;
};

struct xread_param *xrp = NULL;

#endif

/* in-core CW */
struct ICCW {
	struct ht *h;
	struct HT *hi;
	size_t num;
	struct sdup *s;
};

char *progname;
int cmode = 0;

int rindex_type = RIT_BSEARCH;

size_t amem = AMEM;

static void usage(void);
static int toplevel(char **, char const *, char const *, size_t, int, char *, char const *, int);
static int mkdcw(struct ICCW *, RFILES *f, struct iosched *, int);
static int dcw_write(struct ICCW *, int, int, FILE *, void *);
static int qcompar(void const *, void const *);
static int mkdxrr(RFILES *f, struct iosched *, int, DWFILE *[], struct ICCW *);
static int mkdxrc(struct iosched *, int);
#if defined MKDXRCF
static int mkdxrcf(RFILES *f, struct iosched *, int, DWFILE *[], struct ICCW *);
#endif
static int mkdxrfm(RFILES *f, struct iosched *, int, DWFILE *[], struct ICCW *);
static idx_t rv2v(struct dxr_raw_vec const *, struct xr_vec *, DWFILE *[], struct ICCW *);
static ssize_t uniq_elems(struct xr_elem *, ssize_t);
static int init_output(struct iosched *, struct sched *, char const *, int, int, int, int);
static int init_iocb(struct iosched *, struct sched_iocb *, char const *, size_t, int, int, ssize_t, size_t, int, int);
static int finish_output(struct iosched *, struct sched *);
static int make_summary_information(struct sched_iocb *);
#if defined DMKXRC
static int flush_trc_vec(struct sched_iocb *);
#endif
static int flush_iocb_nowait(struct iosched *, struct sched_iocb *);
static int flush_iocb_wait(struct iosched *, struct sched_iocb *);
static int close_iocb(struct iosched *, struct sched_iocb *);
static int write_vec(struct sched_iocb *, idx_t, struct xr_vec const *, ssize_t);
#if defined DMKXRC
static int add_trc_vec(struct sched_iocb *, idx_t, struct xr_vec const *, ssize_t);
#endif
static int output_vec(struct iosched *, struct sched *, struct sched *, idx_t, struct xr_vec const *);
static int scompar(void const *, void const *);
static int distrib_s(struct iosched *, char *, size_t);
static int distrib_ss(struct iosched *, struct sched_iocb *, int, struct srv *, size_t);
static int distrib_sc(struct iosched *, struct srv *, size_t, char *, size_t);
static size_t allocv0(size_t, void **);
static void freev0(void *);
static int mkmonolithic(struct ICCW *);

static void nextf_flush(void);
static void refresh_hash_table(struct ICCW *, void *, size_t);
static int refresh_hash = 0;

extern char const *dfs_host;
extern char const *dfs_proto;
extern void (*rawvec_nextf_hook)(void);

struct timeval start;
static void show_elapsed_time(int, FILE *);

#define	MAXRLIMIT(resource) maxrlimit(resource, #resource)
void maxrlimit(int, char const *);

#if defined X_READ_RAW_VEC
static int parse_xread_param(struct xread_param *, char *);

static int
parse_xread_param(xp, s)
struct xread_param *xp;
char *s;
{
	if (!*s) return -1;
	xp->a.m = strtol(s, &s, 10);
	if (*s!=',') return -2;
	if (!*++s) return -3;
	xp->a.p = strtod(s, &s);
	if (*s!=',') return -4;
	if (!*++s) return -5;
	xp->w.m = strtol(s, &s, 10);
	if (*s!=',') return -6;
	if (!*++s) return -7;
	xp->w.p = strtod(s, &s);
	if (*s) return -8;
	return 0;
}
#endif

static void
usage()
{
	fprintf(stderr, "usage: %s "
	"[-c 0|1|2|3] [-h|-b] [-v0size #] [-compress type] [-sched-conf path] "
#if defined X_READ_RAW_VEC
	"[-xread-param #,#,#,#] "
#endif
	"[-dfs-host host] [-dfs-proto proto] "
	"dirname(GETAROOT) handle [@]path schedule\n", progname);
	exit(1);
}

int
main(argc, argv)
char *argv[];
{
	char *schedule_policy;
	int mtime;
	struct timeval start_t;
	char **freqfiles = NULL;
	char *files, *dirname, *handle;
	size_t v0size = 0;
	int compress = COMPRESS_NONE;
	char sched_conf[MAXPATHLEN] = "sched.conf";

	setvbuf(stdout, NULL, _IONBF, 0);

	progname = basename(argv[0]);
	for (argc--, argv++; argc>0; argc--, argv++) {
		if (!strcmp(*argv, "-h")) {
			rindex_type = RIT_HASH;
		}
		else if (!strcmp(*argv, "-b")) {
			rindex_type = RIT_BSEARCH;
		}
		else if (!strcmp(*argv, "-c") && argc>1) {
			char *endptr;
			argc--;
			argv++;
			cmode = strtol(*argv, &endptr, 10);
			if (*endptr || cmode < 0 || 3 < cmode) {
				usage();
			}
			printf("cmode: %d\n", cmode);
		}
		else if (!strcmp(*argv, "-v0size") && argc>1) {
			char *endptr;
			argc--;
			argv++;
			v0size = strtosize(*argv, &endptr, 10);
			if (*endptr) {
				usage();
			}
			printf("v0size: %s\n", sizetostr(v0size));
		}
		else if (!strcmp(*argv, "-compress") && argc>1) {
			argc--;
			argv++;
			if ((compress = dw_encode_compress(*argv))==-1) {
				usage();
			}
			printf("compress: %s\n", dw_decode_compress(compress));
		}
		else if (!strcmp(*argv, "-sched-conf") && argc>1) {
			argc--;
			argv++;
			strncpy(sched_conf, *argv, sizeof sched_conf);
			printf("sched-conf: %s\n", sched_conf);
		}
#if defined X_READ_RAW_VEC
		else if (!strcmp(*argv, "-xread-param") && argc>1) {
			static struct xread_param xrp0;
			int ec;
			argc--;
			argv++;
			if ((ec=parse_xread_param(&xrp0, *argv))<0) {
				fprintf(stderr, "parse_xread_param failed for [%s], %d\n", *argv, ec);
				usage();
			}
			xrp = &xrp0;
		}
#endif
		else if (!strcmp(*argv, "-dfs-host") && argc>1) {
			argc--;
			argv++;
			dfs_host = strdup(*argv);
			printf("dfs-host: %s\n", dfs_host);
		}
		else if (!strcmp(*argv, "-dfs-proto") && argc>1) {
			argc--;
			argv++;
			dfs_proto = strdup(*argv);
			printf("dfs-proto: %s\n", dfs_proto);
		}
		else {
			break;
		}
	}
	switch (rindex_type) {
	case RIT_HASH:
		printf("rindex_type: HASH\n");
		break;
	case RIT_BSEARCH:
		printf("rindex_type: BINARY\n");
		break;
	default:
		fprintf(stderr, "internal error\n");
		return 1;
	}
	if (argc!=4) {
		usage();
	}

	dirname = argv[0];
	handle = argv[1];
	files=argv[2];
	schedule_policy = argv[3];

#if defined ANON_MTIME
	mtime = ~0;
#else
	gettimeofday(&start_t, NULL);
	mtime = start_t.tv_sec;
#endif

	printf("%s\n", handle);

	if (!(freqfiles = read_files(files))) {
		fprintf(stderr, "%s: read_files failed\n", progname);
		return 1;
	}

#if defined RLIMIT_NOFILE
	MAXRLIMIT(RLIMIT_NOFILE);
#endif

#if defined RLIMIT_CPU
	MAXRLIMIT(RLIMIT_CPU);
#endif

#if defined RLIMIT_VMEM
	MAXRLIMIT(RLIMIT_VMEM);
#endif

#if defined RLIMIT_DATA
	MAXRLIMIT(RLIMIT_DATA);
#endif

	if (toplevel(freqfiles, dirname, handle, v0size, mtime, schedule_policy, sched_conf, compress)==-1) {
		fprintf(stderr, "%s failed\n", progname);
		return 1;
	}
	return 0;
}

static int
toplevel(freqfiles, dirname, handle, v0size, mtime, schedp, sched_conf, compress)
char **freqfiles;
char const *dirname, *handle;
size_t v0size;
char *schedp;
char const *sched_conf;
{
	struct ICCW iccws[2];
	DWFILE *cws[2];
	RFILES *f;
	struct iosched *s, s0;
	FILE *g;
	struct stat sb;
	char *schep;

	s = &s0;
	s->dirname = dirname;
	s->handle = handle;
	s->compress = compress;
	if (schedule_set_policy(s, schedp, sched_conf)==-1) {
		fprintf(stderr, "schedule_set_policy failed\n");
		return -1;
	}

	s->v0size = v0size;

	rawvec_nextf_hook = nextf_flush;

/*
	s->h[0].ls = 1024*64;
	s->h[0].ss = 1024*4;
	s->h[0].rehash = 0;

	s->h[1].ls = 1024*128;
	s->h[1].ss = 1024*4;
	s->h[1].rehash = 1;
*/

	dw_mkpathname(s->path, sizeof s->path, s->dirname, s->handle, DW_SCHE, 0, 1);
	dw_mkpathname(s->dcw_path[0], sizeof s->dcw_path[0], s->dirname, s->handle, DCW_ROW, 0, 0);
	dw_mkpathname(s->dcw_path[1], sizeof s->dcw_path[1], s->dirname, s->handle, DCW_COL, 0, 0);

	if (!(f=init_read_raw_vec(freqfiles))) {
		nperror("init_read_raw_vec");
		return -1;
	}

#if defined X_READ_RAW_VEC
	if (xrp) {
		fprintf(stderr, "xrp: %d %f %d %f\n", xrp->a.m, xrp->a.p, xrp->w.m, xrp->w.p);
		rawvec_set_xread_param(f, xrp->a.m, xrp->a.p, xrp->w.m, xrp->w.p);
	}
#endif

	gettimeofday(&start, NULL);

	if (mkdcw(iccws, f, s, mtime)==-1) {
		fprintf(stderr, "%s: mkdcw failed\n", progname);
		return -1;
	}

	show_elapsed_time(0, stdout);

/* XXX : in-core CW で置換 */
	if (!(cws[0] = dcw_open(s->dcw_path[0], s->handle))) {
		return -1;
	}
	if (!(cws[1] = dcw_open(s->dcw_path[1], s->handle))) {
		return -1;
	}

	if (cmode == 3) {
		if (mkdxrfm(f, s, mtime, cws, iccws)==-1) {
			fprintf(stderr, "%s: mkdxrfm failed\n", progname);
			return -1;
		}
	}
	else if (mkdxrr(f, s, mtime, cws, iccws)==-1) {
		fprintf(stderr, "%s: mkdxrr failed\n", progname);
		return -1;
	}

	show_elapsed_time(0, stdout);
#if defined MKDXRCF
	if (cmode==2 && mkdxrcf(f, s, mtime, cws, iccws)==-1) {
		fprintf(stderr, "%s: mkdxrcf failed\n", progname);
		return -1;
	}
#endif

	finish_read_raw_vec(f);

	dcw_close(cws[0]);
	dcw_close(cws[1]);

	if (cmode==0 && mkdxrc(s, mtime)==-1) {
		fprintf(stderr, "%s: mkdxrc failed\n", progname);
		return -1;
	}

	if (!(g = fopen(s->path, "w"))) {
		nperror(s->path);
		return -1;
	}
	if (print_sched(s, g)==-1) {
		fprintf(stderr, "%s: print_sched failed\n", progname);
		return -1;
	}
	if (fclose(g)==EOF) {
		return -1;
	}

	/* mkdw ga dousa siteiru directory de dwd ga dousa site iruto,
	   sched wo distribute sita tokini, kono tyokuzen de saksei sita
	   g ga, sono dwd ni uwagaki sarete simau. sokode, sched wo
	   ittan memory ni yomikomu. */
	if (stat(s->path, &sb)==-1) {
		nperror(s->path);
		return -1;
	}
	if (!(schep = malloc(sb.st_size))) {
		nperror("malloc");
		return -1;
	}
	if (!(g = fopen(s->path, "r"))) {
		nperror(s->path);
		return -1;
	}
	if (fread(schep, 1, sb.st_size, g)!=sb.st_size) {
		nperror("fread");
		return -1;
	}
	if (fclose(g)==EOF) {
		return -1;
	}

	if (!s->local) {
		if (distrib_s(s, schep, sb.st_size)==-1) {
			return -1;
		}
	}

	free(schep);

	show_elapsed_time(1, stdout);

	return 0;
}

static int
mkdcw(iccws, f, s, mtime)
struct ICCW *iccws;
RFILES *f;
struct iosched *s;
{
	ssize_t ec;
	size_t newrnum, newcnum;
	struct dxr_raw_vec *rv;
	FILE *g;
	ssize_t raw_max_elem_num = 0;

	iccws[0].s = sdup_init((void *)RS_BASE);
	iccws[1].s = sdup_init((void *)CS_BASE);

	iccws[0].h = nht_init((void *)RH_BASE);
	iccws[1].h = nht_init((void *)CH_BASE);

	report_progress_raw_vec(f, stdout, 0);

	if (rewind_read_raw_vec(f)==-1) {
		fprintf(stderr, "%s: rwind_read_vec failed\n", progname);
		return -1;
	}

	for (iccws[0].num=0; (ec=read_raw_vec(f, &rv))>=0; iccws[0].num++) {
		struct hte *rhe, *che;
		struct dxr_raw_elem *ep, *ee, *epp;
if (refresh_hash) refresh_hash_table(iccws, NULL, 0);

		if (ec==0) {
			fprintf(stderr, "\nwarning: empty document: [%s]\n", rv->key);
		}
		switch (rv->ftype) {
		case RAWXR_TYPE_COUNT:
		case RAWXR_TYPE_RAW:
			if (!(rhe = nlt_append(iccws[0].h, rv->key, iccws[0].s))) {
				fprintf(stderr, "%s: !nlt_append rh: [%s]\n", progname, rv->key);
				return -1;
			}
			break;
		default:
			fprintf(stderr, "%s: unknown file type\n", progname);
			return -1;
			break;
		}

		if (raw_max_elem_num < rv->elem_num) {
			raw_max_elem_num = rv->elem_num;
		}

		qsort(rv->elems, rv->elem_num, sizeof *rv->elems, qcompar);

		for (ep=rv->elems, ee=ep+rv->elem_num; ep<ee; ep=epp) {
			freq_t tf;
			tf = ep->tf;
			for (epp=ep+1; epp<ee && !strcmp(epp->name, ep->name); epp++) {
				tf += epp->tf;
if (tf < 0) {
	return -1;
}
			}
			if (!(che = nht_enter(iccws[1].h, ep->name, iccws[1].s))) {
				fprintf(stderr, "%s: !nht_enter ch\n", progname);
				return -1;
			}
			rhe->i.tf += tf;
if (rhe->i.tf < 0) {
	return -1;
}
			che->i.tf += tf;
if (che->i.tf < 0) {
	return -1;
}
			rhe->i.df++;
if (rhe->i.df < 0) {
	return -1;
}
			che->i.df++;
if (che->i.df < 0) {
	return -1;
}
		}
	}
	printf("\n");

	show_elapsed_time(0, stdout);

	if (iccws[0].num==0) {
		fprintf(stderr, "%s: error: incore-row-num==0\n", progname);
		return -1;
	}

	if (ec==-1) {
		fprintf(stderr, "%s: file read error\n", progname);
		return -1;
	}
	printf("make monolithic table\n");

	if (mkmonolithic(iccws) == -1) {
		fprintf(stderr, "%s: mkmonolithic failed\n", progname);
		return -1;
	}

	/* Row table */
	if (iccws[0].num != nht_nent(iccws[0].h)) {
		fprintf(stderr, "%s: error: incore-row-num != ht_nent(rh)\n", progname);
		return -1;
	}
	printf("compacting rtable\n");
	nht_compact(iccws[0].h, &newrnum, 0);
/*	iccws[0].t = iccws[0].h; */
	if (iccws[0].num != newrnum) {
		fprintf(stderr, "%s: internal error: incore-row-num!=newrnum\n", progname);
		return -1;
	}
	printf("row num=%s\n", sizetostr(iccws[0].num));

	show_elapsed_time(0, stdout);

	/* Column table */
	iccws[1].num = nht_nent(iccws[1].h);
	printf("compacting ctable\n");
	nht_compact(iccws[1].h, &newcnum, 1);
/*	iccws[1].t = iccws[1].h; */
	printf("col num=%s, new col num=%s\n", sizetostr(iccws[1].num), sizetostr(newcnum));
	iccws[1].num = newcnum;

	show_elapsed_time(0, stdout);

	printf("scheduling\n");
	if (schedule(s, iccws[0].h, iccws[0].num, iccws[1].h, iccws[1].num, raw_max_elem_num)==-1) {
		fprintf(stderr, "%s: schedule failed\n", progname);
		return -1;
	}

	printf("creating [%s]\n", s->dcw_path[0]);
	if (!(g = fopen(s->dcw_path[0], "w"))) {
		nperror(s->dcw_path[0]);
		return -1;
	}
	if (dcw_write(&iccws[0], DCW_ROW, mtime, g, (void*)RI_BASE)==-1) {
		fprintf(stderr, "%s: dcw_write failed\n", progname);
		(void)fclose(g);
		return -1;
	}
	if (fclose(g)==EOF) {
		return -1;
	}

	printf("creating [%s]\n", s->dcw_path[1]);
	if (!(g = fopen(s->dcw_path[1], "w"))) {
		nperror(s->dcw_path[1]);
		return -1;
	}
	if (dcw_write(&iccws[1], DCW_COL, mtime, g, (void*)CI_BASE)==-1) {
		fprintf(stderr, "%s: dcw_write failed\n", progname);
		(void)fclose(g);
		return -1;
	}
	if (fclose(g)==EOF) {
		return -1;
	}

	return 0;
}

static int
dcw_write(iccw, type, mtime, f, addr)
struct ICCW *iccw;
FILE *f;
void *addr;
{
	struct dcw_hdr h;
	size_t s, i;
	struct bindex_t bi;
	struct rindex_t ri;
	size_t entry_table_size;
	size_t r_index_table_size;
	size_t fullwd_offset;
/*	size_t fwp, pad; */
	struct hte *e;
/*	char *fullwd_base; */
	struct sdup *sd;
	size_t size;

	sd = iccw->s;
	assert(sd->base != NULL);
	if (fill_dcw_hdr(&h, type, mtime)==-1) {
		fprintf(stderr, "fill_dcw_hdr failed\n");
		return -1;
	}

	for (i=0; i<iccw->num; i++) {
		idx_t id;
		id = i + 1;	/* XXX */
		e = hta(iccw->h, i);
		if (e->i.id != id) {
			fprintf(stderr, "%s: scheduler error: incorrect id\n", progname);
			return -1;
		}
	}

	h.nentries = iccw->num;

	h.rindex_type = rindex_type;
	switch (h.rindex_type) {
	case RIT_HASH:
		printf("creating hash-index (%s)\n", sizetostr(iccw->num));
		if (!(iccw->hi = make_hash_index(iccw->h, iccw->num, addr))) {
			return -1;
		}
		r_index_table_size = iccw->hi->size * sizeof ri;
		h.size = iccw->hi->size;
		assert(iccw->num == iccw->hi->nentries);
		break;
	case RIT_BSEARCH:
		printf("creating binary-index (%s)\n", sizetostr(iccw->num));
		if (!(iccw->hi = make_bin_index(iccw->h, iccw->num, addr))) {
			return -1;
		}
		r_index_table_size = (iccw->num+1)*sizeof ri;
		assert(iccw->num == iccw->hi->size);
		assert(iccw->num == iccw->hi->nentries);
		break;
	default:
		return -1;
	}

	entry_table_size = (iccw->num+1)*sizeof bi;
	fullwd_offset = NALIGN(sizeof h + entry_table_size + r_index_table_size, PAGESIZE);

	h.bindex_offset = sizeof h;
	h.rindex_offset = sizeof h + entry_table_size;
	h.fullwd_offset = fullwd_offset;

	if (fseeko(f, fullwd_offset, SEEK_SET)==-1) {
		return -1;
	}
/*
	if (fwrite("", 1, 1, f)!=1) {
		return -1;
	}
*/
	size = sd->pend - sd->base;
	if (fwrite(sd->base, 1, size, f) != size) {
		return -1;
	}
/*
	fwp = 1;
	for (i = 0; i < iccw->num; i++) {
		e = hta(iccw->h, i);
		size_t l = strlen(e->name.p) + 1;
		if (fwrite(e->name.p, 1, l, f) != l) {
			return -1;
		}
		fwp += l;
	}
	if (fwp % PAGESIZE) {
		pad = PAGESIZE - fwp % PAGESIZE;
		for (i = 0; i < pad; i++) {
			if (fwrite("", 1, 1, f) != 1) {
				return -1;
			}
		}
	}
*/

	rewind(f);

	if (fwrite(&h, sizeof h, 1, f)!=1) {
		return -1;
	}

	bi.b = 0;
	if (fwrite(&bi, sizeof bi, 1, f)!=1) {
		return -1;
	}

	for (i = 0; i < iccw->num; i++) {
		e = hta(iccw->h, i);
		bi.b = e->name.p ? e->name.p - sd->base : 0;
		if (fwrite(&bi, sizeof bi, 1, f)!=1) {
			return -1;
		}
	}

	assert(ftello(f) == sizeof h + entry_table_size);
	switch (h.rindex_type) {
	case RIT_HASH:
		for (s=0; s<iccw->hi->size; s++) {
			e = iccw->hi->rr[s];
			if (e) {
				ri.r = e->i.id;
				ri.b = e->name.p ? e->name.p - sd->base : 0;
/*
fprintf(stderr, "%s %s\n", idxtostr(e->i.id), e->name.p ? e->name.p : "(null)");
*/
			}
			else {
				ri.r = 0;
				ri.b = 0;
			}
			if (fwrite(&ri, sizeof ri, 1, f)!=1) {
				return -1;
			}
		}
		break;
	case RIT_BSEARCH:
		ri.r = 0;
		ri.b = 0;
		if (fwrite(&ri, sizeof ri, 1, f)!=1) {
			return -1;
		}
		for (i=0; i<iccw->num; i++) {
			e = iccw->hi->rr[i];
			ri.r = e->i.id;
			ri.b = e->name.p ? e->name.p - sd->base : 0;
/*
fprintf(stderr, "%s %s\n", idxtostr(e->i.id), e->name.p ? e->name.p : "(null)");
*/
			if (fwrite(&ri, sizeof ri, 1, f)!=1) {
				return -1;
			}
		}
		break;
	default:
		return -1;
	}
	assert(ftello(f) == sizeof h + entry_table_size + r_index_table_size);
	assert(ftello(f) <= fullwd_offset);
	return 0;
}

static int
qcompar(va, vb)
void const *va, *vb;
{
	struct dxr_raw_elem const *a, *b;
	a = va;
	b = vb;
	return strcmp(a->name, b->name);
}

extern int dwerrno;

static int
mkdxrr(f, s, mtime, cws, iccws)
RFILES *f;
struct iosched *s;
DWFILE *cws[];
struct ICCW *iccws;
{
	ssize_t ec;
	size_t rnum;
	struct dxr_raw_vec *rv;
	struct xr_vec *v = NULL;
	struct hte *e;

	if (init_output(s, &s->scc[0], s->handle, DXR_ROW, mtime, 0, s->compress)==-1) {
		fprintf(stderr, "mkdxrr: init_output failed (1)\n");
		return -1;
	}
#if defined DMKXRC
	if (cmode==1) {
		if (init_output(s, &s->scc[1], s->handle, DXR_COL, mtime, 1, s->compress)==-1) {
			fprintf(stderr, "init_output failed (2)\n");
			return -1;
		}
	}
#endif
	if (!(v = malloc(sizeof_dxr_vec(s->scc[0].raw_max_elem_num)))) {
		nperror("mkdxrr: malloc");
		return -1;
	}

	report_progress_raw_vec(f, stdout, 0);

	if (rewind_read_raw_vec(f)==-1) {
		fprintf(stderr, "%s: rewind_read_vec failed\n", progname);
		return -1;
	}

	for (rnum=0; (ec=read_raw_vec(f, &rv))>=0; rnum++) {
		idx_t id;
if (refresh_hash) refresh_hash_table(iccws, NULL, 0);

		assert(rv->elem_num <= s->scc[0].raw_max_elem_num);
		if ((id = rv2v(rv, v, cws, iccws)) == 0) {
			fprintf(stderr, "mkdxrr: rv2v failed\n");
			return -1;
		}

		e = hta(s->tg[0].h, id-1);
		assert(id == e->i.id);
		assert(v->elem_num == e->i.df);
		assert(v->freq_sum == e->i.tf);

		if (output_vec(s, &s->scc[0], &s->scc[1], id, v)==-1) {
			fprintf(stderr, "mxdxrr: output_vec failed\n");
			return -1;
		}
	}

	if (ec==-1) {
		fprintf(stderr, "mkdxrr: read_raw_vec failed\n");
		return -1;
	}

	printf("\n");

	if (finish_output(s, &s->scc[0])==-1) {
		fprintf(stderr, "mkdxrr: finish_output failed\n");
		return -1;
	}
#if defined DMKXRC
	if (cmode==1) {
		if (finish_output(s, &s->scc[1])==-1) {
			fprintf(stderr, "mkdxrr: finish_output failed\n");
			return -1;
		}
	}
#endif

	free(v);

	return 0;
}

static int
mkdxrc(s, mtime)
struct iosched *s;
{
	idx_t id;
	struct xr_vec *v;
	struct xr_vec **vtbl;
	void *v0;
	struct xr_vec const *ov;
	size_t p;
	size_t v0size;
	idx_t ide;
	struct hte *e;

	if (init_output(s, &s->scc[1], s->handle, DXR_COL, mtime, 0, s->compress)==-1) {
		fprintf(stderr, "init_output failed (3)\n");
		return -1;
	}

	if (!(v0size = allocv0(s->v0size, &v0))) {
		return -1;
	}

	for (p=0; p<s->scc[0].np; p++) {
		int ss;
		struct sched_iocb *iocb = &s->scc[0].iocb[p];
		if (s->local) {
			if (!(iocb->L.d = dxr_open(iocb->L.path, s->handle))) {
				fprintf(stderr, "dxr_open %s failed\n", iocb->L.path);
				return -1;
			}
			assert(s->scc[0].n == dxr_size(iocb->L.d));
		}
		else {
			if ((ss = nnet_cli(iocb->R.host, iocb->R.serv))==-1) {
				perror(iocb->R.host);
				return -1;
			}

			if (!(iocb->R.g = dwc_dxr_open(ss, s->handle, DXR_ROW, p, -1))) {
				return -1;
			}
			assert(s->scc[0].n == dwc_dxr_size(iocb->R.g));
		}
	}

	for (id = 1; id<=s->scc[1].n; id=ide) {
		idx_t ido, ids, idx;
		size_t total_vs, nvtble;
		ids = id;
/*		ide = id+1; */

		total_vs = 0;
		nvtble = 0;
		for (idx=ids; idx<=s->scc[1].n; idx++) {
			ssize_t elem_num;
			size_t vs;
			e = hta(s->tg[1].h, idx-1);
			elem_num = e->i.df;
/*fprintf(stderr, "C idx = %d, ", idx);*/
/*fprintf(stderr, "elem_num = %s\n", freqtostr(e->i.df));*/
			vs = sizeof_dxr_vec(elem_num);
			if ((total_vs+vs) + (nvtble+1)*sizeof vtbl[0] >= v0size) {
				if (!nvtble) {
					fprintf(stderr, "%s: NG break at vs == %s\n", progname, sizetostr(vs));
					fprintf(stderr, "%s: failed to allocate work-area.\n", progname);
					fprintf(stderr, "%s: increase v0size (==%s) and try again, please.\n", progname, sizetostr(v0size));
					return -1;
				}
				break;
			}
			total_vs += vs;
			nvtble++;
		}
		ide = idx;

		assert(nvtble>0);

		vtbl = (struct xr_vec **)((char *)v0 + v0size - nvtble * sizeof vtbl[0]);
		total_vs = 0;
		for (idx=ids; idx<ids+nvtble; idx++) {
			ssize_t elem_num;
			size_t vs;
			e = hta(s->tg[1].h, idx-1);
			elem_num = e->i.df;
			vs = sizeof_dxr_vec(elem_num);
			vtbl[idx-ids] = (struct xr_vec *)((char *)v0 + total_vs);
			total_vs += vs;
		}
		assert(idx==ide);
		assert((void *)((char *)v0 + total_vs) <= (void *)vtbl);
		printf("~%s", sizetostr(ide-1));
		fflush(stdout);

		for (idx=ids; idx<ide; idx++) {
			ssize_t elem_num;
			e = hta(s->tg[1].h, idx-1);
			elem_num = e->i.df;
			v = vtbl[idx-ids];
			memset(v, 0, sizeof_dxr_vec(elem_num));
		}
		if (!s->local) {
			for (p=0; p<s->scc[0].np; p++) {
				struct sched_iocb *iocb = &s->scc[0].iocb[p];
				if (dwc_dxr_seq_init(iocb->R.g) == -1) {
					fprintf(stderr, "dwc_dxr_seq_init failed\n");
					return -1;
				}
			}
		}

		for (ido = 1; ido<=s->scc[0].n; ido++) {
			for (p=0; p<s->scc[0].np; p++) {
				struct sched_iocb *iocb = &s->scc[0].iocb[p];
				ssize_t k;
				if (s->local) {
					if (dxr_get_vec(iocb->L.d, ido, &ov)==-1) {
						fprintf(stderr, "dxr_get_vec failed\n");
						return -1;
					}
				}
				else {
					if (dwc_dxr_seq_get_vec(iocb->R.g, ido, &ov)==-1) {
						fprintf(stderr, "dwc_dxr_seq_get_vec failed\n");
						return -1;
					}
				}
				for (k=0; k<ov->elem_num; k++) {
					struct xr_elem const *ovk = &ov->elems[k];
					idx = ovk->id;
					if (ids <= idx && idx < ide) {
						freq_t freq;
						struct xr_elem *ve;
						v = vtbl[idx-ids];

						freq = ovk->freq;
						ve = &v->elems[v->elem_num];
						if (ve->id != ido) {
							if (ve->id) {
								v->elem_num++;
								ve++;
e = hta(s->tg[1].h, idx-1);
assert(v->elem_num < e->i.df);
							}
							else {
								assert(v->elem_num == 0);
								assert(v->freq_sum == 0);
							}
							ve->id = ido;
							ve->freq = 0;
						}
/* else fprintf(stderr, "!"); */
						ve->freq += freq;
if (ve->freq < 0) {
	return -1;
}
						v->freq_sum += freq;
if (v->freq_sum < 0) {
	return -1;
}
					}
				}
			}
		}

		show_elapsed_time(2, stdout);

		if (!s->local) {
			for (p=0; p<s->scc[0].np; p++) {
				struct sched_iocb *iocb = &s->scc[0].iocb[p];
				if (dwc_dxr_seq_finish(iocb->R.g) == -1) {
					fprintf(stderr, "dwc_dxr_seq_finish failed\n");
					return -1;
				}
			}
		}

		for (idx = ids; idx<ide; idx++) {
			v = vtbl[idx-ids];
			if (v->elems[v->elem_num].id) {
				v->elem_num++;
			}

			e = hta(s->tg[1].h, idx-1);
/*
fprintf(stderr, "idx = %d e->i.id = %d\n", idx, e->i.id);
fprintf(stderr, "v->elem_num = %d e->i.df = %d\n", v->elem_num, e->i.df);
fprintf(stderr, "v->freq_sum = %d, e->i.tf = %d\n", v->freq_sum, e->i.tf);
*/
			assert(idx == e->i.id);
			assert(v->elem_num == e->i.df);
			assert(v->freq_sum == e->i.tf);

			if (output_vec(s, &s->scc[1], &s->scc[0], idx, v)==-1) {
				return -1;
			}
		}
	}

	for (p=0; p<s->scc[0].np; p++) {
		struct sched_iocb *iocb = &s->scc[0].iocb[p];
		if (s->local) {
			if (dxr_close(iocb->L.d)==-1) {
				return -1;
			}
		}
		else {
			if (dwc_close(iocb->R.g)==-1) {
				return -1;
			}
		}
	}
	printf("\n");

	if (finish_output(s, &s->scc[1])==-1) {
		return -1;
	}
	freev0(v0);
	return 0;
}

#if defined MKDXRCF

static int
mkdxrcf(f, s, mtime, cws, iccws)
RFILES *f;
struct iosched *s;
DWFILE *cws[];
struct ICCW *iccws;
{
	struct dxr_raw_vec *rv;
	idx_t id;
	struct xr_vec *v;
	struct xr_vec **vtbl;
	void *v0;
	struct xr_vec *ov;
	size_t v0size;
	idx_t ide;
refresh_hash = 1;
	struct hte *e;

	if (init_output(s, &s->scc[1], s->handle, DXR_COL, mtime, 0, s->compress)==-1) {
		fprintf(stderr, "init_output failed (4)\n");
		return -1;
	}

	if (!(v0size = allocv0(s->v0size, &v0))) {
		return -1;
	}

	if (!(ov = malloc(sizeof_dxr_vec(s->scc[0].raw_max_elem_num)))) {
		return -1;
	}

	report_progress_raw_vec(f, stdout, 1);

	for (id = 1; id<=s->scc[1].n; id=ide) {
		idx_t ids, idx;
		size_t total_vs, nvtble;
		ids = id;
/*		ide = id+1; */

		total_vs = 0;
		nvtble = 0;
		for (idx=ids; idx<=s->scc[1].n; idx++) {
			ssize_t elem_num;
			size_t vs;
			e = hta(s->tg[1].h, idx-1);
			elem_num = e->i.df;
			vs = sizeof_dxr_vec(elem_num);
			if ((total_vs+vs) + (nvtble+1)*sizeof vtbl[0] >= v0size) {
				if (!nvtble) {
					fprintf(stderr, "%s: NG break at vs == %s\n", progname, sizetostr(vs));
					fprintf(stderr, "%s: failed to allocate work-area.\n", progname);
					fprintf(stderr, "%s: increase v0size (==%s) and try again, please.\n", progname, sizetostr(v0size));
					return -1;
				}
				break;
			}
			total_vs += vs;
			nvtble++;
		}
		ide = idx;

		assert(nvtble>0);

		vtbl = (struct xr_vec **)((char *)v0 + v0size - nvtble * sizeof vtbl[0]);
		total_vs = 0;
		for (idx=ids; idx<ids+nvtble; idx++) {
			ssize_t elem_num;
			size_t vs;
			e = hta(s->tg[1].h, idx-1);
			elem_num = e->i.df;
			vs = sizeof_dxr_vec(elem_num);
			vtbl[idx-ids] = (struct xr_vec *)((char *)v0 + total_vs);
			total_vs += vs;
		}
		assert(idx==ide);
		assert((void *)((char *)v0 + total_vs) <= (void *)vtbl);
		printf("~%s", sizetostr(ide-1));
		fflush(stdout);

		for (idx=ids; idx<ide; idx++) {
			ssize_t elem_num;
			e = hta(s->tg[1].h, idx-1);
			elem_num = e->i.df;
			v = vtbl[idx-ids];
			memset(v, 0, sizeof_dxr_vec(elem_num));
		}

		if (rewind_read_raw_vec(f)==-1) {
			fprintf(stderr, "%s: rwind_read_vec failed\n", progname);
			return -1;
		}

		for (; read_raw_vec(f, &rv)>=0; ) {
			ssize_t k;
			idx_t ido;
if (refresh_hash) refresh_hash_table(iccws, v0, v0size);

			assert(rv->elem_num <= s->scc[0].raw_max_elem_num);
			if ((ido = rv2v(rv, ov, cws, iccws)) == 0) {
				return -1;
			}

			for (k=0; k<ov->elem_num; k++) {
				struct xr_elem *ovk = &ov->elems[k];
				idx = ovk->id;
				if (ids <= idx && idx < ide) {
					freq_t freq;
					struct xr_elem *ve;
					v = vtbl[idx-ids];

					freq = ovk->freq;
					ve = &v->elems[v->elem_num];
					if (ve->id != ido) {
						if (ve->id) {
							v->elem_num++;
							ve++;
e = hta(s->tg[1].h, idx-1);
assert(v->elem_num < e->i.df);
						}
						else {
							assert(v->elem_num == 0);
							assert(v->freq_sum == 0);
						}
						ve->id = ido;
						ve->freq = 0;
					}
/* else fprintf(stderr, "?"); */
					ve->freq += freq;
if (ve->freq < 0) {
	return -1;
}
					v->freq_sum += freq;
if (v->freq_sum < 0) {
	return -1;
}
				}
			}
		}

		show_elapsed_time(2, stdout);

		for (idx = ids; idx<ide; idx++) {
			v = vtbl[idx-ids];
			if (v->elems[v->elem_num].id) {
				v->elem_num++;
			}
			qsort(v->elems, v->elem_num, sizeof *v->elems, ecompar);

			e = hta(s->tg[1].h, idx-1);
			assert(idx == e->i.id);
			assert(v->elem_num == e->i.df);
			assert(v->freq_sum == e->i.tf);

			if (output_vec(s, &s->scc[1], &s->scc[0], idx, v)==-1) {
				return -1;
			}
		}
	}

	printf("\n");
	if (finish_output(s, &s->scc[1])==-1) {
		return -1;
	}
	freev0(v0);
	free(ov);

	printf(" done.\n");
	return 0;
}

#endif

idx_t
rv2v(rv, v, cws, iccws)
struct dxr_raw_vec const *rv;
struct xr_vec *v;
DWFILE *cws[];
struct ICCW *iccws;
{
	idx_t id;
idx_t i;
	freq_t freq_sum;
	df_t elem_num;
	struct dxr_raw_elem const *ep, *ee;
	struct xr_elem *vp;
	switch (rv->ftype) {
	case RAWXR_TYPE_COUNT:
	case RAWXR_TYPE_RAW:
		break;
	default:
		return 0;
	}

	if ((id = dcw_name2id(cws[0], rv->key)) == 0) {
		fprintf(stderr, "lookup failed: \"%s\"\n", rv->key);
		return 0;
	}
i = h_name2id(iccws[0].hi, rv->key);
assert(i == id);

	vp = v->elems;
	freq_sum = 0;
	for (ep=rv->elems, ee=ep+rv->elem_num; ep<ee; ep++) {
		if ((vp->id = dcw_name2id(cws[1], ep->name)) == 0) {
			fprintf(stderr, "lookup failed: \"%s\"\n", ep->name);
			return 0;
		}
i = h_name2id(iccws[1].hi, ep->name);
assert(i == vp->id);
		vp->freq = ep->tf;
		freq_sum += ep->tf;
if (freq_sum < 0) {
	return 0;
}
		vp++;
	}
#if 1
/* Tue Jan 21 14:32:48 JST 2003 */
	elem_num = vp - v->elems;
if (elem_num != vp - v->elems) {
	return 0;
}
	qsort(v->elems, elem_num, sizeof *v->elems, ecompar);
	if ((elem_num = uniq_elems(v->elems, elem_num)) == -1) {
		return 0;
	}
#else
	qsort(v->elems, rv->elem_num, sizeof *v->elems, ecompar);
	elem_num = uniq_elems(v->elems, rv->elem_num);
#endif
	v->elem_num = elem_num;
	v->freq_sum = freq_sum;
	return id;
}

static ssize_t
uniq_elems(v0, elem_num)
struct xr_elem *v0;
ssize_t elem_num;
{
	struct xr_elem *vp, *ve, *vpp, *vps;
	ssize_t ec;
	for (vp=v0, vps=v0, ve=vp+elem_num; vp<ve; vp=vpp) {
		idx_t id0 = vp->id;
		freq_t freq;
		freq = vp->freq;
		for (vpp=vp+1; vpp<ve && vpp->id==id0; vpp++) {
			freq += vpp->freq;
if (freq < 0) {
	return -1;
}
		}
		vps->id = id0;
		vps->freq = freq;
		vps++;
	}
	ec = vps - v0;
	return ec;
}

static int
init_output(s, sc, handle, type, mtime, trc_vec, compress)
struct iosched *s;
struct sched *sc;
char const *handle;
{
	size_t p;
	size_t svs = sc->svec_size;

	if (init_iocb(s, &sc->iocb0, handle, sc->n, type, mtime, -1, svs, 0, compress)==-1) {
		return -1;
	}
	for (p=0; p<sc->np; p++) {
		struct sched_iocb *iocb = &sc->iocb[p];

		if (init_iocb(s, iocb, handle, sc->n, type, mtime, p, svs, trc_vec, compress)==-1) {
			return -1;
		}
	}
	return 0;
}

static int
init_iocb(s, iocb, handle, n, type, mtime, fragment, svec_size, trc_vec, compress)
size_t n;
struct iosched *s;
struct sched_iocb *iocb;
char const *handle;
ssize_t fragment;
size_t svec_size;
{
	struct xr_vec zero = ZEROVEC;
	size_t vec_offset;
	size_t entry_table_size;
	struct dxr_entry bi;
	struct dxr_hdr *h;
	size_t sz;

	iocb->svec = NULL;
	iocb->svec_elem_num = 0;
	sz = sizeof_dxr_vec(iocb->svec_elem_num);
	if (svec_size && !(iocb->svec = malloc(sz))) {
		nperror("malloc");
		return -1;
	}

	if (s->local) {
		h = &iocb->L.dwc.h;

		if (fill_dxr_hdr(h, type, mtime, fragment)==-1) {
			return -1;
		}
		h->nentries = n;
		h->max_elem_num = 0;
		h->max_freq_sum = 0;
		h->total_elem_num = 0;
		h->total_freq_sum = 0;
		h->compress_mode = compress;

		iocb->L.remote.n = n;
		entry_table_size = (n+1)*sizeof bi;
		vec_offset = sizeof *h + entry_table_size;

		h->entry_table_offset = sizeof *h;
		h->vector_table_offset = vec_offset;

		if (!(iocb->L.remote.stream = fopen(iocb->L.path, "w"))) {
			nperror(iocb->L.path);
			return -1;
		}
		if (fseeko(iocb->L.remote.stream, vec_offset, SEEK_SET)==-1) {
			return -1;
		}

		if (fwrite(&zero, sizeof zero, 1, iocb->L.remote.stream)!=1) {
			return -1;
		}
		iocb->L.remote.addr = sizeof zero;

		if (!(iocb->L.remote.bindex = calloc(n+1, sizeof *iocb->L.remote.bindex))) {
			return -1;
		}
#if defined DMKXRC
		iocb->L.remote.trc_vec = trc_vec;
		if (trc_vec) {
			size_t i;
			for (i=1; i<=n; i++) {
				struct cel_t **p;

				p = (struct cel_t **)&iocb->L.remote.bindex[i].addr;
				*p = NULL;
			}
		}
#endif
	}
	else {
		int so;
		if ((so = nnet_cli(iocb->R.host, iocb->R.serv))==-1) {
			perror(iocb->R.host);
			return -1;
		}

		if (!(iocb->R.f = dwc_dxr_create(so, handle, n, type, mtime, fragment, svec_size, trc_vec, compress))) {
			fprintf(stderr, "dwc_dxr_create failed (1)\n");
			return -1;
		}
	}
	return 0;
}

static int
finish_output(s, sc)
struct iosched *s;
struct sched *sc;
{
	size_t p;
	struct sched_iocb *iocb;

printf("flush output: ");
	iocb = &sc->iocb0;
	if (flush_iocb_nowait(s, iocb)==-1) {
		fprintf(stderr, "mkdxrr: finish_output: flush_iocb failed (fragment S)\n");
		return -1;
	}
printf("%s ", ssizetostr(iocb->svec_elem_num));
	free(iocb->svec);

	for (p=0; p<sc->np; p++) {
		iocb = &sc->iocb[p];
		if (flush_iocb_nowait(s, iocb)==-1) {
			fprintf(stderr, "mkdxrr: finish_output: flush_iocb failed (fragment %s)\n", sizetostr(p));
			return -1;
		}
printf("%s ", ssizetostr(iocb->svec_elem_num));
		free(iocb->svec);
	}
printf("done.\n");

printf("wait: ");
	iocb = &sc->iocb0;
	if (flush_iocb_wait(s, iocb)==-1) {
		fprintf(stderr, "mkdxrr: finish_output: flush_iocb failed (fragment S)\n");
		return -1;
	}
printf(".");

	for (p=0; p<sc->np; p++) {
		iocb = &sc->iocb[p];
		if (flush_iocb_wait(s, iocb)==-1) {
			fprintf(stderr, "mkdxrr: finish_output: flush_iocb failed (fragment %s)\n", sizetostr(p));
			return -1;
		}
printf(".");
	}
printf(" done.\n");

printf("close: ");
	iocb = &sc->iocb0;
	if (close_iocb(s, iocb)==-1) {
		fprintf(stderr, "mkdxrr: finish_output: close_iocb failed (fragment S)\n");
		return -1;
	}
printf(".");

	for (p=0; p<sc->np; p++) {
		iocb = &sc->iocb[p];
		if (close_iocb(s, iocb)==-1) {
			fprintf(stderr, "mkdxrr: finish_output: close_iocb failed (fragment %s)\n", sizetostr(p));
			return -1;
		}
printf(".");
	}
printf(" done.\n");
	return 0;
}

static int
make_summary_information(iocb)
struct sched_iocb *iocb;
{
	struct dxr_hdr *h;
	struct dxr_entry *b;
	idx_t id;
	h = &iocb->L.dwc.h;
	b = iocb->L.remote.bindex;
	for (id = 1; id<=h->nentries; id++) {
		ssize_t elem_num = b[id].elem_num;
		freq_t freq_sum = b[id].freq_sum;
if (id < 0) {
	return -1;
}
		if (h->max_elem_num < elem_num) {
			h->max_elem_num = elem_num;
		}
		if (h->max_freq_sum < freq_sum) {
			h->max_freq_sum = freq_sum;
		}
		h->total_elem_num += elem_num;
if (h->total_elem_num < 0) {
	return -1;
}
		h->total_freq_sum += freq_sum;
if (h->total_freq_sum < 0) {
	return -1;
}
	}
if (id < 0) {
	return -1;
}
	return 0;
}

#if defined DMKXRC
static int
flush_trc_vec(iocb)
struct sched_iocb *iocb;
{
	idx_t id;
	struct xr_vec *v = NULL;
	size_t vsize = 0;
	for (id = 1; id<=iocb->L.remote.n; id++) {
		ssize_t i;
/*		size_t size; */
		struct cel_t *q, **p;
		ssize_t size;
		int compress;
		p = (struct cel_t **)&iocb->L.remote.bindex[id].addr;
		for (q=*p, i=0; q; q = q->n, i++) ;
		if ((size = sizeof_dxr_vec(i))>vsize) {
			struct xr_vec *new;
			size = NALIGN(size, PAGESIZE);
			if (!(new = realloc(v, size))) {
				free(v);
				return -1;
			}
			v = new;
			vsize = size;
		}
		v->freq_sum = 0;
		for (q=*p, i=0; q; q = q->n, i++) {
			struct xr_elem *e = &v->elems[i];
			e->id = q->id;
			e->freq = q->freq;
			v->freq_sum += q->freq;
if (v->freq_sum < 0) {
	return -1;
}
		}
		v->elem_num = i;
if (v->elem_num != i) {
	return -1;
}
		qsort(v->elems, v->elem_num, sizeof *v->elems, ecompar);
/*
		size = sizeof_dxr_vec(v->elem_num);
		if (fwrite(v, size, 1, iocb->L.remote.stream)!=1) {
			return -1;
		}
*/
		compress = iocb->L.dwc.h.compress_mode;
		if ((size = wvefc(v, iocb->L.remote.stream, compress))==-1) {
			return -1;
		}
		iocb->L.remote.bindex[id].addr = iocb->L.remote.addr;
		iocb->L.remote.addr += size;
		iocb->L.remote.bindex[id].elem_num = v->elem_num;
		iocb->L.remote.bindex[id].freq_sum = v->freq_sum;
	}
	free(v);
	return 0;
}
#endif

static int
flush_iocb_nowait(s, iocb)
struct iosched *s;
struct sched_iocb *iocb;
{
	if (s->local) {
#if defined DMKXRC
		if (cmode==1) {
			if (iocb->L.remote.trc_vec) {
				if (flush_trc_vec(iocb)==-1) {
					return -1;
				}
			}
		}
#endif

		assert(iocb->L.remote.n == iocb->L.dwc.h.nentries);
		fwrite("\n#\n#\n#\n", 7, 1, iocb->L.remote.stream);
		if (make_summary_information(iocb) == -1) {
			return -1;
		}

		rewind(iocb->L.remote.stream);
		if (fwrite(&iocb->L.dwc.h, sizeof iocb->L.dwc.h, 1, iocb->L.remote.stream)!=1) {
			return -1;
		}

		if (fwrite(iocb->L.remote.bindex, sizeof *iocb->L.remote.bindex, iocb->L.remote.n+1, iocb->L.remote.stream)!=iocb->L.remote.n+1) {
			return -1;
		}
	}
	else {
		if (dwc_flush_nowait(iocb->R.f)==-1) {
			return -1;
		}
	}
	return 0;
}

static int
flush_iocb_wait(s, iocb)
struct iosched *s;
struct sched_iocb *iocb;
{
	if (s->local) {
		return 0;
	}
	else {
		return dwc_flush_wait(iocb->R.f);
	}
}

static int
close_iocb(s, iocb)
struct iosched *s;
struct sched_iocb *iocb;
{
	if (s->local) {
		if (fclose(iocb->L.remote.stream)==EOF) {
			return -1;
		}
	}
	else {
		if (dwc_close(iocb->R.f)==-1) {
			return -1;
		}
	}
	return 0;
}

static int
write_vec(iocb, id, v, fragment)
struct sched_iocb *iocb;
idx_t id;
struct xr_vec const *v;
ssize_t fragment;
{
	if (v->elem_num) {
		if (fragment == WVE_WRITE) {
			ssize_t size;
			int compress = iocb->L.dwc.h.compress_mode;
			if ((size = wvefc(v, iocb->L.remote.stream, compress))==-1) {
				return -1;
			}
			iocb->L.remote.bindex[id].addr = iocb->L.remote.addr;
			iocb->L.remote.addr += size;
		}
		else {
			iocb->L.remote.bindex[id].addr = fragment;
		}
		iocb->L.remote.bindex[id].elem_num = v->elem_num;
		iocb->L.remote.bindex[id].freq_sum = v->freq_sum;
	}
	return 0;
}

#if defined DMKXRC
static int
add_trc_vec(iocb, id, v, fragment)
struct sched_iocb *iocb;
idx_t id;
struct xr_vec const *v;
ssize_t fragment;
{
	ssize_t i;
	if (v->elem_num) {
		if (fragment == WVE_WRITE) {
			assert(iocb->L.remote.trc_vec);
			for (i=0; i<v->elem_num; i++) {
				struct xr_elem const *e = &v->elems[i];
				idx_t j = e->id;
				struct cel_t *cel, **p;

				p = (struct cel_t **)&iocb->L.remote.bindex[j].addr;

				if (!(cel = new_cel())) {
					return -1;
				}
				cel->id = id;
				cel->freq = e->freq;
				cel->n = *p;
				*p = cel;
			}
		}
		else {
			assert(!iocb->L.remote.trc_vec);
			for (i=0; i<v->elem_num; i++) {
				struct xr_elem const *e = &v->elems[i];
				idx_t j = e->id;

				iocb->L.remote.bindex[j].elem_num++;
				iocb->L.remote.bindex[j].freq_sum += e->freq;
if (iocb->L.remote.bindex[j].freq_sum < 0) {
	return -1;
}
			}
		}
	}
	return 0;
}
#endif

static int
output_vec(s, sc, osc, id, v)
struct iosched *s;
struct sched *sc, *osc;
idx_t id;
struct xr_vec const *v;
{
	ssize_t i;
	ssize_t p;

	assert(id != 0);
	assert(id<=sc->n);

	/* clear all sub-vectors */
	for (p=0; p<sc->np; p++) {
		struct sched_iocb *iocb = &sc->iocb[p];
		struct xr_vec *sv = iocb->svec;
		sv->elem_num = 0;
		sv->freq_sum = 0;
	}

	for (i=0; i<v->elem_num; i++) {
		struct xr_vec *sv;
		struct xr_elem *ep;
		struct sched_iocb *iocb;
		struct xr_elem const *e = &v->elems[i];
		idx_t oid = e->id;
		freq_t freq = e->freq;
		p = (*osc->selpp)(osc, oid);
		assert(0<=p);
		assert(p<sc->np);
		iocb = &sc->iocb[p];
		sv = iocb->svec;
		if (sv->elem_num >= iocb->svec_elem_num) {
			ssize_t en;
			size_t sz;
			en = sv->elem_num + sv->elem_num/2 + 1024;
			iocb->svec_elem_num = MIN(en, sc->max_elem_num);
			sz = sizeof_dxr_vec(iocb->svec_elem_num);
			if (!(sv = realloc(iocb->svec, sz))) {
				return -1;
			}
			iocb->svec = sv;
			printf("{%s/%s}", ssizetostr(iocb->svec_elem_num), ssizetostr(sc->max_elem_num));
		}
		assert(sv->elem_num < sc->max_elem_num);
		ep = &sv->elems[sv->elem_num++];
		ep->id = oid;
		ep->freq = freq;
		sv->freq_sum += freq;
if (sv->freq_sum < 0) {
	return -1;
}
	}

	for (p=0; p<sc->np; p++) {
		struct sched_iocb *iocb = &sc->iocb[p];
		if (s->local) {
			if (write_vec(iocb, id, iocb->svec, WVE_WRITE)==-1) {
				return -1;
			}
		}
		else {
			if (dwc_dxr_write_vec(iocb->R.f, id, iocb->svec, WVE_WRITE)==-1) {
				return -1;
			}
		}
		iocb->total_elem_num += iocb->svec->elem_num;
if (iocb->total_elem_num < 0) {
	return -1;
}
	}
	p = (*sc->selpp)(sc, id);
	assert(0<=p);
	assert(p<osc->np);
	if (s->local) {
		if (write_vec(&sc->iocb0, id, v, p)==-1) {
			return -1;
		}
	}
	else {
		if (dwc_dxr_write_vec(sc->iocb0.R.f, id, v, p)==-1) {
			return -1;
		}
	}
	sc->iocb0.total_elem_num += v->elem_num;
if (sc->iocb0.total_elem_num < 0) {
	return -1;
}

#if defined DMKXRC
	if (cmode==1) {
		struct sched_iocb *iocb;
		iocb = &osc->iocb[p];
		if (s->local) {
			if (add_trc_vec(iocb, id, v, WVE_WRITE)==-1) {
				return -1;
			}
			if (add_trc_vec(&osc->iocb0, id, v, p)==-1) {
				return -1;
			}
		}
		else {
			if (dwc_add_trc_vec(iocb->R.f, id, v, WVE_WRITE)==-1) {
				return -1;
			}
			if (dwc_add_trc_vec(osc->iocb0.R.f, id, v, p)==-1) {
				return -1;
			}
		}
		iocb->total_elem_num += v->elem_num;
if (iocb->total_elem_num < 0) {
	return -1;
}
		osc->iocb0.total_elem_num += v->elem_num;
if (osc->iocb0.total_elem_num < 0) {
	return -1;
}
	}
#endif

	return 0;
}

static int
scompar(va, vb)
void const *va, *vb;
{
	struct srv const *a, *b;
	int e;
	a = va;
	b = vb;
	if (e = strcmp(a->serv, b->serv)) {
		return e;
	}
	return strcmp(a->host, b->host);
}

static int
distrib_s(s, schep, size)
struct iosched *s;
char *schep;
size_t size;
{
	struct srv *servers;
	size_t ne;
	struct sched_iocb *iocb;
	size_t i, j, k;
	size_t p;

	ne = s->scc[0].np + s->scc[1].np + 2;

	if (!(servers = calloc(ne, sizeof *servers))) {
		return -1;
	}

	iocb = &s->scc[0].iocb0;
	servers[0].host = iocb->R.host;
	servers[0].serv = iocb->R.serv;
	iocb = &s->scc[1].iocb0;
	servers[1].host = iocb->R.host;
	servers[1].serv = iocb->R.serv;
	i = 2;
	for (p=0; p<s->scc[0].np; p++) {
		iocb = &s->scc[0].iocb[p];
		servers[i].host = iocb->R.host;
		servers[i].serv = iocb->R.serv;
		i++;
	}
	for (p=0; p<s->scc[1].np; p++) {
		iocb = &s->scc[1].iocb[p];
		servers[i].host = iocb->R.host;
		servers[i].serv = iocb->R.serv;
		i++;
	}
	assert(i==ne);

	qsort(servers, i, sizeof *servers, scompar);

	for (i=0, j=0; i<ne; i=k, j++) {
		for (k=i+1; k<ne && scompar(&servers[i], &servers[k])==0; k++) ;
		servers[j] = servers[i];
	}

	if (distrib_ss(s, &s->scc[0].iocb0, DXR_ROW, servers, j)==-1) {
		return -1;
	}
	if (distrib_ss(s, &s->scc[1].iocb0, DXR_COL, servers, j)==-1) {
		return -1;
	}
	if (distrib_sc(s, servers, j, schep, size)==-1) {
		return -1;
	}
	free(servers);
	return 0;
}

static int
distrib_ss(sche, iocb0, type, servers, ne)
struct iosched *sche;
struct sched_iocb *iocb0;
struct srv *servers;
size_t ne;
{
	DWFILE *h;
	int s;
	char const *handle = sche->handle;
	size_t j;
	struct srv server0;

	char *name;

	switch (type) {
	case DXR_ROW:
		name = "xr_row";
		break;
	case DXR_COL:
		name = "xr_col";
		break;
	default:
		fprintf(stderr, __FILN__ ": configuration error\n");
		return -1;
	}

	server0.host = iocb0->R.host;
	server0.serv = iocb0->R.serv;

	if ((s = nnet_cli(server0.host, server0.serv))==-1) {
		perror(server0.host);
		return -1;
	}

	printf("%s.%s.summary < %s >\n", handle, name, server0.host);
	if (!(h = dwc_file_open(s, handle, type, -1))) {
		fprintf(stderr, "dwc_file_open failed\n");
		return -1;
	}

	for (j=0; j<ne; j++) {
		if (scompar(&server0, &servers[j])==0) {
			servers[j].f = NULL;
		}
		else {
			if ((s = nnet_cli(servers[j].host, servers[j].serv))==-1) {
				perror(servers[j].host);
				return -1;
			}
			printf(" %s", servers[j].host);
			if (!(servers[j].f = dwc_file_create(s, handle, type, -1))) {
				return -1;
			}
		}
	}
	printf("\n");

	if (dwc_file_seq_init(h)==-1) {
		fprintf(stderr, "dwc_file_seq_init failed\n");
		return -1;
	}

	for (;;) {
		size_t size;
		char const *b;

		if (dwc_file_seq_recv_blk(h, &b, &size)==-1) {
			fprintf(stderr, "dwc_file_seq_recv_blk failed\n");
			return -1;
		}
		if (size==0) {
			break;
		}

		for (j=0; j<ne; j++) {
			if (servers[j].f && dwc_file_write_blk(servers[j].f, b, size)==-1) {
				fprintf(stderr, "dwc_file_seq_write_blk failed\n");
				return -1;
			}
		}
	}
	if (dwc_file_seq_finish(h)==-1) {
		fprintf(stderr, "dwc_file_seq_finish failed\n");
		return -1;
	}

	if (dwc_close(h)==-1) {
		return -1;
	}
	for (j=0; j<ne; j++) {
		if (servers[j].f) {
			if (dwc_flush(servers[j].f)==-1) {
				return -1;
			}
			if (dwc_close(servers[j].f)==-1) {
				return -1;
			}
		}
	}
	return 0;
}

static int
distrib_sc(sche, servers, ne, schep, size)
struct iosched *sche;
struct srv *servers;
char *schep;
size_t ne, size;
{
	char const *handle = sche->handle;
	size_t j;

	printf("%s.sched >\n", handle);
	for (j=0; j<ne; j++) {
		int s;
		if ((s = nnet_cli(servers[j].host, servers[j].serv))==-1) {
			perror(servers[j].host);
			return -1;
		}

		printf(" %s", servers[j].host);

		if (!(servers[j].f = dwc_file_create(s, handle, DW_SCHE, 0))) {
			return -1;
		}
		if (dwc_file_write_blk(servers[j].f, schep, size)==-1) {
			fprintf(stderr, "dwc_file_seq_write_blk failed\n");
			return -1;
		}
		if (dwc_flush(servers[j].f)==-1) {
			return -1;
		}
		if (dwc_close(servers[j].f)==-1) {
			return -1;
		}
	}
	printf("\n");

	return 0;
}

static size_t
allocv0(size, v0p)
size_t size;
void **v0p;
{
	void *v0, *addr;
	addr = (void *)V0_BASE;	/* XXX */
printf("allocv0: %s, %s\n", sizetostr(size), ptrtoxstr(addr));
	if (addr) {
		if (size == 0) {
			size_t used = mmtotal();
			size = amem - used;
			fprintf(stderr, "allocv0: amem = %sM, used = %sM, v0size = %sM\n", 
				sizetostr(amem/(1024*1024)), 
				sizetostr(used/(1024*1024)),
				sizetostr(size/(1024*1024)));
		}
		if (!(v0 = xmmalc(addr, size))) {
			nperror("malloc v0");
			return 0;
		}
	}
	else {
		v0 = NULL;
		if (size == 0) {
			size = 1024 * 1024 * 512;	/* XXX */
		}
		if (!(v0 = malloc(size))) {
fprintf(stderr, "size = %s\n", sizetostr(size));
			nperror("malloc v0");
			return 0;
		}
	}
	*v0p = v0;
	return size;
}

static void
freev0(v0)
void *v0;
{
	/* do nothing, as program is going to exit. */
}

static int
mkmonolithic(iccws)
struct ICCW *iccws;
{
	struct ht *h;
	FILE *f;
	struct hte *e;
	size_t i;
	size_t fwp, pad;
	char *addr;
	struct sdup *s;
	size_t size, sz[2];
	int d;
	char template[MAXPATHLEN];
	int fd;
	char const *tmpdir;

	if (!(tmpdir = getenv("TMPDIR"))) {
		tmpdir = "/tmp";
	}

	snprintf(template, sizeof template, "%s/mkdw.XXXXXX", tmpdir);
	if ((fd = mkstemp(template)) == -1) {
		perror(template);
		return -1;
	}
	printf("[%s]\n", template);
	unlink(template);
	if (!(f = fdopen(fd, "r+"))) {
		perror(template);
		return -1;
	}

	/* step1: name.p を退避 */
	for (d = 0; d < 2; d++) {
		h = iccws[d].h;
		s = iccws[d].s;
		if (s->monolithic == 1) {
			continue;
		}
		if (fwrite("", 1, 1, f) != 1) {
			perror("mkmonolithic: fwrite");
			return -1;
		}
		fwp = 1;
		for (i = 0; i < h->size; i++) {
			e = hta(h, i);
			if (e->name.p) {
				size_t l = strlen(e->name.p) + 1;
				if (fwrite(e->name.p, 1, l, f) != l) {
					perror("mkmonolithic: fwrite");
					return -1;
				}
				e->name.u = fwp;
				fwp += l;
			}
			else {
				e->name.u = 0;
			}
		}
		if (fwp % PAGESIZE) {
			pad = PAGESIZE - fwp % PAGESIZE;
			for (i = 0; i < pad; i++) {
				if (fwrite("", 1, 1, f) != 1) {
					perror("mkmonolithic: fwrite");
					return -1;
				}
			}
		}
		else {
			pad = 0;
		}
		sz[d] = fwp + pad;
	}

	/* step2: ht 領域を退避 */
	for (d = 0; d < 2; d++) {
		h = iccws[d].h;
		if (h->monolithic == 1) {
			continue;
		}
		for (i = 0; i < h->nbases; i++) {
			if (fwrite(h->bases[i], sizeof *h->bases[i], h->fragsz, f) !=  h->fragsz) {
				perror("mkmonolithic: fwrite");
				return -1;
			}
		}
	}

	rewind(f);

	/* step3: sdup を free */
	for (d = 0; d < 2; d++) {
		s = iccws[d].s;
		if (s->monolithic == 1) {
			continue;
		}
		for (i = 0; i < s->nbases; i++) {
			free(s->bases[i]);
		}
		free(s->bases);
		s->bases = NULL;
		s->bases_size = 0;
	}

	/* step4: nht を free */
	for (d = 0; d < 2; d++) {
		h = iccws[d].h;
		if (h->monolithic == 1) {
			continue;
		}
		for (i = 0; i < h->nbases; i++) {
			free(h->bases[i]);
		}
		free(h->bases);
		h->bases = NULL;
		h->bases_size = 0;
	}

	/* step5: sdup を復元 */
	for (d = 0; d < 2; d++) {
		s = iccws[d].s;
		if (s->monolithic == 1) {
			continue;
		}
		s->monolithic = 2;
		size = sz[d];
		if (!(s->base = malloc(size))) {
			perror("mkmonolithic: malloc");
			return -1;
		}
		s->pend = s->base + size;
		if (fread(s->base, 1, size, f) != size) {
			perror("mkmonolithic: fread");
			return -1;
		}
	}

	/* step6: ht を復元 */
	for (d = 0; d < 2; d++) {
		h = iccws[d].h;
		if (h->monolithic == 1) {
			continue;
		}
		h->monolithic = 2;
		if (!(h->base = malloc(h->size * sizeof *h->base))) {
			perror("mkmonolithic: malloc");
			return -1;
		}
		if (fread(h->base, sizeof *h->base, h->size, f) != h->size) {
			perror("mkmonolithic: fread");
			return -1;
		}
	}

	/* step7: name.p を復元. */
	for (d = 0; d < 2; d++) {
		h = iccws[d].h;
		s = iccws[d].s;
		if (s->monolithic == 1) {
			continue;
		}
		addr = s->base;
		for (i = 0; i < h->size; i++) {
			e = hta(h, i);
			if (e->name.u) {
				e->name.p = addr + e->name.u;
			}
			else {
				e->name.p = NULL;
			}
		}
	}

	fclose(f);

	return 0;
}

static void
nextf_flush()
{
	refresh_hash = 1;
}

static void
refresh_hash_table(iccws, v0, v0size)
struct ICCW *iccws;
void *v0;
size_t v0size;
{
	refresh_hash = 0;
return;
	printf("<");
	if (v0) {
		refresh_mem(v0, v0size);
	}
	sdup_refresh(iccws[0].s);
	sdup_refresh(iccws[1].s);
	nht_refresh(iccws[0].h);
	nht_refresh(iccws[1].h);
	printf(">");
}

#include "mkdxrfm.c"

static void
show_elapsed_time(mode, stream)
FILE *stream;
{
	struct timeval curr, elapsed;

	gettimeofday(&curr, NULL);
	timersub(&curr, &start, &elapsed);
	switch (mode) {
	default:
	case 0:
		fprintf(stream, "elapsed time: %ld\n", elapsed.tv_sec);
		break;
	case 1:
		fprintf(stream, "Elapsed time: %ld\n", elapsed.tv_sec);
		break;
	case 2:
		fprintf(stream, "(%ld)\n", elapsed.tv_sec);
		break;
	}
	fflush(stream);
}

void
maxrlimit(resource, name)
char const *name;
{
	struct rlimit rl;
	if (getrlimit(resource, &rl) == -1) {
		perror(name);
	}
	else if (rl.rlim_cur < rl.rlim_max) {
		rlim_t prev = rl.rlim_cur;
		rl.rlim_cur = rl.rlim_max;
		if (setrlimit(resource, &rl) == -1) {
			perror(name);
		}
		else {
			fprintf(stderr, "%s: %d => %d\n", name, (int)prev, (int)rl.rlim_cur);
		}
	}
}
