/*
 *	Copyright (c) 2005 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: mkdxrfm.c,v 1.6 2007/03/26 23:07:29 nis Exp $";
#endif

struct tpl {
	idx_t w, a;
	freq_t freq;
};

#define	NFRG	256

#define DEBUG_HEAP	0

struct mstream {
	struct {
		size_t i, n;
		struct tpl *v;
	} c;

	struct {
		size_t n;
		FILE *f;
		struct tpl t;
	} s[NFRG];
	size_t ns;

	size_t h[NFRG];
	size_t nh;
};

static int tplcompar(const void *, const void *);
static int sorted(const void *, size_t, size_t, int (*)(const void *, const void *));
static idx_t mgetvec(struct mstream *, struct xr_vec *, size_t);
static idx_t mgetvec_v1(struct mstream *, struct xr_vec *, size_t);
static idx_t mgetvec_f(struct mstream *, struct xr_vec *, size_t);
static int initmini(struct mstream *);
static int mcmp(struct mstream *, size_t, size_t);
static struct tpl findmini(struct mstream *, idx_t);
static int ejectv1(struct tpl *, size_t, struct mstream *);

static int
sorted(base, nmemb, size, compar)
const void *base;
size_t nmemb;
size_t size;
int (*compar)(const void *, const void *);
{
	size_t i;
	const void *a, *b;

	for (i = 0; i < nmemb - 1; i++) {
		a = (char *)base + size * i;
		b = (char *)base + size * (i + 1);
		if (compar(a, b) > 0) {
			return 0;
		}
	}

	return 1;
}

static int
tplcompar(va, vb)
const void *va, *vb;
{
	struct tpl *a, *b;
	a = (struct tpl *)va;
	b = (struct tpl *)vb;
	if (a->w < b->w) {
		return -1;
	}
	else if (a->w > b->w) {
		return 1;
	}
	else if (a->a < b->a) {
		return -1;
	}
	else if (a->a > b->a) {
		return 1;
	}
	return 0;
}

static idx_t
mgetvec(m, v, size)
struct mstream *m;
struct xr_vec *v;
size_t size;
{
	if (m->c.v) {
		return mgetvec_v1(m, v, size);
	}
	else {
		return mgetvec_f(m, v, size);
	}
}

static idx_t
mgetvec_v1(m, v, size)
struct mstream *m;
struct xr_vec *v;
size_t size;
{
	size_t j;
	freq_t sum;
	df_t k;
	idx_t id;

	if (m->c.n <= m->c.i) {
		return 0;
	}
	id = m->c.v[m->c.i].w;
	sum = 0;
	for (k = 0, j = m->c.i; j < m->c.n && m->c.v[m->c.i].w == m->c.v[j].w; j++, k++) {
if (k < 0) {
	return 0;
}
		assert(k < size);	/* overrun check */
		v->elems[k].id = m->c.v[j].a;
		v->elems[k].freq = m->c.v[j].freq;
		sum += m->c.v[j].freq;
if (sum < 0) {
	return 0;
}
		assert(k == 0 || v->elems[k - 1].id != v->elems[k].id);	/* uniqueness check */
	}
if (k < 0) {
	return 0;
}
	v->freq_sum = sum;
	v->elem_num = k;
	m->c.i = j;
	return id;
}

static idx_t
mgetvec_f(m, v, size)
struct mstream *m;
struct xr_vec *v;
size_t size;
{
	freq_t sum;
	df_t k;
	idx_t id;
	struct tpl t;

	if (m->nh == 0) {
		return 0;
	}

	t = findmini(m, 0);
	assert(t.w != 0);

	id = t.w;
	sum = 0;

	k = 0;
	v->elems[k].id = t.a;
	v->elems[k].freq = t.freq;
	sum += t.freq;
if (sum < 0) {
	return 0;
}

	for (k++; ; k++) {
if (k < 0) {
	return 0;
}
		if ((t = findmini(m, id)).w == 0) {
			break;
		}
		assert(t.w == id);

		assert(k < size);	/* overrun check */
		v->elems[k].id = t.a;
		v->elems[k].freq = t.freq;
		sum += t.freq;
if (sum < 0) {
	return 0;
}
		assert(k == 0 || v->elems[k - 1].id != v->elems[k].id);	/* uniqueness check */
	}
if (k < 0) {
	return 0;
}

	v->freq_sum = sum;
	v->elem_num = k;
	return id;
}

int
initmini(m)
struct mstream *m;
{
	size_t i, j, l, r;
	size_t x;
	for (i = 0; i < m->ns; i++) {
		assert(m->s[i].n != 0);
		rewind(m->s[i].f);
		if (fread(&m->s[i].t, sizeof m->s[i].t, 1, m->s[i].f) != 1) {
			return -1;
		}
	}

	m->nh = m->ns;

	for (i = 0; i < m->nh; i++) {
		m->h[i] = i;
	}

/* XXX l = m->nx / 2 de yoi ? */
	l = (m->nh + 1) / 2;
	r = m->nh - 1;
	printf("heap-init\n");

	while (0 < l) {
		l--;
		for (i = l, x = m->h[i]; (j = 2 * i + 1) <= r; i = j) {
			if (j + 1 <= r && mcmp(m, m->h[j], m->h[j + 1]) > 0) {
				j++;
			}
			if (mcmp(m, m->h[j], x) > 0) {
				break;
			}
			m->h[i] = m->h[j];
		}
		m->h[i] = x;
	}
	return 0;
}

static int
mcmp(m, i, j)
struct mstream *m;
size_t i, j;
{
	struct tpl *a, *b;

	a = &m->s[i].t;
	b = &m->s[j].t;

	if (a->w < b->w) {
		return -1;
	}
	else if (a->w > b->w) {
		return 1;
	}
	else if (a->a < b->a) {
		return -1;
	}
	else if (a->a > b->a) {
		return 1;
	}
	return 0;
}

static struct tpl
findmini(m, id)
struct mstream *m;
idx_t id;
{
	size_t i, j, r;
	size_t x;
	struct tpl t;

	if (id != 0 && m->nh == 0) {
		struct tpl null;
		null.w = 0;
		return null;
	}

	assert(m->nh != 0);

#if DEBUG_HEAP
	for (j = 0; j < m->ns; j++) {
		if (m->s[j].f) {
			break;
		}
	}
	assert(j != m->ns);
	t = m->s[j].t;
	for (i = j + 1; i < m->ns; i++) {
		struct tpl *e = &m->s[i].t;
		if (!m->s[i].f) {
			continue;
		}
		if (t.w > e->w || t.w == e->w && t.a > e->a) {
			j = i;
			t = m->s[j].t;
		}
	}
	assert(j == m->h[0]);
#else
	j = m->h[0];
	t = m->s[j].t;
#endif

	if (id != 0 && id != t.w) {
		struct tpl null;
		null.w = 0;
		return null;
	}

	if (--m->s[j].n == 0) {
		fclose(m->s[j].f);
		m->s[j].f = NULL;
		m->h[0] = m->h[m->nh - 1];
		m->nh--;
	}
	else {
		size_t e = fread(&m->s[j].t, sizeof m->s[j].t, 1, m->s[j].f);
		assert(e == 1);
	}

	if (m->nh != 0) {

		r = m->nh - 1;

		for (i = 0, x = m->h[0]; (j = 2 * i + 1) <= r; i = j) {
			if (j + 1 <= r && mcmp(m, m->h[j], m->h[j + 1]) > 0) {
				j++;
			}
			if (mcmp(m, m->h[j], x) > 0) {
				break;
			}
			m->h[i] = m->h[j];
		}
		m->h[i] = x;
	}

	return t;
}

static int
ejectv1(v1, n, m)
struct tpl *v1;
size_t n;
struct mstream *m;
{
	char template[MAXPATHLEN];
	int fd;
	FILE *f;
	char const *tmpdir;

	printf("[");

	if (!(tmpdir = getenv("TMPDIR"))) {
		tmpdir = "/tmp";
	}

	snprintf(template, sizeof template, "%s/mkdwXXXXXX", tmpdir);

	if (nelems(m->s) <= m->ns) {
		fprintf(stderr, "ejectv1: too many fragments\n");
		return -1;
	}
	qsort(v1, n, sizeof *v1, tplcompar);
	if ((fd = mkstemp(template)) == -1) {
		perror("ejectv1: mkstemp");
		return -1;
	}
	unlink(template);
	if (!(f = fdopen(fd, "r+"))) {
		perror("ejectv1: fdopen");
		return -1;
	}
	if (fwrite(v1, sizeof *v1, n, f) != n) {
		perror("ejectv1: fwrite");
		return -1;
	}
	m->s[m->ns].n = n;
	m->s[m->ns].f = f;
	m->ns++;

	printf("%s, %s]", basename(template), sizetostr(n));
	return 0;
}

static int
mkdxrfm(f, s, mtime, cws, iccws)
RFILES *f;
struct iosched *s;
DWFILE *cws[];
struct ICCW *iccws;
{
	idx_t id;
	ssize_t e;
	struct dxr_raw_vec *rv;
	struct xr_vec *v;
	ssize_t vsize, vs1, vs2;
	void *v0;
	size_t v0size, v1size;
	size_t n;
	struct tpl *v1;
	ssize_t k;
	struct timeval curr, elapsed;
	struct mstream ms;

	if (init_output(s, &s->scc[0], s->handle, DXR_ROW, mtime, 0, s->compress)==-1) {
		nlog(LOG_ERR, "mkdxrfm: init_output failed\n");
		return -1;
	}

	if (!(v0size = allocv0(s->v0size, &v0))) {
		fprintf(stderr, "mkdxrfm: allocv0 failed\n");
		return -1;
	}

	v1 = v0;
	v1size = v0size / sizeof *v1;

	vs1 = s->scc[0].raw_max_elem_num;
	for (vs2 = 0, id = 1; id <= s->tg[1].n; id++) {
		struct hte *p = hta(s->tg[1].h, id - 1);
		if (vs2 < p->i.df) {
			vs2 = p->i.df;
		}
	}
	printf("vs1: %s vs2: %s\n", ssizetostr(vs1), ssizetostr(vs2));
	vsize = MAX(vs1, vs2);

	if (!(v = malloc(sizeof_dxr_vec(vsize)))) {
		perror("mkdxrfm: malloc");
		return -1;
	}

	report_progress_raw_vec(f, stdout, 0);

	if (rewind_read_raw_vec(f) == -1) {
		fprintf(stderr, "mkdxrfm: rwind_read_vec failed\n");
		return -1;
	}

	ms.ns = 0;
	for (n = 0; (e = read_raw_vec(f, &rv)) >= 0; ) {
		assert(rv->elem_num <= vsize);

		if ((id = rv2v(rv, v, cws, iccws)) == 0) {
			fprintf(stderr, "mkdxrfm: rv2v failed\n");
			return -1;
		}

		assert(id == hta(s->tg[0].h, id - 1)->i.id);
		assert(v->elem_num == hta(s->tg[0].h, id - 1)->i.df);
		assert(v->freq_sum == hta(s->tg[0].h, id - 1)->i.tf);

		if (output_vec(s, &s->scc[0], &s->scc[1], id, v)==-1) {
			nlog(LOG_ERR, "mkdxrfm: output_vec failed\n");
			return -1;
		}

		for (k = 0; k < v->elem_num; k++) {
			struct tpl t;
			struct xr_elem *xe;
			xe = &v->elems[k];

			if (v1size <= n) {
				if (ejectv1(v1, n, &ms) == -1) {
					fprintf(stderr, "mkdxrfm: ejectv1 failed\n");
					return -1;
				}
				n = 0;
			}

			t.a = id;
			t.w = xe->id;
			t.freq = xe->freq;
			v1[n++] = t;
		}
	}

	gettimeofday(&curr, NULL);
	timersub(&curr, &start, &elapsed);
	printf("(%ld)\n", elapsed.tv_sec);
	fflush(stdout);

	if (e == -1) {
		nlog(LOG_ERR, "mkdxrfm: read_raw_vec failed\n");
		return -1;
	}

	printf("\n");

	if (finish_output(s, &s->scc[0]) == -1) {
		nlog(LOG_ERR, "mkdxrfm: finish_output failed\n");
		return -1;
	}

	if (init_output(s, &s->scc[1], s->handle, DXR_COL, mtime, 0, s->compress)==-1) {
		fprintf(stderr, "mkdxrfm: init_output failed (4)\n");
		return -1;
	}

	if (ms.ns != 0) {
		assert(n != 0);
		if (ejectv1(v1, n, &ms) == -1) {
			fprintf(stderr, "mkdxrfm: ejectv1 failed\n");
			return -1;
		}
		if (initmini(&ms) == -1) {
			fprintf(stderr, "mkdxrfm: initmini failed\n");
			return -1;
		}
		ms.c.i = 0;
		ms.c.n = 0;
		ms.c.v = NULL;
	}
	else {
		qsort(v1, n, sizeof *v1, tplcompar);
		ms.c.i = 0;
		ms.c.n = n;
		ms.c.v = v1;
	}

	for (n = 0; (id = mgetvec(&ms, v, vsize)) != 0; n++) {
		assert(sorted(v->elems, v->elem_num, sizeof *v->elems, ecompar));

		assert(id == hta(s->tg[1].h, id - 1)->i.id);
		assert(v->elem_num == hta(s->tg[1].h, id - 1)->i.df);
		assert(v->freq_sum == hta(s->tg[1].h, id - 1)->i.tf);

		if (output_vec(s, &s->scc[1], &s->scc[0], id, v) == -1) {
			fprintf(stderr, "mkdxrfm: output_vec failed\n");
			return -1;
		}
	}

	printf("%s\n", sizetostr(n));
	if (finish_output(s, &s->scc[1]) == -1) {
		fprintf(stderr, "mkdxrfm: finish_output failed\n");
		return -1;
	}
	freev0(v0);
	free(v);

	printf(" done.\n");
	return 0;
}
