/*
 *	Copyright (c) 2004 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: mmalloc.c,v 1.2 2007/03/21 03:17:41 nis Exp $";
#endif

#include <config.h>

#include <sys/types.h>

#include <sys/mman.h>
#include <stdio.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>

#include "mmalloc.h"

static size_t total = 0;

/* basyo wo sitei site allocate. hikakuteki tiisana memory block you. */
void *
mmalc(addr, size)
void *addr;
size_t size;
{
	void *p;
	if ((p = mmap(addr, size, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON, -1, 0))==MAP_FAILED) {
		fprintf(stderr, "mmap failed: size=%s (%sM)\n", sizetoxstr(size), sizetostr(size/(1024*1024)));
		perror("mmap");
		return NULL;
	}
	total += size;
	madvise(addr, size, MADV_RANDOM);
	fprintf(stderr, "<%s:%s:%sM>", sizetoxstr(size), ptrtoxstr(p), sizetostr(total/(1024*1024)));
	return p;
}

/* basyo wo sitei site allocate. kyodai na memory block you */
void *
xmmalc(addr, size)
void *addr;
size_t size;
{
	void *r = NULL;
	size_t fragsz = 256*1024*1024;
	size_t i, sz;

	for (sz = 0, i = 0; sz < size; sz += fragsz, i++) {
		void *p;
		if ((p = mmap(addr, fragsz, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON, -1, 0))==MAP_FAILED) {
			fprintf(stderr, "mmap failed: size=%s (%sM)\n", sizetoxstr(fragsz), sizetostr(fragsz/(1024*1024)));
			perror("mmap");
			return NULL;
		}
		if (p != addr) {
			fprintf(stderr, "mmap failed: size=%s (!continuous)\n", sizetoxstr(fragsz));
			return NULL;
		}
		if (i == 0) {
			r = p;
		}
		madvise(addr, fragsz, MADV_RANDOM);
		total += fragsz;
		addr += fragsz;
	}
	fprintf(stderr, "<%s:%s:%sM>", sizetoxstr(size), ptrtoxstr(r), sizetostr(total/(1024*1024)));
	return r;
}

/*
 * mmalc de allocate sita block wo free suru.
 * xmmalc de allocate sita block ha free dekinai
 */
void
mmfree(ptr, len)
void *ptr;
size_t len;
{
	if (munmap(ptr, len) == -1) {
		fprintf(stderr, "mmfree: error: ptr=%p, len=%s\n", ptr, sizetostr(len));
	}
	else {
		total -= len;
		fprintf(stderr, "[f:%sM]", sizetoxstr(total/(1024*1024)));
	}
}

size_t
mmtotal()
{
	return total;
}
