/*
 *	Copyright (C) 1996 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: nlog.c,v 1.3 2007/03/21 03:17:41 nis Exp $"; */

#include <config.h>

#include <sys/types.h>

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>

static char const *id = NULL;

static char msgbuf[4096];

#if defined CONST_STDERR
static FILE *stream = stderr;
#define	xxx()
#else
static FILE *stream = NULL;
#define	xxx() do { if (!stream) stream = stderr; } while(0)
#endif

int
nassignlogf(path)
char const *path;
{
	FILE *f;
	xxx();
	if (stream != stderr) {
		fclose(stream);
		stream = stderr;
	}
	if (f = fopen(path, "a+")) {
		stream = f;
		setvbuf(f, NULL, _IONBF, 0);
		return 0;
	}
	return -1;
}

void
nopenlog(ident, logopt, facility)
const char *ident;
{
	id = ident;
}

void
nlog(int priority, char const *message, ...)
{
	va_list ap;
	va_start(ap, message);
	vnlog(priority, message, ap);
	va_end(ap);
}

void
vnlog(int priority, char const *message, va_list ap)
{
	time_t t;
	char *date;
	pid_t pid = getpid();
	int l, ll, i;

	if (!id) {
		vfprintf(stderr, message, ap);
		return;
	}

	time(&t);
	date = ctime(&t);
	date[24] = '\0';
	if ((l = snprintf(msgbuf, sizeof msgbuf, "%s %s[%d]: ", date, id, pid)) < 0) {
		return;
	}

	if ((ll = vsnprintf(msgbuf+l, sizeof msgbuf-l, message, ap))<0) {
		return;
	}

	l += ll;

	for (i=0; msgbuf[i]; i++) {
		int c = msgbuf[i]&0xff;
		if (!(' '<=c && c<='~')) {
			msgbuf[i] = '#';
		}
	}

	xxx();
	if (l>=sizeof msgbuf) {
		fprintf(stream, "%s...\n", msgbuf);
	}
	else {
		fprintf(stream, "%s\n", msgbuf);
	}
}

void
ncloselog()
{
	id = NULL;
	xxx();
	if (stream != stderr) {
		fclose(stream);
		stream = stderr;
	}
}

int
nsetlogmask(maskpri)
{
	return 0;
}

FILE *
nlogstream()
{
	xxx();
	return stream;
}

void
nlogflush()
{
	xxx();
	fflush(stream);
}

void
nperror(s)
char const *s;
{
	char const *msg = strerror(errno);
	if (!s || *s=='\0') {
		fprintf(stderr, "%s", msg);
	}
	else {
		fprintf(stderr, "%s: %s", s, msg);
	}
}

#if 0
main()
{
#if 0
	nopenlog("nlog", 0, 0);
#endif
	nsetlogmask(0);

	nlog(0, "%s", "hello\n");
	nlog(0, "%s", "hello  abc def ghi jkl mno pqr s\n");
	nlog(0, "%s", "hello  abc def ghi jkl mno pqr st\n");
	nlog(0, "%s", "hello  abc def ghi jkl mno pqr stu\n");

	ncloselog();
}
#endif
