/*
 *	Copyright (c) 2004 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: pint.c,v 1.13 2007/03/26 07:49:41 nis Exp $";
#endif

#include <config.h>

#include <sys/types.h>

#include <assert.h>
#include <stdarg.h>
#include <stddef.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/dwerrno.h>

size_t
pack_ints(void *p0, size_t size, void *end, char const *fmt, ...)
{
	char *p = p0, *e = p0 + size, *q;
	char const *f;
	va_list ap;
	va_start(ap, fmt);
	for (f = fmt; *f; p = q, f++) {
#define PUT(type, comtype) \
			do { \
				type *v = va_arg(ap, type *); \
				assert(sizeof (comtype) >= sizeof (type)); \
				q = p + sizeof (comtype); \
				if (e < q) { \
					dwerrno = DWE_OVRN; \
					goto END; \
				} \
				*(type *)p = *(type *)v; \
			} while (0)
		switch (*f) {
		case 'u': PUT(u_int, pintcom_s_t); break;
		case 's': PUT(size_t, pintcom_t); break;
		case 'd': PUT(int, pintcom_s_t); break;
		case 't': PUT(ssize_t, pintcom_t); break;
		case 'f': PUT(freq_t, freq_t); break;
		case 'j': PUT(ssize_t, pintcom_t); break;
		case 'i': PUT(idx_t, idx_t); break;
		default:
			dwerrno = DWE_OVRN;
			goto END;
		}
	}
	va_end(ap);
END:
	if (end) {
		*(void **)end = p;
	}
	return f - fmt;
}

size_t
unpack_ints(void const *p0, size_t size, void *end, char const *fmt, ...)
{
	char const *p = p0, *e = p0 + size, *q;
	char const *f;
	va_list ap;
	va_start(ap, fmt);
	for (f = fmt; *f; p = q, f++) {
#define GET(type, comtype) \
			do { \
				type *v = va_arg(ap, type *); \
				assert(sizeof (comtype) >= sizeof (type)); \
				q = p + sizeof (comtype); \
				if (e < q) { \
					dwerrno = DWE_OVRN; \
					goto END; \
				} \
				*(type *)v = *(type *)p; \
			} while (0)
		switch (*f) {
		case 'u': GET(u_int, pintcom_s_t); break;
		case 's': GET(size_t, pintcom_t); break; /* XXX check overflow */
		case 'd': GET(int, pintcom_s_t); break;
		case 't': GET(ssize_t, pintcom_t); break; /* XXX check overflow */
		case 'f': GET(freq_t, freq_t); break;
		case 'j': GET(ssize_t, pintcom_t); break;
		case 'i': GET(idx_t, idx_t); break;
		default:
			dwerrno = DWE_OVRN;
			goto END;
		}
	}
	va_end(ap);
END:
	if (end) {
		*(void const **)end = p;
	}
	return f - fmt;
}

#include <string.h>
#include "pint_typecheck.c"
