#include <config.h>

#include <sys/types.h>

#include <libgen.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

char *progname;

void usage(void);
char const *fmtc2typename(int);
void genpc(char const *);
void genph(char const *);
void genuc(char const *);
void genuh(char const *);

void
usage()
{
	fprintf(stderr, "usage: %s fmt uh|ph|uc|pc", progname);
	exit(1);
}

main(argc, argv)
char *argv[];
{
	progname = basename(argv[0]);
	char const *fmt = argv[1];
	char const *mode = argv[2];
	if (argc != 3) {
		usage();
	}
	if (!strcmp(mode, "ph")) {
		genph(fmt);
	}
	else if (!strcmp(mode, "pc")) {
		genpc(fmt);
	}
	else if (!strcmp(mode, "uh")) {
		genuh(fmt);
	}
	else if (!strcmp(mode, "uc")) {
		genuc(fmt);
	}
	else {
		usage();
	}
	return 0;
}

char const *
fmtc2typename(c)
{
	switch (c) {
	case 'u': return "u_int";
	case 's': return "size_t";
	case 'd': return "int";
	case 't': return "ssize_t";
	case 'f': return "freq_t";
	case 'i': return "idx_t";
	}
	return "???";
}

void
genph(fmt)
char const *fmt;
{
	printf("size_t pack_ints_%s(void *, size_t, void *, char const *", fmt);
	for (; *fmt; fmt++) {
		printf(", %s *", fmtc2typename(*fmt));
	}
	printf(");\n");
}

void
genpc(fmt)
char const *fmt;
{
	size_t l = strlen(fmt);
	int i;

	printf("\n"
		"size_t\n"
		"pack_ints_%s(p0, size, end, fmt", fmt);

	for (i = 0; i < l; i++) {
		printf(", v%d", i);
	}

	printf(")\n"
		"void *p0;\n"
		"size_t size;\n"
		"void *end;\n"
		"char const *fmt;\n");

	for (i = 0; i < l; i++) {
		printf("%s *v%d;\n", fmtc2typename(fmt[i]), i);
	}

	printf("{\n"
		"	if (strcmp(fmt, \"%s\")) {\n"
		"		return 0;\n"
		"	}\n"
		"	return pack_ints(p0, size, end, fmt",
		fmt);

	for (i = 0; i < l; i++) {
		printf(", v%d", i);
	}

	printf(");\n"
		"}\n");
}

void
genuh(fmt)
char const *fmt;
{
	printf("size_t unpack_ints_%s(void const *, size_t, void *, char const *", fmt);
	for (; *fmt; fmt++) {
		printf(", %s *", fmtc2typename(*fmt));
	}
	printf(");\n");
}

void
genuc(fmt)
char const *fmt;
{
	size_t l = strlen(fmt);
	int i;

	printf("\n"
		"size_t\n"
		"unpack_ints_%s(p0, size, end, fmt", fmt);

	for (i = 0; i < l; i++) {
		printf(", v%d", i);
	}

	printf(")\n"
		"void const *p0;\n"
		"size_t size;\n"
		"void *end;\n"
		"char const *fmt;\n");

	for (i = 0; i < l; i++) {
		printf("%s *v%d;\n", fmtc2typename(fmt[i]), i);
	}

	printf("{\n"
		"	if (strcmp(fmt, \"%s\")) {\n"
		"		return 0;\n"
		"	}\n"
		"	return unpack_ints(p0, size, end, fmt",
		fmt);

	for (i = 0; i < l; i++) {
		printf(", v%d", i);
	}

	printf(");\n"
		"}\n");
}
