/*	$Id: priq.c,v 1.6 2007/03/29 07:37:38 nis Exp $	*/

/*-
 * Copyright (c) 2006 Shingo Nishioka.
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Id: priq.c,v 1.6 2007/03/29 07:37:38 nis Exp $";
#endif

#include <config.h>

#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "priq.h"

struct priq *
priq_creat(size, compar)
size_t size;
int (*compar)(const void *, const void *);
{
	return priq_creat_p(size, compar, priq_dfl_prop_f, priq_dfl_prop_b);
}

struct priq *
priq_creat_p(size, compar, prop_f, prop_b)
size_t size;
int (*compar)(const void *, const void *);
void (*prop_f)(struct priq *);
void (*prop_b)(struct priq *);
{
	struct priq *q;
	if (!(q = calloc(1, sizeof *q))) {
		return NULL;
	}
	if (!(q->tmp = calloc(1, size))) {
		free(q);
		return NULL;
	}
	q->esize = size;
	q->size = 0;
	q->n = 0;
	q->q = NULL;
	q->compar = compar;
	q->prop_f = prop_f;
	q->prop_b = prop_b;
	return q;
}

void
priq_free(q)
struct priq *q;
{
	free(q->tmp);
	free(q->q);
	free(q);
}

int
priq_emptyp(q)
struct priq *q;
{
	return q->n == 0;
}

ssize_t
priq_nmemb(q)
struct priq *q;
{
	return q->n;
}

#define	A(prq, idx)	((char *)(prq)->q + (prq)->esize * (idx))
#define	C1(prq, dst, src)	(memmove((dst), (src), (prq)->esize))

ssize_t
priq_enq(q, d)
struct priq *q;
void *d;
{
	if (q->n >= q->size) {
		void *new;
		size_t newsize;
		newsize = q->n + 128;
		if (!(new = realloc(q->q, newsize * q->esize))) {
			return -1;
		}
		q->size = newsize;
		q->q = new;
	}
	C1(q, A(q, q->n), d);
	q->n++;
	(*q->prop_b)(q);
	return q->n;
}

void *
priq_top(q)
struct priq *q;
{
	return q->q;
}

ssize_t
priq_deq(q)
struct priq *q;
{
	if (q->n == 0) {
		return -1;
	}
	--q->n;
	if (q->n == 0) {
		free(q->q);
		q->size = 0;
		q->q = NULL;
		return 0;
	}
	C1(q, A(q, 0), A(q, q->n));
	(*q->prop_f)(q);
	return q->n;
}

void *
priq_rplatop(q, d)
struct priq *q;
void *d;
{
	if (q->n == 0) {
		return NULL;
	}
	C1(q, A(q, 0), d);
	(*q->prop_f)(q);
	return q->q;
}

#define compar_dfl(e, a, b, q)	e = ((*q->compar)(a, b))

PROP_F(priq_dfl_prop_f, compar_dfl)
PROP_B(priq_dfl_prop_b, compar_dfl)
