/*	$Id: priq.h,v 1.5 2006/09/25 01:58:51 nis Exp $	*/

/*-
 * Copyright (c) 2006 Shingo Nishioka.
 * All rights reserved.
 */

struct priq {
	size_t esize, size, n;
	void *q, *tmp;
	int (*compar)(const void *, const void *);
	void (*prop_f)(struct priq *);
	void (*prop_b)(struct priq *);
};

struct priq *priq_creat(size_t, int (*)(const void *, const void *));
struct priq *priq_creat_p(size_t, int (*)(const void *, const void *), void (*)(struct priq *), void (*)(struct priq *));
void priq_free(struct priq *);
int priq_emptyp(struct priq *);
ssize_t priq_nmemb(struct priq *);
ssize_t priq_enq(struct priq *, void *);
void *priq_top(struct priq *);
ssize_t priq_deq(struct priq *);
void *priq_rplatop(struct priq *, void *);

void priq_dfl_prop_f(struct priq *);
void priq_dfl_prop_b(struct priq *);

#define	PRIQ_EMPTYP(priq)	((priq)->n == 0)
#define	PRIQ_NMEMB(priq)	((priq)->n)
#define	PRIQ_TOP(priq)		((priq)->q)

#define PROP_F(name, cmp) \
void \
name(q) \
struct priq *q; \
{ \
	size_t i, j, n, esize; \
	n = q->n; \
	esize = q->esize; \
	for (i = 0; (j = i * 2 + 1) < n; i = j) { \
		void *a, *b, *t; \
		int e; \
		a = (char *)q->q + esize * i; \
		b = (char *)q->q + esize * j; \
		if (j + 1 < n) { \
			void *c; \
			c = (char *)q->q + esize * (j + 1); \
			cmp(e, b, c, q); \
			if (e > 0) { \
				j++; \
				b = c; \
			} \
		} \
		cmp(e, a, b, q); \
		if (e <= 0) { \
			break; \
		} \
		t = q->tmp; \
		memmove(t, a, esize); \
		memmove(a, b, esize); \
		memmove(b, t, esize); \
	} \
}

#define PROP_B(name, cmp) \
void \
name(q) \
struct priq *q; \
{ \
	size_t i, j, esize; \
	int (*compar)(const void *, const void *); \
	if (q->n == 0) { \
		return; \
	} \
	esize = q->esize; \
	compar = q->compar; \
	for (j = q->n - 1; j > 0; j = i) { \
		void *a, *b, *t; \
		int e; \
		i = (j - 1) / 2; \
		a = (char *)q->q + esize * i; \
		b = (char *)q->q + esize * j; \
		cmp(e, a, b, q); \
		if (e <= 0) { \
			break; \
		} \
		t = q->tmp; \
		memmove(t, a, esize); \
		memmove(a, b, esize); \
		memmove(b, t, esize); \
	} \
}

#define PROP_F_TYPED(name, cmp, type) \
void \
name(q) \
struct priq *q; \
{ \
	size_t i, j, n; \
	type a; \
	type b; \
	type t; \
	int e; \
	n = q->n; \
	for (i = 0; (j = i * 2 + 1) < n; i = j) { \
		a = (type)q->q + i; \
		b = (type)q->q + j; \
		if (j + 1 < n) { \
			type c; \
			c = (type)q->q + (j + 1); \
			cmp(e, b, c, q); \
			if (e > 0) { \
				j++; \
				b = c; \
			} \
		} \
		cmp(e, a, b, q); \
		if (e <= 0) { \
			break; \
		} \
		t = q->tmp; \
		*t = *a; \
		*a = *b; \
		*b = *t; \
	} \
}

#define PROP_B_TYPED(name, cmp, type) \
void \
name(q) \
struct priq *q; \
{ \
	size_t i, j; \
	type a; \
	type b; \
	type t; \
	int e; \
	if (q->n == 0) { \
		return; \
	} \
	for (j = q->n - 1; j > 0; j = i) { \
		i = (j - 1) / 2; \
		a = (type)q->q + i; \
		b = (type)q->q + j; \
		cmp(e, a, b, q); \
		if (e <= 0) { \
			break; \
		} \
		t = q->tmp; \
		*t = *a; \
		*a = *b; \
		*b = *t; \
	} \
}
