/*
 *      Copyright (c) 2003 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: ptest.c,v 1.7 2007/03/26 00:17:57 nis Exp $";
#endif

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/uio.h>

#include <assert.h>
#include <errno.h>
#if defined HAVE_LIBGEN_H
#include <libgen.h>
#endif
#include <math.h>
#include <netdb.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/timer.h>
#include <geta/dci.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/daui.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwsched.h>
#include <geta/dwam.h>

static int ptest(char const *, char const *, FILE *, size_t, size_t, size_t);
static struct syminfo *setq(size_t, freq_t, ssize_t);

char *progname = NULL;

int
main(argc, argv)
char *argv[];
{
	char const *dirname, *handle;
	size_t nq, nd, ntimes;

	progname = basename(argv[0]);

	if (argc != 6) {
		fprintf(stderr, "usage: %s dir handle nq nd #times\n", progname);
		return 1;
	}

	srandom(0);

	dirname = argv[1];
	handle = argv[2];
	nq = strtosize(argv[3], NULL, 10);
	nd = strtosize(argv[4], NULL, 10);
	ntimes = strtosize(argv[5], NULL, 10);

	if (ptest(dirname, handle, stdout, nq, nd, ntimes)==-1) {
		return 1;
	}

	return 0;
}

int
ptest(dirname, handle, stream, reqnq, reqnd, ntimes)
char const *dirname, *handle;
FILE *stream;
size_t reqnq, reqnd, ntimes;
{
	size_t i;
	int dir, type;
	ssize_t nq, nd, total;
	WAM *w;
	struct syminfo *q, *d;
	ssize_t maxid;
	freq_t maxtf;

	if (wam_init(dirname)==-1) {
		return -1;
	}
	for (i=0; i<ntimes; i++) {

		if (!(w = wam_open(handle))) {
			nperror("wam_open");
			return -1;
		}
		if (reqnd==0) {
			reqnd = wam_size(w, WAM_ROW);
		}
		maxid = wam_size(w, WAM_COL);

		maxtf = 2;
		q = setq(reqnq, maxtf, maxid);
		nq = reqnq;
		dir = WAM_COL;
		type = WT_SMARTWA;
		nd = reqnd;
		d = wsh(q, nq, w, dir, type, &nd, &total);
		free(q);
		free(d);

		wam_close(w);
	}
	return 0;
}

struct query {
	idx_t id;
	size_t tf;
};

static struct syminfo *
setq(reqnq, maxtf, maxid)
size_t reqnq;
freq_t maxtf;
ssize_t maxid;
{
	struct syminfo *q;
	size_t i;
	q = calloc(reqnq, sizeof *q);
	for (i=0; i<reqnq; i++) {
		struct syminfo *r = &q[i];
		r->id = random()%(maxid/8)+1;
		r->attr = WSH_NONE;
		r->TF = r->TF_d = r->DF = r->DF_d = random()%(maxtf-1)+1; /* XXX freq_t -> size_t wo keisan siteha ikenai */
		r->weight = 0;
	}
	return q;
}
