/*
 *	Copyright (C) 1999 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: rawvec.c,v 1.5 2007/03/24 02:52:58 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
/* #include <sys/stat.h> */

#include <assert.h>
#include <ctype.h>
#if defined HAVE_LIBGEN_H
#include <libgen.h>
#endif
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dxr.h>

#include "dwsys.h"
#include "dxrP.h"

#define	BSIZE		8192
#define	BSIZEINC	8192

#include "dfs.h"

struct dw_rfiles {
	char **files;
	char **filep;
	SFILE *f;
	size_t size;
	size_t nthf;	/* == filep - files */
	struct {
		FILE *stream;
		int mode;
		size_t col;
	} report_progress;
#if defined X_READ_RAW_VEC
	struct {
		freq_t e;
		size_t cnt;
		struct {
			char state[256];
			int m;
			double sel_p;
		} a, w;
		struct dxr_raw_vec *dvec[128];
		char template[32];
	} x;
#endif
};

static int open_current_f(RFILES *);
static void close_current_f(RFILES *);
static int find_next_f(RFILES *);
static ssize_t read_raw_vec_q(RFILES *, struct dxr_raw_vec **);
static int cnrr(char **, size_t *, char **);
static int reset_statics(void);

#if defined X_READ_RAW_VEC
static void xfreevec(struct dxr_raw_vec *);
static struct dxr_raw_vec *xdupvec(RFILES *, struct dxr_raw_vec *, size_t);
static char *xdupstr(char const *, char const *, ssize_t);
#endif

void (*rawvec_nextf_hook)(void) = NULL;

RFILES *
init_read_raw_vec(files)
char **files;
{
	RFILES *f;
	char **filep;
#if defined X_READ_RAW_VEC
	size_t i;
#endif

	if (!(f = calloc(1, sizeof *f))) {
		nperror("calloc f");
		return NULL;
	}

	f->f = NULL;

	for (filep = files; *filep; filep++) ;
	if (!(f->files = calloc(filep - files + 1, sizeof *f->files))) {
		nperror("calloc f->files");
		free(f);
		return NULL;
	}

	for (filep = f->files; *files; files++, filep++) {
		if (!(*filep = strdup(*files))) {
			finish_read_raw_vec(f);
			return NULL;
		}
	}
	*filep = NULL;
	f->report_progress.stream = NULL;
	f->report_progress.mode = 0;

#if defined X_READ_RAW_VEC
	f->x.a.m = 1;
	f->x.a.sel_p = 0;
	f->x.w.m = 1;
	f->x.w.sel_p = 0;
	for (i=0; i<nelems(f->x.dvec); i++) {
		f->x.dvec[i] = NULL;
	}
	strlcpy(f->x.template, "-%d", sizeof f->x.template);

	f->x.e = -1;
	f->x.cnt = 0;
#endif

	return f;
}

#if defined X_READ_RAW_VEC
/* kakuritu p de, kizi wo erabi, sono kizi wo m bai suru
 * a.m: hukusei ga okotta toki ni, a.m bai ni naru.
 * a.p: kono kakuritu de hukusei ga okoru.
 * w.m: saisyo no m ko no tango wo hukusei suru.
 * w.p: misiyou.
 */
void
rawvec_set_xread_param(f, am, ap, wm, wp)
RFILES *f;
double ap, wp;
{
	f->x.a.m = am;
	f->x.a.sel_p = ap;
	f->x.w.m = wm;
	f->x.w.sel_p = wp;
}
#endif

static int
open_current_f(f)
RFILES *f;
{
	char *path;
	path = *f->filep;

	if (f->report_progress.stream) {
		if (f->report_progress.mode == 0) {
			char const *p;
			size_t l;
			p = basename(path);
			l = strlen(p);
			if ((f->report_progress.col += l+2)>68) {
				fprintf(f->report_progress.stream, "\n");
				f->report_progress.col = l+2;
			}
			fprintf(f->report_progress.stream, "[%s,", p);
		}
		else {
			fprintf(f->report_progress.stream, ".");
		}
		fflush(f->report_progress.stream);
	}
	if (!(f->f = sfopen(path, "r"))) {
		nperror(path);
		return -1;
	}
	if ((f->size = sfsize(f->f)) == -1) {
		close_current_f(f);
		return 0;
	}
	return 1;
}

static void
close_current_f(f)
RFILES *f;
{
	if (f->f) {
		sfclose(f->f);
		f->f = NULL;
	}
}

static int
find_next_f(f)
RFILES *f;
{
	for (; *f->filep; f->filep++, f->nthf++) {
		switch (open_current_f(f)) {
		default:
			fprintf(stderr, "find_next_f: internal error\n");
			/* fallthru */
		case -1:
			return -1;
		case 0:
			break;
		case 1:
			return 0;
		}
	}
	return 0;
}

void
report_progress_raw_vec(f, stream, mode)
RFILES *f;
FILE *stream;
{
	f->report_progress.stream = stream;
	f->report_progress.mode = mode;
}

int
rewind_read_raw_vec(f)
RFILES *f;
{
	close_current_f(f);

	if (reset_statics()==-1) {
		return -1;
	}

	f->filep = f->files;
	f->nthf = 0;

	f->report_progress.col = 0;

#if defined X_READ_RAW_VEC
	f->x.e = -1;
	f->x.cnt = 0;
        initstate(0, f->x.a.state, sizeof f->x.a.state);
        initstate(0, f->x.w.state, sizeof f->x.w.state);
#endif

	return find_next_f(f);
}

int
finish_read_raw_vec(f)
RFILES *f;
{
	char **p;
	close_current_f(f);
	for (p = f->files; *p; p++) {
		free(*p);
	}
	memset(f->files, 0, (p - f->files + 1) * sizeof *f->files);
	free(f->files);
	memset(f, 0, sizeof *f);
	free(f);
	return 0;
}

/*
 * vec wo caller ga free suru hituyou ha nai
 * caller ga, elems wo hakai sitemo kamawanai!
 * subete no baai ni oite, sentou no space
 * (COUNT no baai niha suuzi ni tuduku space ) to, '\n' ga tori nozokareru
 * gyou ha subete '\n' nomi de kirarete iru to minasu
 * '\r' nado ha hozon sareru
 * subete no vector wo yomi owaruto, -2 wo kaesu
 */
#if defined X_READ_RAW_VEC
ssize_t
read_raw_vec(f, vec)
RFILES *f;
struct dxr_raw_vec **vec;
{
	size_t i;
	if (f->x.cnt-->0) {
		*vec = f->x.dvec[f->x.cnt];
		return f->x.e;
	}
restart:
	if (!*f->filep) {
		return -2;	/* no more vector */
	}
	if ((f->x.e = read_raw_vec_q(f, vec))<0) {
		return f->x.e;
	}
	setstate(f->x.a.state);
	if (f->x.a.sel_p > (double)random()/0x7fffffff) {
		if (f->x.a.m==0) {	/* zero bai suru baai. */
			goto restart;
		}
		/* 1 bai izyou */
		f->x.cnt = f->x.a.m-1;
		for (i=0; i<f->x.cnt; i++) {
			xfreevec(f->x.dvec[i]);
			if (!(f->x.dvec[i] = xdupvec(f, *vec, i))) {
				return -1;
			}
		}
	}
	else {
		f->x.cnt = 0;
	}
	return f->x.e;
}

static void
xfreevec(v)
struct dxr_raw_vec *v;
{
	size_t i;
	if (!v) {
		return;
	}
	free((char *)v->key);
	for (i=0; i<v->elem_num; i++) {
		free((char *)v->elems[i].name);
	}
	free(v);
}

static struct dxr_raw_vec *
xdupvec(f, v, i)
RFILES *f;
struct dxr_raw_vec *v;
size_t i;
{
	struct dxr_raw_vec *n;
	size_t j;
	size_t size;

	setstate(f->x.w.state);

	size = sizeof *v + sizeof v->elems[0] * v->elem_num;
	if (!(n = malloc(size))) {
		return NULL;
	}
	bcopy(v, n, size);
	n->key = xdupstr(f->x.template, n->key, i);
	for (j=0; j<n->elem_num; j++) {
		n->elems[j].name = xdupstr(f->x.template, n->elems[j].name, j<f->x.w.m?i:-1);
	}
	return n;
}

static char *
xdupstr(template, s, i)
char const *template, *s;
ssize_t i;
{
	char *n;
	char x[128];
	size_t size;

	if (i==-1) {
		return strdup(s);
	}
	snprintf(x, sizeof x, template, i);

	size = strlen(s) + strlen(x) + 1;
	if (!(n = malloc(size))) {
		return NULL;
	}
	strcpy(n, s);
	strcat(n, x);
	return n;
}
#else
ssize_t
read_raw_vec(f, vec)
RFILES *f;
struct dxr_raw_vec **vec;
{
	if (!*f->filep) {
		return -2;	/* no more vector */
	}
	return read_raw_vec_q(f, vec);
}
#endif

static struct {
	struct dxr_raw_vec *v;
	ssize_t bsize;
	char *fullwd;
	size_t fsize;
} qs = { NULL, 0, NULL, 0};

static ssize_t
read_raw_vec_q(f, vec)
RFILES *f;
struct dxr_raw_vec **vec;
{
	char *path = *f->filep;	/* for error message */
	struct dxr_raw_elem *ep;
	size_t re;
	int eofp;

	char *fp, *r;

	int ftype;
	size_t pos;
	ssize_t ne;
	size_t l, off;
	size_t key_len;

	char *p;
	char *p0;	/* for error message */

	pos = sfpos(f->f);

	fp = qs.fullwd;
	re = qs.fsize /* - (fp - qs.fullwd) */;
	if (re < BSIZEINC) {
		if (cnrr(&qs.fullwd, &qs.fsize, &fp)==-1) {
			fprintf(stderr, "rrvq: memory allocation failed\n");
			return -1;
		}
		assert(fp == qs.fullwd);
		re = qs.fsize /* - (fp - qs.fullwd) */;
	}

	assert(re>1);
	/* key ha 8k yori mizikai kotoga youkyuu sarete irunode, kokode mugen tyou fgets wo yobu hituyou ha nai */
	if (!sfgets(fp, re, f->f)) {
		fprintf(stderr, "rrvq: fgets failed: %s:%s\n", path, sizetostr(pos));
		return -1;
	}
	l = key_len = strlen(fp);
	if (l==0 || fp[l-1] != '\n') {
		fprintf(stderr, "rrvq: error: no newline character: %s:%s\n", path, sizetostr(pos));
	}

	switch (ftype = *fp) {
	case RAWXR_TYPE_COUNT:
	case RAWXR_TYPE_RAW:
	case RAWXR_TYPE_ASCII:
		break;
	default:
		fprintf(stderr, "rrvq: error: no selector: %s:%s\n", path, sizetostr(pos));
		return -1;
	}

	fp += l;
	off = l;
	for (ne=0; !(eofp = f->size <= pos + off); ne++) {
		int c;
		re = qs.fsize - (fp - qs.fullwd);
		if (re < BSIZEINC) {
			if (cnrr(&qs.fullwd, &qs.fsize, &fp)==-1) {
				fprintf(stderr, "rrvq: memory allocation failed\n");
				return -1;
			}
			re = qs.fsize - (fp - qs.fullwd);
		}
		assert(re>1);

		if ((c = sfpeekc(f->f)) == EOF) {
			goto EXIT;
		}

		switch (ftype) {
		case RAWXR_TYPE_COUNT:
			if (!isdigit(c)) {
				goto EXIT;
			}
			break;
		case RAWXR_TYPE_RAW:
		case RAWXR_TYPE_ASCII:
			if (c != ' ') {
				goto EXIT;
			}
			break;
		default:	/* never happen */
			break;
		}

		if (!sfgets(fp, re, f->f)) {
			fprintf(stderr, "rrvq: fgets failed: %s:%s\n", path, sizetostr(pos+off));
			return -1;
		}
		l = strlen(fp);

		/* mugen-tyou fgets */
		while ((re = qs.fsize - (fp+l - qs.fullwd)) == 1 && fp[l-1]!='\n') {
			size_t l2;
			if (re<BSIZEINC) {
				if (cnrr(&qs.fullwd, &qs.fsize, &fp)==-1) {
					fprintf(stderr, "rrvq: memory allocation failed\n");
					return -1;
				}
				re = qs.fsize - (fp+l - qs.fullwd);
			}
			assert(re>1);
			if (!sfgets(fp+l, re, f->f)) {
				fprintf(stderr, "rrvq: fgets failed: %s:%s\n", path, sizetostr(pos+off));
				return -1;
			}
			if ((l2 = strlen(fp+l))==0) {
				break;
			}
			l += l2;
		}
		/* end: mugen tyou fgets */

		if (l==0 || fp[l-1]!='\n') {
			size_t i;
			fprintf(stderr, "rrvq: error: no newline character: %s:%s\n", path, sizetostr(pos+off));
			for (i=0; fp[i]; i++) {
				fprintf(stderr, "%03o ", fp[i]);
			}
			fprintf(stderr, "\n");
			return -1;
		}
		fp += l;
		off += l;
	}
EXIT:

	if (eofp) {
		assert(pos+off == f->size);
	}

	*fp = '\0';

	if (!qs.v || ne > qs.bsize) {
		void *new;
		ssize_t newsize;
		newsize = NALIGN(ne, BSIZEINC);
/*
		fprintf(stderr, "*%d/%d ", newsize, ne);
*/
		if (!(new = realloc(qs.v, sizeof_dxr_raw_vec(newsize)))) {
			zfree(qs.v);
			qs.bsize = 0;
			nperror("rrvq: realloc");
			return -1;
		}
		qs.bsize = newsize;
		qs.v = new;
	}

	p0 = p = qs.fullwd;
	qs.v->key = p+1;
	qs.v->ftype = ftype;
	qs.v->pos = pos;
	qs.v->key_len = key_len;
	qs.v->elem_len = off - key_len;
	qs.v->frag = f->nthf;

	if (!(p = index(p, '\n'))) {
		fprintf(stderr, "rrvq: Internal error\n");
		return -1;
	}
	*p++ = '\0';
	if (!*qs.v->key) {
		fprintf(stderr, "rrvq: error: empty key: %s:%s\n", path, sizetostr(pos));
		return -1;
	}

	qs.v->elem_num = ne;

	for (ep = qs.v->elems; *p; ep++, p=r) {
		freq_t freq;
		char *s = NULL;
		if (!(r = index(p, '\n'))) {
			fprintf(stderr, "rrvq: Internal error\n");
			return -1;
		}
		*r++ = '\0';
		freq = 0;
		switch (ftype) {
			char *s0, *s1, *endp;
		case RAWXR_TYPE_COUNT:
			if (!isdigit(*p&0xff)) {
				fprintf(stderr, "rrvq: error: a digit expected: %s:%s\n", path, sizetostr(pos+p-p0));
				return -1;
			}
			if (!(s1 = index(s0 = p, ' '))) {
				fprintf(stderr, "rrvq: error: a space expected: %s:%s\n", path, sizetostr(pos+p-p0));
			}
			*s1++ = '\0';
			freq = strtofreq(s0, &endp, 10);
			if (!s0 || !*s0 || *endp) {
				fprintf(stderr, "rrvq: error: not a number: %s:%s\n", path, sizetostr(pos+p-p0));
				return -1;
			}
			s = s1;
			break;
		case RAWXR_TYPE_RAW:
			freq = 1;
			/* fallthru */
		case RAWXR_TYPE_ASCII:
			if (*p!=' ') {
				fprintf(stderr, "rrvq: error: space expected: %s:%s\n", path, sizetostr(pos+p-p0));
				return -1;
			}
			s = p+1;
			break;
		default:
			/* never happen */
			break;
		}
		if (!*s) {
			fprintf(stderr, "rrvq: error: a non-whitespace character expected: %s:%s\n", path, sizetostr(pos+p-p0));
			return -1;
		}
		ep->tf = freq;
		ep->name = s;
	}

	if (eofp) {	/* reached at the end of this file */
		if (f->report_progress.stream) {
			char s[32];
			snprintf(s, sizeof s, "%s", sizetostr(f->size));
			if (f->report_progress.mode == 0) {
				f->report_progress.col += strlen(s)+2;
				fprintf(f->report_progress.stream, "%s] ", s);
				fflush(f->report_progress.stream);
			}
		}
		close_current_f(f);
		f->filep++;
		f->nthf++;
		if (rawvec_nextf_hook) {
			(*rawvec_nextf_hook)();
		}
		if (find_next_f(f)==-1) {
			return -1;
		}
	}

	*vec = qs.v;
	return qs.v->elem_num;
}

static int
cnrr(f, s, p)
char **f, **p;
size_t *s;
{
	char *new_f;
	size_t new_s, off;

	new_s = *s + BSIZEINC*2;
	off = *p - *f;

/*
	fprintf(stderr, "cnrr: info: f=%p, *s=%d, p=%p\n", *f, *s, *p);
*/
	if (!(new_f = realloc(*f, new_s))) {
		nperror("cnrr: realloc");
		zfree(*f);
		*p = NULL;
		*s = 0;
		return -1;
	}
	*f = new_f;
	*s = new_s;
	*p = new_f + off;

	return 0;
}

static int
reset_statics()
{

	if (qs.v) {
		free(qs.v);
		/* assert(qs.bsize); */
	}
	else {
		assert(qs.bsize==0);
	}

	if (qs.fullwd) {
		free(qs.fullwd);
		assert(qs.fsize);
	}
	else {
		assert(qs.fsize==0);
	}

	if (qs.bsize) {
		void *p;
		p = qs.v;
		if (!(qs.v = malloc(sizeof_dxr_raw_vec(qs.bsize)))) {
			nperror("reset_statics: malloc qs.v");
			return -1;
		}
/*
		fprintf(stderr, "reset_statics: info: qs.v: %p->%p\n", p, qs.v);
*/
	}

	if (qs.fsize) {
		void *p;
		p = qs.fullwd;
		if (!(qs.fullwd = malloc(qs.fsize))) {
			nperror("reset_statics: malloc qs.fullwd");
			return -1;
		}
/*
		fprintf(stderr, "reset_statics: info: qs.fullwd: %p->%p\n", p, qs.fullwd);
*/
	}

	return 0;
}
