/*
 *	Copyright (C) 1997 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: readfiles.c,v 1.2 2007/03/21 03:17:41 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/stat.h>

#include <assert.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dxr.h>

char **
read_files(path)
char const *path;
{
	FILE *f = NULL;
	char **a;
	size_t i, na;

	if (*path=='@') {
		char buf[MAXPATHLEN];

		if (!(f = fopen(path+1, "r"))) {
			nperror(path+1);
			goto ERROR;
		}

		for (na=0; fgets(buf, sizeof buf, f); na++) ;

		if (!(a = calloc(na+1, sizeof *a))) {
			nperror("calloc");
			goto ERROR;
		}

		rewind(f);

		for (i=0; fgets(buf, sizeof buf, f); i++) {
			char *p;
			assert(i<na);
			if (p = rindex(buf, '\n')) {
				*p = '\0';
			}
			if (!(a[i] = strdup(buf))) {
				nperror("strdup");
				goto ERROR;
			}
		}

		fclose(f);
	}
	else {
		na = 1;
		if (!(a = calloc(na+1, sizeof *a))) {
			nperror("calloc");
			goto ERROR;
		}
		i = 0;
		if (!(a[i] = strdup(path))) {
			nperror("strdup");
			goto ERROR;
		}

		i++;
	}

	assert(i==na);
	a[i] = NULL;

#if 0
	for (i=0; i<na; i++) {
		struct stat sb;
		if (stat(a[i], &sb)==-1) {
			nperror(a[i]);
			goto ERROR;
		}
		if (sb.st_size==0) {
			fprintf(stderr, "warning: read_files: [%s] is empty\n", a[i]);
		}
	}
#endif

	return a;

ERROR:
	if (f) {
		fclose(f);
	}
	return NULL;
}
