/*
 *	Copyright (c) 2005 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: rht.c,v 1.5 2007/03/21 22:34:27 nis Exp $";
#endif

#include <config.h>

#include <sys/param.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>

#include "sdup.h"
#include "nht.h"
#include "nsxhash.h"
#include "rht.h"
#include "mmalloc.h"

static struct HT *h_init(size_t, void *);
static void h_free(struct HT *);
static struct hte *h_enter(struct HT *, struct hte *);
static int h_pcompar(void const *, void const *);
static struct hte **h_lookup(struct HT *, char const *);

static struct HT *
h_init(size, addr)
size_t size;
void *addr;
{
	struct HT *h;
	size_t i;
	if (!(h = calloc(1, sizeof *h))) {
		return NULL;
	}
	h->size = size;
	if (addr) {
		h->rr = mmalc(addr, h->size * sizeof *h->rr);
	}
	else {
		h->rr = malloc(h->size * sizeof *h->rr);
	}
	if (!h->rr) {
		free(h);
		return NULL;
	}
	for (i = 0; i < size; i++) {
		h->rr[i] = NULL;
	}
	h->nentries = 0;
	return h;
}

static void
h_free(h)
struct HT *h;
{
	mmfree(h->rr, h->size * sizeof *h->rr);
	free(h);
}

static struct hte *
h_enter(h, r)
struct HT *h;
struct hte *r;
{
	hash_t s;
	size_t c;
	size_t size = h->size;
	char *p = r->name.p;

	c = 0;
	for (s = nsxhash(size, p, c++); ; s = nsxhash(size, p, c++)) {
		struct hte **e = &h->rr[s];
		if (!*e) {	/* new */
			h->nentries++;
			*e = r;
			return *e;
		}
		else if (!strcmp((*e)->name.p, p)) {
			/* exists */
			fprintf(stderr, "!duplicated entry: [%s]\n", p);
			return NULL;
		}
		/* conflict */
	}
	fprintf(stderr, "sorry, hash table is full\n");
	return NULL;
}

#define	HASH	0
#define	BIN	1

struct HT *
make_hash_index(r, n, addr)
struct ht *r;
size_t n;
void *addr;
{
	size_t i;
	struct HT *h;
	if (!(h = h_init(MAX(THRESHOLD_1(n), 1024), addr))) {
		fprintf(stderr, "make_hash_index fained (0)\n");
		return NULL;
	}
	h->type = HASH;
	for (i=0; i<n; i++) {
		if (!h_enter(h, hta(r, i))) {
			fprintf(stderr, "make_hash_index fained (1)\n");
			return NULL;
		}
	}

	return h;
}

struct HT *
make_bin_index(t, n, addr)
struct ht *t;
size_t n;
void *addr;
{
	struct HT *h;
	size_t i;

	if (!(h = h_init(n, addr))) {
		fprintf(stderr, "make_bin_index fained (0)\n");
		return NULL;
	}
	h->type = BIN;
	for (i=0; i < n; i++) {
		h->rr[i] = hta(t, i);
	}
	h->nentries = n;
#if HAVE_HEAPSORT
	if (heapsort(h->rr, n, sizeof *h->rr, h_pcompar)==-1) {
		return NULL;
	}
#else
	qsort(h->rr, n, sizeof *h->rr, h_pcompar);
#endif
	for (i=0; i<n-1; i++) {
		if (!strcmp(h->rr[i]->name.p, h->rr[i+1]->name.p)) {
			fprintf(stderr, "duplicated entry: [%s]\n", h->rr[i]->name.p);
			return NULL;
		}
	}
	return h;
}

static int
h_pcompar(va, vb)
void const *va, *vb;
{
	struct hte * const *a;
	struct hte * const *b;
	a = va;
	b = vb;
	return strcmp((*a)->name.p, (*b)->name.p);
}

idx_t
h_name2id(h, name)
struct HT *h;
char const *name;
{
	struct hte **p;
	switch (h->type) {
		struct hte key0, *key;
	case BIN:
		key = &key0;
	/* bsearch (and h_pcompar) ga key wo hakai surukoto ha nai. warning ha musi suru koto */
		key->name.p = name;
		p = bsearch(&key, h->rr, h->nentries, sizeof *h->rr, h_pcompar);
		break;
	case HASH:
		p = h_lookup(h, name);
		break;
	default:
		return 0;
	}
	if (p) {
		return (*p)->i.id;
	}
	return 0;
}

static struct hte **
h_lookup(h, name)
struct HT *h;
char const *name;
{
	hash_t s;
	size_t c;
	size_t size = h->size;
	char const *p = name;

	c = 0;
	for (s = nsxhash(size, p, c++); ; s = nsxhash(size, p, c++)) {
		struct hte **e = &h->rr[s];
		if (!*e) {	/* new */
			return NULL;
		}
		else if (!strcmp((*e)->name.p, p)) {
			/* found */
			return e;
		}
		/* conflict */
	}
	fprintf(stderr, "sorry, hash table is full\n");
	return NULL;
}
