/*
 *	Copyright (c) 2005 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: sdup.c,v 1.2 2007/03/21 03:17:41 nis Exp $";
#endif

#include <config.h>

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <assert.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dwerrno.h>

#include "dwsys.h"
#include "sdup.h"
#include "mmalloc.h"

#define	BASES_SIZE	1024			/* total 8G bytes of fullword */
#define	FULLWDFRGSZ	(8 * 1024 * 1024)	/* 8M Bytes for each fragment */

static int sdup_extend(struct sdup *);

struct sdup *
sdup_init(addr)
void *addr;
{
	struct sdup *s;
	if (!(s = calloc(1, sizeof *s))) {
		return NULL;
	}
	s->base = s->end = s->pend = addr;
	s->nbases = 0;
	s->monolithic = addr != NULL;
	switch (s->monolithic) {
	case 0:
		s->bases_size = BASES_SIZE;
		if (!(s->bases = calloc(s->bases_size, sizeof *s->bases))) {
			free(s);
			return NULL;
		}
		break;
	case 1:
		s->bases_size = 0;
		s->bases = NULL;
		break;
	default:
		fprintf(stderr, "sdup: internal error\n");
		return NULL;
	}
	s->fragsz = FULLWDFRGSZ;
	if (!sdup(s, "")) {
		return NULL;
	}
	return s;
}

void
sdup_free(s)
struct sdup *s;
{
	char *p;
	size_t i;

	switch (s->monolithic) {
	case 0:
		for (i = 0; i < s->nbases; i++) {
			free(s->bases[i]);
		}
		free(s->bases);
		break;
	case 1:
		for (p = s->base; p < s->pend; p += s->fragsz) {
			mmfree(p, s->fragsz);
		}
		break;
	case 2:
		free(s->base);
		break;
	default:
		fprintf(stderr, "sdup: internal error\n");
		break;
	}
	free(s);
}

static int
sdup_extend(s)
struct sdup *s;
{
	void *p;
	switch (s->monolithic) {
	case 0:
		if (s->nbases >= s->bases_size) {
			fprintf(stderr, "sdup_extend: error: too many frags\n");
			return -1;
		}
		if (!(p = malloc(s->fragsz))) {
			perror("sdup_extend: mmap");
			return -1;
		}
		s->bases[s->nbases++] = p;
		s->end = p;
		s->pend = p + s->fragsz;
		break;
	case 1:
		if (!(p = mmalc(s->pend, s->fragsz))) {
			fprintf(stderr, "sdup_extend: error: !mmalc\n");
			return -1;
		}
		if (p != s->pend) {
			fprintf(stderr, "sdup_extend: error: !continuous(%s %s)\n", ptrtoxstr(p), ptrtoxstr(s->pend));
			return -1;
		}
		s->pend += s->fragsz;
		break;
	case 2:
		fprintf(stderr, "sdup_extend failed: !extensible\n");
		return -1;
	default:
		fprintf(stderr, "sdup: internal error\n");
		return -1;
	}
	return 0;
}

char *
sdup(s, p)
struct sdup *s;
char const *p;
{
	size_t l;
	char *r;

	l = strlen(p) + 1;

	if (s->fragsz < l) {
		fprintf(stderr, "too long word\n");
		return NULL;	/* too long word */
	}
	if (s->pend - s->end < l) {
		if (sdup_extend(s)==-1) {
			return NULL;
		}
	}
assert(l <= s->pend - s->end);
	r = s->end;
	s->end += l;
	memmove(r, p, l);
	return r;
}

void
sdup_refresh(s)
struct sdup *s;
{
	char *p;
	size_t i;
	switch (s->monolithic) {
	case 0:
		for (i = 0; i < s->nbases; i++) {
			refresh_mem(s->bases[i], s->fragsz);
		}
		break;
	case 1:
		for (p = s->base; p < s->pend; p += s->fragsz) {
			refresh_mem(p, s->fragsz);
		}
		break;
	case 2:
		refresh_mem(s->base, s->pend - s->base);
		break;
	default:
		fprintf(stderr, "sdup: internal error\n");
		break;
	}
}
