/*
 *	Copyright (C) 2001 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: swam.c,v 1.8 2007/03/26 00:17:57 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <geta/dtypes.h>
#include <geta/dwut.h>
#include <geta/dw.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/daui.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwsched.h>
#include <geta/dci.h>
#include <geta/dwam.h>

#include "sexp.h"
#include "swam.h"

#if ! defined MAX
#define	MAX(a,b)	((a)>(b)?(a):(b))
#endif

extern struct SEXP *nil;

static void put_nil(FILE *);
static void put_int(int, FILE *);
static void put_icdf(ssize_t, FILE *);
static void put_freq(freq_t, FILE *);
static void put_string(char const *, FILE *);
static void put_syminfo(struct syminfo *, ssize_t, WAM *, int, FILE *);
static void put_vec(WAM *, struct xr_vec const *, int, FILE *);
static void put_name2id(WAM *, int, char const *, FILE *);
static void err(FILE *, char const *, ...);
static int stodir(sexp, int *);
static int stowtype(sexp, int *);
static int stod(sexp, double *);
static int stoi(sexp, int *);
static int stosize(sexp, size_t *);
static int stodf(sexp, ssize_t *);
static int stofreq(sexp, freq_t *);
static int stoidx(sexp, idx_t *);
static int stow(sexp, WAM **);
static int fillq(sexp, struct syminfo *);
static int stosym(sexp, struct syminfo **, ssize_t *);

static int xs_wam_open(char const *);
static int xs_wam_close(int);
static WAM *xs_wam_desc(int);
static size_t iwl(struct syminfo *, size_t, size_t, WAM *, int);

static WAM *dt[128];

static char const *weight_types[] = WEIGHT_TYPE_NAMES;

/*
#include "jis.c"
*/

static int
xs_wam_open(handle)
char const *handle;
{
	int d;
	static int initialized = 0;
	if (!initialized) {
		for (d=0; d<nelems(dt); d++) {
			dt[d] = NULL;
		}
		initialized = 1;
	}
	for (d=0; d<nelems(dt); d++) {
		if (!dt[d]) {
			if (dt[d] = wam_open(handle)) {
				return d;
			}
			else {
				return -1;
			}
		}
	}
	return -1;
}

static int
xs_wam_close(d)
{
	if (d<0 || nelems(dt)<=d || !dt[d]) {
		return -1;
	}
	wam_close(dt[d]);
	dt[d] = NULL;
	return 0;
}

static WAM *
xs_wam_desc(d)
{
	if (d<0 || nelems(dt)<=d) {
		return NULL;
	}
	return dt[d];
}

static size_t
iwl(terms, nkwords, takano_iw_limit, w, dir)
struct syminfo *terms;
size_t nkwords, takano_iw_limit;
WAM *w;
{
	int i, j;

	if (nkwords<=0) {
		return nkwords;
	}
	/* note: terms[j..nkwords] ni gomi ga nokoru. */
	for (i=0, j=0; i<nkwords; i++) {
		if (wam_elem_num(w, dir, terms[i].id) <= takano_iw_limit) {
			terms[j++] = terms[i];
		}
	}
	return MAX(j, 1);
}

static void
put_nil(stream)
FILE *stream;
{
	fprintf(stream, "#f");
}

static void
put_idx_t(u, stream)
idx_t u;
FILE *stream;
{
	fprintf(stream, "%s", idxtostr(u));
}

static void
put_int(i, stream)
int i;
FILE *stream;
{
	fprintf(stream, "%d", i);
}

static void
put_icdf(i, stream)
ssize_t i;
FILE *stream;
{
	fprintf(stream, "%s", ssizetostr(i));
}

static void
put_freq(i, stream)
freq_t i;
FILE *stream;
{
	fprintf(stream, "%s", freqtostr(i));
}

static void
put_string(s, stream)
char const *s;
FILE *stream;
{
	if (s) {
		sexp_put_string(s, stream);
	}
	else {
		put_nil(stream);
	}
}

#define	MAXPREC "%24.16e"

static void
put_syminfo(s, ns, wam, dir, stream)
struct syminfo *s;
ssize_t ns;
WAM *wam;
FILE *stream;
{
	size_t i;

	fprintf(stream, "(");
	for (i=0; i<ns; i++) {
		struct syminfo *p = s+i;
		fprintf(stream, "#(%s ", idxtostr(p->id));
		put_string(wam_id2name(wam, dir, p->id), stream);
		fprintf(stream, " %s %s %s %s " MAXPREC " %d)", freqtostr(p->TF), freqtostr(p->TF_d), freqtostr(p->DF), freqtostr(p->DF_d), p->weight, p->attr);
	}

	fprintf(stream, ")");
}

static void
put_vec(wam, vec, rdir, stream)
WAM *wam;
struct xr_vec const *vec;
FILE *stream;
{
	ssize_t i;
	fprintf(stream, "(");
	fprintf(stream, "(%s %s)", dftostr(vec->elem_num), freqtostr(vec->freq_sum));
	for (i=0; i<vec->elem_num; i++) {
		struct xr_elem const *e = vec->elems+i;
		fprintf(stream, "#(%s %s ", idxtostr(e->id), freqtostr(e->freq));
		put_string(wam_id2name(wam, rdir, e->id), stream);
		fprintf(stream, ")");
	}
	fprintf(stream, ")");
}

static void
put_name2id(w, mode, name, stream)
WAM *w;
char const *name;
FILE *stream;
{
	idx_t e;
#if defined JIS_WAM
	char const *p;

	for (p=name; *p; p++) {
		if (*p&0x80) {
			break;
		}
	}

	if (*p) {
		int l;
		static char *q = NULL;
		static int qlen = 0;

		l = etojs(name, NULL, 0)+1;
		if (qlen<l) {
			if (!(q = realloc(q, l))) {
				qlen = 0;
				err(stream, "realloc");
				return;
			}
			qlen = l;
		}
		etojs(name, q, l);
		name = q;
	}
#endif
	if ((e = wam_name2id(w, mode, name)) != 0) {
		put_idx_t(e, stream);
	}
	else {
		put_nil(stream);
	}
}

static void
err(FILE *stream, char const *msg, ...)
{
	va_list ap;
	va_start(ap, msg);
	fprintf(stderr, "error: ");
	vfprintf(stderr, msg, ap);
	fprintf(stderr, "\n");
	va_end(ap);

	put_nil(stream);
}

static int
stodir(s, d)
sexp s;
int *d;
{
	char const *p;

	if (!(p = symbol_name(s))) {
		return -1;
	}
	if (!strcasecmp("wam_row", p)) {
		*d = WAM_ROW;
		return 0;
	}
	else if (!strcasecmp("wam_col", p)) {
		*d = WAM_COL;
		return 0;
	}
	return -1;
}

static int
stowtype(s, d)
sexp s;
int *d;
{
	char const *p;
	int i;

	if (!(p = symbol_name(s))) {
		return -1;
	}
	for (i=0; i<nelems(weight_types); i++) {
		if (!strcasecmp(weight_types[i], p)) {
			*d = i;
			return 0;
		}
	}
	return -1;
}

static int
stod(s, d)
sexp s;
double *d;
{
	char const *p;
	char *e;

	if (!(p = symbol_name(s))) {
		return -1;
	}
	*d = strtod(p, &e);
	if (!(*p && !*e)) {
		return -1;
	}
	return 0;
}

static int
stoi(s, d)
sexp s;
int *d;
{
	char const *p;
	char *e;

	if (!(p = symbol_name(s))) {
		return -1;
	}
	*d = strtol(p, &e, 10);
	if (!(*p && !*e)) {
		return -1;
	}
	return 0;
}

static int
stosize(s, d)
sexp s;
size_t *d;
{
	char const *p;
	char *e;

	if (!(p = symbol_name(s))) {
		return -1;
	}
	*d = strtol(p, &e, 10);
	if (!(*p && !*e)) {
		return -1;
	}
	return 0;
}

static int
stodf(s, d)
sexp s;
ssize_t *d;
{
	char const *p;
	char *e;

	if (!(p = symbol_name(s))) {
		return -1;
	}
	*d = strtodf(p, &e, 10);
	if (!(*p && !*e)) {
		return -1;
	}
	return 0;
}

static int
stofreq(s, d)
sexp s;
freq_t *d;
{
	char const *p;
	char *e;

	if (!(p = symbol_name(s))) {
		return -1;
	}
	*d = strtofreq(p, &e, 10);
	if (!(*p && !*e)) {
		return -1;
	}
	return 0;
}

static int
stoidx(s, d)
sexp s;
idx_t *d;
{
	char const *p;
	char *e;

	if (!(p = symbol_name(s))) {
		return -1;
	}
	*d = strtoidx(p, &e, 10);
	if (*p && !*e) {
		return 0;
	}
	return -1;
}

static int
stow(s, w)
sexp s;
WAM **w;
{
	int d;

	if (stoi(s, &d)==-1) {
		return -1;
	}
	if (!(*w = xs_wam_desc(d))) {
		return -1;
	}
	return 0;
}

static int
fillq(s, q)
sexp s;
struct syminfo *q;
{
	sexp p, c;
	if (length(s)!=8) {
		return -1;
	}
	p = s;

	c = car(p); p = cdr(p);
	stoidx(c, &q->id);

	c = car(p); p = cdr(p);
	/* ignore */

	c = car(p); p = cdr(p);
	stofreq(c, &q->TF);

	c = car(p); p = cdr(p);
	stofreq(c, &q->TF_d);

	c = car(p); p = cdr(p);
	stofreq(c, &q->DF);

	c = car(p); p = cdr(p);
	stofreq(c, &q->DF_d);

	c = car(p); p = cdr(p);
	stod(c, &q->weight);

	c = car(p); p = cdr(p);
	stoi(c, &q->attr);

	return 0;
}

static int
stosym(s, qp, nqp)
sexp s;
struct syminfo **qp;
ssize_t *nqp;
{
	ssize_t nq;
	struct syminfo *q;
	sexp p;
	ssize_t i;

	if ((nq = length(s))<=0) {
		return -1;
	}
	if (!(q = calloc(nq, sizeof *q))) {
		return -1;
	}

	for (p=s, i=0; p!=nil; p=cdr(p), i++) {
		assert(i<nq);
		if (fillq(car(p), &q[i])==-1) {
			free(q);
			return -1;
		}
	}

	*qp = q;
	*nqp = nq;
	return 0;
}

/* kokokara hontai */

void
swam_init(argc, argv, stream)
sexp argv;
FILE *stream;
{
	char const *projectroot = NULL;
	int e;

	if (argc!=1) {
		err(stream, "argc=%d, root=%p", argc, projectroot);
		return;
	}
	if (!(projectroot = symbol_name(nth(argv, 0)))) {
		err(stream, "argc=%d, root=%p", argc, projectroot);
		return;
	}

	if (!strcmp(projectroot, "")) {
		projectroot = NULL;
	}
	if ((e = wam_init(projectroot))==-1) {
		err(stream, "");
		return;
	}
	put_int(e, stream);
}

void
swam_open(argc, argv, stream)
sexp argv;
FILE *stream;
{
	char const *handle;
	int e;

	if (argc!=1) {
		err(stream, "");
		return;
	}
	if (!(handle = symbol_name(nth(argv, 0)))) {
		err(stream, "");
		return;
	}
	if ((e = xs_wam_open(handle))==-1) {
		err(stream, "");
		return;
	}
	put_int(e, stream);
}

void
swam_close(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int d;
	if (argc!=1) {
		err(stream, "");
		return;
	}
	if (stoi(nth(argv, 0), &d)==-1) {
		err(stream, "");
		return;
	}
	if (xs_wam_close(d)==-1) {
		err(stream, "");
		return;
	}
	put_int(0, stream);
}

void
swam_size(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	WAM *w;
	ssize_t e;
	if (argc!=2) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if ((e = wam_size(w, mode))==-1) {
		err(stream, "");
		return;
	}
	put_icdf(e, stream);
}

void
swam_elem_num(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	idx_t id;
	WAM *w;
	ssize_t e;
	if (argc!=3) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if (stoidx(nth(argv, 2), &id)==-1) {
		err(stream, "");
		return;
	}
	if ((e=wam_elem_num(w, mode, id))==-1) {
		err(stream, "");
		return;
	}
	put_icdf(e, stream);
}

void
swam_freq_sum(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	idx_t id;
	WAM *w;
	freq_t e;
	if (argc!=3) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if (stoidx(nth(argv, 2), &id)==-1) {
		err(stream, "");
		return;
	}
	if ((e=wam_freq_sum(w, mode, id))==-1) {
		err(stream, "");
		return;
	}
	put_freq(e, stream);
}

void
swam_max_elem_num(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	WAM *w;
	ssize_t e;
	if (argc!=2) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if ((e=wam_max_elem_num(w, mode))==-1) {
		err(stream, "");
		return;
	}
	put_icdf(e, stream);
}

void
swam_max_freq_sum(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	WAM *w;
	freq_t e;
	if (argc!=2) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if ((e=wam_max_freq_sum(w, mode))==-1) {
		err(stream, "");
		return;
	}
	put_freq(e, stream);
}

void
swam_total_elem_num(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	WAM *w;
	freq_t e;
	if (argc!=2) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if ((e=wam_total_elem_num(w, mode))==-1) {
		err(stream, "");
		return;
	}
	put_freq(e, stream);
}

void
swam_total_freq_sum(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	WAM *w;
	freq_t e;
	if (argc!=2) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if ((e=wam_total_freq_sum(w, mode))==-1) {
		err(stream, "");
		return;
	}
	put_freq(e, stream);
}

void
swam_id2name(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	idx_t id;
	WAM *w;
	if (argc!=3) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if (stoidx(nth(argv, 2), &id)==-1) {
		err(stream, "");
		return;
	}
	put_string(wam_id2name(w, mode, id), stream);
}

void
swam_name2id(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	char const *name;
	WAM *w;
	if (argc!=3) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if (!(name = symbol_name(nth(argv, 2)))) {
		err(stream, "");
		return;
	}
	put_name2id(w, mode, name, stream);
}

void
swam_get_vec(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	idx_t id;
	WAM *w;
	ssize_t e;
	struct xr_vec const *vec;

	if (argc!=3) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if (stoidx(nth(argv, 2), &id)==-1) {
		err(stream, "");
		return;
	}

	if ((e = wam_get_vec(w, mode, id, &vec))!=-1) {
		put_vec(w, vec, WAM_REVERT_DIRECTION(mode), stream);
	}
	else {
		err(stream, "wam_get_vec failed");
		return;
	}
}

void
swam_wsh(argc, argv, stream)
sexp argv;
FILE *stream;
{
	struct syminfo *q = NULL;
	int mode, weight_type;
	size_t takano_iw_limit;
	ssize_t nq, nd, maxnd;
	struct syminfo *dp = NULL;
	ssize_t total;
	WAM *w, *target_wam;

	if (argc!=7) {
		err(stream, "!argc");
		return;
	}
	if (stow(nth(argv, 1), &w)==-1) {
		err(stream, "!w");
		return;
	}
	if (stodir(nth(argv, 2), &mode)==-1) {
		err(stream, "!mode");
		return;
	}
	if (stowtype(nth(argv, 3), &weight_type)==-1) {
		err(stream, "!weight_type");
		return;
	}
	if (stodf(nth(argv, 4), &maxnd)==-1) {
		err(stream, "!maxnd");
		return;
	}
	if (stosize(nth(argv, 5), &takano_iw_limit)==-1) {
		err(stream, "!takano_iw_limit");
		return;
	}
	if (stow(nth(argv, 6), &target_wam)==-1) {
		err(stream, "!target_wam");
		return;
	}

	if (stosym(nth(argv, 0), &q, &nq)==-1) {
		err(stream, "!q");
		return;
	}
	/* koko ikou, return siteha naranai. dp, q wo free surutame. */

	nd = maxnd;
	if (dp = wsh(q, nq, w, mode, weight_type, &nd, &total)) {
		qsort(dp, nd, sizeof *dp, sym_compar);
		if (takano_iw_limit) {
			nd = iwl(dp, nd, takano_iw_limit, target_wam, WAM_REVERT_DIRECTION(mode));
		}
		fprintf(stream, "(");
		put_syminfo(dp, nd, w, WAM_REVERT_DIRECTION(mode), stream);
		fprintf(stream, " ");
		put_icdf(total, stream);
		fprintf(stream, ")");
	}
	else {
		err(stream, "wsh");
	}
	free(dp);
	free(q);
}

void
swam_getprop(argc, argv, stream)
sexp argv;
FILE *stream;
{
	WAM *f;
	char const *handle, *path, *p;
	idx_t id;
	int mode;

	if (argc!=4) {
		err(stream, "");
		return;
	}
	if (!(handle = symbol_name(nth(argv, 0)))) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if (!(path = symbol_name(nth(argv, 2)))) {
		err(stream, "");
		return;
	}
	if (stoidx(nth(argv, 3), &id)==-1) {
		err(stream, "");
		return;
	}

	if (f = wam_prop_open(handle, mode, path)) {
		if (wam_prop_gets(f, id, &p)>0) {
			put_string(p, stream);
		}
		else {
			err(stream, "");
		}
		wam_close(f);
	}
	else {
		err(stream, "");
	}
}

void
swam_wstem(argc, argv, stream)
sexp argv;
FILE *stream;
{
	ssize_t nd;
	struct syminfo *dp = NULL;
	char const *text, *path;
	int mode;
	WAM *w;
	if (argc!=4) {
		err(stream, "");
		return;
	}
	if (!(text = symbol_name(nth(argv, 0)))) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 1), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 2), &mode)==-1) {
		err(stream, "");
		return;
	}
	if (!(path = symbol_name(nth(argv, 3)))) {
		err(stream, "");
		return;
	}

	if ((dp = wstem(text, w, mode, path, &nd))) {
		put_syminfo(dp, nd, w, mode, stream);
	}
	else {
		err(stream, "");
	}
	free(dp);
}

#warning broken
void
swam_wstemv(argc, argv, stream)
sexp argv;
FILE *stream;
{
	char **av = NULL;
	ssize_t nd;
	struct syminfo *dp = NULL;
	char const *text, *path;
	int mode;
	WAM *w;

	if (argc!=4) {
		err(stream, "");
		return;
	}
	if (!(text = symbol_name(nth(argv, 0)))) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 1), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 2), &mode)==-1) {
		err(stream, "");
		return;
	}
	if (!(path = symbol_name(nth(argv, 3)))) {
		err(stream, "");
		return;
	}

/*
	av = calloc(?);
	av[0] = path;
*/

	if ((dp = wstemv(text, w, mode, path, av, &nd))) {
		put_syminfo(dp, nd, w, mode, stream);
	}
	else {
		err(stream, "");
	}
	free(dp);
	free(argv);
}


void
swam_get_weight_types(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int i;

	if (argc!=0) {
		err(stream, "");
		return;
	}

	fprintf(stream, "(");
	for (i=0; i < nelems(weight_types); i++) {
		fprintf(stream, "%s", weight_types[i]);
		if (i + 1 < nelems(weight_types)) {
			fprintf(stream, " ");
		}
	}
	fprintf(stream, ")");
}

#if USE_CICONF
void
swam_ci_value(argc, argv, stream)
sexp argv;
FILE *stream;
{
	char const *handle, *name;
	if (argc!=2) {
		err(stream, "");
		return;
	}
	if (!(handle = symbol_name(nth(argv, 0)))) {
		err(stream, "");
		return;
	}
	if (!(name = symbol_name(nth(argv, 1)))) {
		err(stream, "");
		return;
	}

	put_string(ci_value(handle, name), stream);
}
#endif

void
swam_idl2namel(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	WAM *w;
	sexp s, p;
	idx_t id;
	if (argc!=3) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if ((s = nth(argv, 2))==nil || !pairp(s)) {
		err(stream, "");
		return;
	}

	for (p=s; p!=nil; p=cdr(p)) {
		if (stoidx(car(p), &id)==-1) {
			err(stream, "");
			return;
		}
	}

	fprintf(stream, "(");
	for (p=s; p!=nil; p=cdr(p)) {
		if (stoidx(car(p), &id)==-1) {
			err(stream, "unexpected error!");
			return;
		}
		put_string(wam_id2name(w, mode, id), stream);
		if (cdr(p)!=nil) {
			fprintf(stream, " ");
		}
	}
	fprintf(stream, ")");
}

void
swam_namel2idl(argc, argv, stream)
sexp argv;
FILE *stream;
{
	int mode;
	WAM *w;
	sexp s, p;
	char const *name;
	if (argc!=3) {
		err(stream, "");
		return;
	}
	if (stow(nth(argv, 0), &w)==-1) {
		err(stream, "");
		return;
	}
	if (stodir(nth(argv, 1), &mode)==-1) {
		err(stream, "");
		return;
	}
	if ((s = nth(argv, 2))==nil || !pairp(s)) {
		err(stream, "");
		return;
	}

	for (p=s; p!=nil; p=cdr(p)) {
		if (!(name = symbol_name(car(p)))) {
			err(stream, "");
			return;
		}
	}

	fprintf(stream, "(");
	for (p=s; p!=nil; p=cdr(p)) {
		if (!(name = symbol_name(car(p)))) {
			err(stream, "unexpected error!");
			return;
		}
		put_name2id(w, mode, name, stream);
		if (cdr(p)!=nil) {
			fprintf(stream, " ");
		}
	}
	fprintf(stream, ")");
}
