/*
 *	Copyright (c) 2001 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: token.c,v 1.1.1.1 2005/06/20 07:13:48 nis Exp $";
#endif

#include <config.h>

#include <sys/types.h>
#include <sys/uio.h>

#include <ctype.h>
#include <stdio.h>
#include <unistd.h>

#include "sexp.h"

static int my_getchar(void);
static int fillc(int);
static int shiftc(void);

static struct {
	int mys;
	char *p, *ep;
	int eofp;
	char buf[4096];
} gst = { -1, NULL, NULL, 0};

static struct {
	size_t nc;
	int c[3];
} cst = { 0, { EOF, EOF, EOF }};

void
sexp_setd(s)
{
	gst.mys = s;
	gst.p = gst.ep = NULL;
	gst.eofp = 0;
	cst.nc = 0;
	cst.c[0] = cst.c[1] = cst.c[2] = EOF;
}

static int
my_getchar()
{
	if (gst.eofp || gst.mys==-1) {
		return EOF;
	}
	while (gst.p==gst.ep) {
		ssize_t c;
		if ((c = read(gst.mys, gst.buf, sizeof gst.buf))<=0) {
			gst.eofp = 1;
			return EOF;
		}
		gst.p = gst.buf;
		gst.ep = gst.p + c;
	}
	return *gst.p++ & 0xff;
}

static int
fillc(n)
{
	size_t i;

	if (n<0 || 3<n) {
		return EOF;
	}
	for (i=cst.nc; i<n; i++) {
		cst.c[i] = my_getchar();
	}
	cst.nc = n;
	return cst.c[0];
}

static int
shiftc()
{
	switch (cst.nc) {
	case 3:
		cst.c[0] = cst.c[1];
		cst.c[1] = cst.c[2];
		cst.nc = 2;
		break;
	case 2:
		cst.c[0] = cst.c[1];
		cst.nc = 1;
		break;
	case 1:
	case 0:
		cst.c[0] = my_getchar();
		cst.nc = 1;
		break;
	default:
		fprintf(stderr, "never happen\n");
		return EOF;
	}
	return cst.c[0] & 0xff;
}

int
gettoken(token, size)
char *token;
{
	char *t, *ep;
	int c, c1;
	int type;

	t = token;
	ep = token+size;
	c = shiftc();

	if (c==EOF) {
		type = EOF;
	}
	else if (c=='\n' || c=='(' || c==')') {
		*t++ = c;
		type = c;
	}
	else if (c==' ' || c=='\t') {
		for (; t<ep-1; ) {
			*t++ = c;
			fillc(2);
			c1 = cst.c[1];
			if (!(c1==' ' || c1=='\t')) {
				break;
			}
			c = shiftc();
		}
		type = ' ';
	}
	else if (c==';') {
		for (; c!=EOF && t<ep-1; ) {
			*t++ = c;
			if (c=='\n') {
				break;
			}
			c = shiftc();
		}
		type = ';';
	}
	else if (c=='#' && fillc(2)!=EOF && cst.c[1]=='(') {
		*t++ = c;
		c = shiftc();
		*t++ = c;
		type = 'v';
	}
	else if (c=='#' && fillc(2)!=EOF && cst.c[1]=='\\') {
		*t++ = c;
		c = shiftc();
		*t++ = c;
		c = shiftc();
		if (c!=EOF && isalpha(c)) {
			for (; t<ep-1; ) {
				*t++ = c;
				fillc(2);
				c1 = cst.c[1];
				if (c1==EOF || !isalpha(c1)) {
					break;
				}
				c = shiftc();
			}
		}
		else if (c!=EOF) {
			*t++ = c;
		}
		type = 't';
	}
	else if (c=='"') {
#define	SHIFT_SB	0
#define	SHIFT_MB	1
		int shift = SHIFT_SB;
/*		*t++ = c; */
		c = shiftc();
		for (; c!=EOF && t<ep-1; ) {
			*t++ = c;
			if (c==033 && fillc(3)!=EOF && ((c1=cst.c[1])=='$' || c1=='(') && cst.c[2]=='B') {
				c = shiftc();
				*t++ = c;
				c = shiftc();
				*t++ = c;
				if (c1=='$') {
					shift = SHIFT_MB;
				}
				else {
					shift = SHIFT_SB;
				}
			}
			else if (shift==SHIFT_MB) {
				if ((c = shiftc())==EOF) {
					break;
				}
				*t++ = c;
			}
			else if (c=='\\' && fillc(2)!=EOF && (c1=cst.c[1]) && (c1=='"' || c1=='\\')) {
				c = shiftc();
				*(t-1) = c;
			}
			else if (c=='\\' && fillc(2)!=EOF && isdigit(cst.c[1])) {
				/* for MIT-scheme */
				int cc = 0;
				c = shiftc();	/* 1 mozi me */
				cc = c - '0';
				fillc(2);
				if (isdigit(cst.c[1])) {
					c = shiftc();	/* 2 mozi me */
					cc = cc*8 + c - '0';
					fillc(2);
					if (isdigit(cst.c[1])) {
						c = shiftc();	/* 3 mozi me */
						cc = cc*8 + c - '0';
					}
				}
				*(t-1) = cc;
			}
			else if (c=='"') {
				t--;
				break;
			}
			c = shiftc();
		}
		type = '"';
	}
	else {
		for (; t<ep-1; ) {
			*t++ = c;
			fillc(2);
			c1 = cst.c[1];
			if (c1==EOF || c1=='\n' || c1=='(' || c1==')' || c1==' ' || c1=='\t' || c1==';' || (c1=='#' && fillc(3)!=EOF && (cst.c[2]=='(' || cst.c[2]=='\\')) || c1=='"') {
				break;
			}
			c = shiftc();
		}
		type = 't';
	}
	*t = '\0';
	return type;
}
