/*
 *	Copyright (C) 1999 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: wflid.c,v 1.3 2007/03/29 03:21:19 nis Exp $"; */

#if 0
#define	PROF
#endif

#include <config.h>

#include <sys/param.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include "sexp.h"
#include "swam.h"

#if ! defined SA_RESETHAND
#define SA_RESETHAND	0
#endif

#if ! defined AF_LOCAL
#define	AF_LOCAL	AF_UNIX
#endif

static void usage(void);
static void sigflg(int, int, int);
static void sigchld(int);
static void serve(int);

char *progname;

static void
usage()
{
	fprintf(stderr, "usage: %s [-local|-inet]\n", progname);
}

static void
sigflg(sig, set, reset)
{
	struct sigaction sa;
	sigaction(sig, NULL, &sa);
	sa.sa_flags |= set;
	sa.sa_flags &= ~reset;
	sigaction(sig, &sa, NULL);
}

static void
sigchld(n)
{
	pid_t pid;
	int status;

	while ((pid = wait3(&status, WNOHANG, NULL)) > 0) ;
}

int
main(argc, argv)
char *argv[];
{
#if defined TCP_NODELAY
	int tmp=1;
#endif
	int ls;
	int domain;
	struct sockaddr *myaddr;
	struct sockaddr_in myaddr_in;
	struct sockaddr_un myaddr_un;
#if HAVE_SOCKLEN_T
	socklen_t addrlen;
#else
	size_t addrlen;
#endif

	if (progname = rindex(argv[0], '/')) {
		progname++;
	}
	else {
		progname = argv[0];
	}

	if (argc!=2) {
		usage();
		return 1;
	}
	if (!strcmp(argv[1], "-local")) {
		domain = AF_LOCAL; }
	else if (!strcmp(argv[1], "-inet")) {
		domain = AF_INET;
	}
	else {
		usage();
		return 1;
	}

	signal(SIGCHLD, sigchld);
	sigflg(SIGCHLD, SA_RESTART, SA_RESETHAND);
	signal(SIGPIPE, SIG_IGN);

	if (domain == AF_LOCAL) {
		char *path = "/tmp/wflis";
		int sun_len;

		unlink(path);
		sun_len = sizeof myaddr_un.sun_family + strlen(path)+1;
		bzero(&myaddr_un, sizeof myaddr_un);
		myaddr_un.sun_family = AF_LOCAL;
#if HAVE_SUN_LEN
		myaddr_un.sun_len = sun_len;
#endif
		*myaddr_un.sun_path = '\0';
		strncat(myaddr_un.sun_path, path, sizeof myaddr_un.sun_path);
		myaddr = (struct sockaddr *)&myaddr_un;
		addrlen = sun_len;
		fprintf(stderr, "domain=AF_LOCAL\n");
	}
	else if (domain == AF_INET) {
		int port = 2905;

		bzero(&myaddr_in, sizeof myaddr_in);
		myaddr_in.sin_family = AF_INET;
		myaddr_in.sin_addr.s_addr = INADDR_ANY;
		myaddr_in.sin_port = htons(port);
		myaddr = (struct sockaddr *)&myaddr_in;
		addrlen = sizeof myaddr_in;
		fprintf(stderr, "domain=AF_INET\n");
	}
	else {
		return 1;
	}

	if ((ls = socket(domain, SOCK_STREAM, 0))==-1) {
		perror("socket");
		return 1;
	}

	if (bind(ls, myaddr, addrlen)==-1) {
		perror("bind");
		return 1;
	}
	if (listen(ls, 2)==-1) {
		perror("listen");
		return 1;
	}

	for (;;) {
		struct sockaddr_un peeraddr_un;
		struct sockaddr_in peeraddr_in;
		struct sockaddr *peeraddr;
		int s;

		if (domain==AF_LOCAL) {
			bzero(&peeraddr_un, sizeof peeraddr_un);
			peeraddr = (struct sockaddr *)&peeraddr_un;
			addrlen = sizeof peeraddr_un;
		}
		else if (domain==AF_INET) {
			bzero(&peeraddr_in, sizeof peeraddr_in);
			peeraddr = (struct sockaddr *)&peeraddr_in;
			addrlen = sizeof peeraddr_in;
		}
		else {
			return 1;
		}
		if ((s = accept(ls, peeraddr, &addrlen))==-1) {
			perror("accept");
			if (errno!=EINTR) {
				sleep(1);
			}
			continue;
		}

#if defined TCP_NODELAY
		(void)setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *)&tmp, sizeof tmp);
		fprintf(stderr, "nodelay\n");
#endif
		fprintf(stderr, "accept\n");
#if ! defined PROF
		switch (fork()) {
		case -1:
			break;
		case 0:
			close(ls);
#endif
			serve(s);
			close(s);
#if ! defined PROF
			fprintf(stderr, "exit\n");
			_exit(0);
		default:
			close(s);
		}
#else
		exit(0);
#endif
	}
	return 0;
}

static void
serve(s)
{
	FILE *f;
	if (!(f = fdopen(s, "a"))) {
		return;
	}
	sexp_setd(s);

	for (;;) {
		sexp r;
		if (r = sexp_read()) {
			if (docmd(r, f) == -1) {
				break;
			}
			fprintf(f, "\n");
		}
		else {
			fprintf(f, "()\n");
			break;
		}
		fflush(f);
		relall();
	}
	return;
}
