/*
 *	Copyright (C) 2001 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: simple_unifdef.c,v 1.1.1.1 2005/06/20 07:13:48 nis Exp $"; */

#include <assert.h>
#include <ctype.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

char *progname;

static void usage(void);
static int dcompar(const void *, const void *);
static void reset_stack(void);
static int pop(void);
static void push(int);
static int empty(void);
static int read_config_h(FILE *);
static int process_file(FILE *);

#define	DISPLAY	1
#define	KILL	2
#define	SUSPEND	3
#define	BUFSIZE	1024

struct def {
	char *name;
	int value;
} *defines = NULL;

size_t ndefines = 0;

static void
usage()
{
	fprintf(stderr, "usage: %s config.h file\n", progname);
	exit(1);
}

static int
dcompar(va, vb)
const void *va, *vb;
{
	struct def const *a, *b;
	a = va;
	b = vb;
	return strcmp(a->name, b->name);
}

#define	nelems(e)	(sizeof(e)/sizeof*(e))

static int stack[64];
static int sp;

static void
reset_stack()
{
	sp = nelems(stack);
}

static int
empty()
{
	return sp==nelems(stack);
}

static int
pop()
{
	if (sp>=nelems(stack)) {
		fprintf(stderr, "stack empty\n");
		exit(2);
	}
	return stack[sp++];
}

static void
push(v)
{
	if (sp<0) {
		fprintf(stderr, "stack overflow\n");
		exit(3);
	}
	stack[--sp] = v;
}

static int
read_config_h(f)
FILE *f;
{
	size_t i;
	static char buf[BUFSIZE];

	for (ndefines=0; fgets(buf, sizeof buf, f); ) {
		if (!strncmp(buf, "#define", 7)) {
			ndefines++;
		}
	}
	rewind(f);
	if (!(defines = calloc(ndefines, sizeof *defines))) {
		perror("calloc");
		return 1;
	}
	for (i=0; fgets(buf, sizeof buf, f); ) {
		if (!strncmp(buf, "#define", 7)) {
			char *p, *name, *value;
			for (p=buf+7; isspace(*p); p++) ;
			name = p;
			for (; *p && !isspace(*p); p++) ;
			if (*p) {
				*p++ = '\0';
			}
			for (; isspace(*p); p++) ;
			value = p;
			for (; *p && !isspace(*p); p++) ;
			if (*p) {
				*p++ = '\0';
			}
			if (!(defines[i].name = malloc(strlen(name)+1))) {
				perror("malloc");
				return 1;
			}
			strcpy(defines[i].name, name);
			defines[i].value = atoi(value);
			i++;
		}
	}
	assert(i==ndefines);

	qsort(defines, ndefines, sizeof *defines, dcompar);
	return 0;
}

static int
process_file(f)
FILE *f;
{
	static char buf[BUFSIZE], work[BUFSIZE];
	int state;

	reset_stack();
	state = DISPLAY;
	for (; fgets(buf, sizeof buf, f); ) {
		if (!strncmp(buf, "#if", 3)) {
			struct def key, *v;
			char *name, *p;
			push(state);
			strcpy(work, buf);
			if (!strncmp(work, "#if ! ", 6)) {
				name = work+6;
				for (p=name; *p && !isspace(*p); p++) ;
				*p = '\0';
				key.name = name;
				v = bsearch(&key, defines, ndefines, sizeof key, dcompar);
				if (v) {
					if (v->value) {
						state = SUSPEND;
					}
					else {
						state = KILL;
					}
				}
			}
			else if (!strncmp(work, "#if ", 4)) {
				name = work+4;
				for (p=name; *p && !isspace(*p); p++) ;
				*p = '\0';
				key.name = name;
				v = bsearch(&key, defines, ndefines, sizeof key, dcompar);
				if (v) {
					if (!v->value) {
						state = SUSPEND;
					}
					else {
						state = KILL;
					}
				}
			}
		}
		if (state==DISPLAY) {
			fputs(buf, stdout);
		}
		else if (state==KILL) {
			if (strncmp(buf, "#if ", 4) && strncmp(buf, "#endif", 6)) {
				fputs(buf, stdout);
			}
		}
		if (!strncmp(buf, "#endif", 6)) {
			state = pop();
		}
	}
	if (!empty()) {
		fprintf(stderr, "stack not empty\n");
		return -1;
	}
	return 0;
}

main(argc, argv)
char *argv[];
{
	FILE *f;
	char *configh, *file;

	if (progname = rindex(argv[0], '/')) {
		progname++;
	}
	else {
		progname = argv[0];
	}
	for (argc--, argv++; argc>0; argc--, argv++) {
		if (!strcmp(*argv, "-h")) {
			usage();
		}
		else {
			break;
		}
	}
	if (argc!=2) {
		usage();
	}
	configh = argv[0];
	file = argv[1];
	if (!(f = fopen(configh, "r"))) {
		perror(configh);
		return 1;
	}
	if (read_config_h(f)==-1) {
		return 5;
	}
	fclose(f);
	if (!(f = fopen(file, "r"))) {
		perror(file);
		return 1;
	}
	if (process_file(f)==-1) {
		return 4;
	}
	fclose(f);
	exit(0);
}
