/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: tutil.c,v 1.12 2007/03/29 08:34:55 nis Exp $"; */

#define RORDER_WSH

#include <config.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <assert.h>
#include <errno.h>
#include <math.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dwerrno.h>
#include <geta/timer.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwsched.h>
#include <geta/dwam.h>

#include "tutil.h"
#include "nnet.h"

#define	MAXNODES	64

#if 0
static int connect_rem(struct sym_vec *, size_t, int);
#endif
static int fcompar(void const *, void const *);
static struct syminfo *dw_wsh_merge(struct sym_vec *, size_t, ssize_t *);
static ssize_t safe_request_num(ssize_t, double, int, size_t);
static ssize_t **dw_wcm_merge(struct sym_vec *, size_t, ssize_t);

struct gtim {
	struct timeval ave, best, worst, rcons;
	struct timeval elas[8];
	struct timeval com, mrg;
/*
	int n;
*/
};

struct gtim gtim;

struct gtim *
get_gtim()
{
	struct gtim *r;
	if (!(r = calloc(1, sizeof *r))) {
		return NULL;
	}
	memcpy(r, &gtim, sizeof *r);
	return r;
}

void
merge_gtim(a, b, c)
struct gtim const *a, *b;
struct gtim *c;
{
	size_t i;
#define	MSET(memb) do {\
	if (timercmp(&a->memb, &b->memb, <)) { \
		c->memb = a->memb; \
	} \
	else { \
		c->memb = b->memb; \
	} \
} while (0)
	MSET(ave);
	MSET(best);
	MSET(worst);
	MSET(rcons);
	MSET(com);
	MSET(mrg);
	for (i=0; i<nelems(a->elas); i++) {
		MSET(elas[i]);
	}
}

void
clear_timers()
{
	size_t i;
	struct gtim *g = &gtim;

	timerclear(&g->ave);
	timerclear(&g->best);
	timerclear(&g->worst);
	timerclear(&g->rcons);
	timerclear(&g->com);
	timerclear(&g->mrg);
	for (i=0; i<nelems(g->elas); i++) {
		timerclear(&g->elas[i]);
	}
}

void
print_timers(f, elap)
FILE *f;
struct timeval *elap;
{
	print_gtim(f, elap, &gtim);
}

void
print_gtim(f, elap, g)
FILE *f;
struct timeval *elap;
struct gtim *g;
{
	size_t i;

#define	PPP(stream, name, var) \
	(fprintf((stream), " (" #name " "), \
	 timerprint((stream), (var)), \
	 fprintf((stream), ")"))

	PPP(f, ela, elap);
	PPP(f, com, &g->com);
	PPP(f, mrg, &g->mrg);
	PPP(f, bes, &g->best);
	PPP(f, ave, &g->ave);
	PPP(f, wor, &g->worst);
	PPP(f, sum, &g->rcons);
#undef PPP

	fprintf(f, " (elas");
	for (i=0; i<nelems(g->elas); i++) {
		if (timerisset(&g->elas[i])) {
			fprintf(f, " (%s ", sizetostr(i));
			timerprint(f, &g->elas[i]);
			fprintf(f, ")");
		}
	}
	fprintf(f, ")");
}

extern int nnet_errno;

int
rcxsp_connect_n_open(rp, fault_tolerant)
struct rcxsp *rp;
{
	int s;
#if 0
	fprintf(stderr, "connecting to %s:%s\n", rp->host, rp->serv);
#endif

	if (fault_tolerant && dlm_match(rp->host, rp->serv)) {
		fprintf(stderr, "warnig: %s:%s is registered in the down-list\n", rp->host, rp->serv);
		return -1;
	}

	if ((s = nnet_cli(rp->host, rp->serv))==-1) {
		fprintf(stderr, "%s: %s", rp->host, strerror(errno));
		if (nnet_errno == ETIMEDOUT) {
			fprintf(stderr, "%s:%s %s\n", rp->host, rp->serv, strerror(errno));
			(void)dlm_add(rp->host, rp->serv);
		}
		return -1;
	}
	if (!(rp->f = dwc_dxr_open(s, rp->i.handle, rp->i.type, rp->i.fragment, rp->i.cascade))) {
		close(s);
		fprintf(stderr, "%s: %s", rp->host, strerror(errno));
		return -1;
	}
	rp->s = s;	/* for random-order-wsh */
	dwc_tcp_nodelay(rp->f);
#if 1
	syslog(LOG_DEBUG, "connected to %s:%s.%s\n", rp->host, rp->serv, ssizetostr(rp->i.fragment));
#endif
	return 0;
}

ssize_t
mksum(d, n, sump, sum_i)
ssize_t n;
ssize_t sum_i;
struct syminfo *d;
float **sump;
{
	ssize_t i;
	float *sum;
	ssize_t nsum;

	qsort(d, n, sizeof *d, sym_compar);

	nsum = (n-1)/sum_i+1;
	if (!(sum = malloc(nsum * sizeof *sum))) {
		fprintf(stderr, "malloc: %s", strerror(errno));
		dwerrno = errno;
		return -1;
	}
	for (i = 0; i<nsum; i++) {
		ssize_t j;
		j = (i+1)*sum_i-1;
		if (j>=n) {
			assert(i+1 >= nsum);
			j = n-1;
		}
		sum[i] = d[j].weight;
	}
	*sump = sum;
	return nsum;
}

static int
fcompar(va, vb)
void const *va, *vb;
{
	float const *a, *b;

	a = va;
	b = vb;
	if (!fequal(*a, *b)) {
		/* descending order */
		if (*a < *b) {
			return 1;
		}
		else if (*a > *b) {
			return -1;
		}
	}
	return 0;
}

ssize_t
makesv(my_nodeid, np_all, rems, svr, q, fault_tolerant)
ssize_t my_nodeid;
size_t np_all;
struct rcxsp **rems;
struct sym_vec **svr;
struct qqs *q;
{
	int need_pack = 0;
	ssize_t p, np;
	ssize_t children[MAXNODES], descendant[MAXNODES];
	struct sym_vec *sv;
	int cascade;
	struct rcxsp *rem;

/* if (fault_tolerant) fprintf(stderr, "makesv: FAULTTOLERANT\n"); */

	if (np_all>MAXNODES) {
		fprintf(stderr, __FILN__ ": node # limit exceeded");
		dwerrno = DWE_LIMIT;
		return -1;
	}

	cascade = my_nodeid!=-2;
/* if (cascade) fprintf(stderr, "makesv: CASCADE\n"); */
	rem = rems[cascade];

	np = make_children_list(my_nodeid, np_all, children, 0);
/* fprintf(stderr, "np_all=%d, np=%d", np_all, np); */
	assert(0<=np);
	if (!(sv = calloc(np+1, sizeof *sv))) {
		fprintf(stderr, __FILN__ ": calloc: %s", strerror(errno));
		dwerrno = errno;
		return -1;
	}
	for (p=0; p<np; p++) {
		ssize_t j, k, nd;
		double prob;
		struct sym_vec *svp = &sv[p];

		j = children[p];
		assert(0<=j);
		assert(j<np_all);
		assert(j != my_nodeid);
		svp->nodeid = j;
		svp->rem = &rem[j];
		svp->q = NULL;
		prob = rem[j].p;
		if (cascade) {
			nd = make_children_list(j, np_all, descendant, 1);
			for (k=0; k<nd; k++) {
				prob += rem[descendant[k]].p;
			}
		}
		svp->p = prob;
		svp->active = 1;
	}

	for (p=0; p<np; p++) {
		struct sym_vec *svp = &sv[p];
		struct rcxsp *rp = svp->rem;
		/* ssize_t fragment; */

		/* note: svp->nodeid ha fragment no id ni hitosii */
		/* fragment = svp->nodeid; */
		if (rp && !rp->f && rcxsp_connect_n_open(rp, fault_tolerant)==-1) {
			if (fault_tolerant) {
syslog(LOG_DEBUG, "%s:%s is down. skipped.", rp->host, rp->serv);
				need_pack = 1;
				svp->active = 0;
				continue;
			}
			fprintf(stderr, __FILN__ ": rcxsp_connect_n_open failed");
			return -1;
		}
	}
	if (need_pack) {
		double ptotal;
		ssize_t i = 0;
		if (cascade || q) {
			fprintf(stderr, "internal error\n");
			return -1;
		}
		for (p = 0, ptotal = 0; p < np; p++) {
			struct sym_vec *svp = &sv[p];
			if (svp->active) {
				sv[i++] = *svp;
				ptotal += svp->p;
			}
		}
		assert(np != i);
		assert(ptotal != 0);
		if (i == 0) {	/* all nodes are down */
			return -1;
		}
		np = i;
		for (i = 0, ptotal = 0; i < np; i++) {
			struct sym_vec *svp = &sv[p];
			svp->p /= ptotal;
		}
	}

	if (my_nodeid>=0) {	/* (called from dwdserv.c:cascade_init) */
		struct sym_vec *svp = &sv[np++];
		if (!q) {
			fprintf(stderr, __FILN__ ": !q");
			dwerrno = DWE_INVAL;
			return -1;
		}
		assert(0<=my_nodeid);
		assert(my_nodeid<np_all);
		svp->nodeid = my_nodeid;
		svp->rem = NULL;
		svp->q = q;
		svp->p = rem[my_nodeid].p;
		svp->active = 1;
	}
	*svr = sv;
	assert(0<np);
	return np;
}

struct syminfo *
rdwsh(q, nq, sv, np, type, nd, tnd, totalp, sum_i, prob_mode, fault_tolerant)
struct syminfo const *q;
struct sym_vec *sv;
ssize_t *nd;
ssize_t *totalp;
ssize_t nq, tnd, sum_i;
ssize_t np;
{
	ssize_t total;
	struct timeval comm_start, comm_end, best, worst, sum;
	struct timeval merge_start, merge_end;
	struct syminfo *d;
	size_t p;
	ssize_t nsss;
#if defined XDEBUG
fprintf(stderr, "%s: tutil: rdwsh: start\n", udate());
#endif

	if (prob_mode) {
/* fprintf(stderr, "rdwsh: PROBM: %d\n", prob_mode); */
#if 0
static char s[512];
static int cnt=0;
*s='\0';
cnt++;
#endif
		for (p=0; p<np; p++) {
			int b = prob_mode;
			struct sym_vec *svp = &sv[p];
			svp->srn = safe_request_num(tnd, svp->p, b, np);
#if 0
printf("nd=%s, p=%e, a=0.%d srn=%s\n", ssizetostr(*nd), svp->p, b-1, ssizetostr(svp->srn));
#endif

#if 0
if (cnt<4) sprintf(s+strlen(s), "%s: %f %s ", sizetostr(svp->nodeid), svp->p, dftostr(svp->srn));
#endif
		}
#if 0
if (cnt<4) fprintf(stderr, "%s", s);
#endif
	}
	else {
		for (p=0; p<np; p++) {
			struct sym_vec *svp = &sv[p];
			svp->srn = *nd;
			assert((*nd) == tnd);
		}
	}

	total = 0;

	gettimeofday(&comm_start, NULL);

	for (p=0; p<np; p++) {
		struct sym_vec *svp = &sv[p];
		struct rcxsp *rp = svp->rem;

		if (rp) {
#if defined XDEBUG
fprintf(stderr, "%s: tutil: rdwsh: send request to child %d\n", udate(), p);
#endif
			if (dwc_wshS(rp->f, q, nq, type, svp->srn, tnd, sum_i, prob_mode)==-1) {
				fprintf(stderr, __FILN__ ": dwc_wshS failed");
				return NULL;
			}

#if defined RORDER_WSH
			svp->received = 0;
#endif
		}
	}

	nsss = 0;

	for (p=0; p<np; p++) {
		struct sym_vec *svp = &sv[p];
		struct qqs *qq = svp->q;

		if (qq) {
			struct timeval wsh_start, wsh_end;
			ssize_t t;
			svp->n = *nd;
			gettimeofday(&wsh_start, NULL);
			if (!(svp->s = dw_wsh(q, nq, qq->xr, qq->qs, qq->ds, type, &svp->n, &t))) {
				fprintf(stderr, __FILN__ ": dw_wsh failed");
				return NULL;
			}
			if (sum_i) {
				if ((svp->nsum = mksum(svp->s, svp->n, &svp->sum, sum_i))==-1) {
				fprintf(stderr, __FILN__ ": mksum failed");
					return NULL;
				}
				nsss += svp->nsum;
			}
			gettimeofday(&wsh_end, NULL);
			timersub(&wsh_end, &wsh_start, &svp->elapsed);
			total += t;
		}
	}

	if (sum_i) {
		ssize_t maxns;
		float *sss, piv;
		ssize_t i;
		ssize_t nn;
/* fprintf(stderr, "rdwsh: SUMI: %d\n", sum_i); */

		maxns = ((*nd)-1)/sum_i+1;

/* XXX RORDER ni subeki? */
		for (p=0; p<np; p++) {
			struct sym_vec *svp = &sv[p];
			struct rcxsp *rp = svp->rem;

			if (rp) {

				if ((svp->nsum = dwc_wshRs(rp->f, &svp->sum, maxns))==-1) {
					fprintf(stderr, __FILN__ ": dwc_wshRs (from %s) failed", rp->host);
					return NULL;
				}
				nsss += svp->nsum;
			}
		}

		assert(sizeof *sv[0].sum==sizeof *sss);

		if (!(sss = malloc(nsss * sizeof *sss))) {
			fprintf(stderr, __FILN__ ": malloc failed");
			dwerrno = errno;
			return NULL;
		}
		for (p=0, i = 0; p<np; p++) {
			struct sym_vec *svp = &sv[p];

			memmove(sss+i, svp->sum, svp->nsum * sizeof *svp->sum);
			i += svp->nsum;
		}
		assert(i==nsss);

/*
全部で N 個の玉があり, それぞれに数が記されている. 
これらの玉を np 個のグループに分割し, 
さらに, それぞれのグループ内で大きい順に sum_i 個ずつ取り箱に入れる. 
したがって, 最後の箱には, 1 〜 sum_i 個の玉が入ることになる. 
それぞれの箱の表面にはその箱に入っている玉に記された数のうち, 
最小のものを記す. 
最後に, 全部の箱を集め, その箱の表面に記された数の大きい順に並べる. 
どの箱がどのグループに属するものなのかは分からない. 
 */

/*
さて, これらの箱から nn 個を選び, 少なくとも nd 個の玉を含むようにするには
nn は最低いくらでなければならないか. 
(多くとも np 個の箱については sum_i 個以下の玉しか入っていないことに注意. )

np >= nd ならば,  np 個の箱を取れば良い. 
そうでない場合, まず,  np 個の箱を取る. すると, 最悪であと(nd-np)個の玉が
必要だから, (nd-np-1)/sum_i+1 個の箱を取れば良い. 結局, 
np + (nd-np-1)/sum_i + 1 個の箱を取れば良い. 

*/
		if (np >= (*nd)) {
			nn = *nd;
		}
		else {
			nn = np + ((*nd) - np - 1)/sum_i + 1;
		}
		if (nn>nsss) {
			nn = nsss;
		}
		if (nn==nsss) {
			qsort(sss, nn, sizeof *sss, fcompar);
			piv = sss[nn-1];
		}
		else {
			heapselect(sss, nn, nsss, sizeof *sss, fcompar);
			piv = sss[0];
		}

/*
全部の箱をその表面に記されている数の大きい順(同じ数があれば, それらのうち
それを選ぶかは任意)に nn 個の箱を, 選んできた. 
このとき, それらのうちで最小の数が piv であった. 
(すなわち, この nn 個の箱の中の玉に記されているどの数も piv より小さくは無い)
従って, piv 以上の玉だけを集めれば, 全ての玉の(等しいものも含めて)
大きい順に nd 個調べることができる. 
 */

		free(sss);

		for (p=0; p<np; p++) {
			struct sym_vec *svp = &sv[p];
			struct rcxsp *rp = svp->rem;
			ssize_t n, nreq;

/*
これからは, 各グループに関する知識が利用可能である. 
各グループから, それぞれいくつづつ取れば良いか. 
 */

			for (n = 0; n<svp->nsum && svp->sum[n]>=piv; n++) ;
/*
このグループで表面に piv 以上の数が記された箱は調べなければならない. 
さらに, 次の箱に sum_i-1 個の玉に, piv 以上の数が記されている可能性が
ある. 
結局, このグループからは
nreq = (n+1)*sum_i-1
個の玉(n+1個の箱, ただし, 最後の箱からは最大 sum_i-1 個の玉) を調べなければ
ならない. 
 */

			free(svp->sum);
			nreq = (n+1)*sum_i-1;

			if (rp) {
				if (dwc_wshRn(rp->f, nreq)==-1) {
					fprintf(stderr, __FILN__ ": dwc_wshRn failed");
					return NULL;
				}
			}
			else {
				if (svp->n > nreq) { 
					svp->n = nreq;
				}
			}
		}
	}

#if defined RORDER_WSH

	for (;;) {
		fd_set readfds;
		int nfds;

		FD_ZERO(&readfds);
		nfds = 0;
		for (p=0; p<np; p++) {
			struct sym_vec *svp = &sv[p];
			struct rcxsp *rp = svp->rem;
			if (rp && !svp->received) {
				FD_SET(rp->s, &readfds);
				if (nfds <= rp->s) {
					nfds = rp->s+1;
				}
			}
		}
		if (nfds==0) {
			break;
		}

		if (select(nfds, &readfds, NULL, NULL, NULL)==-1) {
			fprintf(stderr, __FILN__ ": select: %s", strerror(errno));
			dwerrno = errno;
			return NULL;
		}

		for (p=0; p<np; p++) {
			struct sym_vec *svp = &sv[p];
			struct rcxsp *rp = svp->rem;
			if (rp && FD_ISSET(rp->s, &readfds)) {
				ssize_t t;
				assert(!svp->received);
#if defined XDEBUG
fprintf(stderr, "%s: tutil: rdwsh: receive answer from child %d\n", udate(), p);
#endif
				if (!(svp->s = dwc_wshR(rp->f, &svp->n, &t, &svp->elapsed))) {
					fprintf(stderr, __FILN__ ": dwc_wshR failed");
					return NULL;
				}
				svp->received = 1;
				total += t;
			}
		}
	}
#else
	for (p=0; p<np; p++) {
		struct sym_vec *svp = &sv[p];
		struct rcxsp *rp = svp->rem;
		ssize_t t;

#if defined XDEBUG
fprintf(stderr, "%s: tutil: rdwsh: receive answer from child %d\n", udate(), p);
#endif
		if (rp) {
/* XXX : arg 3 ---> freq_t ? */
			if (!(svp->s = dwc_wshR(rp->f, &svp->n, &t, &svp->elapsed))) {
				fprintf(stderr, __FILN__ ": dwc_wshR failed");
				return NULL;
			}
			total += t;
		}
	}
#endif

	gettimeofday(&comm_end, NULL);
	merge_start = comm_end;
	timersub(&comm_end, &comm_start, &comm_end);
	timeradd(&comm_end, &gtim.com, &gtim.com);

	if (!(d = dw_wsh_merge(sv, np, nd))) {
		fprintf(stderr, __FILN__ ": dw_wsh_merge failed");
		return NULL;
	}
	gettimeofday(&merge_end, NULL);
	timersub(&merge_end, &merge_start, &gtim.mrg);

	timerclear(&best);
	timerclear(&worst);
	timerclear(&sum);
	for (p = 0; p<np; p++) {
		struct sym_vec *svp = &sv[p];

		timeradd(&svp->elapsed, &gtim.rcons, &gtim.rcons);
		timeradd(&svp->elapsed, &sum, &sum);
		if (p==0 || timercmp(&best, &svp->elapsed, >)) {
			best = svp->elapsed;
		}
		if (timercmp(&worst, &svp->elapsed, <)) {
			worst = svp->elapsed;
		}
		if (p < nelems(gtim.elas)) {
			timeradd(&svp->elapsed, &gtim.elas[p], &gtim.elas[p]);
		}
		free(svp->s);
	}
	timeradd(&best, &gtim.best, &gtim.best);
	timeradd(&worst, &gtim.worst, &gtim.worst);
	if (p) {
		timeridiv(&sum, p, &sum);
	}
	timeradd(&sum, &gtim.ave, &gtim.ave);

	if (totalp) {
		*totalp = total;
	}
#if defined XDEBUG
fprintf(stderr, "%s: tutil: rdwsh: finish\n", udate());
#endif

	return d;
}

static struct syminfo *
dw_wsh_merge(sv, n, nd)
struct sym_vec *sv;
size_t n;
ssize_t *nd;
{
	struct syminfo *d;
	ssize_t len, totlen;
	size_t i;
	ssize_t j;

	totlen = 0;
	for (i=0; i<n; i++) {
		struct sym_vec *svp = &sv[i];
		len = svp->n;
		totlen += len;
	}

	if (totlen==0) {
		/* empty! */
		*nd = 0;
		if (!(d = malloc(sizeof *d))) {	/* dummy entry */
			fprintf(stderr, "malloc: %s", strerror(errno));
			dwerrno = errno;
			return NULL;
		}
		return d;
	}

	if (!(d = malloc(totlen * sizeof *d))) {
		fprintf(stderr, "malloc: %s", strerror(errno));
		dwerrno = errno;
		return NULL;
	}

	j = 0;
	for (i=0; i<n; i++) {
		struct sym_vec *svp = &sv[i];
		len = svp->n;
		memmove(&d[j], svp->s, len * sizeof *d);
		j += len;
	}

	if ((*nd) > totlen) {
		*nd = totlen;
	}
	else {
		heapselect(d, (*nd), totlen, sizeof *d, sym_compar);
	}

	return d;
}

#if 0
static int
safe_request_num(n, p, b)
double p;
{
	double a;
	int k;
	double q, s, t;

	a = 1.0-1.0/b;
	q = 1 - p;
	s = t = pow(q, n);
	if (s==0) {
		return n;
	}
	for (k=0; k<n && s<a; k++) {
		t *= (n - k) * p / ((k + 1) * q);
		s += t;
	}
	return k;
}
#endif

static ssize_t
safe_request_num(n, p, b, l)
ssize_t n;
size_t l;
double p;
{
	double a;
	ssize_t k;
	double q, s, t;
	double r;

	a = 1.0-1.0/b;
	r = 1.0/b;
	q = 1 - p;
	s = t = pow(q, n);
	if (s==0) {
		return n;
	}
	for (k = 0; k<n && 1.0-s>r/l; k++) {
		t *= (n - k) * p / ((k + 1) * q);
		s += t;
	}
	return k;
}

ssize_t **
rdwcm(words, nwords, docs, ndocs, sv, np, fault_tolerant)
struct syminfo const *words;
struct syminfo const *docs;
struct sym_vec *sv;
ssize_t nwords, ndocs;
ssize_t np;
{
	ssize_t **d;
	size_t p;
#if defined XDEBUG
fprintf(stderr, "%s: tutil: rdwcm: start\n", udate());
#endif
fprintf(stderr, "tutil: rdwcm: start\n");

fprintf(stderr, "tutil: rdwcm: distributing...\n");
	for (p=0; p<np; p++) {
		struct sym_vec *svp = &sv[p];
		struct rcxsp *rp = svp->rem;

		if (rp) {
#if defined XDEBUG
fprintf(stderr, "%s: tutil: rdwcm: send request to child %s\n", udate(), sizetostr(p));
#endif
fprintf(stderr, "tutil: rdwcm: send request to child %s\n", sizetostr(p));
			if (dwc_wcmS(rp->f, words, nwords, docs, ndocs)==-1) {
				fprintf(stderr, __FILN__ ": dwc_wcmS failed");
				return NULL;
			}

#if defined RORDER_WSH
			svp->received = 0;
#endif
		}
	}

fprintf(stderr, "tutil: rdwcm: local...\n");
	for (p=0; p<np; p++) {
		struct sym_vec *svp = &sv[p];
		struct qqs *qq = svp->q;

		if (qq) {
			if (!(svp->c = dw_cm(words, nwords, docs, ndocs, qq->xr))) {
				fprintf(stderr, __FILN__ ": dxr_cm failed");
				return NULL;
			}
		}
	}
fprintf(stderr, "tutil: rdwcm: gathering...\n");


#if defined RORDER_WSH

	for (;;) {
		fd_set readfds;
		int nfds;

		FD_ZERO(&readfds);
		nfds = 0;
		for (p=0; p<np; p++) {
			struct sym_vec *svp = &sv[p];
			struct rcxsp *rp = svp->rem;
			if (rp && !svp->received) {
				FD_SET(rp->s, &readfds);
				if (nfds <= rp->s) {
					nfds = rp->s+1;
				}
			}
		}
		if (nfds==0) {
			break;
		}

		if (select(nfds, &readfds, NULL, NULL, NULL)==-1) {
			fprintf(stderr, __FILN__ ": select: %s", strerror(errno));
			dwerrno = errno;
			return NULL;
		}

		for (p=0; p<np; p++) {
			struct sym_vec *svp = &sv[p];
			struct rcxsp *rp = svp->rem;
			if (rp && FD_ISSET(rp->s, &readfds)) {
				assert(!svp->received);
#if defined XDEBUG
fprintf(stderr, "%s: tutil: rdwcm: receive answer from child %s\n", udate(), sizetostr(p));
#endif
fprintf(stderr, "tutil: rdwcm: receive answer from child %s (1)\n", sizetostr(p));
				if (!(svp->c = dwc_wcmR(rp->f))) {
					fprintf(stderr, __FILN__ ": dwc_wcmR failed");
					return NULL;
				}
				svp->received = 1;
			}
		}
	}
#else
	for (p=0; p<np; p++) {
		struct sym_vec *svp = &sv[p];
		struct rcxsp *rp = svp->rem;

#if defined XDEBUG
fprintf(stderr, "%s: tutil: rdwcm: receive answer from child %d\n", udate(), p);
#endif
fprintf(stderr, "tutil: rdwcm: receive answer from child %d (2)\n", p);
		if (rp) {
			if (!(svp->c = dwc_wcmR(rp->f))) {
				fprintf(stderr, __FILN__ ": dwc_wcmR failed");
				return NULL;
			}
		}
	}
#endif
fprintf(stderr, "tutil: rdwcm: merging...\n");

	d = dw_wcm_merge(sv, np, nwords);

	for (p = 0; p<np; p++) {
		struct sym_vec *svp = &sv[p];
		free(svp->c);
	}

#if defined XDEBUG
fprintf(stderr, "%s: tutil: rdwcm: finish\n", udate());
#endif
fprintf(stderr, "tutil: rdwcm: finish\n");

	return d;
}

static ssize_t **
dw_wcm_merge(sv, np, nw)
struct sym_vec *sv;
size_t np;
ssize_t nw;
{
	size_t p;
	ssize_t i, j;
	ssize_t **cm;
	if (!(cm = dw_matrix(nw, nw, sizeof **cm))) {
		return NULL;
	}
	for (i = 0; i<nw; i++) {
		for (j = 0; j<nw; j++) {
			cm[i][j] = 0;
		}
	}
	for (p=0; p<np; p++) {
		struct sym_vec *svp = &sv[p];
		ssize_t **c;
		c = svp->c;
		for (i = 0; i<nw; i++) {
			for (j = 0; j<nw; j++) {
				cm[i][j] += c[i][j];
			}
		}
	}
	return cm;
}
