/*
 *	Copyright (C) 1996 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: wsh.c,v 1.10 2007/03/26 00:17:57 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>

#include <assert.h>
#if defined HAVE_DLFCN_H
#include <dlfcn.h>
#endif
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <math.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dwerrno.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#if defined WAM_WSH
#include <geta/dwam.h>
#endif
#if defined WITH_WSHC
#include <geta/wshc.h>
#endif

#include "dxrP.h"
#include "dcsP.h"
#include "bex.h"

struct qvec {
	/* vec, elemendp ha, section 2 de access kanou */
	struct xr_vec const *vec;
	struct xr_elem const *elemendp;
	ssize_t elem_num;
	freq_t freq_sum;
	/* wq ha section 2 de keisan sareru. section 3, 5 de access kanou */
	double wq;
	/* q, elemp ha, section 3, 5 de siyou kanou */
	struct syminfo const *q;
	struct xr_elem const *elemp;
	/* elemp1, elemp1id ha heap no kanri ni motii rareru. user ha access huka */
	struct xr_elem const *elemp1;
	idx_t elemp1id;
};

inline static int acompar(void const *, void const *);
inline static int dcompar(void const *, void const *);
inline static void a_reheap(struct qvec **, size_t, size_t);
inline static void d_reheap(struct syminfo **, size_t, size_t);
static int bool_type(struct syminfo const *, ssize_t, ssize_t *);

#if ! defined WAM_WSH
#define	DXR_HANDLE(d)			dxr_handle(g->dxr.d)
#define	DXR_SIZE(d)			dxr_size(g->dxr.d)
#define	DXR_TOTAL_ELEM_NUM(d)		dxr_total_elem_num(g->dxr.d)
#define	DXR_ELEM_NUM(d, id)		dxr_elem_num(g->dxr.d, id)
#define	DXR_FREQ_SUM(d, id)		dxr_freq_sum(g->dxr.d, id)
#define	DXR_GET_VEC(d, id, v)		dxr_get_vec(g->dxr.d, id, v)
#define	DXR_GET_VEC_DUP(d, id, v)	dxr_get_vec_dup(g->dxr.d, id, v)
#define	DXR_NEED_DUP_P(d)		dxr_need_dup_p(g->dxr.d)
#else
#define	DXR_HANDLE(d)			wam_handle(g->wam)
#define	DXR_SIZE(d)			wam_size(g->wam, g->dir.d)
#define	DXR_TOTAL_ELEM_NUM(d)		wam_total_elem_num(g->wam, g->dir.d)
#define	DXR_ELEM_NUM(d, id)		wam_elem_num(g->wam, g->dir.d, id)
#define	DXR_FREQ_SUM(d, id)		wam_freq_sum(g->wam, g->dir.d, id)
#define	DXR_GET_VEC(d, id, v)		wam_get_vec(g->wam, g->dir.d, id, v)
#define	DXR_GET_VEC_DUP(d, id, v)	ERROR NO TRANSLATION
#define	DXR_NEED_DUP_P(d)		ERROR NO TRANSLATION
#endif

#if ! defined SO_WSH
#if ! defined WAM_WSH
int
wsh_init()
{
	init_hypGsum();
	return 0;
}
#endif
#endif

inline static int
acompar(va, vb)
void const *va, *vb;
{
	struct qvec const *a, *b;
	a = va;
	b = vb;

	if (a->elemp1id < b->elemp1id) {
		return -1;
	}
	else if (a->elemp1id > b->elemp1id) {
		return 1;
	}
	else if (a < b) {
		return -1;
	}
	else if (a > b) {
		return 1;
	}
	return 0;
}

/*
 * この関数は compar.c にある sym_compar と同じでなければならない.
 * しかし, (gcc などの拡張機能である) inline を使える場合は使いたいので,
 * ここにべつの名前で置いている. 修正する場合は注意されたい.
 */
static inline int
dcompar(va, vb)
void const *va, *vb;
{
	struct syminfo const *a, *b;
	a = va;
	b = vb;
	if (!fequal(a->weight, b->weight)) {
		/* descending order */
		if (a->weight < b->weight) {
			return 1;
		}
		else if (a->weight > b->weight) {
			return -1;
		}
	}
	else if (a->id < b->id) {
		return -1;
	}
	else if (a->id > b->id) {
		return 1;
	}
	return 0;
}

#if 0
inline static void
a_reheap(a, l, na)
struct qvec **a;
{
	size_t i, j;
	struct qvec *x;
	i = l;
	x = a[i];
	while ((j=2*i) <= na) {
		if (j<na && acompar(a[j], a[j+1])>0) {
			j++;
		}
		if (acompar(a[j], x)>=0) {
			break;
		}
		a[i] = a[j];
		i = j;
	}
	a[i] = x;
}
#else
static void
a_reheap(a, l, na)
struct qvec **a;
size_t l, na;
{
	size_t i, j;
	struct qvec *x;
	idx_t aid, bid;
	i = l;
	x = a[i];
	while ((j=2*i) <= na) {
		if (j<na) {
			aid = a[j]->elemp1id;
			bid = a[j+1]->elemp1id;
			if (aid < bid) {
				/* ok */
			}
			else if (aid > bid) {
				j++;
			}
			else if (a[j] < a[j+1]) {
				/* ok */
			}
			else if (a[j] > a[j+1]) {
				j++;
			}
		}
		aid = a[j]->elemp1id;
		bid = x->elemp1id;
		if (aid < bid) {
			/* ok */
		}
		else if (aid > bid) {
			break;
		}
		else if (a[j] < x) {
			/* ok */
		}
		else if (a[j] > x) {
			break;
		}
		a[i] = a[j];
		i = j;
	}
	a[i] = x;
}
#endif

inline static void
d_reheap(dh, l, nd)
struct syminfo **dh;
size_t l, nd;
{
	size_t i, j;
	struct syminfo *x;
	i=l;
	x = dh[i];
	while ((j=2*i) <= nd) {
		if (j<nd && dcompar(dh[j], dh[j+1])<0) {
			j++;
		}
		if (dcompar(dh[j], x)<=0) {
			break;
		}
		dh[i] = dh[j];
		i = j;
	}
	dh[i] = x;
}

#define	private

struct _g {
	/* args */
	struct syminfo const *q;
	ssize_t qlen;
#if ! defined WAM_WSH
	struct {
		DWFILE *q, *qs, *ds;
	} dxr;
#else
	WAM *wam;
	struct {
		int q, qs, ds;
	} dir;
#endif
	int weight_type;
	private ssize_t *ndp;
	private ssize_t *totalp;
#if defined WAM_WSH
	void *cookie;
#endif

	char const *handle;
	private	ssize_t total;
	private	struct syminfo *d;
		/* kono pointer ga sasu ryouiki ha
		  nd + 1 youso kakuho sareteori, saigo no 1 youso
		(i.e. d[nd]) ha, itizi ryouiki to site motii rareru */
	private	struct syminfo **dh;	/* d wo heap kanri surtameno hairetu.
					heap ha 1 kara kazoeruto benri nanode,
				dh[1..nd] wo tukau. (dh[0] ha tukawanai) */
	private	ssize_t nd;	/* mitukeru saidai kosuu. *ndp de atae rareru */
	private	ssize_t nnd;	/* mitukatta kosuu */

	ssize_t total_N_q, total_N_d;
	freq_t TF_q;
	freq_t total_TF_q, total_TF_d;

	private	struct qvec *qvec;
	private	size_t qveclen;

	struct qvec **qvec_h;

	/* heap */
	private	struct qvec **a;
	private	size_t na;

	/* boolean */
#define	BOOL_OFF	0
#define	BOOL_PSEUDO	1
#define	BOOL_BEX	2
	private struct {
		int type;
		size_t num_and_q, num_not_q, num_org_q;
		int *attrs;
		struct bex *bex;
		size_t bex_len;
	} bool;
#if ! defined WAM_WSH
	private int need_dup;
#endif
	union {
#if ! defined SO_WSH
#include "wt/s0.c"
#else
#include "sowt/s0.c"
#endif
	} u;
};

static struct syminfo *wsh_toplevel(struct _g *);
static int init_globals(struct _g *);
static int p2(struct _g *);
static int init_heap(struct _g *);
static int p3(struct _g *);
static idx_t p3x(struct _g *, size_t *, size_t *, size_t *, size_t *, int *);
static int calc_sigma(struct _g *, struct syminfo *, size_t);
static int normalize(struct _g *, struct syminfo *);
static int addoc(struct _g *, struct syminfo *);

static struct xr_vec zerovec = {1, 0, {{0, 0}}};

#define	Bstr(x, y)	x ## _ ## y
#define	Cstr(x)		# x

#if ! defined WAM_WSH
#define	FN_NAME_0	dw_wsh
#else
#define	FN_NAME_0	wwsh
#endif
#define	FN_NAME_STR	FN_NAME_SSTR(FN_NAME_0)
#define	FN_NAME_SSTR(x)	Cstr(x)

#define	FN_NAME_APPEND(x, y)	Bstr(x, y)

#if ! defined SO_WSH
#define	FN_NAME	FN_NAME_0
#else
#include "sowt/sname.c"
#define	FN_NAME	FN_NAME_APPEND(FN_NAME_0, SO_NAME)
#endif

#define	SO_WT_BASE	0x10000

struct syminfo *
FN_NAME
#if ! defined WAM_WSH
	(q, qlen, dxr_q, dxr_qs, dxr_ds, weight_type, ndp, totalp)
#else
	(q, qlen, wam, dir_q, weight_type, ndp, totalp, cookie)
#endif

#undef	FN_NAME
#undef	SO_NAME
#undef	FN_NAME_APPEND
#undef	Bstr

struct syminfo const *q;
ssize_t qlen;
#if ! defined WAM_WSH
DWFILE *dxr_q, *dxr_qs, *dxr_ds;
#else
WAM *wam;
#endif
private ssize_t *ndp;
private ssize_t *totalp;
#if defined WAM_WSH
void *cookie;
#endif
{
	struct _g g0, *g=&g0;

#if ! defined SO_WSH
	if (weight_type>=SO_WT_BASE) {
		char name[32];
		char path[MAXPATHLEN];
		int soid = weight_type - SO_WT_BASE;
		char const *sodir = "/tmp";
#if ! defined WAM_WSH
		char const *prefix = "libdwsh-";
#else
		char const *prefix = "libwwsh-";
#endif
		void *dl;
		void *(*v)();
		snprintf(path, sizeof path, "%s/%s%08x.so", sodir, prefix, soid);
		snprintf(name, sizeof name, FN_NAME_STR "_%08x", soid);
		if (!(dl = dlopen(path, RTLD_LAZY|RTLD_GLOBAL))) {
			fprintf(stderr, "\n%s: %s\n", path, dlerror());
			return NULL;
		}
		if (!(v = dlsym(dl, name))) {
			fprintf(stderr, "\n%s: %s\n", name, dlerror());
			return NULL;
		}
		return v
#if ! defined WAM_WSH
		 (q, qlen, dxr_q, dxr_qs, dxr_ds, weight_type, ndp, totalp);
#else
		 (q, qlen, wam, dir_q, weight_type, ndp, totalp, cookie);
#endif

	}
#endif
#undef	FN_NAME_0
#undef	FN_NAME_STR
#undef	FN_NAME_SSTR
#undef	Cstr

	g->q = q;
	g->qlen = qlen;
#if ! defined WAM_WSH
	g->dxr.q = dxr_q;
	g->dxr.qs = dxr_qs;
	g->dxr.ds = dxr_ds;
#else
	g->wam = wam;
	g->dir.q = g->dir.qs = dir_q;
	g->dir.ds = WAM_REVERT_DIRECTION(dir_q);
#endif
	g->weight_type = weight_type;
	g->ndp = ndp;
	g->totalp = totalp;
#if defined WAM_WSH
	g->cookie = cookie;
#endif

	g->d = NULL;
	g->dh = NULL;
	g->qvec = NULL;
	g->qveclen = 0;
	g->qvec_h = NULL;
	g->a = NULL;
	g->handle = DXR_HANDLE(q);

#if ! defined WAM_WSH
	g->need_dup = DXR_NEED_DUP_P(q);
#endif
	g->bool.bex = NULL;
	g->bool.attrs = NULL;

	return wsh_toplevel(g);
}

static struct syminfo *
wsh_toplevel(g)
struct _g *g;
{
	size_t i;
	if (init_globals(g)==-1) {
		goto error;
	}
	if (p2(g)==-1) {
		goto error;
	}
	for (i = 0; i < g->qveclen; i++) {
		if (g->qvec[i].elem_num != 0) {
			break;
		}
	}
	if (g->qveclen == 0 || i == g->qveclen) {
		/* trivial case: empty result! */
		*g->ndp = 0;
		if (g->totalp) {
			*g->totalp = 0;
		}
		goto cleanup;
	}

	if (init_heap(g)==-1) {
		goto error;
	}

	if (p3(g)==-1) {
		goto error;
	}
cleanup:

	if (
#if ! defined WAM_WSH
		g->need_dup && 
#endif
		g->qvec) {
		size_t i;
		for (i=0; i<g->qveclen; i++) {
			if (g->qvec[i].vec != &zerovec) {
				free((void *)g->qvec[i].vec);
			}
		}
	}

	zfree(g->qvec_h);
	zfree(g->qvec);
	zfree(g->a);
	zfree(g->dh);
	bex_free(g->bool.bex), g->bool.bex = NULL;
	zfree(g->bool.attrs);

	return g->d;

error:
	zfree(g->d);
	g->d = NULL;
	goto cleanup;
}

static int
init_globals(g)
struct _g *g;
{
	ssize_t bex_start;
	if ((g->nd = *g->ndp)<=0) {
		fprintf(stderr, "wsh: invalid # of result requested (%s)\n", ssizetostr(g->nd));
		dwerrno = DWE_INVAL;
		return -1;
	}

	if (!(g->d = malloc((g->nd+1) * sizeof *g->d))) {
		fprintf(stderr, "wsh: Cannot alloc memory for g->d(%sbytes)\n", sizetostr((g->nd+1) * sizeof *g->d));
		dwerrno = errno;
		return -1;
	}

	if (g->qlen<=0) {
		fprintf(stderr, "wsh: qlen<=0\n");
		dwerrno = DWE_INVAL;
		return -1;
	}

	g->bool.type = bool_type(g->q, g->qlen, &bex_start);
	g->bool.num_and_q = g->bool.num_not_q = g->bool.num_org_q = 0;
	if (g->bool.type==BOOL_BEX) {
		g->bool.bex_len = g->qlen - bex_start;
		g->qlen = bex_start;
		if (!(g->bool.attrs = calloc(g->bool.bex_len, sizeof *g->bool.attrs))) {
			perror("calloc g->bool.bex_len");
			return -1;
		}
	}
	else {
		g->bool.bex_len = 0;
	}

	g->total_N_q = DXR_SIZE(qs);
	g->total_N_d = DXR_SIZE(ds);	/* # of different documents/words in the corpus */
	g->total_TF_q = DXR_TOTAL_ELEM_NUM(qs);
	g->total_TF_d = DXR_TOTAL_ELEM_NUM(ds);

	if (!(g->qvec = malloc((g->qlen + g->bool.bex_len) * sizeof *g->qvec))) {
		fprintf(stderr, "wsh: Cannot alloc memory for g->qvec\n");
		dwerrno = errno;
		return -1;
	}
	if (!(g->qvec_h = malloc((g->qlen + g->bool.bex_len) * sizeof *g->qvec_h))) {
		fprintf(stderr, "wsh: Cannot alloc memory for g->qvec_h\n");
		dwerrno = errno;
		return -1;
	}

	{
#if ! defined SO_WSH
#include "wt/s1.c"
#else
#include "sowt/s1.c"
#endif
	}
	return 0;
}

static int
bool_type(q, len, bex_start)
struct syminfo const *q;
ssize_t len, *bex_start;
{
	ssize_t i;
	int pseudo = 0;
	for (i = 0; i<len; i++) {
		struct syminfo const *qp;
		qp = q+i;
		if (qp->attr==WSH_AND || qp->attr==WSH_NOT || qp->attr==WSH_ORG) {
			pseudo = 1;
		}
		else if (qp->attr==BEX_START) {
			*bex_start = i;
			return BOOL_BEX;
		}
		else if (qp->attr==WSH_OR) {
			/* do nothing */
		}
		else {
			return BOOL_OFF;
		}
	}
	if (pseudo) {
		return BOOL_PSEUDO;
	}
	return BOOL_OFF;
}

/* lookup all words/documents in q */
static int
p2(g)
struct _g *g;
{
	struct syminfo const *qp;
	private	struct syminfo const *qe;
	struct qvec *qvecp;

	for (g->TF_q = 0, qp=g->q, qe=g->q + g->qlen + g->bool.bex_len, qvecp=g->qvec; qp<qe; qp++) {
		double wq;
		{
			ssize_t e;
			struct xr_vec const *v;
			size_t qpoff = qp-g->q;
			if (g->bool.type == BOOL_BEX && qpoff>=g->qlen && qp->id==0) {
				e = 0;
			}
#if ! defined WAM_WSH
			else if (g->need_dup) {
				struct xr_vec const *u;
				e = DXR_GET_VEC_DUP(q, qp->id, &u);
				v = u;
			}
#endif
			else {
				e = DXR_GET_VEC(q, qp->id, &v);
			}
#if defined WAM_WSH
			if (e>0 && !(v = dxr_dup(v))) {
				e = -1;
			}
#endif
			if (e<=0) {
				qvecp->vec = &zerovec;
				e = 1;
			}
			else {
				qvecp->vec = v;
			}
			g->qveclen++;	/* qvecp->vec wo set sitara, qveclen wo suguni increment siteoku */
			assert(e==qvecp->vec->elem_num);
		}
		switch (g->bool.type) {
			size_t qvoff;
		case BOOL_PSEUDO:
			if (qp->attr==WSH_AND) {
				g->bool.num_and_q++;
			}
			if (qp->attr==WSH_NOT) {
				g->bool.num_not_q++;
			}
			if (qp->attr==WSH_ORG) {
				g->bool.num_org_q++;
			}
			break;
		case BOOL_BEX:
			qvoff = qvecp-g->qvec;
			if (qvoff >= g->qlen) {
				assert(qvoff < g->qlen + g->bool.bex_len);
				g->bool.attrs[qvoff - g->qlen] = qp->attr;
			}
			break;
		}
		if (qvecp->vec!=&zerovec) {
			qvecp->elem_num = DXR_ELEM_NUM(qs, qp->id);
			qvecp->freq_sum = DXR_FREQ_SUM(qs, qp->id);
		}
		else {
			qvecp->elem_num = 0;
			qvecp->freq_sum = 0;
		}
		g->TF_q += qvecp->freq_sum;
		qvecp->elemp1 = qvecp->vec->elems;
		qvecp->elemp1id = qvecp->elemp1->id;
		qvecp->elemendp = qvecp->vec->elems + qvecp->vec->elem_num;
		wq = 0.0;
		if (qvecp->vec!=&zerovec) {
#if ! defined SO_WSH
#include "wt/s2.c"
#else
#include "sowt/s2.c"
#endif
		}
		qvecp->wq = wq;
		qvecp->q = qp;
		qvecp++;	/* NOTE: qvecp to qveclen ha susumu timing ga sukosi tigau. */
	}
	assert(qvecp-g->qvec==g->qveclen);
	assert(g->qveclen == g->qlen + g->bool.bex_len);
	if (g->bool.type==BOOL_BEX) {
		if (!(g->bool.bex = bex_compile(g->bool.attrs, g->bool.bex_len))) {
			return -1;
		}
	}
	return 0;
}

static int
init_heap(g)
struct _g *g;
{
	size_t i, l;

	if (!(g->a = malloc((g->qveclen+1) * sizeof *g->a))) {
			/* (worst) heap size is qveclen+1 */
		fprintf(stderr, "wsh: Cannot alloc memory for a\n");
		dwerrno = errno;
		return -1;
	}

	for (i=0, l=1; i<g->qveclen; i++) {
		if (g->qvec[i].vec != &zerovec) {
			g->a[l++] = &g->qvec[i];
		}
	}
	g->na = l-1;	/* initial heap size */

	for (l=g->na/2; l>0; l--) {
		a_reheap(g->a, l, g->na);
	}
	return 0;
}

/* phase-3 loop */ 
static int
p3(g)
struct _g *g;
{
	private	int finish;
	struct syminfo *dp;

	double M_LN2i;
	double log_total_N_q;

	log_total_N_q = log((double)g->total_N_q);
	M_LN2i = 1/M_LN2;

	for (dp=g->d, g->nnd=0, finish=0, g->total=0; !finish; ) {
		idx_t id;
		size_t qvec_h_len;
		size_t num_and_d, num_not_d, num_org_d;

		if (g->bool.type==BOOL_BEX) {
			bex_clear(g->bool.bex);
		}

		if ((id = p3x(g, &qvec_h_len, &num_and_d, &num_not_d, &num_org_d, &finish)) == 0) {
			return -1;
		}

		switch (g->bool.type) {
		case BOOL_PSEUDO:
			if (g->bool.num_and_q != num_and_d || num_not_d) {
				goto s5_break;
			}
			if (g->bool.num_org_q && !num_org_d) {
				goto s5_break;
			}
			break;
		case BOOL_BEX:
			if (!bex_check(g->bool.bex)) {
				goto s5_break;
			}
			break;
		}

		/* add a document/word to `g->d[*].id' */
		{
#if ! defined SO_WSH
#include "wt/s5.c"
#else
#include "sowt/s5.c"
#endif
		}

		dp->id = id;
		dp->DF_d = qvec_h_len;
		dp->DF = DXR_ELEM_NUM(ds, id);
		dp->TF = DXR_FREQ_SUM(ds, id);
		calc_sigma(g, dp, qvec_h_len);
		normalize(g, dp);
		g->total++;
		if (addoc(g, dp)) {
			dp++;
		}

	s5_break:;	/* eject label for s5.c */
	}
	*g->ndp = g->nnd;
	if (g->totalp) {
		*g->totalp = g->total;
	}
	return 0;
}

static idx_t
p3x(g, qvec_h_len_p, num_and_d_p, num_not_d_p, num_org_d_p, finish_p)
struct _g *g;
size_t *qvec_h_len_p, *num_and_d_p, *num_not_d_p, *num_org_d_p;
int *finish_p;
{
	struct qvec *qvecp;
	idx_t id;

	size_t qvec_h_len;
	size_t num_and_d, num_not_d, num_org_d;

	qvecp = g->a[1];

	num_and_d = num_not_d = num_org_d = 0;
	qvec_h_len = 0;

	assert(qvecp->elemp1id!=0);
	for (id = qvecp->elemp1id; ; ) {
		size_t qvoff = qvecp-g->qvec;
		if (g->bool.type == BOOL_BEX && qvoff>=g->qlen) {
			bex_setnth(g->bool.bex, qvoff-g->qlen);
		}
		else {
			g->qvec_h[qvec_h_len++] = qvecp;
		}
		if (g->bool.type == BOOL_PSEUDO) {
			switch (qvecp->q->attr) {
			case WSH_OR:
				break;
			case WSH_AND:
				num_and_d++;
				break;
			case WSH_NOT:
				num_not_d++;
				break;
			case WSH_ORG:
				num_org_d++;
				break;
			default:
				fprintf(stderr, "wsh: invalid q->attr(%d)\n", qvecp->q->attr);
				dwerrno = DWE_INVAL;
				return 0;
			}
		}

		qvecp->elemp = qvecp->elemp1++;
		if (qvecp->elemendp<=qvecp->elemp1) {
			/* exhausted */
			if (g->na==1) {
				/* all elemp1 were exhausted */
				*finish_p = 1;
				break;
			}
			g->a[1] = g->a[g->na--];
		}
		else {
			qvecp->elemp1id = qvecp->elemp1->id;
		}
		a_reheap(g->a, 1, g->na);
		if ((qvecp = g->a[1])->elemp1id != id) {
			break;
		}
	}

	assert(qvec_h_len <= g->qveclen);
	assert(id!=0);
	*qvec_h_len_p = qvec_h_len;
	*num_and_d_p = num_and_d;
	*num_not_d_p = num_not_d;
	*num_org_d_p = num_org_d;
	return id;
}

static int
calc_sigma(g, dp, qvec_h_len)
struct _g *g;
struct syminfo *dp;
size_t qvec_h_len;
{
	private	register freq_t dp_TF_d;
	private	register double dp_weight;
	private int i;

	dp_TF_d = 0;
	dp_weight = 0.0;
	for (i=0; i<qvec_h_len; i++) {
		struct qvec *qvecp;
		qvecp = g->qvec_h[i];
		dp_TF_d += qvecp->elemp->freq;

		if (g->weight_type != WT_NONE) {
			double wd = 0.0;
			{
#if ! defined SO_WSH
#include "wt/s3.c"
#else
#include "sowt/s3.c"
#endif
			}
			dp_weight += qvecp->wq * wd;
		}
	}
	dp->TF_d = dp_TF_d;
	dp->weight = dp_weight;
	dp->attr = WSH_NONE;
	return 0;
}

static int
normalize(g, dp)
struct _g *g;
struct syminfo *dp;
{
	double p_DF, p_TF, p_DF_d, p_TF_d;
	double norm;

	p_DF = dp->DF;
	p_TF = dp->TF;
	p_TF_d = dp->TF_d;
	p_DF_d = dp->DF_d;
	norm = 1.0;
	{
#if ! defined SO_WSH
#include "wt/s4.c"
#else
#include "sowt/s4.c"
#endif
	}
	dp->weight /= norm;
	return 0;
}

static int
addoc(g, dp)
struct _g *g;
struct syminfo *dp;
{
	if (g->nnd==g->nd) {
		struct syminfo *sp;
		/* dp ha &g->d[g->nd](i.e. temporary ryouiki) wo sasite iru hazu */
		assert(dp==&g->d[g->nd]);
		sp = g->dh[1];		/* g->d[0..g->nd-1] no uti, weight ga saisyou no mono */

		if (dcompar(dp, sp)<0) {
			/* *sp wo sute, kono youso wo heap ni sounyu. */
			*sp = *dp;
			/* heap wo tukuri naosu */
			d_reheap(g->dh, 1, g->nd);
		}
		/* else nani mo sinai */
		return 0;
	}
	else {	/* tanzyuun ni tuika */
		g->nnd++;
		if (g->nnd==g->nd) {	/* tyoudo ippai ni natta */
			int l;
			/* dp ha &g->d[g->nd] wo sasite iru. kore ikou, dp ha ugokasanai */

			if (!(g->dh = malloc((g->nd+1) * sizeof *g->dh))) {
				fprintf(stderr, "wsh: Cannot alloc memory for g->dh(%sbytes)\n", sizetostr((g->nd+1) * sizeof *g->dh));
				dwerrno = errno;
				return -1;
			}
			g->dh[0] = NULL;
			for (l=1; l<=g->nd; l++) {
				g->dh[l] = &g->d[l-1];
			}
			for (l=g->nd/2; l>0; l--) {
				d_reheap(g->dh, l, g->nd);
			}
		}
		return 1;	/* dp++; */
	}
}
