/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: wstem.c,v 1.5 2007/03/24 02:52:58 nis Exp $"; */

#include <config.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dxr.h>
#include <geta/dwsh.h>
#include <geta/dwam.h>

/*extern char **environ;*/

struct syminfo *
wstem(text, w, dir, path, ndp)
char const *text;
WAM *w;
char const *path;
ssize_t *ndp;
{
	char *argv[3];

	argv[0] = (char *)path;
	argv[1] = "-1";
	argv[2] = NULL;
	return wstemv(text, w, dir, path, argv, ndp);
}

struct syminfo *
wstemv(text, w, dir, path, argv, ndp)
char const *text;
WAM *w;
char const *path;
char * const argv[];
ssize_t *ndp;
{
	struct itimerval it = {{0, 0}, {0, 0}}, ot;

	int status;
	pid_t cpid, wpid;
	int e;
/*
 *	wstem(writer) -> fds0[1] == fds0[0] =dup2=> 0 -> path (child)
 *	wstem(parent) <- fds1[0] == fds1[1] =dup2=> 1 <- path
 */
	int fds0[2];
	int fds1[2];
	FILE *f;

	size_t i, j, k;
	struct syminfo *p, *r;
	char *buf;
	size_t l, buflen;
	size_t rmax;

	r = NULL;
	buf = NULL;
	f = NULL;
	fds0[0] = fds0[1] = fds1[0] = fds1[1] = -1;

	l = strlen(text);
	rmax = l * 4;
	buflen = l + 513;

	if (!(buf = malloc(buflen))) {
		nperror("malloc buf");
		goto error;
	}
	if (!(r = calloc(rmax, sizeof *r))) {
		nperror("calloc r");
		goto error;
	}
	if (pipe(fds0)==-1) {
		nperror("pipe");
		goto error;
	}
	if (pipe(fds1)==-1) {
		nperror("pipe");
		goto error;
	}

	switch (cpid = fork()) {
	case -1:
		goto error;
	case 0:		/* child */
		setitimer(ITIMER_VIRTUAL, &it, &ot);

		close(fds1[0]), fds1[0] = -1;
		close(fds0[1]), fds0[1] = -1;
		if (fds0[0]!=0) {
			dup2(fds0[0], 0);
			close(fds0[0]), fds0[0] = -1;
		}
		if (fds1[1]!=1) {
			dup2(fds1[1], 1);
			close(fds1[1]), fds1[1] = -1;
		}
		execv(path, argv/*, environ*/);
		nperror(path);
		_exit(1);
		break;
	default:	/* parent */
		close(fds0[0]), fds0[0] = -1;
		close(fds1[1]), fds1[1] = -1;
		break;
	}

	switch (wpid = fork()) {
	case -1:
		goto error;
	case 0:		/* writer */
		setitimer(ITIMER_VIRTUAL, &it, &ot);
		if (write(fds0[1], text, l)!=l) {
			goto error;
		}
		close(fds0[1]), fds0[1] = -1;
		close(fds1[0]), fds1[0] = -1;
		_exit(0);
		break;
	default:	/* parent */
		close(fds0[1]), fds0[1] = -1;
		break;
	}

	if (!(f = fdopen(fds1[0], "r"))) {
		goto error;
	}
	fds1[0] = -1;

	for (i=0; fgets(buf, buflen, f); ) {
		idx_t id;
		char *q;
		if (q = index(buf, '\n')) {
			*q = '\0';
		}
		if ((id = wam_name2id(w, dir, buf))) {
			if (i>=rmax) {
				goto error;
			}
			r[i++].id = id;
		}
	}
	fclose(f), f = NULL;

	qsort(r, i, sizeof *r, sym_icompar);

	for (j=0, p=r; j<i; j=k, p++) {
		for (k=j+1; r[j].id == r[k].id; k++) ;
		p->id = r[j].id;
		p->TF = wam_freq_sum(w, dir, p->id);
		p->TF_d = k-j;
		p->weight = 0.0;
		p->attr = WSH_OR;
	}
	i = p-r;

	if ((e = wait4(cpid, &status, 0, NULL))==-1) {
/*
		fprintf(stderr, "wait4 C=%d: %s\n", (int)cpid, strerror(errno));
*/
	}
	if ((e = wait4(wpid, &status, 0, NULL))==-1) {
/*
		fprintf(stderr, "wait4 W=%d: %s\n", (int)wpid, strerror(errno));
*/
	}

	*ndp = i;

cleanup:
	zfree(buf);
	if (fds0[0] != -1) {
		close(fds0[0]);
	}
	if (fds0[1] != -1) {
		close(fds0[1]);
	}
	if (fds1[0] != -1) {
		close(fds1[0]);
	}
	if (fds1[1] != -1) {
		close(fds1[1]);
	}
	if (f) {
		fclose(f);
	}

	return r;

error:
	zfree(r);
	r = NULL;	/* important! */
	goto cleanup;
}
