/*
 *	Copyright (C) 1999 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: f2c.c,v 1.5 2007/03/24 02:52:58 nis Exp $"; */

#include <config.h>

#include <sys/param.h>

#include <ctype.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dwut.h>

char *progname;
static char verbose = 0;

static void usage(void);
static int process_a_file(char *, int, FILE *);
static void process_a_portion(int, char *, char *[], FILE *, char *, size_t);
static char *trans_line(char *, int);

static void
usage()
{
	fprintf(stderr, "usage: %s [-v] file...\n", progname);
	exit(1);
}

int
main(argc, argv)
char *argv[];
{
	int step;

	if (progname = rindex(argv[0], '/')) {
		progname++;
	}
	else {
		progname = argv[0];
	}
	for (argc--, argv++; argc>0; argc--, argv++) {
		if (!strcmp(*argv, "-h")) {
			usage();
			exit(1);
		}
		else if (!strcmp(*argv, "-v")) {
			verbose = 1;
		}
		else {
			break;
		}
	}
	if (argc==0) {
		usage();
	}
	for (step = 0; step <= 5; step++) {
		char **files;
		size_t nfiles;
		FILE *g;
		char oname[MAXPATHLEN];

		snprintf(oname, sizeof oname, "s%d.c", step);
		if (!(g = fopen(oname, "w"))) {
			perror(oname);
			exit(1);
		}
		if (step!=0) {
			fprintf(g, "\t/* Step %d */\n", step);
			fprintf(g, "\tswitch (g->weight_type) {\n");
			fprintf(g, "\tcase WT_NONE: {\n");
			fprintf(g, "\t\t} break;\n\n");
		}
		for (nfiles=argc, files=argv; nfiles>0; nfiles--, files++) {
			if (process_a_file(*files, step, g)==-1) {
				exit(1);
			}
		}
		if (step!=0) {
			fprintf(g, "\tdefault:\n");
			fprintf(g, "\t\tfprintf(stderr, \"step%d: undefined weight type: %%d\\n\", g->weight_type);\n", step);
			fprintf(g, "\t\tbreak;\n");
			fprintf(g, "\t}\n");
		}
		fclose(g);
	}
	
	exit(0);
}

static int
process_a_file(path, step, stream)
char *path;
FILE *stream;
{
	int found;
	size_t l;
	FILE *f;
	char buf[1024];
#define	L_NAME	"name: "
	char *typename;
	char step_name[32];
	char *lines[256];
	size_t i;
	size_t orig_line;

	if (!(f = fopen(path, "r"))) {
		perror(path);
		return -1;
	}

	orig_line = 0;

	for (typename = NULL, l = strlen(L_NAME); fgets(buf, sizeof buf, f); ) {
		orig_line++;
		if (!strncasecmp(buf, L_NAME, l)) {
			char *p;
			if (p = rindex(buf, '\n')) {
				*p = '\0';
			}
			if (!(typename = strdup(buf + l))) {
				perror("strdup typename");
			}
if (verbose) {
	fprintf(stderr, "typename: %s\n", typename);
}
			break;
		}
	}
	if (!typename) {
		fprintf(stderr, "cannot find for \"%s\" in [%s]\n", L_NAME, path);
		return -1;
	}

	snprintf(step_name, sizeof step_name, "%d:", step);
	l = strlen(step_name);
	for (found = 0; fgets(buf, sizeof buf, f); ) {
		orig_line++;
		if (!strncasecmp(buf, step_name, l)) {
			found = 1;
			break;
		}
	}
	if (!found) {
		fprintf(stderr, "cannot find for \"%s\" in [%s]\n", step_name, path);
		return -1;
	}

	orig_line++;
if (verbose) {
	fprintf(stderr, "file %s, portion %s\n", path, step_name);
}
	for (i = 0; i<nelems(lines)-1; ) {
		if (!fgets(buf, sizeof buf, f)) {
			break;
		}

		if (isdigit(*buf&0xff) && *(buf+1)==':') {
			break;
		}
if (verbose) {
	fprintf(stderr, "file %s, line %s\n", path, buf);
}

		if (!(lines[i++] = strdup(buf))) {
			perror("strdup lines[i]");
			return -1;
		}

	}
	lines[i] = NULL;

	process_a_portion(step, typename, lines, stream, path, orig_line);

	for (i=0; lines[i]; i++) {
		free(lines[i]);
	}
	free(typename);

	fclose(f);
	return 0;
}

static void
process_a_portion(step, typename, lines, stream, orig_name, orig_line)
char *typename, *lines[];
FILE *stream;
char *orig_name;
size_t orig_line;
{
	size_t i;

	if (step!=0) {
		fprintf(stream, "\tcase %s: {\n", typename);
	}
	fprintf(stream, "#line %d \"%s\"\n", (int)orig_line, orig_name);
	for (i=0; lines[i]; i++) {
		char *p;
		p = trans_line(lines[i], step);
		fprintf(stream, "\t%s", p);
	}
	if (step!=0) {
		fprintf(stream, "\t\t} break;\n\n");
	}
}

struct rr {
	char *from, *to;
};

struct rr rr2[] = {
	{"TF(t|q)", "((double)qp->TF_d)"},
	{"TF_s(t)", "((double)qp->TF)"},
	{"DF(t)", "((double)qvecp->elem_num)"},
	{"TF(t)", "((double)qvecp->freq_sum)"},
	{"idf(t)", "(log1p(g->total_N_d / (double)qvecp->elem_num))"},
	{"wq(t|q)", "wq"},	/* lvalue */
	{NULL, NULL}};

struct rr rr3[] = {
	{"TF(t|d)", "((double)qvecp->elemp->freq)"},
	{"TF(.|d)", "((double)dp->TF)"},
	{"DF(d|q)", "((double)dp->DF_d)"},
	{"DF(.|d)", "((double)dp->DF)"},
	{"wq(t|q)", "(qvecp->wq)"},
	{"wd(t|d)", "wd"},	/* lvalue */
	{NULL, NULL}};

struct rr rr4[] = {
	{"TF(.|d)", "(p_TF)"},
	{"DF(.|d)", "(p_DF)"},
	{"norm(d)", "norm"},	/* lvalue */
	{NULL, NULL}};

struct rr *rrs[] = {
	NULL,
	NULL,
	rr2,
	rr3,
	rr4,
	NULL
};

static char *
trans_line(l, step)
char *l;
{
	struct rr *rp, *r;
	static char buf[1024], work[1024];

	if (0<=step && step<=5) {
		rp = rrs[step];
	}
	else {
		fprintf(stderr, "f2c: internal error\n");
		exit(1);
	}

	if (!rp) {
		return l;
	}

	strlcpy(buf, l, sizeof buf);
again:
	for (r=rp; r->from; r++) {
		char *s, *e;
		if (s = strstr(buf, r->from)) {
			size_t buf_len, from_len, to_len;
			buf_len = strlen(buf);
			from_len = strlen(r->from);
			to_len = strlen(r->to);
if (verbose) {
	fprintf(stderr, "<%s", buf);
}
			e = s + from_len;
			bcopy(buf, work, s-buf);
			bcopy(r->to, work+(s-buf), to_len);
			bcopy(e, work+(s-buf)+to_len, buf_len-(e-buf)+1);
			strcpy(buf, work);
if (verbose) {
	fprintf(stderr, ">%s", buf);
}

			goto again;
		}
	}

	return buf;
}

#if ! HAVE_STRLCPY
size_t
strlcpy(dst, src, size)
char *dst;
const char *src;
size_t size;
{
	size_t l, c;
	c = l = strlen(src);
	if (size==0) {
		return l;
	}
	if (c > size-1) {
		c = size-1;
	}
	memcpy(dst, src, c);
	dst[c] = '\0';
	return l;
}
#endif
