/*
 *	Copyright (C) 2001 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: xcgen.c,v 1.3 2007/03/29 07:37:38 nis Exp $"; */

#include <config.h>

#include <sys/param.h>

#include <assert.h>
#include <stdio.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>

struct foo {
	char *name;
	size_t dbits[32];
};

static int prtbl(struct foo *);
static int prinfo(struct foo *);
static int encode(struct foo *);
static int decode(struct foo *);

#define	nelems(e)	(sizeof (e)/sizeof *(e))

#if BITLIST_BPW == 8
typedef uint64_t val_t;
#define	valtostr	ulltostr
#define	valtoxstr	ulltoxstr
#else
typedef uint32_t val_t;
#define	valtostr	ultostr
#define	valtoxstr	ultoxstr
#endif

main()
{
	size_t i;
	struct foo foos[] = {
#if BITLIST_BPW == 8
		{"Z36", {0, 1, 1, 1,  2,  3,  6, 32, 64, -1}},
		{"NIB", {10, 12, 13, 14, 15, 15, 17, 32, 64, -1}},
		{"B36", {3, 9, 15, 21, 27, 32, 64, -1}},
#else
		{"Z36", {0, 1, 1, 1,  2,  3,  6, 32, -1}},
		{"NIB", {10, 12, 13, 14, 15, 15, 17, 32, -1}},
		{"B36", {3, 9, 15, 21, 27, 32, -1}},
#endif
/*
 * {"Z36", {0, 3, 9, 15, 21, 27, 32, 64, -1}},
 * {"NIB", {7, 13, 16, 19, 22, 25, 28, 32, 64, -1}},
 * {"NIB", {15, 15, 15, 15, 15, 15, 15, 32, 64, -1}},
 */
	};

	for (i=0; i<nelems(foos); i++) {
		prtbl(&foos[i]);
	}
	return 0;
}

static int
prtbl(z)
struct foo *z;
{
	prinfo(z);
	encode(z);
	decode(z);
	return 0;
}

static int
prinfo(z)
struct foo *z;
{
	size_t db;
	size_t i;
	val_t v = 1;
	char s[81];

	printf("/* %s\n", z->name);
	for (i=0; (db = z->dbits[i])!=-1; i++) {
		int lastp = z->dbits[i+1]==-1;
		size_t j;
		val_t n, l;
		size_t ib = i+1;
		for (n=1, j=0; j<db; j++, n*=2) ;
		l = v+n;

		if (lastp) {
			snprintf(s, sizeof s, "0x%s--", valtoxstr(v));
		}
		else {
			if (db>0) {
				snprintf(s, sizeof s, "0x%s--0x%s", valtoxstr(v), ultoxstr(l-1));
			}
			else {
				snprintf(s, sizeof s, "0x1");
			}
		}
		printf(" * %d: %-24s %d, %d, %d\n", (int)i, s, (int)ib, (int)db, (int)(ib+db));
		v = l;
	}
	printf(" */\n");
	return 0;
}

static int
encode(z)
struct foo *z;
{
	size_t db;
	size_t i;
	val_t v = 1;

	printf("#define	ASSIGN_%s(D) do {\\\n", z->name);
	printf("\tregister bitlist_reg_t v_v = (D);\\\n");
	for (i=0; (db = z->dbits[i])!=-1; i++) {
		int lastp = z->dbits[i+1]==-1;
		int firstp = i==0;
		size_t j;
		val_t n, l;
		size_t ib = i+1;
		for (n=1, j=0; j<db; j++, n*=2) ;
		l = v+n;

		if (lastp) {
			assert(!firstp);
			assert(db>0);
			assert(ib>0);
			printf("\telse {\\\n");
			printf("\t\tBITLIST_PUTBITS(~(bitlist_reg_t)0, %d, &b);\\\n", (int)(ib-1));
			printf("\t\tv_v -= 0x%s;\\\n", valtoxstr(v));
			printf("\t\tBITLIST_PUTBITS(v_v, %d, &b);\\\n", (int)db);
			printf("\t}\\\n");
		}
		else {
			if (firstp) {
				printf("\t");
			}
			else {
				printf("\telse ");
			}
			printf("if (v_v < 0x%s) {\\\n", ultoxstr(l));
			printf("\t\tBITLIST_PUTBITS(~(bitlist_reg_t)1, %d, &b);\\\n", (int)ib);

			if (db>0) {
				printf("\t\tv_v -= 0x%s;\\\n", valtoxstr(v));
				printf("\t\tBITLIST_PUTBITS(v_v, %d, &b);\\\n", (int)db);
			}
			else {
				assert(v==1);
			}
			printf("\t}\\\n");

		}
		v = l;
	}
	printf("} while (0)\n");
	return 0;
}

static int
decode(z)
struct foo *z;
{
	size_t db;
	size_t i;
	val_t v = 1;

	printf("#define	RETRIEVE_%s(D) do {\\\n", z->name);
	printf("\tregister bitlist_reg_t v_v = 0;\\\n");
	for (i=0; (db = z->dbits[i])!=-1; i++) {
		int lastp = z->dbits[i+1]==-1;
		int firstp = i==0;
		size_t j;
		val_t n, l;
		for (n=1, j=0; j<db; j++, n*=2) ;
		l = v+n;

		if (lastp) {
			assert(!firstp);
			/* assert(db>0); */
			/* assert(ib>0); */
			printf("\tBITLIST_GETBITS(v_v, %d, &b);\\\n", (int)db);
			printf("\t(D) = v_v + 0x%s;\\\n", valtoxstr(v));
		}
		else {
			printf("\tBITLIST_GETBITS(v_v, 1, &b);\\\n");
			printf("\tif (v_v==0) { \\\n");
			if (db>0) {
				printf("\t\tBITLIST_GETBITS(v_v, %d, &b);\\\n", (int)db);
				printf("\t\t(D) = v_v + 0x%s;\\\n", valtoxstr(v));
			}
			else {
				printf("\t\t(D) = 1;\\\n");
			}
			printf("\t\tbreak;\\\n");
			printf("\t}\\\n");

		}
		v = l;
	}
	printf("} while (0)\n");
	return 0;
}
