/*
 *	Copyright (C) 1997 Advanced Research Laboratory, Hitachi, Ltd.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: xcompress.c,v 1.5 2007/03/26 07:32:14 nis Exp $"; */

/* comress/uncompress */

#if defined NEED_COMPRESS || defined NEED_UNCOMPRESS

#include "bitlist.h"

#define	CB_WORSTSIZE(elem_num)	((elem_num)*12)

#endif

#if BITLIST_BPW == 8
#include "z36-8.c"
#else
#include "z36-4.c"
#endif

#if defined NEED_COMPRESS

#define	INIT_ASSIGN(P0, L0)	BITLIST_INIT_PUTBITS(&b, (P0), (L0))

#define	FINISH_ASSIGN()		BITLIST_FINISH_PUTBITS(&b)

static ssize_t dxr_compress_vec(struct xr_vec const *, struct dxr_zvec *, size_t, int);

static ssize_t
dxr_compress_vec(vec, zvec, size, mode)
struct xr_vec const *vec;
struct dxr_zvec *zvec;
size_t size;
{
	struct _bitlist_t b;
	struct xr_elem const *q;
	size_t i;
	u_char *l;
	ssize_t elem_num;

	elem_num = vec->elem_num;
	q = vec->elems;

	l = (u_char *)zvec->zelems + size;
	INIT_ASSIGN(zvec->zelems, l);

	switch (mode) {
	case COMPRESS_NIB:
		for (i=0; i<elem_num; i++, q++) {
			ASSIGN_NIB(q->id);
			ASSIGN_NIB(q->freq);
		}
		break;
	case COMPRESS_B36:
		for (i=0; i<elem_num; i++, q++) {
			ASSIGN_B36(q->id);
			ASSIGN_B36(q->freq);
		}
		break;
	case COMPRESS_Z36:
		for (i=0; i<elem_num; i++, q++) {
			ASSIGN_NIB(q->id);
			ASSIGN_Z36(q->freq);
		}
		break;
	default:
		dwerrno = DWE_BADR;
		return -1;
	}
	FINISH_ASSIGN();
	if (b.p>l) {
		dwerrno = DWE_LIMIT;
		return -1;
	}
	return b.p - (u_char *)zvec;
}

#endif

#undef INIT_ASSIGN
#undef FINISH_ASSIGN

#if defined NEED_UNCOMPRESS

#define	INIT_RETRIEVE(P0)	BITLIST_INIT_GETBITS(&b, (P0))

static int dxr_uncompress_vec(struct xr_vec *, struct dxr_zvec const *, int, struct dfilt *);

static int
dxr_uncompress_vec(vec, zvec, mode, f)
struct xr_vec *vec;
struct dxr_zvec const *zvec;
struct dfilt *f;
{
	struct _bitlist_t b;
	struct xr_elem *q;
	size_t i;
	ssize_t elem_num;

	elem_num = vec->elem_num;
	q = vec->elems;

	INIT_RETRIEVE(zvec->zelems);

	switch (mode) {
	case COMPRESS_NIB:
		if (f) {
			freq_t freq_sum = 0;
			for (i=0; i<elem_num; i++) {
				RETRIEVE_NIB(q->id);
				RETRIEVE_NIB(q->freq);
				if (f && (*f->screened)(f->cookie, q->id)) {
					continue;
				}
				freq_sum += q->freq;
				q++;
			}
			vec->elem_num = q - vec->elems;
			vec->freq_sum = freq_sum;
		}
		else {
			for (i=0; i<elem_num; i++) {
				RETRIEVE_NIB(q->id);
				RETRIEVE_NIB(q->freq);
				q++;
			}
		}
		break;
	case COMPRESS_B36:
		if (f) {
			freq_t freq_sum = 0;
			for (i=0; i<elem_num; i++) {
				RETRIEVE_B36(q->id);
				RETRIEVE_B36(q->freq);
				if (f && (*f->screened)(f->cookie, q->id)) {
					continue;
				}
				freq_sum += q->freq;
				q++;
			}
			vec->elem_num = q - vec->elems;
			vec->freq_sum = freq_sum;
		}
		else {
			for (i=0; i<elem_num; i++) {
				RETRIEVE_B36(q->id);
				RETRIEVE_B36(q->freq);
				q++;
			}
		}
		break;
	case COMPRESS_Z36:
		if (f) {
			freq_t freq_sum = 0;
			for (i=0; i<elem_num; i++) {
				RETRIEVE_NIB(q->id);
				RETRIEVE_Z36(q->freq);
				if (f && (*f->screened)(f->cookie, q->id)) {
					continue;
				}
				freq_sum += q->freq;
				q++;
			}
			vec->elem_num = q - vec->elems;
			vec->freq_sum = freq_sum;
		}
		else {
			for (i=0; i<elem_num; i++) {
				RETRIEVE_NIB(q->id);
				RETRIEVE_Z36(q->freq);
				q++;
			}
		}
		break;
	default:
		dwerrno = DWE_BADR;
		return -1;
	}
	return 0;
}

#undef INIT_RETRIEVE

#endif

#undef ASSIGN_B36
#undef RETRIEVE_B36

#undef ASSIGN_NIB
#undef RETRIEVE_NIB

#undef ASSIGN_Z36
#undef RETRIEVE_Z36
