/*
 *      Copyright (c) 2003 Shingo Nishioka.  All rights reserved.
 */

#if 0
static char rcsid[] = "$Id: btest.c,v 1.2 2006/10/20 07:57:00 nis Exp $";
#endif

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/uio.h>

#include <assert.h>
#include <errno.h>
#include <libgen.h>
#include <math.h>
#include <netdb.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/nlog.h>
#include <geta/dwut.h>
#include <geta/timer.h>
#include <geta/dw.h>
#include <geta/dci.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/daui.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwsched.h>
#include <geta/dwam.h>

static int btest(char const *, char const *, FILE *);
static struct syminfo *setq(int *);
static struct syminfo *sete(WAM *, int *);
static void print_syminfo(struct syminfo const *, int, WAM *, int, FILE *);

char *progname = NULL;

main(argc, argv)
char *argv[];
{
	char const *dirname, *handle;

	progname = basename(argv[0]);

	if (argc != 3) {
		fprintf(stderr, "usage: %s dir handle\n", progname);
		return 1;
	}

	dirname = argv[1];
	handle = argv[2];

	if (btest(dirname, handle, stdout)==-1) {
		return 1;
	}

	return 0;
}

btest(dirname, handle, stream)
char const *dirname, *handle;
FILE *stream;
{
	int dir, type, nq, nd, ne, total;
	WAM *w;
	struct syminfo *q, *d, *e;

	if (wam_init(dirname)==-1) {
		return -1;
	}

	if (!(w = wam_open(handle))) {
		nperror("wam_open");
		return -1;
	}

	q = setq(&nq);
nq = 0;
	e = sete(w, &ne);

	dir = WAM_COL;
	type = WT_SMARTWA;
	nd = wam_size(w, WAM_ROW);
	d = bex_wsh(q, nq, w, dir, type, &nd, &total, e, ne);
	print_syminfo(d, nd, w, WAM_ROW, stream);
	free(q);
	free(d);
	free(e);

	wam_close(w);
	return 0;
}

struct query {
	int id, tf;
};

static struct syminfo *
setq(nq)
int *nq;
{
	struct query o[] = {{1, 1},};
	struct syminfo *q;
	int i;
	*nq = nelems(o);
	q = calloc(*nq, sizeof *q);
	for (i=0; i<*nq; i++) {
		struct syminfo *r = &q[i];
		struct query *p = &o[i];
		r->id = p->id;
		r->attr = WSH_NONE;
		r->TF = r->TF_d = r->DF = r->DF_d = p->tf;
		r->weight = 0;
	}
	return q;
}

struct expr {
	char *name;
	int op;
};

static struct syminfo *
sete(w, ne)
WAM *w;
int *ne;
{
	struct expr o[] = {
#include "xt.h"
	};
	struct syminfo *e;
	int i;
	*ne = nelems(o);
	e = calloc(*ne, sizeof *e);
	for (i=0; i<*ne; i++) {
		struct syminfo *r = &e[i];
		struct expr *p = &o[i];
		if (p->name) {
			r->id = wam_name2id(w, WAM_COL, p->name);
			r->attr = 0;
		}
		else {
			r->id = 0;
			r->attr = p->op;
		}
		r->TF = r->TF_d = r->DF = r->DF_d = 0;
		r->weight = 0;
	}
	return e;
}

static void
print_syminfo(d, nd, w, dir, stream)
struct syminfo const *d;
WAM *w;
FILE *stream;
{
	int i;
	if (!d) {
		fprintf(stream, "(null)\n");
		return;
	}

	for (i=0; i<nd; i++) {
		struct syminfo const *r = &d[i];
		fprintf(stream, "%s\n", wam_id2name(w, dir, r->id));
	}
}
