;;;
;;;     Copyright (c) 2003 Shingo Nishioka.  All rights reserved.
;;;

;;; rcsid="$Id: mkxt.scm,v 1.1.1.1 2005/06/20 07:13:48 nis Exp $"

(define (nvars c n)
  (if (> n 0)
      (cons c (nvars (integer->char (+ 1 (char->integer c))) (- n 1)))
      '()))

(define (com n)
  (do ((i 0 (+ 1 i))
       (r '()))
      ((> i n) r)
    (set! r (append r (cnb (nvars #\a n) i)))))

(define (cnb l n)
  (if (< (length l) n)
      '()
      (if (= n 0) '(())
	  (append (let ((e (car l)))
		    (map (lambda (x) (cons e x))
			 (cnb (cdr l) (- n 1))))
		  (cnb (cdr l) n)))))

(define (pp-frq c)
  (for-each (lambda (e) (if (not (null? e)) (pp-ent e))) c))

(define (pp-ttl c)
  (for-each (lambda (e) (if (not (null? e)) (pp-ent-ttl e))) c))

(define (pp-ent e)
  (display "@")
  (display (list->string e))
  (display "\n")
  (let ((n (+ 8 (char->integer (car e)))))
    (for-each (lambda (w)
		(display (number->string n))
		(set! n (+ n 1))
		(display " ") (display w) (display "\n")) e)))

(define (pp-ent-ttl e)
  (display "%")
  (display (list->string e))
  (display "\n ")
  (display (list->string e))
  (display "\n"))

(define (expr->c e)
  (let ((l (apply append
		  (map (lambda (c)
			 (case c
			   ((#\&) (string->list "&&"))
			   ((#\|) (string->list "||"))
			   (else (list c))))
		       (string->list e)))))
    (list->string l)))

(define (expr->or e)
  (let ((l (apply append
		  (map (lambda (c)
			 (case c
			   ((#\( #\) #\& #\| #\!) '())
			   (else (list c))))
		       (string->list e)))))
    (apply string-append
	   (cdr (apply append
		       (map (lambda (c) (list "||" (list->string (list c))))
			    l))))))

(define (expr->h e)
  (let ((l (apply append
		  (map (lambda (c)
			 (case c
			   ((#\( #\) #\| #\& #\!)
			    (append (string->list "{0, '")
				    (list c)
				    (string->list "'},\n")))
			   (else
			    (append (string->list "{\"")
				    (list c)
				    (string->list "\", 0},\n")))))
		       (string->list e)))))
    (list->string l)))

(define (cpp-expr test-expr)
  (display (expr->h test-expr)))

(define (cpp-frq c test-expr l)
  (display "#include <stdio.h>\n\nmain()\n{\n\tchar buf[8192];\n\twhile (fgets(buf, sizeof buf, stdin)) {\n\t\tint ")
  (for-each (lambda (e) (display e) (display ", ")) l)
  (display "sentinel;\n\t\tchar name[128];\n\t\tsscanf(buf, \"")
  (for-each (lambda (e) (display "%d ")) l)
  (display "%s\\n\", ")
  (for-each (lambda (e) (display "&") (display e) (display ", ")) l)
  (display "name);\n")
  (display (string-append "\t\tif ((" (expr->or test-expr) ") && (" (expr->c test-expr) ")) {\n\t\t\tprintf(\"%s\\n\", name);\n\t\t}\n\t}\n\treturn 0;\n}\n")))

(define (cpp-dat c l)
  (for-each (lambda (e) (if (not (null? e)) (cpp-dent e l))) c))

(define (cpp-dent e l)
  (for-each (lambda (w) (display (string-append (if (memv w e) "1" "0") " "))) l)
  (display (string-append (list->string e) "\n")))

(define (main argv)
  (let* ((n (string->number (cadr argv)))
	 (c (com n)))
    (cond ((string=? (car argv) "frq")
	   (pp-frq c))
          ((string=? (car argv) "ttl")
	   (pp-ttl c))
          ((string=? (car argv) "d")
	   (cpp-dat c (nvars #\a n)))
	  ((string=? (car argv) "expr")
	   (cpp-expr (caddr argv)))
	  ((string=? (car argv) "c")
	   (cpp-frq c (caddr argv) (nvars #\a n))))))

(main (cdddr *argv*))
