/*
 *	Copyright (C) 2001 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: zvec.c,v 1.4 2007/03/26 07:32:14 nis Exp $"; */

#include <config.h>

#if 1
#define DEBUG_ZVEC	1
#endif

#include <sys/param.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <assert.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dwerrno.h>
#include <geta/timer.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/dwc.h>

#include "dwsys.h"
#include "sdup.h"
#include "dfi.h"
#include "dcwP.h"
#include "dxrP.h"
#include "zvec.h"

#define NEED_COMPRESS
#if defined DEBUG_ZVEC
#define NEED_UNCOMPRESS
#endif
#include "xcompress.c"

ssize_t
wvefc(v, stream, compress)
struct xr_vec const *v;
FILE *stream;
{
	ssize_t size;
	void const *p;
	if (compress==COMPRESS_NONE) {
		size = sizeof_dxr_vec(v->elem_num);
		p = v;
	}
	else {
		static struct dxr_zvec *z = NULL;
		static size_t zsize = 0, zsizeb = 0;
		if (v->elem_num > zsize) {
			struct dxr_zvec *new;
			size_t newsize, newsizeb;
			newsize = NALIGN(v->elem_num, 128);
			newsizeb = CB_WORSTSIZE(newsize);

			if (!(new = realloc(z, newsizeb))) {
				dwerrno = errno;
fprintf(stderr, "wvefc: realloc failed\n");
				return -1;
			}
			zsize = newsize;
			zsizeb = newsizeb;
			z = new;
		}
		if ((size = dxr_compress_vec(v, z, zsizeb, compress))==-1) {
			fprintf(stderr, "sorry, zvec overruns");
fprintf(stderr, "wvefc: zvec overruns\n");
			return -1;
		}
#if defined DEBUG_ZVEC && 0
{
	size_t i;
	fprintf(stderr, "\n{%s,%s}", dftostr(v->elem_num), freqtostr(v->freq_sum));
	for (i=0; i<v->elem_num; i++) {
		fprintf(stderr, " %s,%s", idxtostr(v->elems[i].id), freqtostr(v->elems[i].freq));
	}
	fprintf(stderr, "\n{%s}", sizetostr(size));
	for (i=0; i<size; i++) {
		fprintf(stderr, " %02x", z->zelems[i]);
	}
	fprintf(stderr, "\n");
}
#endif

#if defined DEBUG_ZVEC
{
	size_t i;
	struct xr_vec *n;
/*
	fprintf(stderr, "***[%d,%d]", v->elem_num, size);
*/
	if (n = malloc(sizeof_dxr_vec(v->elem_num))) {
		n->elem_num = v->elem_num;
		if (dxr_uncompress_vec(n, z, compress, NULL)==-1) {
			fprintf(stderr, "E");
		}
		for (i=0; i<v->elem_num; i++) {
			if (v->elems[i].id != n->elems[i].id) {
				fprintf(stderr, "ZVEC ERROR: i%s.%s\n", idxtostr(v->elems[i].id), idxtostr(n->elems[i].id));
			}
			if (v->elems[i].freq != n->elems[i].freq) {
				fprintf(stderr, "ZVEC ERROR: f%s.%s\n", freqtostr(v->elems[i].freq), freqtostr(n->elems[i].freq));
			}
		}
		free(n);
	}
	else {
		fprintf(stderr, "#");
	}
}
#endif
		p = z;
	}
	if (size && fwrite(p, size, 1, stream)!=1) {
		dwerrno = errno;
fprintf(stderr, "wvefc: fwrite failed: %p, %s, %s\n", p, sizetostr(size), strerror(errno));
		return -1;
	}
	return size;
}
#if 0

static struct {
	int i;
	char const *name;
} cnames[] = {
#define	foo(n)	{COMPRESS_##n, #n}
	foo(NONE),
	foo(NIB),
	foo(B36),
/*
	{COMPRESS_NONE, "none"},
	{COMPRESS_NIB, "nib"},
	{COMPRESS_B36, "b36"},
*/
#undef foo
};

int
encode_compress(c)
char const *c;
{
	size_t i;
	for (i=0; i<nelems(cnames); i++) {
		if (!strcasecmp(c, cnames[i].name)) {
			return cnames[i].i;
		}
	}
	fprintf(stderr, "unknown compression type: [%s]\n", c);
	fprintf(stderr, "available compression types are (case insensitive):");
	for (i=0; i<nelems(cnames); i++) {
		fprintf(stderr, " %s", cnames[i].name);
	}
	fprintf(stderr, "\n");
	dwerrno = DWE_BADR;
	return -1;
}

char const *
decode_compress(n)
{
	size_t i;
	for (i=0; i<nelems(cnames); i++) {
		if (cnames[i].i == n) {
			return cnames[i].name;
		}
	}
	dwerrno = DWE_BADR;
	return NULL;
}
#endif
